/************************************************************************
 *   This is part of Dancer, a text relay daemon
 *   Copyright (C) 2002 Andrew Suffield <asuffield@users.sourceforge.net>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

#ifdef FD_H
#else
#define FD_H
#include <dancerd/config.h>
#include <dancerd/modules.h>

/** \file fd.h
 * \brief File descriptor polling and callbacks
 */

/** \defgroup fd File descriptor polling and callbacks
 * \ingroup main
 * File descriptors can be registered, and added to a group of
 * handlers. When read/write/close events occur on a file descriptor,
 * the appropriate callback from the handler group will be invoked
 */

extern void io_loop(void) noreturn_attribute;

/* Servers get polled once every loop */
#define FD_CLASS_SERVER 0
/* Clients get polled unless load is heavy (ie, something called suspend_client_fds) */
#define FD_CLASS_CLIENT 1
/* Local fds are for things which aren't sockets, like the stderr pipe. We poll these once each loop */
#define FD_CLASS_LOCAL 2

#define FD_CLASS_MAX 2

typedef int fd_read_handler(int, void *);
typedef int fd_write_handler(int, void *);
/* A close handler does not need to deregister the fd, this is done
 * automatically after the close handler is called.
 */
typedef void fd_close_handler(int, void *);

extern void register_fd_handler_group(struct module *, const char *, int,
				      fd_read_handler *, fd_write_handler *, fd_close_handler *);
extern void deregister_fd_handler_group(const char *, int);
extern void register_fd_handler(int, int, const char *, int, void *);
extern void deregister_fd_handler(int);
extern void update_fd_data(int, void *);

extern void suspend_client_fds(void);
extern void resume_client_fds(void);

/* When reads/writes are blocked, the fd will not be polled for
 * read/write, even if it normally has a handler function.
 *
 * This is necessary in particular for writing. If an fd is ready for
 * writing but there is no data pending to be written, it should be
 * blocked for writing so that the write handler is not endlessly
 * called, using up all the available processing power.
 *
 * It is acceptable to simple unblock every time the relevant internal
 * buffer has data ready, and to block every time the write handler is
 * called when the buffer is empty.
 */
extern void block_fd_reads(int);
extern void block_fd_writes(int);
extern void unblock_fd_reads(int);
extern void unblock_fd_writes(int);

/* Set the maximum number of fds. This is initialised from getrlimit
 * at startup, if something calls setrlimit then it must be changed.
 * The maximum can never decrease, this function will ignore all such
 * calls.
 */
extern void set_max_fd(int);

extern int set_nonblocking_fd(int);

#endif
