/************************************************************************
 *   This is part of Dancer, a text relay daemon
 *   Copyright (C) 2002 Andrew Suffield <asuffield@users.sourceforge.net>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

#ifdef FD_INTERNAL_H
struct fd_handler;
#else
#define FD_INTERNAL_H
#include <dancerd/config.h>

/** \file fd_internal.h
 * \brief Interface between fd.c and the various polling loops themselves
 */

/** \defgroup fd_guts Interface between fd.c and the various polling loops themselves
 * \ingroup mainguts
 * The polling loop used is selected at configure time, and must
 * implement all the functions in this group
 *
 * @{
 */

/** \brief Handler group
 *
 * \par
 * Every registered handler group has one of these; fd.c will store
 * them and setup pointers into each struct fd_handler as
 * necessary. It should be considered read-only by the polling loop
 * implementation
 */
struct fd_handler_funcs
{
  /** Name of the module this group belongs to */
  char *module_name;
  /** Index of this group in the module */
  int group;
  /** Associated module handle */
  struct module *module;
  /** Function to be called when a descriptor is ready for reading */
  fd_read_handler *rh;
  /** Function to be called when a descriptor is ready for writing */
  fd_write_handler *wh;
  /** Function to be called when a descriptor has been closed */
  fd_close_handler *ch;
};

/** \brief Registered file descriptor and the associated handlers
 *
 * \par
 * fd.c will create one of these for every fd registered, and keep
 * them updated
 */
struct fd_handler
{
  int fd;
  int priority_class;
  /** Points to the current group of handler functions. This may
   * change without notification, so should always be used directly
   * and not cached
   */
  struct fd_handler_funcs *handlers;
  /** True if this fd should not be polled for reading */
  int block_read;
  /** True if this fd should not be polled for writeing */
  int block_write;
  /** Data handle which should be passed to all the handler functions */
  void *data;
};

/** True if fds in the client class should be polled */
extern int fd_poll_clients;
/** Array of handler data, indexed by fd number */
extern struct fd_handler *fd_table;

/** Called when a new fd is registered */
extern void internal_add_fd(struct fd_handler *);
/** Called when an fd is being removed */
extern void internal_remove_fd(struct fd_handler *);

/** \brief Main polling loop
 *
 * \par
 * Called once after initialisation. This function should not return
 */
extern void internal_fd_poll(void) noreturn_attribute;

/** Called when the maximum fd is increased */
extern void fd_internal_set_max_fd(unsigned int);

extern void internal_block_fd_reads(struct fd_handler *);
extern void internal_block_fd_writes(struct fd_handler *);
extern void internal_unblock_fd_reads(struct fd_handler *);
extern void internal_unblock_fd_writes(struct fd_handler *);

/** @}
 */

#endif
