/************************************************************************
 *   This is part of Dancer, a text relay daemon
 *   Copyright (C) 2002 Andrew Suffield <asuffield@users.sourceforge.net>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

#ifdef LOG_H
struct logger;
#else
#define LOG_H
#include <dancerd/config.h>
#include <stdarg.h>

/** \file log.h
 * \brief Logging system
 */

/** \defgroup log Logging system
 * \ingroup main
 * The logging system provides a simple API for modules to log
 * messages, which are passed to other modules to be stored or
 * displayed.
 *
 * It does not itself write to any files; it only distributes messages
 * to other places.
 *
 * @{
 */ 

/* This is an opaque struct. No peeking. Use the access functions/macros */
struct logger;

/* Debugging data is extremely verbose and may violate privacy. Use with caution. */
#define LOG_PRIORITY_DEBUG	0
/* Trace data tracks changes, updates, and so on (protects privacy) */
#define LOG_PRIORITY_TRACE	1
/* This is the default priority for most logs, any random stuff */
#define LOG_PRIORITY_INFO	2
/* Warnings are undesireable but acceptable things */
#define LOG_PRIORITY_WARN	3
/* Errors are things which are definitely wrong */
#define LOG_PRIORITY_ERROR	4
/* Critical errors cause dancer to exit */
#define LOG_PRIORITY_CRITICAL	5

const char *priority_name(int) pure_attribute;

/* If a log writer returns 0, errno will be checked, the logger
 * closed, and this event will be logged (assuming any loggers are
 * left)
 */
typedef int log_writer_func(const char *, int, void *, const char *, va_list);
/* If the filter returns 0 then this logger is skipped */
typedef int log_filter_func(const char *, int, void *, const char *, va_list);
/* A log reopener attempts to close and reopen the log, for SIGHUP and
 * similar (allows log rotation)
 *
 * If this fails it returns 0, and the log will be closed (and this
 * event logged)
 */
typedef int log_reopen_func(void *);
typedef void log_close_func(void *);

/* These manipulate individual loggers */
#define logger_data(A) ((A)->data)
extern void close_logger(struct logger *);

/* This creates new loggers in the system */
/* If this fails, it returns NULL, and does not call the close handler */
extern struct logger *register_logger(const char *, log_writer_func *, log_filter_func *,
				      log_reopen_func *, log_close_func *, void *);

/* This sends log messages in */
extern void log(const char *, int, const char *, ...)
     printf_format_attribute(3,4);

/** @}
 */

#endif
