/************************************************************************
 *   This is part of Dancer, a text relay daemon
 *   Copyright (C) 2002 Andrew Suffield <asuffield@users.sourceforge.net>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

#ifdef MODULES_H
struct module;
struct module_dependency;
#else
#define MODULES_H
#include <dancerd/config.h>

/* The symbol macros require llist and log */
#include <dancerd/llist.h>
#include <dancerd/log.h>

/* For NULL */
#include <stddef.h>

/** \file modules.h
 * \brief Module system interface
 */

/** \defgroup modulesystem Module system interface
 * \ingroup main
 * The interface to the loadable module system, providing mechanisms
 * for loading/unloading modules, dependencies between modules, and
 * symbol resolution across modules
 *
 * A series of preprocessor macros are defined to assist in the
 * writing of loadable modules. These macros enforce a strict sequence
 * of declaration:
 *
 * \code
 * DEFINE_NAMESPACE(namespace, version)
 * ...
 * DECLARE_SYMBOL(value_type, namespace, symbol_name, argument_list)
 * ...
 * DECLARE_SYMBOL(value_type2, namespace, symbol_name2, argument_list2)
 * ...
 * END_NAMESPACE(namespace)
 * \endcode
 * 
 * After this, the symbols may be referenced or defined. A symbol is
 * defined like this:
 *
 * \code
 * DEFINE_SYMBOL(value_type, namespace, symbol_name, argument_list)
 * \endcode
 *
 * This macro should be used as if it were equivalent to the token sequence:
 *
 * value_type namespace::symbol_name argument_list
 *
 * So, if the symbol is a function, it might be defined:
 * \code
 * DEFINE_SYMBOL(int, foo, bar, (int argc, char **argv))
 * {
 *   while (argc--)
 *     printf("%s\n", argv++);
 * }
 * \endcode
 *
 * If it were a variable, it might be defined:
 * DEFINE_SYMBOL(int, foo, baz,) = 1;
 *
 * No storage-class specifiers may be used with symbols; being a
 * symbol is itself a storage class. Symbols do not have default
 * values; if they are defined without giving them a value (legal for
 * variables, illegal for functions) then their value is unspecified.
 *
 * If a symbol is defined in this module (ie, with an DEFINE_SYMBOL)
 * then it may be referenced with the LSYM (local SYM) macro:
 * \code
 * LSYM(namespace, symbol_name)
 * \endcode
 * For example:
 * \code
 * if (LSYM(foo, bar)(0, NULL) == 1)
 *   return LSYM(foo, baz);
 * else
 *   return 0;
 * \endcode
 *
 * Normally, DECLARE_SYMBOL should be used only in headers, and
 * DEFINE_SYMBOL should not be used in headers.
 *
 * To use a symbol that is defined in a different module, the
 * following sequence of macros must be seen:
 * \code
 * DECLARE_SYMBOL(value_type, namespace, symbol_name, argument_list)
 * ...
 * IMPORTED_SYMBOL(namespace, name)
 * \endcode
 * Normally, a header will be included with the DECLARE_SYMBOL macro
 * in it. IMPORTED_SYMBOL should not be used in headers, because a
 * symbol may only be imported once per compilation unit.
 *
 * Once a symbol has been imported, it can be referenced with the SYM
 * macro, which behaves like LSYM.
 *
 * If IMPORTED_WEAK_SYMBOL is used instead of IMPORTED_SYMBOL, then the
 * symbol will be imported only when it is available. This means that
 * the module which defines the symbol can be freely loaded and
 * unloaded. The macro SYM_PRESENT(namespace, name) can be used as an
 * expression which is true when the symbol is available, and false
 * when the symbol is not available. The value of this macro will not
 * change while a module is being executed, so it does not have to be
 * tested more than once in the same function (although it does not
 * matter if it is).
 *
 * If a symbol is referenced while it is not available (ie, do
 * anything with SYM(namespace, name) while SYM_PRESENT(namespace,
 * name) is false) then the daemon will probably segfault.
 *
 * In one of the compilation units in a given module, after
 * SETUP_MODULE has been called, you should #include "symtab", and
 * generate symtab by running $(top_srcdir)/scripts/make-symtab over
 * all the source files. This will add all the necessary calls to
 * IMPORT_SYMBOL, EXPORT_SYMBOL, and a suitable SYMBOL_TABLE.
 *
 * In all the other compilation units that use imported symbols, you
 * should #include "file.c.imptab" (replacing with the name of the
 * current file). imptab is also generated by make-symtab, when the
 * --imptab option is specified.
 *
 * @{
 */

struct module;

/* This returns a unique non-NULL handle if the named module is
 * loaded and fully initialised, or NULL if it is not
 */
extern struct module *find_module(const char *name);

/* These two queue up events */
extern void load_module(const char *name, const char *version);
extern void unload_module(const char *name);
/* And this one dequeues them and processes them */
extern void poll_module_events(void);

extern void add_module_path(const char *);

typedef void module_unloaded_func(void *);
extern void register_module_unload_handler(struct module *interested, struct module *m, module_unloaded_func *, void *);
extern void deregister_module_unload_handler(struct module *interested, struct module *m, module_unloaded_func *, void *);

/** \brief Module dependency
 *
 * Declares a dependency on another module. Don't use this directly;
 * use the MODULE_DEPS macro.
 *
 * If the version field is NULL, the dependency is not on a specific
 * version.
 *
 * \code
 * MODULE_DEPS = 
 *   {
 *     {"module1", "version"},
 *     {"module2", NULL},
 *     {NULL, NULL}
 *   };
 * \endcode
 */
struct module_dependency
{
  /** Name of the module to depend on */
  const char *name;
  /** Minimum version that satisfies this dependency (NULL for any version) */
  const char *version;
};

/* Passed its own handle and an arbitrary pointer. Returns MODULE_LOAD_OK or MODULE_LOAD_FAILED */
typedef int module_init_func(void *);
/* Passed the version and init function of the module being loaded to
 * replace it, or two NULLs if it is being unloaded
 *
 * Returns MODULE_LOAD_OK if the new module loaded, MODULE_LOAD_FAILED
 * if it failed. Returns MODULE_UNLOAD_OK if there was no new module
 * to load and the unload was successful, returns MODULE_UNLOAD_FAILED
 * otherwise.
 *
 * Returns MODULE_LOAD_FUBAR if it failed to init the new module, and
 * then failed to restore the current one to a working state.
 *
 * When MODULE_INIT is called from the module system itself:
 * A return of:              Will cause:
 * MODULE_LOAD_OK            Successful load logged
 * MODULE_LOAD_FAILED        Load failure logged. MODULE_DEINIT called.
 *                           Module unloaded.
 *
 * Likewise, for MODULE_DEINIT:
 * MODULE_LOAD_OK            Log success. Unload old module.
 * MODULE_LOAD_FAILED        Log failure.
 *                           MODULE_DEINIT called for new module.
 *                           Unload new module.
 * MODULE_UNLOAD_FAILED      Log failure.
 *                           Unload new module
 * MODULE_UNLOAD_OK          Log success.
 *                           Unload module.
 *                           If an init function was passed to the deinit,
 *                            call it now with a NULL argument.
 * MODULE_LOAD_FUBAR         Log critical failure.
 *                           MODULE_DEINIT on the new module called.
 *                           Unload new module.
 *                           MODULE_DEINIT on the old module called again,
 *                            with its own version and init function as
 *                            arguments.
 *                           If that succeeds, log it and stop.
 *                           Otherwise, unload old module.
 *
 * Note that MODULE_DEINIT *must* be idempotant and universally safe
 * when called with NULL arguments. No matter how many times it is
 * called, when it returns it must ensure that all resources are
 * released.
 *
 * If MODULE_DEINIT can return MODULE_LOAD_FUBAR, it must handle the
 * case where it is passed its own init function. Otherwise it does
 * not need to do this. (This means there is a chance to get back into
 * a working state even after a FUBAR).
 */
typedef int module_deinit_func(const char *, module_init_func *);

#define MODULE_INIT _dancer__module_init
#define MODULE_DEINIT _dancer__module_deinit
extern module_init_func MODULE_INIT;
extern module_deinit_func MODULE_DEINIT;

#define MODULE_LOAD_OK 11
#define MODULE_LOAD_FAILED 12
#define MODULE_UNLOAD_FAILED 13
#define MODULE_UNLOAD_OK 14
#define MODULE_LOAD_FUBAR 15

/***********************************************************************************
 ***********************************************************************************/

/* Brace yourself. It gets ugly from here on */

/** \brief Element in a symbol table
 *
 * Don't use this directly; use the SYMBOL_TABLE, IMPORT_SYMBOL,
 * EXPORT_SYMBOL, and SYMBOL macros.
 *
 * IMPORT_SYMBOL(foo, sym1)
 * EXPORT_SYMBOL(bar, sym2)
 * SYMBOL_TABLE = 
 *   {
 *     SYMBOL(foo, sym1),
 *     SYMBOL(bar, sym2),
 *     NULL
 *   };
 *
 * Or, more simply, use IMPORTED_SYMBOL and EXPORTED_SYMBOL and use
 * make-symtab.
 */
struct pending_symbol
{
  /** Namespace the symbol is in */
  const char *namespace;
  /** Version of the namespace */
  const char *version;
  /** Name of the symbol */
  const char *name;
  /** Value to export the symbol as */
  void *ptr;
  /** Location to import the symbol to */
  void **sym_pointer;
  /** True if the symbol must always be available while this module is loaded */
  int required;
  /** True if this is an IMPORT, false if this is an EXPORT */
  int import;
};

/* Here we go... */

#define SETUP_MODULE(NAME, MODVERSION) \
 const char _dancer__module_name[] = NAME; \
 const char _dancer__module_version[] = MODVERSION; \
 const char _dancer__module_dancer_version[] = VERSION; \
 struct module *_dancer__module_handle = NULL

#define RELOAD_INTERFACE(VERSION) const int _dancer__module_reload_interface = VERSION

/* This one is like SETUP_MODULE, but you call it from all the other files (like you would extern) */
#define EXTERN_MODULE() \
 extern char _dancer__module_name[]; \
 extern char _dancer__module_version[]; \
 extern struct module *_dancer__module_handle

#define MYNAME _dancer__module_name
#define MYVERSION _dancer__module_version
#define MYHANDLE _dancer__module_handle

#define MODULE_DEPS struct module_dependency _dancer__module_deps[]

#define SYMNAME(NAMESPACE, NAME) symbol__ ## NAMESPACE ## __ ## NAME
#define PENDINGNAME(NAMESPACE, NAME) pending_symbol__ ## NAMESPACE ## __ ## NAME
#define SYMTYPENAME(NAMESPACE, NAME) NAMESPACE ## __ ## NAME ## __type
#define VERSIONNAME(NAMESPACE) _dancer__namespace_version__ ## NAMESPACE
#define LSYM(NAMESPACE, NAME) NAMESPACE ## __ ## NAME ## __symbol

#define NAMESPACE_STARTNAME(NAMESPACE) _dancer__ ## NAMESPACE ## __start
#define NAMESPACE_ENDNAME(NAMESPACE) _dancer__ ## NAMESPACE ## __end
#define NAMESPACE_END2NAME(NAMESPACE) _dancer__ ## NAMESPACE ## __end2
#ifdef HAVE_MACRO_PRAGMA
#define PRAGMA(X) _Pragma(#X)
#else
#define PRAGMA(X)
#endif
#define POISON_SYMBOL(X) PRAGMA(GCC poison X)

/**
 * \internal
 * \brief Assert that the namespace has been defined but not closed at this point
 * \param NAMESPACE namespace to check
 * \param FROB arbitrary unique identifier
 *
 * This macro will cause compiling to fail if the namespace is not
 * currently "open".
 */
#define CHECK_NAMESPACE_OPEN(NAMESPACE, FROB) \
 typedef NAMESPACE_STARTNAME(NAMESPACE) _dancer__ ## NAMESPACE ## __start_check__ ## FROB
/**
 * \internal
 * \brief Assert that the namespace has been defined and closed
 * \param NAMESPACE namespace to check
 * \param FROB arbitrary unique identifier
 *
 * This macro will cause compiling to fail if the namespace is not
 * currently "closed".
 */
#define CHECK_NAMESPACE_CLOSED(NAMESPACE, FROB) \
 typedef NAMESPACE_ENDNAME(NAMESPACE) _dancer__ ## NAMESPACE ## __start_check__ ## FROB

#define DECLARE_SYMBOL(VALUE_TYPE, NAMESPACE, NAME, PARAMETERS) \
 CHECK_NAMESPACE_OPEN(NAMESPACE, NAME ## _exported); \
 typedef VALUE_TYPE SYMTYPENAME(NAMESPACE, NAME) PARAMETERS; \
 extern SYMTYPENAME(NAMESPACE, NAME) LSYM(NAMESPACE, NAME)
#define DEFINE_NAMESPACE(NAMESPACE, VERSION) \
 static const char VERSIONNAME(NAMESPACE)[] = #VERSION; \
 typedef int NAMESPACE_STARTNAME(NAMESPACE)
#define END_NAMESPACE(NAMESPACE) \
 typedef NAMESPACE_STARTNAME(NAMESPACE) NAMESPACE_ENDNAME(NAMESPACE); \
 POISON_SYMBOL(_dancer__ ## NAMESPACE ## __start) \
 typedef NAMESPACE_ENDNAME(NAMESPACE) NAMESPACE_END2NAME(NAMESPACE)

/** \brief Declare the named symbol as exported from this module */
#define EXPORT_SYMBOL(NAMESPACE, NAME) \
 CHECK_NAMESPACE_CLOSED(NAMESPACE, NAME ## _export_symbol); \
 static struct pending_symbol PENDINGNAME(NAMESPACE, NAME) = \
  {#NAMESPACE, VERSIONNAME(NAMESPACE), #NAME, (void *)&LSYM(NAMESPACE, NAME), NULL, 0, 0}

#define IMPORT_SYMBOL_(NAMESPACE, NAME, REQUIRED) \
 CHECK_NAMESPACE_CLOSED(NAMESPACE, NAME ## _import); \
 SYMTYPENAME(NAMESPACE, NAME) (*SYMNAME(NAMESPACE, NAME)); \
 static struct pending_symbol PENDINGNAME(NAMESPACE, NAME) = \
  {#NAMESPACE, VERSIONNAME(NAMESPACE), #NAME, NULL, (void **)&SYMNAME(NAMESPACE, NAME), REQUIRED, 1}; \
 extern SYMTYPENAME(NAMESPACE, NAME) LSYM(NAMESPACE, NAME)

/** \brief Import the named symbol into the current module and require its presence */
#define IMPORT_SYMBOL(NAMESPACE, NAME) IMPORT_SYMBOL_(NAMESPACE, NAME, 1)
/** \brief Import the named symbol into the current module if it is available */
#define IMPORT_WEAK_SYMBOL(NAMESPACE, NAME) IMPORT_SYMBOL_(NAMESPACE, NAME, 0)

/** \brief Define the list of symbols in this module */
#define SYMBOL_LIST struct pending_symbol *_dancer__module_symbols[]
/** \brief Member of the SYMBOL_LIST */
#define SYMBOL(NAMESPACE, NAME) (&PENDINGNAME(NAMESPACE, NAME))

/** \brief Declare the named symbol as imported into this module */
#define IMPORTED_SYMBOL(NAMESPACE, NAME) \
 CHECK_NAMESPACE_CLOSED(NAMESPACE, NAME ## _imported); \
 extern SYMTYPENAME(NAMESPACE, NAME) (*SYMNAME(NAMESPACE, NAME))

#define IMPORTED_WEAK_SYMBOL(NAMESPACE, NAME) \
 CHECK_NAMESPACE_CLOSED(NAMESPACE, NAME ## _imported); \
 extern SYMTYPENAME(NAMESPACE, NAME) (*SYMNAME(NAMESPACE, NAME))

/** \brief Define a symbol */
#define DEFINE_SYMBOL(VALUE_TYPE, NAMESPACE, NAME, PARAMETERS) \
 CHECK_NAMESPACE_CLOSED(NAMESPACE, NAME ## _export); \
 VALUE_TYPE LSYM(NAMESPACE, NAME) PARAMETERS

#define SYM(NAMESPACE, NAME) (*SYMNAME(NAMESPACE, NAME))
#define SYM_PRESENT(NAMESPACE, NAME) SYMNAME(NAMESPACE, NAME)

/** @}
 */

#endif
