/************************************************************************
 *   This is part of Dancer, a text relay daemon
 *   Copyright (C) 2002 Andrew Suffield <asuffield@users.sourceforge.net>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

/** \file clock.c
 * Caches current time to avoid expensive time() syscalls
 */

#include <dancerd/config.h>
#include <dancerd/clock.h>
#include <dancerd/log.h>
#include <dancerd/gettext.h>

#include <errno.h>
#include <string.h>

/** \brief Time when the current event loop started
 *
 * Measured in seconds since the epoch. This should be used over
 * get_real_time() wherever possible
 */
time_t current_time;
/** \brief Time when the current event loop started, with high clock resolution
 *
 * Measured in microseconds and seconds since the epoch. This should
 * be used over get_real_time_hires() wherever possible
 */
struct time_hires current_time_hires;

/** \brief Get the current time at 1-second resolution
 * \return current time since the epoch, in seconds
 */
time_t
get_real_time(void)
{
  return time(NULL);
}

/** \brief Get the current time at the maximum available resolution
 * \param t set to the current time when the function returns
 *
 * \note
 *
 * Accuracy is limited to whatever the hardware can provide. Many
 * systems do not provide microsecond resolution.
 */
void
get_real_time_hires(struct time_hires *t)
{
  struct timeval tv;
  if (gettimeofday(&tv, NULL) == -1)
    log("core", LOG_PRIORITY_ERROR, N_("get_real_time_hires: gettimeofday() failed: %s"), strerror(errno));
  t->sec = tv.tv_sec;
  t->usec = tv.tv_usec;
}

/** \brief Updates current_time and current_time_hires
 */
void
update_clock(void)
{
  current_time = get_real_time();
  get_real_time_hires(&current_time_hires);
}

/** \brief Returns the current date/time as a string
 */
const char *
get_real_time_str(void)
{
  time_t t;
  struct tm *tm;
  static char buf[256];
  time(&t);
  tm = gmtime(&t);
  if (!strftime(buf, 256, "%a, %b %d %Y %T %Z", tm))
    buf[0] = '\0';
  return buf;
}
