/************************************************************************
 *   This is part of Dancer, a text relay daemon
 *   Copyright (C) 2002 Andrew Suffield <asuffield@users.sourceforge.net>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

/** \file ctor.c
 * Global constructors and destructors in the main binary
 *
 * ctor provides a global construction/destruction system with
 * dependencies, so that ctors/dtors are called in the correct order.
 *
 * This should not be used in modules; instead, let the module
 * dependencies themselves sort it out.
 */

#include <dancerd/config.h>
#include <dancerd/ctor.h>
#include <dancerd/llist.h>
#include <dancerd/xmalloc.h>
#include <dancerd/xstring.h>

#include <stddef.h>
#include <assert.h>

static struct llist *pending_constructors = NULL, *pending_destructors = NULL, *constructed = NULL, *destroyed = NULL;
static int done_global_construction = 0;

struct pending_ctor
{
  constructor_func *cf;
  char *name;
  struct llist *deps;
};

struct pending_dtor
{
  destructor_func *df;
  char *name;
  struct llist *deps;
};

static void
check_init(void)
{
  if (!pending_constructors)
    pending_constructors = new_llist();
  if (!pending_destructors)
    pending_destructors = new_llist();
  if (!constructed)
    constructed = new_llist();
  if (!destroyed)
    destroyed = new_llist();
}

/** \brief Register a function to be called during global construction
 * \param cf the constructor to call
 * \param name name of this constructor
 * \param deps NULL-terminated array of names of dependencies
 *
 * If deps is NULL, this is treated the same as *deps being NULL (ie,
 * no dependencies)
 *
 * If all the elements in deps have already been constructed, then the
 * constructor will be called immediately.
 *
 * If all the elements in deps have \b not already been constructed, then
 * the constructor will be delayed until global construction occurs
 * (immediately after main() begins) and all its dependencies have
 * been constructed.
 */

void
register_constructor(constructor_func *cf, const char *name, const char *deps[])
{
  if (!cf || !name)
    return;
  assert(!done_global_construction);
  check_init();
  if (deps)
    {
      const char **dep;
      int not_ready = 0;

      /* Check to see if we have any unfulfilled dependencies */
      for (dep = deps; *dep; dep++)
        {
          if (!find_llist_match(constructed, &string_match, *dep))
            {
              not_ready = 1;
              break;
            }
        }
      /* Queue it up for processing later */
      if (not_ready)
        {
          struct pending_ctor *pc = xmalloc(sizeof(struct pending_ctor));
          /* Copy the data */
          pc->cf = cf;
          pc->name = xstrdup(name);
          pc->deps = new_llist();
          for (dep = deps; *dep; dep++)
            add_llist_head(pc->deps, xstrdup(*dep));
          add_llist_head(pending_constructors, pc);
          return;
        }
    }
  /* If we got here, all dependencies (if any) are satisfied. So call the constructor */
  add_llist_head(constructed, xstrdup(name));
  (*cf)();
}

/** \brief Register a function to be called during global destruction
 * \param df the destructor to call
 * \param name name of this destructor
 * \param deps NULL-terminated array of names of dependencies
 *
 * If deps is NULL, this is treated the same as *deps being NULL (ie,
 * no dependencies)
 *
 * The destructor will be called at global destruction, after all its
 * dependencies have been destroyed.
 */

void
register_destructor(destructor_func *df, const char *name, const char *deps[])
{
  struct pending_dtor *pd;
  const char **dep;
  assert(!done_global_construction);
  if (!df || !name)
    return;
  check_init();
  /* A destructor is never called immediately; it must wait until
   * global destruction
   */
  pd = xmalloc(sizeof(struct pending_dtor));
  /* Copy the data */
  pd->df = df;
  pd->name = xstrdup(name);
  pd->deps = new_llist();
  for (dep = deps; *dep; dep++)
    add_llist_head(pd->deps, xstrdup(*dep));
  add_llist_head(pending_destructors, pd);
}

/** \internal
 * Called from main() as the first thing that happens
 */
void
do_global_construction(void)
{
  assert(!done_global_construction);

  /* This function is horribly inefficient. But it works, and it doesn't need to be fast */
  check_init();
  while (llist_head(pending_constructors))
    {
      struct llist_node *n, *next;
      int changed = 0;
      /* Loop over all the pending constructors */
      for (n = llist_head(pending_constructors); n; n = next)
        {
          struct pending_ctor *pc = llist_node_data(n);
          struct llist_node *dep;
          int not_ready = 0;
          next = llist_node_next(n);
          for (dep = llist_head(pc->deps); dep && llist_node_data(dep); dep = llist_node_next(dep))
            {
              if (!find_llist_match(constructed, &string_match, llist_node_data(dep)))
                {
                  not_ready = 1;
                }
            }
          if (!not_ready)
            {
              (*pc->cf)();
              add_llist_head(constructed, pc->name);
              destroy_llist(pc->deps, &xfree);
              xfree(pc);
              delete_llist_node(n);
              changed = 1;
            }
        }
      /* There is a series programming error if changed is not true;
       * it means one or more constructors have dependencies which
       * cannot be satisfied
       */
      assert(changed);
    }
  /* These are no longer needed; free them */
  free_llist(pending_constructors);
  pending_constructors = NULL;
  destroy_llist(constructed, &xfree);
  constructed = NULL;

  done_global_construction = 1;
}

/** \internal
 * Called (somehow) via atexit() when the daemon exits
 */
void
do_global_destruction(void)
{
  assert(!pending_constructors);
  assert(!constructed);
  assert(done_global_construction);

  if (!destroyed)
    destroyed = new_llist();

  while (pending_destructors && llist_head(pending_destructors))
    {
      struct llist_node *n, *next;
      int changed = 0;
      /* Loop over all the pending destructors */
      for (n = llist_head(pending_destructors); n; n = next)
        {
          struct pending_dtor *pd = llist_node_data(n);
          struct llist_node *dep;
          int not_ready = 0;
          next = llist_node_next(n);
          for (dep = llist_head(pd->deps); dep && llist_node_data(dep); dep = llist_node_next(dep))
            {
              if (!find_llist_match(destroyed, &string_match, llist_node_data(dep)))
                {
                  not_ready = 1;
                }
            }
          if (!not_ready)
            {
              (*pd->df)();
              add_llist_head(destroyed, pd->name);
              destroy_llist(pd->deps, &xfree);
              xfree(pd);
              delete_llist_node(n);
              changed = 1;
            }
        }
      /* There is a series programming error if changed is not true;
       * it means one or more destructors have dependencies which
       * cannot be satisfied
       */
      assert(changed);
    }

  /* Now we've finished with these, too */
  free_llist(pending_destructors);
  pending_destructors = NULL;
  destroy_llist(destroyed, &xfree);
  destroyed = NULL;
}
