/************************************************************************
 *   This is part of Dancer, a text relay daemon
 *   Copyright (C) 2002 Andrew Suffield <asuffield@users.sourceforge.net>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

/** \file error_log.c
 *
 * A simple logging method which sends all log messages for \em "core"
 * to a file
 */
#include <dancerd/config.h>
#include <dancerd/log.h>
#include <dancerd/dbuffer.h>
#include <dancerd/xmalloc.h>
#include <dancerd/error_log.h>
#include <dancerd/fd.h>
#include <dancerd/paths.h>
#include <dancerd/clock.h>
#include <dancerd/gettext.h>

#include <assert.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdio.h>
#include <errno.h>

/* An error log is a simple logger which writes events from "core" to
 * a text file.
 *
 * The error log also provides a pipe which logs all \n-terminated
 * strings sent there with priority ERROR. This is used in main.c to
 * replace stderr.
 */

struct error_log_data
{
  int fd;
  int priority;
  int use_logfile;
};

struct error_log_pipe_data
{
  struct logger *logger;
  struct dbuffer *buf;
};

static int
error_log_writer(const char *facility, int priority, void *data,
                 const char *format, va_list args)
{
  struct error_log_data *d = data;
  char buf[1024];
  if (d->fd < 0)
    return 0;
  /* First the prefix */
  snprintf(buf, 1024, "%s %s [%s]: ", get_real_time_str(), facility, priority_name(priority));
  if (write(d->fd, buf, strlen(buf)) == -1)
    return 0;
  /* Then the translated message */
  vsnprintf(buf, 1024, gettext(format), args);
  if (write(d->fd, buf, strlen(buf)) == -1)
    return 0;
  /* Then a newline */
  if (write(d->fd, "\n", 1) == -1)
    return 0;
  else
    return 1;
}

static int
error_log_filter(const char *facility, int priority, void *data,
                 const char *format unused_attribute, va_list args unused_attribute)
{
  struct error_log_data *d = data;
  return (strcmp(facility, "core") == 0) && (priority >= d->priority);
}

static int
error_log_reopen(void *data)
{
  struct error_log_data *d = data;
  close(d->fd);
  if (d->use_logfile)
    d->fd = open(logfile, O_CREAT | O_WRONLY | O_APPEND, S_IRUSR | S_IWUSR);
  else
    d->fd = -1;
  if (d->fd == -1)
    return 0;
  else
    return 1;
}

static void
error_log_close(void *data)
{
  struct error_log_data *d = data;
  close(d->fd);
  xfree(d);
}

static struct logger *
register_error_logger(int fd, int priority, int use_logfile)
{
  struct error_log_data *data = xmalloc(sizeof(struct error_log_data));
  data->use_logfile = use_logfile;
  data->fd = fd;
  data->priority = priority;
  return register_logger("Core error log", &error_log_writer, &error_log_filter,
			 &error_log_reopen, &error_log_close, data);
}

static int
error_log_pipe_read_handler(int fd, void *data)
{
  struct error_log_pipe_data *pipe_data = data;
  char buf[1024];
  ssize_t len = read(fd, buf, 1024);

  if (len == 0)
    {
      /* Who closed the pipe? Ah well, shut down the error pipe */
      close(fd);
      xfree(data);
    }
  else if (len == -1)
    {
      if (errno != EINTR)
	return 0;
    }
  else
    {
      assert(len >= 0);
      dbuffer_add(pipe_data->buf, buf, (size_t)len);
    }

  if ((len = dbuffer_find_char(pipe_data->buf, '\n')) != -1)
    {
      if (len > 0)
	{
	  dbuffer_get(pipe_data->buf, buf, (len <= 1023) ? (size_t)len : 1023);
	  buf[(len <= 1023) ? len : 1023] = '\0';
	  log("core", LOG_PRIORITY_ERROR, buf);
	}
      assert(len >= 0);
      dbuffer_remove(pipe_data->buf, (size_t)(len + 1));
    }

  return 1;
}

static void
error_log_pipe_close_handler(int fd, void *data)
{
  struct error_log_pipe_data *pipe_data = data;
  close(fd);
  free_dbuffer(pipe_data->buf);
  xfree(pipe_data);
}

/** \brief Open the error log
 * \param argv0 first member of the argv array passed to main()
 * \param priority minimum priority of messages to be logged
 * \param use_logfile if false, errors are logged to stderr instead of to a file
 *
 * \note
 * This function may only be called from main.c
 */
int
open_error_log(char *argv0, int priority, int use_logfile)
{
  static int called = 0;
  int fd;
  int error_log_pipe[2];
  struct logger *error_logger;
  struct error_log_pipe_data *pipe_data;

  /* This function may be called zero or one times only */
  if (called)
    abort();
  called = 1;

  if (use_logfile)
    {
      /* Write errors to the relevant file */
      fd = open(logfile, O_CREAT | O_WRONLY | O_APPEND, S_IRUSR | S_IWUSR);
      if (fd == -1)
	{
	  fprintf(stderr, _("%s: couldn't open %s for writing: %s\n"), argv0, logfile, strerror(errno));
	  exit(1);
	}
    }
  else
    {
      /* Write errors to original stderr (we're going to replace fd 2 with a pipe) */
      fd = dup(2);
      if (fd == -1)
	{
	  fprintf(stderr, _("%s: couldn't duplicate stderr: %s\n"), argv0, strerror(errno));
	  exit(1);
	}
    }

  error_logger = register_error_logger(fd, priority, use_logfile);

  /* Create the pipe which will replace stderr */
  if (pipe(error_log_pipe) == -1)
    {
      fprintf(stderr, _("%s: couldn't create error log pipe: %s\n"), argv0, strerror(errno));
      exit(1);
    }

  /* Allocate the pipe data struct and register the fd */
  pipe_data = xmalloc(sizeof(struct error_log_pipe_data));
  pipe_data->logger = error_logger;
  pipe_data->buf = new_dbuffer();
  register_fd_handler_group(NULL, "core.error_log", 0,
			    &error_log_pipe_read_handler, NULL, &error_log_pipe_close_handler);
  register_fd_handler(error_log_pipe[0], FD_CLASS_LOCAL, "core.error_log", 0, pipe_data);

  return error_log_pipe[1];
}
