/************************************************************************
 *   This is part of Dancer, a text relay daemon
 *   Copyright (C) 2002 Andrew Suffield <asuffield@users.sourceforge.net>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

/** \file event.c
 *
 * Generic string-based event handling system. This can be used for
 * two unrelated modules to signal each other without knowing anything
 * about the others existance.  When an event is raised, all the
 * registered handlers will be called in the sequence they were
 * registered
 */

#include <dancerd/config.h>
#include <dancerd/event.h>
#include <dancerd/llist.h>
#include <dancerd/xmalloc.h>
#include <dancerd/xstring.h>
#include <dancerd/ctor.h>

/** \brief Opaque event type
 *
 * An opaque type which represents an event. It is guaranteed that
 * only one event struct exists for a given name
 */
struct event
{
  /** \internal Name of the event */
  char *name;
  /** \internal List of handlers to invoke when the event is raised
   * This list contains data of type event_handler_func*
   */
  struct llist *handlers;
};

struct llist *events = NULL;

/** \brief Register an event handler
 * \param e event to register for
 * \param h function to be added to the list of handlers
 */
void
add_event_handler(struct event *e, event_handler_func *h)
{
  add_llist_tail(e->handlers, (void *)h);
}

/** \brief Unregister an event handler
 * \param e event to unregister from
 * \param h function to be deleted from the list of handlers
 */
void
delete_event_handler(struct event *e, event_handler_func *h)
{
  delete_llist_data(e->handlers, (void *)h);
}

static void
call_event_handler(void *d1, void *d2)
{
  event_handler_func *h = (event_handler_func *)d1;
  struct event *e = d2;
  (*h)(e);
}

/** \brief Raises the given event
 * \param e event to raise
 *
 * Causes all the handlers for this event to be called immediately;
 * when the function returns they will all have completed.
 */
void
raise_event(struct event *e)
{
  foreach_llist(e->handlers, &call_event_handler, e);
}

static int
match_event(const void *d1, const void *d2)
{
  const struct event *e = d1;
  const char *s = d2;
  return strcmp(e->name, s) == 0;
}

/** \brief Selects an event by name
 * \param name name of the event
 * \return pointer to the event
 *
 * In order to save expensive string lookups, they are split out into
 * this function. It will always return the same pointer for the same
 * string, so the return value can be cached indefinitely.
 */
struct event *
event(const char *name)
{
  if (!events)
    events = new_llist();

  struct llist_node *n = find_llist_match(events, &match_event, name);
  struct event *e;
  if (n)
    return llist_node_data(n);

  e = xmalloc(sizeof(struct event));
  e->name = xstrdup(name);
  e->handlers = new_llist();
  add_llist_head(events, e);
  return e;
}
