/************************************************************************
 *   This is part of Dancer, a text relay daemon
 *   Copyright (C) 2002 Andrew Suffield <asuffield@users.sourceforge.net>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

/** \file fd.c
 * \brief File descriptor polling and callbacks
 *
 * File descriptors can be registered, and added to a group of
 * handlers. When read/write/close events occur on a file descriptor,
 * the appropriate callback from the handler group will be invoked
 */

#include <dancerd/config.h>
#include <dancerd/fd.h>
#include <dancerd/fd_internal.h>
#include <dancerd/xmalloc.h>
#include <dancerd/xstring.h>
#include <dancerd/llist.h>
#include <dancerd/log.h>
#include <dancerd/modules.h>
#include <dancerd/ctor.h>
#include <dancerd/gettext.h>

#ifdef NBLOCK_SYSV
#include <sys/ioctl.h>
#endif
#include <assert.h>
#include <stdio.h>
#include <sys/resource.h>
#include <unistd.h>
#include <errno.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>

int fd_poll_clients = 1;
struct fd_handler *fd_table = NULL;
int fd_table_size = 0;
static struct llist *handler_funcs;

#if defined(RLIMIT_OFILE) && !defined(RLIMIT_NOFILE)
# define RLIMIT_NOFILE RLIMIT_OFILE
#endif

static inline void
clear_fd_handler(struct fd_handler *h)
{
  h->fd = -1;
  h->priority_class = -1;
  h->handlers = NULL;
  h->block_read = h->block_write = 1;
}

GLOBAL_CONSTRUCTOR(main_fd, NULL)
{
  struct rlimit rlim;
  int max_open_files = 0;
  int i;
  if (getrlimit(RLIMIT_NOFILE, &rlim) == -1)
    {
      fprintf(stderr, N_("getrlimit() failed to get the maximum number of open files: %s\n"), strerror(errno));
      exit(0);
    }
  if (rlim.rlim_cur == RLIM_INFINITY)
    {
      /* Drat, need a firm figure. How about the hard limit? */
      if (rlim.rlim_max == RLIM_INFINITY)
	{
	  /* Is this even possible? Let's try sysconf, that might tell us what the kernel is capable of */
	  if ((max_open_files = sysconf(_SC_OPEN_MAX)) == -1)
	    {
	      /* Give up. Go set a ulimit, we need a number. */
	      fprintf(stderr, N_("Couldn't determine the maximum number of open files: %s\n"), strerror(errno));
	      exit(1);
	    }
	}
      else
	max_open_files = rlim.rlim_max;
    }
  else
    max_open_files = rlim.rlim_cur;
  fd_table_size = max_open_files;
  fd_table = xmalloc(fd_table_size * sizeof(struct fd_handler));
  for (i = 0; i < fd_table_size; i++)
    clear_fd_handler(&fd_table[i]);
  assert(fd_table_size >= 0);
  fd_internal_set_max_fd((unsigned int)fd_table_size);

  if (!handler_funcs)
    handler_funcs = new_llist();
}

GLOBAL_DESTRUCTOR(main_fd, "main_modules", NULL)
{
  destroy_llist(handler_funcs, &xfree);
}

/** \brief Set the maximum fd number
 * \param new_max the new maximum
 *
 * Sets the maximum integer value an fd can have. The memory occupied
 * by the internal data structures is directly proportional to this
 * maximum. It can never be decreased, only increased.
 *
 * If necessary, it will be increased dynamically when fds are
 * registered, by calling this function.
 *
 * \note
 * This function is quite slow
 */
void
set_max_fd(int new_max)
{
  struct fd_handler *new_table;
  int i;
  if (new_max <= fd_table_size)
    return;
  new_table = xmalloc(new_max * sizeof(struct fd_handler));
  memcpy(new_table, fd_table, fd_table_size * sizeof(struct fd_handler));
  xfree(fd_table);
  for (i = fd_table_size; i < new_max; i++)
    clear_fd_handler(&fd_table[i]);
  fd_table = new_table;
  fd_table_size = new_max;
  assert(fd_table_size >= 0);
  fd_internal_set_max_fd((unsigned int)fd_table_size);
}

/** \brief Master IO loop
 *
 * This function is called from main() after initialisation has
 * completed. If it returns, the daemon will exit.
 */
void
io_loop(void)
{
  internal_fd_poll();
}

struct hf_data
{
  const char *module_name;
  int group;
};

static int
handler_funcs_match(const void *d1, const void *d2)
{
  const struct fd_handler_funcs *hf1 = d1;
  const struct hf_data *hf2 = d2;
  return (hf1->group == hf2->group) && (strcmp(hf1->module_name, hf2->module_name) == 0);
}

static inline struct fd_handler_funcs *
get_handler_funcs(const char *module_name, int group)
{
  struct fd_handler_funcs *hf;
  struct hf_data hd;
  struct llist_node *n;

  hd.module_name = module_name;
  hd.group = group;

  if (!handler_funcs)
    handler_funcs = new_llist();

  n = find_llist_match(handler_funcs, &handler_funcs_match, &hd);
  if (n)
    hf = llist_node_data(n);
  else
    {
      hf = xmalloc(sizeof(struct fd_handler_funcs));
      hf->module_name = xstrdup(module_name);
      hf->group = group;
      hf->rh = NULL;
      hf->wh = NULL;
      hf->ch = NULL;
      add_llist_head(handler_funcs, hf);
    }
  return hf;
}

static void
deregister_fd_handler_group_(void *d)
{
  struct fd_handler_funcs *hf = d;
  hf->rh = NULL;
  hf->wh = NULL;
  hf->ch = NULL;
  deregister_module_unload_handler(NULL, hf->module, &deregister_fd_handler_group_, hf);
}

/** \brief Deregister the given handler group
 * \param module_name name of the module this group is associated with
 * \param group index of this group
 *
 * When a handler group is deregistered, any fds which are associated
 * with that group will continue to be polled, but no handlers will be
 * called for them until a new group with the same index and module
 * name is registered
 */
void
deregister_fd_handler_group(const char *module_name, int group)
{
  struct fd_handler_funcs *hf;
  hf = get_handler_funcs(module_name, group);
  deregister_fd_handler_group_(hf);
}

/** \brief Register a handler group
 * \param m module handle that is registering the group
 * \param module_name name of the module that is registering the group
 * \param group index of the group
 * \param rh function to be called on reads
 * \param wh function to be called on writes
 * \param ch function to be called on closes
 *
 * A handler group is referenced by a module_name/group pair, which
 * uniquely identifies it.
 *
 * When the given module handle (if any) is unloaded, all fd handlers
 * will be deregistered automatically. This handle should therefore
 * point to the module which contains the handler functions, or NULL
 * if they are in the main binary
 */
void
register_fd_handler_group(struct module *m, const char *module_name, int group, 
			  fd_read_handler *rh, fd_write_handler *wh, fd_close_handler *ch)
{
  struct fd_handler_funcs *hf;
  hf = get_handler_funcs(module_name, group);
  hf->module = m;
  hf->rh = rh;
  hf->wh = wh;
  hf->ch = ch;
  register_module_unload_handler(NULL, m, &deregister_fd_handler_group_, hf);
}

/** \brief Put a file descriptor into non-blocking mode
 * \param fd File descriptor to modify
 * \return true on success, false on failure
 *
 * On failure, errno will contain the error code.
 */
#ifdef NBLOCK_SYSV
int
set_nonblocking_fd(int fd)
{
  int res = 1;

  if (ioctl(fd, FIONBIO, &res) == -1)
    return 0;
  else
    return 1;
}
#else /* !NBLOCK_SYSV */
int
set_nonblocking_fd(int fd)
{
  int flags = 0;

  flags = fcntl(fd, F_GETFL, 0);
  if (flags == -1)
    return 0;

#ifdef NBLOCK_POSIX
  flags |= O_NONBLOCK;
#endif
#ifdef NBLOCK_BSD
  flags |= O_NDELAY;
#endif

  if (fcntl(fd, F_SETFL, flags) == -1)
    return 0;
  else
    return 1;
}
#endif /* !NBLOCK_SYSV */

/** \brief Register a file descriptor
 * \param fd file descriptor to register
 * \param priority_class priority class this fd is in (see below)
 * \param module_name name of the module the handler group is in
 * \param group index of the handler group
 * \param data arbitrary value to be passed to all handlers for this fd
 *
 * The priority class can be one of FD_CLASS_SERVER, FD_CLASS_CLIENT,
 * or FD_CLASS_LOCAL. Server fds are those belonging to remote
 * servers, and will always be given priority over those associated
 * with clients. Local fds are always handled immediately and without
 * any magic; the local class is intended for things like
 * stdin/stdout.
 */
void
register_fd_handler(int fd, int priority_class,
		    const char *module_name, int group, void *data)
{
  struct fd_handler_funcs *hf;

  if (fd < 0)
    return;
  if (fd >= fd_table_size)
    {
      /* Bugger. Increase to match */
      log("core", LOG_PRIORITY_WARN, N_("Attempting to register fd %d with max set to %d, increasing max"), fd, fd_table_size - 1);
      set_max_fd(fd);
    }
  /* This is bad. Try to recover as best we can. */
  if (fd_table[fd].fd != -1)
    {
      /* We can't call the old close handler here, because it will
       * close() the fd. Log the problem and carry on, hope it doesn't
       * matter
       */
      log("core", LOG_PRIORITY_ERROR, N_("BUG: register_fd_handler() called with already-registered fd %d"), fd);
      deregister_fd_handler(fd);
    }

  if (!set_nonblocking_fd(fd))
    {
      log("core", LOG_PRIORITY_ERROR, N_("Couldn't set fd %d non-blocking: %s"), fd, strerror(errno));
      return;
    }

  hf = get_handler_funcs(module_name, group);
  fd_table[fd].fd = fd;
  fd_table[fd].priority_class = priority_class;
  fd_table[fd].handlers = hf;
  fd_table[fd].block_read = fd_table[fd].block_write = 0;
  fd_table[fd].data = data;
  internal_add_fd(&fd_table[fd]);
}

/** \brief Change the arbitrary data value for an fd
 * \param fd fd to change
 * \param data new arbitrary value to be passed to all handlers for this fd
 */
void
update_fd_data(int fd, void *data)
{
  if ((fd < 0) || (fd_table_size <= fd))
    return;
  if (fd_table[fd].fd == -1)
    return;
  fd_table[fd].data = data;
}

/** \brief Deregister a file descriptor
 * \param fd descriptor to deregister
 *
 * No further handlers will be called for this fd.
 */
void
deregister_fd_handler(int fd)
{
  if ((fd < 0) || (fd_table_size <= fd))
    return;
  if (fd_table[fd].fd == fd)
    internal_remove_fd(&fd_table[fd]);
  clear_fd_handler(&fd_table[fd]);
}

/** \brief Disable polling of fds in the client class
 */
void
suspend_client_fds(void)
{
  fd_poll_clients = 0;
}

/** \brief Enable polling of fds in the client class
 */
void
resume_client_fds(void)
{
  fd_poll_clients = 1;
}

/** \brief Disable polling for read on the given fd
 */
void
block_fd_reads(int fd)
{
  if ((fd < 0) || (fd_table_size <= fd))
    return;
  if (fd_table[fd].fd != -1)
    {
      fd_table[fd].block_read = 1;
      internal_block_fd_reads(&fd_table[fd]);
    }
}

/** \brief Disable polling for write on the given fd
 */
void
block_fd_writes(int fd)
{
  if ((fd < 0) || (fd_table_size <= fd))
    return;
  if (fd_table[fd].fd != -1)
    {
      fd_table[fd].block_write = 1;
      internal_block_fd_writes(&fd_table[fd]);
    }
}

/** \brief Enable polling for read on the given fd
 */
void
unblock_fd_reads(int fd)
{
  if ((fd < 0) || (fd_table_size <= fd))
    return;
  if (fd_table[fd].fd != -1)
    {
      fd_table[fd].block_read = 0;
      internal_unblock_fd_reads(&fd_table[fd]);
    }
}

/** \brief Enable polling for write on the given fd
 */
void
unblock_fd_writes(int fd)
{
  if ((fd < 0) || (fd_table_size <= fd))
    return;
  if (fd_table[fd].fd != -1)
    {
      fd_table[fd].block_write = 0;
      internal_unblock_fd_writes(&fd_table[fd]);
    }
}

