/************************************************************************
 *   This is part of Dancer, a text relay daemon
 *   Copyright (C) 2002 Andrew Suffield <asuffield@users.sourceforge.net>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

#define MODPREFIX core

#include <dancerd/config.h>
#include <dancerd/fd.h>
#include <dancerd/fd_internal.h>
#include <dancerd/timer.h>
#include <dancerd/clock.h>
#include <dancerd/signal_handler.h>
#include <dancerd/xmalloc.h>
#include <dancerd/modules.h>
#include <dancerd/log.h>
#include <dancerd/ctor.h>
#include <dancerd/gettext.h>

#include <sys/poll.h>
#include <errno.h>
#include <string.h>
#include <assert.h>

struct fd_index
{
  /* Index of fd -> position in pollfd */
  unsigned int *index;
  struct pollfd *pollfd_list, *working_pollfd_list;
  /* Number of elements allocated */
  unsigned int pollfd_list_size;
  /* Number of elements used */
  unsigned int pollfd_list_used;
};

static unsigned int fd_internal_max_fds = 0;
static struct fd_index fd_index[FD_CLASS_MAX + 1];

GLOBAL_CONSTRUCTOR(main_fd_poll, NULL)
{
  int i;
  for (i = 0; i <= FD_CLASS_MAX; i++)
    {
      fd_index[i].index = NULL;
      fd_index[i].pollfd_list = fd_index[i].working_pollfd_list = NULL;
      fd_index[i].pollfd_list_size = fd_index[i].pollfd_list_used = 0;
    }
}

static inline void
internal_add_fd_class(struct fd_handler *fh, struct fd_index *fi)
{
  int i = fi->pollfd_list_used++;
  assert(fi->pollfd_list_used <= fi->pollfd_list_size);
  fi->pollfd_list[i].fd = fh->fd;
  fi->pollfd_list[i].events = 0;
  if (fh->handlers->rh)
    fi->pollfd_list[i].events |= POLLIN;
  if (fh->handlers->wh)
    fi->pollfd_list[i].events |= POLLOUT;
  fi->pollfd_list[i].revents = 0;
  fi->index[fh->fd] = i;
}

void
internal_add_fd(struct fd_handler *fh)
{
  internal_add_fd_class(fh, &fd_index[fh->priority_class]);
}

static inline void
internal_remove_fd_class(struct fd_handler *fh, struct fd_index *fi)
{
  unsigned int i = fi->index[fh->fd], j;
  if (i == 0)
    return;
  assert(fi->pollfd_list_used > 0);
  memmove(&fi->pollfd_list[i], &fi->pollfd_list[i+1], (fi->pollfd_list_used - (i + 1)) * sizeof(struct pollfd));
  fi->pollfd_list_used--;
  fi->index[fh->fd] = 0;
  /* Update the index to reflect this */
  for (j = 0; j < fd_internal_max_fds; j++)
    if (fi->index[j] > i)
      fi->index[j]--;
}

void
internal_remove_fd(struct fd_handler *fh)
{
  internal_remove_fd_class(fh, &fd_index[fh->priority_class]);
}

static inline void
internal_fd_poll_class(struct fd_index *fi)
{
  int active_fds;
  unsigned int target_fds;
  unsigned int i;
  unsigned int j;
  if (fi->pollfd_list_used == 0)
    return;
  /* Save this value out front, because it can change when we call
   * close handlers/deregister_fd_handler()
   */
  target_fds = fi->pollfd_list_used;
  memcpy(fi->working_pollfd_list, fi->pollfd_list, fi->pollfd_list_used * sizeof(struct pollfd));
  if ((active_fds = poll(fi->working_pollfd_list, target_fds, 100)) == -1)
    {
      switch (errno)
	{
	case EINTR:
	  /* Ignore this */
	  return;
	case EBADF:
	  /* Damn, serious error. Find the bad fds */
	  for (j = 0; j < fi->pollfd_list_used; j++)
	    if (fi->working_pollfd_list[j].revents & POLLNVAL)
	      {
		int fd = fi->working_pollfd_list[j].fd;
		log("core", LOG_PRIORITY_ERROR, N_("internal_fd_poll_class: invalid fd %d in list, removing"), fd);
		/* Call the close handler */
		(*fd_table[fd].handlers->ch)(fd, fd_table[fd].data);
		/* This modifies fi->pollfd_list, but working_pollfd_list will remain intact */
		/* fi->pollfd_list_used does not, however, so we saved it in target_fds */
		deregister_fd_handler(fd);
	      }
	  break;
	default:
	  log("core", LOG_PRIORITY_CRITICAL, N_("internal_fd_poll_class: poll() failed: %s"), strerror(errno));
	  /* This is either "out of memory" or "segmentation fault". Either way, die. */
	  exit(1);
	}
    }
  /* OK, run over the fd list */
  for (i = 0; i < target_fds; i++)
    {
      int fd = fi->working_pollfd_list[i].fd;
      /* The fd must be valid and in this table at the start, but the
       * callbacks could have invalidated it, so we'll skip any such
       * fds
       */
      if (fd_table[fd].fd == -1)
        continue;
      if (!fd_table[fd].handlers)
        continue;
      assert(fd_table[fd].fd != -1);
      assert(fd_table[fd].handlers);
      if ((fi->working_pollfd_list[i].revents & POLLIN) && fd_table[fd].handlers->rh)
	if (!(*fd_table[fd].handlers->rh)(fd, fd_table[fd].data))
	  {
            /* It is possible that the read handler caused the fd to
             * be deregistered, so handle that case safely
             */
	    if (fd_table[fd].handlers && fd_table[fd].handlers->ch)
	      (*fd_table[fd].handlers->ch)(fd, fd_table[fd].data);
	    deregister_fd_handler(fd);
            continue;
	  }
      if (fd_table[fd].fd == -1)
        continue;
      if ((fi->working_pollfd_list[i].revents & POLLOUT) && fd_table[fd].handlers->wh)
	if (!(*fd_table[fd].handlers->wh)(fd, fd_table[fd].data))
	  {
	    if (fd_table[fd].handlers && fd_table[fd].handlers->ch)
	      (*fd_table[fd].handlers->ch)(fd, fd_table[fd].data);
	    deregister_fd_handler(fd);
            continue;
	  }
      if (fd_table[fd].fd == -1)
        continue;
      if ((fi->working_pollfd_list[i].revents & POLLHUP)
	  || (fi->working_pollfd_list[i].revents & POLLERR))
	{
	  if (fd_table[fd].handlers && fd_table[fd].handlers->ch)
	    (*fd_table[fd].handlers->ch)(fd, fd_table[fd].data);
	  deregister_fd_handler(fd);
          continue;
	}
    }
}

void
internal_fd_poll(void)
{
  for(;;)
    {
      poll_timer();
      update_clock();
      poll_module_events();
      poll_signals();
      internal_fd_poll_class(&fd_index[FD_CLASS_LOCAL]);
      internal_fd_poll_class(&fd_index[FD_CLASS_SERVER]);
      if (fd_poll_clients)
	internal_fd_poll_class(&fd_index[FD_CLASS_CLIENT]);
    }
}

void
fd_internal_set_max_fd(unsigned int new_max)
{
  unsigned int *new_index;
  struct pollfd *new_pollfds;
  int i;
  if (new_max < fd_internal_max_fds)
    return;
  for (i = 0; i <= FD_CLASS_MAX; i++)
    {
      /* First the index */
      new_index = xmalloc(new_max * sizeof(unsigned int));
      if (fd_index[i].index)
	memcpy(new_index, fd_index[i].index, fd_internal_max_fds * sizeof(unsigned int));
      memset(new_index + fd_internal_max_fds, 0, (new_max - fd_internal_max_fds) * sizeof(unsigned int));
      xfree(fd_index[i].index);
      fd_index[i].index = new_index;

      /* Next the pollfds */
      new_pollfds = xmalloc(new_max * sizeof(struct pollfd));
      if (fd_index[i].pollfd_list)
	memcpy(new_pollfds, fd_index[i].pollfd_list, fd_index[i].pollfd_list_size * sizeof(struct pollfd));
      xfree(fd_index[i].pollfd_list);
      fd_index[i].pollfd_list = new_pollfds;

      /* And the scratch space */
      xfree(fd_index[i].working_pollfd_list);
      fd_index[i].working_pollfd_list = xmalloc(new_max * sizeof(struct pollfd));

      fd_index[i].pollfd_list_size = new_max;
      fd_internal_max_fds = new_max;
    }
}

void
internal_block_fd_reads(struct fd_handler *fh)
{
  struct fd_index *fi = &fd_index[fh->priority_class];
  unsigned int i = fi->index[fh->fd];
  fi->pollfd_list[i].events &= ~POLLIN;
}

void
internal_block_fd_writes(struct fd_handler *fh)
{
  struct fd_index *fi = &fd_index[fh->priority_class];
  unsigned int i = fi->index[fh->fd];
  fi->pollfd_list[i].events &= ~POLLOUT;
}

void
internal_unblock_fd_reads(struct fd_handler *fh)
{
  struct fd_index *fi = &fd_index[fh->priority_class];
  unsigned int i = fi->index[fh->fd];
  if (fh->handlers->rh)
    fi->pollfd_list[i].events |= POLLIN;
}

void
internal_unblock_fd_writes(struct fd_handler *fh)
{
  struct fd_index *fi = &fd_index[fh->priority_class];
  unsigned int i = fi->index[fh->fd];
  if (fh->handlers->wh)
    fi->pollfd_list[i].events |= POLLOUT;
}

