/************************************************************************
 *   This is part of Dancer, a text relay daemon
 *   Copyright (C) 2002 Andrew Suffield <asuffield@users.sourceforge.net>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

#include <dancerd/config.h>
#include <dancerd/fd.h>
#include <dancerd/dbuffer.h>
#include <dancerd/fdbuffer.h>
#include <dancerd/log.h>
#include <dancerd/xmalloc.h>
#include <dancerd/ctor.h>
#include <dancerd/gettext.h>

#include <assert.h>
#include <stdio.h>
#include <stdarg.h>
#include <errno.h>
#include <string.h>
#include <unistd.h>

#define WRITE_CHUNK 2048
#define READ_CHUNK 2048

struct fdbuffer
{
  struct dbuffer *readbuf, *writebuf;
  fdbuffer_read_handler *rh;
  int fd;
  struct module *module;
};

static int
fdbuffer_fd_read_handler(int fd, void *data)
{
  struct fdbuffer *fdb = data;
  char buf[READ_CHUNK];
  ssize_t len;

  if ((len = read(fd, buf, READ_CHUNK)) == -1)
    {
      switch (errno)
	{
	  /* If we get this then the fd is not behaving in a normal
	   * manner; it's returning ready-for-reading but isn't
	   * ready
	   */
	case EAGAIN:
	  block_fd_reads(fd);
	  return 1;
	default:
	  log("core", LOG_PRIORITY_ERROR, N_("Error reading from buffered fd %d: %s"), fd, strerror(errno));
	  return 0;
	}
    }

  assert(len >= 0);
  dbuffer_add(fdb->readbuf, buf, (size_t)len);

  if (fdb->rh)
    (*fdb->rh)(fdb, fdb->readbuf, buf, (size_t)len);
  return 1;
}

static int
fdbuffer_fd_write_handler(int fd, void *data)
{
  struct fdbuffer *fdb = data;
  char buf[WRITE_CHUNK];
  size_t buf_len;
  ssize_t wrote_len;

  if (!dbuffer_len(fdb->writebuf))
    {
      block_fd_writes(fd);
      return 1;
    }

  buf_len = dbuffer_extract(fdb->writebuf, buf, WRITE_CHUNK);

  if ((wrote_len = write(fd, buf, buf_len)) == -1)
    {
      switch (errno)
	{
	case EAGAIN:
	  block_fd_writes(fd);
	  return 1;
	default:
	  log("core", LOG_PRIORITY_ERROR, N_("Error writing to buffered fd %d: %s"), fd, strerror(errno));
	  return 0;
	}
    }

  return 1;
}

static void
fdbuffer_fd_close_handler(int fd unused_attribute, void *data)
{
  struct fdbuffer *fdb = data;
  fdb->fd = -1;
  free_dbuffer(fdb->writebuf);
  fdb->writebuf = NULL;
  if (fdb->rh)
    (*fdb->rh)(fdb, fdb->readbuf, NULL, 0);
}

static void
detach_read_handler(void *d)
{
  struct fdbuffer *fdb = d;
  if (fdb->module)
    deregister_module_unload_handler(NULL, fdb->module, &detach_read_handler, fdb);
  fdb->module = NULL;
  fdb->rh = NULL;
}

struct fdbuffer *
setup_fdbuffer(struct module *m, int fd, int priority_class, fdbuffer_read_handler *rh)
{
  struct fdbuffer *fdb;

  if (fd < 0)
    return NULL;

  fdb = xmalloc(sizeof(struct fdbuffer));
  fdb->readbuf = new_dbuffer();
  fdb->writebuf = new_dbuffer();
  fdb->fd = fd;
  fdb->rh = rh;
  fdb->module = m;
  if (m)
    register_module_unload_handler(NULL, m, &detach_read_handler, fdb);

  register_fd_handler(fd, priority_class, "core.fdbuffer", rh ? 0 : 1, fdb);

  return fdb;
}

void
attach_fdbuffer_read_handler(struct module *m, struct fdbuffer *fdb, fdbuffer_read_handler *rh)
{
  detach_read_handler(fdb);
  fdb->module = m;
  fdb->rh = rh;
}

GLOBAL_CONSTRUCTOR(main_fdbuffer, NULL)
{
  register_fd_handler_group(NULL, "core.fdbuffer", 0, &fdbuffer_fd_read_handler,
			    &fdbuffer_fd_write_handler, &fdbuffer_fd_close_handler);
  register_fd_handler_group(NULL, "core.fdbuffer", 1, NULL,
			    &fdbuffer_fd_write_handler, &fdbuffer_fd_close_handler);
}

void
destroy_fdbuffer(struct fdbuffer *fdb)
{
  if (!fdb)
    return;

  if (fdb->module)
    deregister_module_unload_handler(NULL, fdb->module, &detach_read_handler, fdb);

  if (fdb->fd >= 0)
    deregister_fd_handler(fdb->fd);

  free_dbuffer(fdb->readbuf);
  free_dbuffer(fdb->writebuf);
  xfree(fdb);
}

void
write_fdbuffer(struct fdbuffer *fdb, const char *data, size_t len)
{
  if (!fdb || !data)
    return;
  if (fdb->writebuf)
    dbuffer_add(fdb->writebuf, data, len);
  if (fdb->fd >= 0)
    unblock_fd_writes(fdb->fd);
}

void
printf_fdbuffer(struct fdbuffer *fdb, const char *format, ...)
{
  static char buf[32768];
  va_list args;

  va_start(args, format);
  vsnprintf(buf, 32768, format, args);
  va_end(args);

  write_fdbuffer(fdb, buf, strlen(buf));
}

size_t
fdbuffer_write_pending(struct fdbuffer *fdb)
{
  return fdb ? dbuffer_len(fdb->writebuf) : 0;
}

int
fdbuffer_closed(struct fdbuffer *fdb)
{
  return (fdb && (fdb->fd >= 0)) ? 0 : 1;
}
