/************************************************************************
 *   This is part of Dancer, a text relay daemon
 *   Copyright (C) 2002 Andrew Suffield <asuffield@users.sourceforge.net>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

#include <dancerd/config.h>
#include <dancerd/hashtable.h>
#include <dancerd/xmalloc.h>
#include <dancerd/xstring.h>
#include <dancerd/log.h>
#include <dancerd/gettext.h>

#include <assert.h>
#include <stddef.h>
#include <stdint.h>

struct hashedpair
{
  char *key;
  void *value;
  struct hashedpair *next;
};

struct hashtable
{
  hash_func *hf;
  uint32_t size;
  struct hashedpair *table;
};

struct hashtable *
new_hashtable(hash_func *hf, uint32_t size)
{
  struct hashtable *table;
  uint32_t i;
  if (!hf || !size)
    return NULL;
  table = xmalloc(sizeof(struct hashtable));
  table->size = size;
  table->hf = hf;
  table->table = xmalloc(size * sizeof(struct hashedpair));
  for (i = 0; i < size; i++)
    {
      table->table[i].key = NULL;
      table->table[i].value = NULL;
      table->table[i].next = NULL;
    }
  return table;
}

void
free_hashtable(struct hashtable *table)
{
  uint32_t i;
  if (!table)
    return;
  for (i = 0; i < table->size; i++)
    {
      struct hashedpair *hp = &table->table[i], *next;
      xfree(hp->key);
      for (hp = hp->next; hp; hp = next)
        {
          next = hp->next;
          xfree(hp->key);
          xfree(hp);
        }
    }
  xfree(table->table);
  xfree(table);
}

static inline struct hashedpair *
hashtable_find(struct hashtable *table, const char *key, int add)
{
  uint32_t hash = table->hf(key) % table->size;
  struct hashedpair *hp, *last = NULL;

  if (table->table[hash].key)
    for (hp = &table->table[hash]; hp; hp = hp->next)
      {
        if (strcmp(hp->key, key) == 0)
          return hp;
        last = hp;
      }

  if (!add)
    return NULL;

  if (last)
    {
      last->next = xmalloc(sizeof(struct hashedpair));
      last->next->key = xstrdup(key);
      last->next->value = NULL;
      last->next->next = NULL;
      return last->next;
    }
  else
    {
      table->table[hash].key = xstrdup(key);
      return &table->table[hash];
    }
}

void
hashtable_insert(struct hashtable *table, const char *key, void *value)
{
  struct hashedpair *hp = hashtable_find(table, key, 1);
  hp->value = value;
}

void *
hashtable_lookup(struct hashtable *table, const char *key)
{
  struct hashedpair *hp = hashtable_find(table, key, 0);
  if (hp)
    return hp->value;
  else
    return NULL;
}

void
hashtable_delete(struct hashtable *table, const char *key)
{
  uint32_t hash = table->hf(key) % table->size;
  struct hashedpair *hp, *last = NULL;

  if (table->table[hash].key)
    for (hp = &table->table[hash]; hp; hp = hp->next)
      {
        if (strcmp(hp->key, key) == 0)
          {
            if (last)
              {
                last->next = hp->next;
                xfree(hp->key);
                xfree(hp);
              }
            else
              {
                xfree(table->table[hash].key);
                if (hp->next)
                  table->table[hash] = *hp->next;
                else
                  {
                    table->table[hash].key = NULL;
                    table->table[hash].value = NULL;
                    table->table[hash].next = NULL;
                  }
              }
            return;
          }
        last = hp;
      }
}

uint32_t
hash_ascii(const char *value)
{
  uint32_t hash = 0;
  char c;
  int i = 8;

  assert(value);

  while ((c = *value++) && i--)
    {
      hash ^= c;
      hash <<= 4;
    }
  return hash;
}

int
hashtable_check_sanity(struct hashtable *table)
{
  uint32_t i;
  if (!table)
    return 1;
  for (i = 0; i < table->size; i++)
    {
      struct hashedpair *hp;
      if (table->table[i].key)
        for (hp = &table->table[i]; hp; hp = hp->next)
          {
            uint32_t hash = table->hf(hp->key) % table->size;
            if (i != hash)
              {
                log("core", LOG_PRIORITY_WARN, N_("Sanity check failed on hashtable at %p: hash mismatch on key %s"),
                    (void*)table, hp->key);
                return 0;
              }
          }
    }
  return 1;
}
