/************************************************************************
 *   This is part of Dancer, a text relay daemon
 *   Copyright (C) 2002 Andrew Suffield <asuffield@users.sourceforge.net>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

#include <dancerd/config.h>
#include <dancerd/log.h>
#include <dancerd/llist.h>
#include <dancerd/xmalloc.h>
#include <dancerd/xstring.h>
#include <dancerd/ctor.h>
#include <dancerd/gettext.h>

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>
#include <stdarg.h>

struct logger
{
  char *name;
  log_writer_func *writer;
  log_filter_func *filter;
  log_reopen_func *reopen;
  log_close_func *close;
  void *data;
};

struct llist *loggers;

GLOBAL_CONSTRUCTOR(main_log, NULL)
{
  loggers = new_llist();
}

struct logger *
register_logger(const char *name, log_writer_func *lwriter, log_filter_func *lfilter,
		log_reopen_func *lreopen, log_close_func *lclose, void *data)
{
  struct logger *new_logger;
  if (!name || !lwriter || !lfilter || !lreopen || !lclose)
    return NULL;
  new_logger = xmalloc(sizeof(struct logger));
  new_logger->name = xstrdup(name);
  new_logger->writer = lwriter;
  new_logger->filter = lfilter;
  new_logger->reopen = lreopen;
  new_logger->close = lclose;
  new_logger->data = data;
  add_llist_head(loggers, new_logger);
  return new_logger;
}

void
close_logger(struct logger *l)
{
  if (!l)
    return;
  delete_llist_data(loggers, l);
  l->close(l->data);
  xfree(l->name);
  xfree(l);
}

static void
try_log(struct logger *logger, const char *facility, int priority, const char *format, va_list args)
{
  if (!(*logger->filter)(facility, priority, logger->data, format, args))
    return;

  if (!(*logger->writer)(facility, priority, logger->data, format, args))
    {
      int saved_errno = errno;
      (*logger->close)(logger->data);
      delete_llist_data(loggers, logger);
      if (saved_errno)
	log("core", LOG_PRIORITY_ERROR, N_("Logger %s failed to write: %s"), logger->name, strerror(saved_errno));
      else
	log("core", LOG_PRIORITY_ERROR, N_("Logger %s failed to write, closing (no error code)"), logger->name);
      xfree(logger->name);
      xfree(logger);
    }
}

void
log(const char *facility, int priority, const char *format, ...)
{
  va_list args;

  if (loggers)
    {
      struct llist_node *n;
      for (n = llist_head(loggers); n; n = llist_node_next(n))
        {
          va_start(args, format);
          try_log(llist_node_data(n), facility, priority, format, args);
          va_end(args);
        }
    }
  else
    {
      fprintf(stderr, _("log: called before initialisation has completed\n"));
      abort();
    }
}

const char *
priority_name(int priority)
{
  static const char *names[] =
    {
      "debug", "trace", "info", "warn", "error", "critical"
    };
  if (priority < 0)
    priority = 0;
  if (priority > LOG_PRIORITY_CRITICAL)
    priority = LOG_PRIORITY_CRITICAL;
  return names[priority];
}
