/************************************************************************
 *   This is part of Dancer, a text relay daemon
 *   Copyright (C) 2002 Andrew Suffield <asuffield@users.sourceforge.net>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

/** \file magic_printf.c
 * Add magic to printf format strings
 */

#include <dancerd/config.h>
#include <dancerd/magic_printf.h>
#include <dancerd/gettext.h>
#include <dancerd/xstring.h>
#include <dancerd/xmalloc.h>

#include <sys/types.h>
#include <stdio.h>
#include <stdlib.h>
#include <printf.h>
#include <locale.h>

/* The point of all this is that translation is delayed until the
 * point where vsnprintf is called, rather than happening when strings
 * are first passed in. This means we can say things like:
 *
 *  log(MYNAME, LOG_PRIORITY_WARN, "Couldn't do foo: %T", errstr);
 *
 * And the string will be translated individually for each logger,
 * rather than once at the top level - so different loggers can use
 * different locales.
 */

static printf_function translated_string;
static int
translated_string(FILE *stream, const struct printf_info *info,
                  const void *const *args)
{
  const char *string;
  char *saved_language = NULL;
  int len;

  if (info->alt)
    {
      const char *language = args[0];
      string = args[1];
      /* NULL language means to leave it alone. "" means native locale */
      if (language)
        {
          saved_language = xstrdup(setlocale(LC_MESSAGES, NULL));
          setlocale(LC_MESSAGES, language);
        }
    }
  else
    {
      /* No language means current locale */
      string = args[0];
    }

  /* Work around the gettext bug that returns the translation
   * descriptor for "" instead of another ""
   *
   * Also return a translated (null) for NULL
   */
  len = fprintf(stream, "%*s", info->left ? -info->width : info->width,
                string ? (string[0] ? gettext(string) : "") : _("(null)"));

  if (saved_language)
    {
      setlocale(LC_MESSAGES, saved_language);
      xfree(saved_language);
    }

  return len;
}

static printf_arginfo_function translated_string_arginfo;
static int
translated_string_arginfo(const struct printf_info *info, size_t n, int *argtypes)
{
  if (info->alt)
    {
      /* A locale will be specified */
      if (n > 0)
        argtypes[0] = PA_STRING;
      if (n > 1)
        argtypes[1] = PA_STRING;
      return 2;
    }
  else
    {
      /* Use the native locale */
      if (n > 0)
        argtypes[0] = PA_STRING;
      return 1;
    }
}

void
init_printf(void)
{
  if (register_printf_function('T', &translated_string, &translated_string_arginfo) == -1)
    {
      fprintf(stderr, _("init_printf: register_printf_function failed for %%T\n"));
      exit(1);
    }
}
