/************************************************************************
 *   This is part of Dancer, a text relay daemon
 *   Copyright (C) 2002 Andrew Suffield <asuffield@users.sourceforge.net>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

#define MODPREFIX core

#include <dancerd/config.h>
#include <dancerd/paths.h>
#include <dancerd/modules.h>
#include <dancerd/error_log.h>
#include <dancerd/log.h>
#include <dancerd/llist.h>
#include <dancerd/fd.h>
#include <dancerd/clock.h>
#include <dancerd/version.h>
#include <dancerd/pidfile.h>
#include <dancerd/ctor.h>
#include <dancerd/main.h>
#include <dancerd/gettext.h>
#include <dancerd/magic_printf.h>

#include <sys/types.h>
#include <getopt.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <locale.h>

#ifdef HAVE_GETOPT_LONG
static struct option const longopts[] =
  {
    {"root-directory", required_argument, NULL, 'd'},
    {"no-fork", no_argument, NULL, 'n'},
    {"fork", no_argument, NULL, 'N'},
    {"no-error-log", no_argument, NULL, 'e'},
    {"error-log", no_argument, NULL, 'E'},
    {"log-level", optional_argument, NULL, 'x'},
    {"module", required_argument, NULL, 'm'},
    {"module-path", required_argument, NULL, 'M'},
    {"version", no_argument, NULL, 'V'},
    {"copyright", no_argument, NULL, 'L'},
    {"license", no_argument, NULL, 'L'},
    {"help", no_argument, NULL, 'h'},
    {NULL, no_argument, NULL, 0}
  };
#endif

static int no_fork = 0;
static int log_priority = LOG_PRIORITY_WARN;
static char *root_directory = NULL;
static int use_error_log = 2;
/** \brief True if stdin and stdout are available for a module to use.
 *
 * When a module loads to use these, it should set this variable to be
 * false, and restore it on unload.
 */
int have_stdio = 1;

static inline void show_usage(void) noreturn_attribute;
static inline void show_version(void) noreturn_attribute;
static inline void show_license(void) noreturn_attribute;

static inline void
show_usage(void)
{
  fprintf(stderr, _("Usage: dancer [-LnNhveE] [-d directory] [-m module] [-M path]\n"));
#ifdef HAVE_GETOPT_LONG
  fprintf(stderr, _("       [--{no-,}fork] [--help] [--root-directory directory] [--module name]\n"));
  fprintf(stderr, _("       [--module-path name] [--{no-,}error-log] [--version] [--license]\n"));
#endif
  fprintf(stderr, _("Report bugs to <dancer-devel@linuxfr.org>\n"));
  exit(0);
}

static inline void
show_version(void)
{
  fprintf(stderr, "dancer %s (%s)\n", dancer_version, serial_number);
  exit(0);
}

static inline void
show_license(void)
{
  int i;
  for (i = 0; license[i]; i++)
    fprintf(stderr, "%s\n", license[i]);
  exit(0);
}

#ifdef HAVE_GETOPT_LONG
#define do_getopt(X) getopt_long(argc, argv, X, longopts, NULL)
#else
#define do_getopt(X) getopt(argc, argv, X)
#endif

static void
process_argv(int argc, char *argv[])
{
  int optc;

  while ((optc = do_getopt("?d:x:m:M:cnNeEVhL")) != -1)
    {
      switch (optc)
	{
	case 'd':
	  if (optarg)
	    root_directory = optarg;
	  break;
	case 'n':
	  no_fork = 1;
	  /* Turn this off when -n is set unless -e or -E occurs someplace */
	  if (use_error_log == 2)
	    use_error_log = 0;
	  break;
	case 'N':
	  no_fork = 0;
	  break;
	case 'e':
	  use_error_log = 0;
	  break;
	case 'E':
	  use_error_log = 1;
	  break;
	case 'x':
	  if (optarg)
	    log_priority = strtoul(optarg, NULL, 0);
	  else
	    if (log_priority > LOG_PRIORITY_DEBUG)
	      log_priority--;
	  break;
	case 'm':
	  if (optarg)
	    load_module(optarg, NULL);
	  else
	    show_usage();
	  break;
	case 'M':
	  if (optarg)
	    add_module_path(optarg);
	  else
	    show_usage();
	  break;
	case 'V':
	  show_version();
	case 'h':
        case '?':
	  show_usage();
	case 'L':
	  show_license();
	default:
	  abort();
	}
    }
}

int
main(int argc, char *argv[])
{
  int error_log_fd;

  textdomain(PACKAGE);
  /* We must run all of these in POSIX mode */
  setlocale(LC_CTYPE, "POSIX");
  setlocale(LC_COLLATE, "POSIX");
  setlocale(LC_MONETARY, "POSIX");
  setlocale(LC_NUMERIC, "POSIX");
  setlocale(LC_TIME, "POSIX");

  /* These can be localised */
  setlocale(LC_MESSAGES, "");

  do_global_construction();
  if (atexit(&do_global_destruction) != 0)
    {
      perror(_("Failed to register global destruction handler"));
      exit(1);
    }
  process_argv(argc, argv);

  set_root(root_directory);
  bindtextdomain(PACKAGE, localedir);
  textdomain(PACKAGE);

  /* Load this after locale handling is ready */
  init_printf();

  error_log_fd = open_error_log(argv[0], log_priority, use_error_log);
  /* Now we daemonise ourselves - fork, and close any stray fds from
   * the parent
   */
  if (!no_fork)
    {
      pid_t pid;

      switch ((pid = fork()))
	{
	case 0:
	  if (!write_pidfile(argv))
	    exit(1);
	  break;
	case -1:
	  fprintf(stderr, _("%s: fork() failed: %s\n"), argv[0], strerror(errno));
	  exit(1);
	default:
	  exit(0);
	}

      /* Only the daemon child reaches this point */
      if (freopen("/dev/null", "r", stdin) == NULL)
	{
	  fprintf(stderr, _("%s: couldn't reopen stdin to /dev/null: %s\n"), argv[0], strerror(errno));
	}

      if (freopen("/dev/null", "w", stdout) == NULL)
	{
	  fprintf(stderr, _("%s: couldn't reopen stdout to /dev/null: %s\n"), argv[0], strerror(errno));
	}

      have_stdio = 0;
    }
  /* Replace stderr as well */
  if (dup2(error_log_fd, 2) == -1)
    {
      fprintf(stderr, _("%s: couldn't reopen stderr to error pipe: %s\n"), argv[0], strerror(errno));
      exit(1);
    }
  close(error_log_fd);

  update_clock();
  add_module_path(module_dir);
  log("core", LOG_PRIORITY_INFO, _("Startup complete"));
  io_loop();
  log("core", LOG_PRIORITY_CRITICAL, _("Main io loop exited. Shutting down."));
  return 0;
}
