/************************************************************************
 *   This is part of Dancer, a text relay daemon
 *   Copyright (C) 2002 Andrew Suffield <asuffield@users.sourceforge.net>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

#include <dancerd/config.h>
#include <dancerd/modules.h>
#include <dancerd/llist.h>
#include <dancerd/log.h>
#include <dancerd/xmalloc.h>
#include <dancerd/xstring.h>
#include <dancerd/ctor.h>

#include <sys/types.h>
#include <ltdl.h>
#include <dlfcn.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <dirent.h>
#include <link.h>

#include "modules_internal.h"

struct module_handle
{
  lt_dlhandle ltdl;
  void *dl;
};

struct llist *search_paths;
const char *last_error = NULL;

GLOBAL_CONSTRUCTOR(main_modules_loader, NULL)
{
  lt_dlmalloc = &xmalloc;
  lt_dlfree = &xfree;
  search_paths = new_llist();

  if (lt_dlinit() != 0)
    {
      fprintf(stderr, "lt_dlinit() failed: %s\n", lt_dlerror());
      exit(1);
    }
}

GLOBAL_DESTRUCTOR(main_modules_loader, "main_modules", NULL)
{
  lt_dlexit();
  destroy_llist(search_paths, xfree);
}

static char *
find_object(const char *name, const char *version)
{
  struct llist_node *n, *next;
  char *filename = NULL;
  char *la_filename;
  char *so_filename;
  char *filename_prefix;
  if (version)
    {
      size_t len = strlen(name) + 1 + strlen(version) + 3 + 1;
      filename_prefix = NULL;
      la_filename = xmalloc(len);
      so_filename = xmalloc(len);
      snprintf(la_filename, len, "%s-%s.la", name, version);
      snprintf(so_filename, len, "%s-%s.so", name, version);
      log("core", LOG_PRIORITY_DEBUG, "Looking for %s or %s", la_filename, so_filename);
    }
  else
    {
      size_t len = strlen(name) + 1 + 1;
      filename_prefix = xmalloc(len);
      la_filename = xmalloc(len + 2);
      so_filename = xmalloc(len + 2);
      snprintf(filename_prefix, len, "%s-", name);
      snprintf(la_filename, len + 3, "%s.la", name);
      snprintf(so_filename, len + 3, "%s.so", name);
      log("core", LOG_PRIORITY_DEBUG, "Looking for %s or %s, or prefix %s", la_filename, so_filename, filename_prefix);
    }
  for (n = llist_head(search_paths); n; n = next)
    {
      char *path = llist_node_data(n);
      DIR *dir;
      struct dirent *ent;
      next = llist_node_next(n);

      dir = opendir(path);
      if (!dir)
        {
          log("core", LOG_PRIORITY_ERROR, "Failed to open %s: %s", path, strerror(errno));
          if ((errno == ENOTDIR) || (errno == ENOENT))
            {
              log("core", LOG_PRIORITY_ERROR, "Invalid module search path entry %s, removing", path);
              delete_llist_node(n);
            }
          continue;
        }
      while ((ent = readdir(dir)))
        {
          if ((strcmp(ent->d_name, la_filename) == 0) || (strcmp(ent->d_name, so_filename) == 0))
            {
              filename = xmalloc(strlen(path) + 1 + strlen(ent->d_name) + 1);
              sprintf(filename, "%s/%s", path, ent->d_name);
              break;
            }
          if (filename_prefix && (strncmp(ent->d_name, filename_prefix, strlen(filename_prefix)) == 0))
            {
              if ((strlen(ent->d_name) > 3) &&
                  (((ent->d_name[strlen(ent->d_name) - 3] == '.') &&
                    (ent->d_name[strlen(ent->d_name) - 2] == 'l') &&
                    (ent->d_name[strlen(ent->d_name) - 1] == 'a')) ||
                   ((ent->d_name[strlen(ent->d_name) - 3] == '.') &&
                    (ent->d_name[strlen(ent->d_name) - 2] == 's') &&
                    (ent->d_name[strlen(ent->d_name) - 1] == 'o'))
                   )
                  )
                {
                  filename = xmalloc(strlen(path) + 1 + strlen(ent->d_name) + 1);
                  sprintf(filename, "%s/%s", path, ent->d_name);
                  break;
                }
            }
        }
      closedir(dir);
      if (filename)
        break;
    }
  xfree(la_filename);
  xfree(so_filename);
  xfree(filename_prefix);
  return filename;
}

struct module_handle *
load_module_object(const char *name, const char *version)
{
  struct module_handle *handle = NULL;
  char *filename = find_object(name, version);
  if (!filename)
    {
      last_error = strerror(ENOENT);
      return NULL;
    }
  log("core", LOG_PRIORITY_DEBUG, "Trying to open %s as a module", filename);

  if ((strlen(filename) > 3) &&
      (filename[strlen(filename) - 3] == '.') &&
      (filename[strlen(filename) - 2] == 'l') &&
      (filename[strlen(filename) - 1] == 'a'))
    {
      lt_dlhandle ltdl = lt_dlopen(filename);
      xfree(filename);
      if (!ltdl)
        {
          last_error = lt_dlerror();
          return NULL;
        }

      handle = xmalloc(sizeof(struct module_handle));
      handle->ltdl = ltdl;
      handle->dl = NULL;
    }
  else
    {
      void *dl = dlopen(filename, RTLD_NOW);
      xfree(filename);
      if (!dl)
        {
          last_error = dlerror();
          return NULL;
        }

      handle = xmalloc(sizeof(struct module_handle));
      handle->ltdl = NULL;
      handle->dl = dl;
    }
  return handle;
}

void *
get_object_symbol(struct module_handle *handle, const char *sym)
{
  if (handle->ltdl)
    {
      void *p = lt_dlsym(handle->ltdl, sym);
      if (!p)
        last_error = lt_dlerror();
      return p;
    }
  if (handle->dl)
    {
      void *p = dlsym(handle->dl, sym);
      if (!p)
        last_error = dlerror();
      return p;
    }
  return NULL;
}

extern void unload_module_object(struct module_handle *handle)
{
  /* Just throw the errors away, there's nothing we can do with
   * them. The dancer module system currently operates on the
   * assumption that unloading a module at the system level cannot
   * sanely fail
   */
  if (handle->ltdl)
    {
      if (lt_dlclose(handle->ltdl) != 0)
        lt_dlerror();
    }
  if (handle->dl)
    {
      if (dlclose(handle->dl) != 0)
        dlerror();
    }
  xfree(handle);
}

const char *
get_loader_error(void)
{
  const char *e = last_error;
  last_error = NULL;
  return e;
}

void
add_module_path(const char *path)
{
  char *p = xstrdup(path);
  if (p[strlen(p) - 1] == '/')
    p[strlen(p) - 1] = '\0';
  add_llist_tail(search_paths, p);
  log("core", LOG_PRIORITY_INFO, "Adding %s to module search path", path);
}
