/************************************************************************
 *   This is part of Dancer, a text relay daemon
 *   Copyright (C) 2002 Andrew Suffield <asuffield@users.sourceforge.net>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

#include <dancerd/config.h>
#include <dancerd/modules.h>
#include <dancerd/log.h>
#include <dancerd/xmalloc.h>
#include <dancerd/xstring.h>
#include <dancerd/paths.h>
#include <dancerd/ctor.h>
#include <dancerd/version.h>

#include <assert.h>
#include <stdlib.h>

#include "modules_internal.h"

static struct llist *pending_events, *modules;

struct unload_handler
{
  module_unloaded_func *mu;
  void *data;
  struct module *interested;
  /* For your multiply-recursive pleasure: uhi_node is equal to the
   * node pointing to this in interested->unload_handler_interests
   */
  struct llist_node *uhi_node;
};

GLOBAL_CONSTRUCTOR(main_modules, NULL)
{
  pending_events = new_llist();
  modules = new_llist();
}

static int
module_match(const void *d1, const void *d2)
{
  const struct module *m = d1;
  const char *s = d2;
  return strcmp(m->name, s) == 0;
}

static inline struct module *
internal_find_module(const char *name)
{
  struct llist_node *n;
  n = find_llist_match(modules, &module_match, name);
  if (!n)
    return NULL;
  else
    return llist_node_data(n);
}

static void
free_unload_handler_interest(struct llist_node *n)
{
  xfree(llist_node_data(n));
  delete_llist_node(n);
}

static int
match_dependant(const void *d1, const void *d2)
{
  const struct module_dependant *dep = d1;
  const struct module *m = d2;
  return dep->from == m;
}

/* Unlink a module from all things that reference it, and free it,
 * calling unload handlers as appropriate
 */
static void
deregister_module(struct module *m)
{
  const struct module_dependency *module_deps;
  /* Remove it from the main list of modules */
  delete_llist_data(modules, m);

  /* Each module has a list of things it depends on ("dependencies")
   * and a list of things that depend on it ("dependants")
   */
  if (m->dependencies)
    /* For each dependency */
    for (module_deps = m->dependencies; module_deps->name; module_deps++)
      {
        /* Delete it from the list of dependants of the offending module */
        struct module *m2 = find_module(module_deps->name);
        struct llist_node *n;
        assert(m2);
        n = find_llist_match(m2->dependants, &match_dependant, m);
        if (n)
          {
            struct module_dependant *dep = llist_node_data(n);
            xfree(dep);
            delete_llist_node(n);
          }
      }

  /* We empty this list out by hand, rather than calling
   * destroy_llist.
   *
   * Otherwise, we get into problems when unload handlers call back
   * into the module system.
   *
   * This way the list is "safe" whenever an unload handler is called.
   */
  if (m->unload_handlers)
    {
      struct llist_node *n;
      while ((n = llist_head(m->unload_handlers)))
        {
          struct unload_handler *uh = llist_node_data(n);
          (*uh->mu)(uh->data);
          /* If the things we are interested in have not changed... */
          if ((n == llist_head(m->unload_handlers)) && (uh == llist_node_data(n)))
            {
              /* Free the lot */
              if (uh->uhi_node)
                delete_llist_node(uh->uhi_node);
              xfree(uh);
              delete_llist_node(n);
            }
        }
      free_llist(m->unload_handlers);
      m->unload_handlers = NULL;
    }
  destroy_llist(m->unload_handler_interests, free_unload_handler_interest);
  destroy_llist(m->dependants, xfree);

  unimport_all_symbols(m);
  unexport_all_symbols(m);

  xfree(m->version);
  xfree(m->name);

  /* ...and free the struct */
  xfree(m);
}

/* Attach the given list of dependencies to the module */
static void
add_deps(struct module *m, const struct module_dependency *module_deps)
{
  m->dependencies = module_deps;
  /* Now add to the dependants list of everything else */
  if (m->dependencies)
    for (module_deps = m->dependencies; module_deps->name; module_deps++)
      {
        struct module *m2 = find_module(module_deps->name);
        struct module_dependant *dep = xmalloc(sizeof(struct module_dependant));
        assert(m2);
        dep->from = m;
        dep->version = module_deps->version;
        add_llist_tail(m2->dependants, dep);
      }
}

static struct module *
create_module(const char *name, const char *version)
{
  struct module *m;

  if (!name)
    return NULL;
  m = xmalloc(sizeof(struct module));
  m->name = xstrdup(name);
  m->version = xstrdup(version);
  m->permanent = 0;
  m->imported_symbols = new_llist();
  m->exported_symbols = new_llist();
  m->unload_handlers = new_llist();
  m->unload_handler_interests = new_llist();
  m->dependants = new_llist();
  m->dependencies = NULL;
  return m;
}

static struct module *
register_module(const char *name, const char *version)
{
  struct module *m;

  if (!name)
    return NULL;

  m = create_module(name, version);
  add_llist_tail(modules, m);
  return m;
}

void
register_module_unload_handler(struct module *interested, struct module *m, module_unloaded_func *mu, void *data)
{
  struct unload_handler *uh;
  if (!mu || !m)
    return;

  uh = xmalloc(sizeof(struct unload_handler));
  uh->mu = mu;
  uh->data = data;
  uh->interested = interested;
  add_llist_head(m->unload_handlers, uh);
  if (interested)
    {
      add_llist_head(interested->unload_handler_interests, llist_head(m->unload_handlers));
      uh->uhi_node = llist_head(interested->unload_handler_interests);
    }
  else
    uh->uhi_node = NULL;
}

static int
match_module_unload_handler(const void *d1, const void *d2)
{
  const struct unload_handler *uh1 = d1, *uh2 = d2;
  return (uh1->mu == uh2->mu) && (uh1->data == uh2->data) && (uh1->interested == uh2->interested);
}

void
deregister_module_unload_handler(struct module *interested, struct module *m, module_unloaded_func *mu, void *data)
{
  struct llist_node *n;
  struct unload_handler uh;
  if (!m)
    return;
  uh.mu = mu;
  uh.data = data;
  uh.interested = interested;
  while ((n = find_llist_match(m->unload_handlers, &match_module_unload_handler, &uh)))
    {
      struct unload_handler *puh = llist_node_data(n);
      if (puh->uhi_node)
	delete_llist_node(puh->uhi_node);
      puh->uhi_node = NULL;
      delete_llist_node(n);
    }
}

struct module_event
{
  int load;
  char *name;
  char *version;
};

void
load_module(const char *name, const char *version)
{
  struct module_event *e;
  if (!name)
    return;
  e = xmalloc(sizeof(struct module_event));
  e->load = 1;
  e->name = xstrdup(name);
  e->version = xstrdup(version);
  add_llist_tail(pending_events, e);
}

void
unload_module(const char *name)
{
  struct module_event *e;
  if (!name)
    return;
  e = xmalloc(sizeof(struct module_event));
  e->load = 0;
  e->name = xstrdup(name);
  add_llist_tail(pending_events, e);
}

struct module *
find_module(const char *name)
{
  struct module *m = internal_find_module(name);
  return m;
}

static inline int
check_module(const char *name, const char *version, struct module_handle *handle)
{
  const char *module_name = get_object_symbol(handle, "_dancer__module_name");
  const char *module_version = get_object_symbol(handle, "_dancer__module_version");
  const char *module_dancer_version = get_object_symbol(handle, "_dancer__module_dancer_version");

  if (!module_name)
    {
      log("core", LOG_PRIORITY_ERROR, "Module %s does not contain a module name, aborting", name);
      return 0;
    }
  if (!module_version)
    {
      log("core", LOG_PRIORITY_ERROR, "Module %s does not contain a module version, aborting", name);
      return 0;
    }
  if (!module_dancer_version)
    {
      log("core", LOG_PRIORITY_ERROR, "Module %s does not contain a dancer version, aborting", name);
      return 0;
    }
  if (strcmp(module_dancer_version, dancer_version) != 0)
    {
      log("core", LOG_PRIORITY_ERROR, "Module %s compiled for dancer version %s, but this is %s, aborting",
          name, module_dancer_version, dancer_version);
      return 0;
    }
  if (strcmp(module_name, name) != 0)
    {
      log("core", LOG_PRIORITY_ERROR, "Loaded module %s, but its embedded name is %s, aborting", name, module_name);
      return 0;
    }
  if (version)
    if (strcmp(module_version, version) != 0)
      {
        log("core", LOG_PRIORITY_ERROR, "Loaded module %s version %s, but its embedded version is %s, aborting",
            name, version, module_version);
        return 0;
      }
  return 1;
}

static inline int
check_deps(struct module *old_module, const char *new_version, const struct module_dependency *module_deps)
{
  struct llist_node *n;

  if (old_module)
    for (n = llist_head(old_module->dependants); n; n = llist_node_next(n))
      {
        struct module_dependant *dep = llist_node_data(n);
        if (dep->version)
          if (xstrverscmp(new_version, dep->version) < 0)
            {
              log("core", LOG_PRIORITY_ERROR, "Module %s depends on %s version %s or later, cannot downgrade to %s",
                  dep->from->name, old_module->name, dep->version, new_version);
              return 0;
            }
      }

  if (module_deps)
    for (; module_deps->name; module_deps++)
      {
        struct module *m = find_module(module_deps->name);
        if (!m)
          {
            if (module_deps->version)
              log("core", LOG_PRIORITY_ERROR, "Dependency on %s version %s not satisfied",
                  module_deps->name, module_deps->version);
            else
              log("core", LOG_PRIORITY_ERROR, "Dependency on %s not satisfied", module_deps->name);
            return 0;
          }
        if (module_deps->version)
          if (xstrverscmp(m->version, module_deps->version) < 0)
            {
              log("core", LOG_PRIORITY_ERROR, "Dependency on %s version %s not satisfied (version %s is loaded)",
                  module_deps->name, module_deps->version, m->version);
              return 0;
            }
      }

  return 1;
}

/* handle will be unloaded by the caller if this returns 0 */
static inline int
do_init_module(const char *name, const char *version, struct module_handle *handle)
{
  module_init_func *init = (module_init_func *)get_object_symbol(handle, "_dancer__module_init");
  const char *module_version = get_object_symbol(handle, "_dancer__module_version");
  struct module **module_handle = get_object_symbol(handle, "_dancer__module_handle");
  const struct module_dependency *module_deps = get_object_symbol(handle, "_dancer__module_deps");
  const struct pending_symbol **module_symbols = get_object_symbol(handle, "_dancer__module_symbols");
  struct module *m;
  int x;

  /* First, check that we have the module we were looking for */
  if (!check_module(name, version, handle))
    return 0;

  /* Now, check that all our dependencies are satisfied */
  if (!check_deps(NULL, NULL, module_deps))
    {
      log("core", LOG_PRIORITY_ERROR, "Dependencies for %s not satisfied, aborting", name);
      return 0;
    }

  m = register_module(name, module_version);
  m->handle = handle;
  if (module_handle)
    *module_handle = m;

  /* Process the symbol table */
  if (module_symbols)
    {
      if (!process_symbol_table(NULL, m, module_symbols))
        {
          log("core", LOG_PRIORITY_ERROR, "Failed to process all symbols for %s, aborting", name);
          deregister_module(m);
          return 0;
        }
    }

  add_deps(m, module_deps);
  /* Call the init function */
  if (init)
    switch (x = (*init)(NULL))
      {
      case MODULE_LOAD_OK:
        return 1;
      case MODULE_LOAD_FAILED:
        /* Darn */
        log("core", LOG_PRIORITY_ERROR, "Module %s init function failed, aborting", name);
        deregister_module(m);
        return 0;
      default:
        log("core", LOG_PRIORITY_ERROR, "Weird return %d from module %s deinit function, unloading", x, name);
        deregister_module(m);
        return 0;
      }

  return 1;
}

/* new_handle must be unloaded by this function if it fails */
static inline int
do_swap_modules(struct module *old_module, const char *new_name, const char *new_version, struct module_handle *new_handle)
{
  module_init_func *init = (module_init_func *)get_object_symbol(new_handle, "_dancer__module_init");
  module_deinit_func *deinit = (module_deinit_func *)get_object_symbol(old_module->handle, "_dancer__module_deinit");
  const char *module_version = get_object_symbol(new_handle, "_dancer__module_version");
  struct module **module_handle = get_object_symbol(new_handle, "_dancer__module_handle");
  const struct module_dependency *module_deps = get_object_symbol(new_handle, "_dancer__module_deps");
  const struct pending_symbol **module_symbols = get_object_symbol(new_handle, "_dancer__module_symbols");
  struct module *new_module;
  int x;

  /* First, check the new module is the one we expected */
  if (!check_module(new_name, new_version, new_handle))
    {
      unload_module_object(new_handle);
      return 0;
    }

  /* Now, check that all the new module's dependencies are satisfied */
  if (!check_deps(old_module, new_version, module_deps))
    {
      log("core", LOG_PRIORITY_ERROR, "Dependencies for %s not satisfied, aborting", new_name);
      unload_module_object(new_handle);
      return 0;
    }

  new_module = create_module(new_name, module_version);
  new_module->handle = new_handle;

  if (module_handle)
    *module_handle = new_module;

  /* Process the new symbol table */
  if (module_symbols)
    {
      /* When we swap modules, this is a bit more complex
       *
       * First, we need to check that the new symbol table will
       * satisfy any required symbols that the old one exports
       */
      if (check_busy_symbols(old_module, 1, module_symbols))
        {
          log("core", LOG_PRIORITY_ERROR, "Cannot load new module %s version %s, it does not export some required symbols that the old version %s is providing", new_name, new_version, old_module->version);
          return 0;
        }

      /* OK, we can let process_symbol_table swap over the two modules */
      if (!process_symbol_table(old_module, new_module, module_symbols))
        {
          log("core", LOG_PRIORITY_ERROR, "Failed to process all symbols for %s, aborting", new_name);
          deregister_module(new_module);
          unload_module_object(new_handle);
          return 0;
        }
    }

  add_deps(new_module, module_deps);
  /* Call the deinit function */
  if (deinit)
    {
      /* We replace the module handle in the node with the new one, to preserve its ordering */
      struct llist_node *n;
      switch (x = (*deinit)(new_version, init))
        {
        case MODULE_LOAD_OK:
          {
            struct module_handle *old_handle = old_module->handle;
            log("core", LOG_PRIORITY_INFO, "Module %s version %s successfully replaced version %s",
                new_name, new_version, old_module->version);
            n = find_llist_data(modules, old_module);
            llist_node_data_lvalue(n) = new_module;
            new_module->dependants = old_module->dependants;
            old_module->dependants = NULL;
            deregister_module(old_module);
            unload_module_object(old_handle);
            return 1;
          }
        case MODULE_LOAD_FAILED:
          /* Darn */
          log("core", LOG_PRIORITY_ERROR, "Module %s (new version, %s) init function failed, aborting",
              new_name, new_version);
          deinit = (module_deinit_func *) get_object_symbol(new_handle, "_dancer__module_deinit");
          if (deinit)
            (*deinit)(NULL, NULL);
          deregister_module(new_module);
          unload_module_object(new_handle);
          return 0;
        case MODULE_UNLOAD_OK:
          {
            struct module_handle *old_handle = old_module->handle;
            log("core", LOG_PRIORITY_INFO, "Module %s (old version, %s) unloaded, completing load of new module",
                new_name, old_module->version);
            n = find_llist_data(modules, old_module);
            llist_node_data_lvalue(n) = new_module;
            new_module->dependants = old_module->dependants;
            old_module->dependants = NULL;
            deregister_module(old_module);
            unload_module_object(old_handle);
            /* Now call the new init function directly */
            if (init)
              switch ((*init)(NULL))
                {
                case MODULE_LOAD_OK:
                  log("core", LOG_PRIORITY_INFO, "Successfully loaded module %s version %s", new_name, new_version);
                  return 1;
                case MODULE_LOAD_FAILED:
                  /* Darn */
                  log("core", LOG_PRIORITY_ERROR, "Module %s init function failed, aborting", new_name);
                  deregister_module(new_module);
                  unload_module_object(new_handle);
                  return 0;
                }
          }
        case MODULE_UNLOAD_FAILED:
          log("core", LOG_PRIORITY_ERROR, "Module %s (old version, %s) failed to unload, aborting",
              new_name, old_module->version);
          deregister_module(new_module);
          unload_module_object(new_handle);
          return 0;
        case MODULE_LOAD_FUBAR:
          log("core", LOG_PRIORITY_CRITICAL, "Module %s unload went FUBAR, attempting to recover", new_name);
          deinit = (module_deinit_func *) get_object_symbol(new_handle, "_dancer__module_deinit");
          if (deinit)
            (*deinit)(NULL, NULL);
          deregister_module(new_module);
          unload_module_object(new_handle);
          init = (module_init_func *)get_object_symbol(old_module->handle, "_dancer__module_init");
          deinit = (module_deinit_func *)get_object_symbol(old_module->handle, "_dancer__module_deinit");
          if (deinit)
            {
              if ((*deinit)(old_module->version, init) == MODULE_LOAD_OK)
                {
                  log("core", LOG_PRIORITY_WARN, "Got old module back to sane state, continuing");
                  return 0;
                }
            }
          {
            struct module_handle *old_handle = old_module->handle;
            log("core", LOG_PRIORITY_WARN, "Failed to reach a sane state. Unloading old module and hoping for the best.");
            n = find_llist_data(modules, old_module);
            llist_node_data_lvalue(n) = new_module;
            new_module->dependants = old_module->dependants;
            old_module->dependants = NULL;
            deregister_module(old_module);
            unload_module_object(old_handle);
            return 1;
          }
        default:
          /* This probably means the new module failed to load, so unload it */
          log("core", LOG_PRIORITY_ERROR, "Weird return %d from module %s deinit function, unloading", x, new_name);
          deregister_module(new_module);
          unload_module_object(new_handle);
          break;
        }
    }
  else
    {
      struct llist_node *n;
      /* Hmm, no old deinit. Call the new init function instead */
      if (init)
        switch (x = (*init)(NULL))
          {
          case MODULE_LOAD_OK:
            {
              struct module_handle *old_handle = old_module->handle;
              log("core", LOG_PRIORITY_INFO, "Module %s version %s successfully replaced version %s",
                  new_name, new_version, old_module->version);
              n = find_llist_data(modules, old_module);
              llist_node_data_lvalue(n) = new_module;
              new_module->dependants = old_module->dependants;
              old_module->dependants = NULL;
              deregister_module(old_module);
              unload_module_object(old_handle);
              return 1;
            }
          case MODULE_LOAD_FAILED:
            /* Darn */
            log("core", LOG_PRIORITY_ERROR, "Module %s init function failed, aborting", new_name);
            deregister_module(new_module);
            unload_module_object(new_handle);
            return 0;
          default:
            log("core", LOG_PRIORITY_ERROR, "Weird return %d from module %s deinit function, unloading", x, new_name);
            deregister_module(new_module);
            unload_module_object(new_handle);
            break;
          }
    }

  return 1;
}

static inline int
do_load_module(const char *name, const char *version)
{
  struct module_handle *handle;
  struct module *m;

  assert(!version || *version);

  if (version)
    log("core", LOG_PRIORITY_INFO, "Trying to load module %s-%s", name, version);
  else
    log("core", LOG_PRIORITY_INFO, "Trying to load module %s", name);

  handle = load_module_object(name, version);

  if (!handle)
    {
      const char *err = get_loader_error();
      if (err)
        log("core", LOG_PRIORITY_ERROR, "Couldn't load module %s: %s", name, err);
      else
        log("core", LOG_PRIORITY_ERROR, "Couldn't load module %s", name);
      return 0;
    }

  m = internal_find_module(name);
  if (!m)
    {
      /* The module is not already loaded. Initialise it right away
       * and register it
       */
      if (!do_init_module(name, version, handle))
        {
          unload_module_object(handle);
          return 0;
        }

      if (version)
        log("core", LOG_PRIORITY_INFO, "Successfully loaded module %s version %s", name, version);
      else
        log("core", LOG_PRIORITY_INFO, "Successfully loaded module %s", name);
      return 1;
    }
  else
    {
      /* There's already a module of that name loaded. Can we do an
       * in-place reload?
       */
      const char *old_reload_interface = get_object_symbol(m->handle, "_dancer__module_reload_interface");
      const char *new_reload_interface = get_object_symbol(handle, "_dancer__module_reload_interface");
      if (m->permanent)
        {
          log("core", LOG_PRIORITY_ERROR, "Cannot load a new copy of module %s, it is permanently resident", name);
          return 0;
        }
      if (!old_reload_interface || !new_reload_interface || (strcmp(old_reload_interface, new_reload_interface) != 0))
        {
          if (!old_reload_interface)
            log("core", LOG_PRIORITY_ERROR, "Module %s version %s does not support in-place reloading, unload first",
                m->name, m->version);
          else if (!new_reload_interface)
            log("core", LOG_PRIORITY_ERROR, "Module %s version %s does not support in-place reloading, unload first",
                name, version);
          else
            log("core", LOG_PRIORITY_ERROR, "Module %s versions %s and %s have differing interface versions, unload first",
                name, m->version, version);
          return 0;
        }
      return do_swap_modules(m, name, version, handle);
    }
}

static inline void
do_unload_module(const char *name)
{
  struct module *m = internal_find_module(name);
  struct module_handle *handle;
  struct llist_node *n;
  module_deinit_func *deinit;

  if (!m)
    {
      log("core", LOG_PRIORITY_ERROR, "Not unloading module %s, since it isn't loaded", name);
      return;
    }
  if (m->permanent)
    {
      log("core", LOG_PRIORITY_ERROR, "Cannot unload module %s, it is permanently resident", name);
      return;
    }
  if (check_busy_symbols(m, 1, NULL))
    {
      log("core", LOG_PRIORITY_ERROR, "Cannot unload module %s, symbols are in use", name);
      return;
    }
  if ((n = llist_head(m->dependants)))
    {
      struct module_dependant *dep = llist_node_data(n);
      log("core", LOG_PRIORITY_ERROR, "Cannot unload module %s, %s depends on it", name, dep->from->name);
      return;
    }
  handle = m->handle;
  if ((deinit = (module_deinit_func *)get_object_symbol(handle, "_dancer__module_deinit")))
    {
      int x;
      switch ((x = (*deinit)(NULL, NULL)))
        {
        case MODULE_UNLOAD_OK:
        case MODULE_LOAD_OK:
          deregister_module(m);
          unload_module_object(handle);
          log("core", LOG_PRIORITY_INFO, "Unloaded module %s", name);
          break;
        case MODULE_UNLOAD_FAILED:
          log("core", LOG_PRIORITY_ERROR, "Unload of module %s failed", name);
          break;
        case MODULE_LOAD_FUBAR:
          log("core", LOG_PRIORITY_CRITICAL, "Module %s unload went FUBAR, unloading anyway", name);
          deregister_module(m);
          unload_module_object(handle);
          break;
        default:
          log("core", LOG_PRIORITY_ERROR, "Weird return %d from module %s deinit function, unloading anyway", x, name);
          deregister_module(m);
          unload_module_object(handle);
          break;
        }
    }
  else
    {
      deregister_module(m);
      unload_module_object(handle);
      log("core", LOG_PRIORITY_INFO, "Unloaded module %s", name);
    }
}

void
poll_module_events(void)
{
  while (llist_head(pending_events))
    {
      struct module_event *e = llist_node_data(llist_head(pending_events));
      if (e->load)
	do_load_module(e->name, e->version);
      else
	do_unload_module(e->name);
      xfree(e->name);
      xfree(e);
      delete_llist_node(llist_head(pending_events));
    }
}

GLOBAL_DESTRUCTOR(main_modules, NULL)
{
  while (llist_head(modules))
    {
      struct llist_node *n, *prev;
      int changed = 0;
      /* Work through this list backwards, since the last module added
       * is probably also the first one we can remove (modules don't
       * have to be unloaded in LIFO order, but LIFO should always
       * work)
       */
      for (n = llist_tail(modules); n; n = prev) 
        {
          struct module *m = llist_node_data(n);
          prev = llist_node_prev(n);
          if (m->handle)
            {
              /* Don't try to unload modules which things still depend on */
              if (!llist_head(m->dependants))
                {
                  /* Copy this because do_unload_module will destroy it */
                  char *name = xstrdup(m->name);
                  do_unload_module(name);
                  if (!find_module(name))
                    changed = 1;
                  xfree(name);
                }
            }
          else
            deregister_module(m);
        }
      /* If we tried them all and couldn't unload any, we're stuck,
       * dump core
       */
      if (!changed)
        abort();
    }
  destroy_llist(modules, &xfree);
  destroy_llist(pending_events, &xfree);
}
