/************************************************************************
 *   This is part of Dancer, a text relay daemon
 *   Copyright (C) 2002 Andrew Suffield <asuffield@users.sourceforge.net>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

#include <dancerd/config.h>
#include <dancerd/modules.h>
#include <dancerd/llist.h>
#include <dancerd/ctor.h>
#include <dancerd/xstring.h>
#include <dancerd/log.h>

#include "modules_internal.h"

struct symbol
{
  struct module *exporter;
  char *namespace, *name, *version;
  struct llist *imports;
  void *ptr;
};

struct imported_symbol
{
  struct symbol *sym;
  struct module *module;
  int required;
  void **sym_pointer;
};

static struct llist *symbols;

GLOBAL_CONSTRUCTOR(main_module_symbols, NULL)
{
  symbols = new_llist();
}

GLOBAL_DESTRUCTOR(main_module_symbols, NULL)
{
  destroy_llist(symbols, &xfree);
}

struct symbol_match_data
{
  const char *namespace, *name, *version;
};

static int
match_symbol(const void *d1, const void *d2)
{
  const struct symbol *s1 = d1;
  const struct symbol_match_data*s2 = d2;
  return (strcmp(s1->namespace, s2->namespace) == 0)
    && (strcmp(s1->name, s2->name) == 0)
    && (strcmp(s1->version, s2->version) == 0);
}

static struct symbol *
find_symbol(const char *namespace, const char *version, const char *name)
{
  struct symbol_match_data s;
  struct llist_node *n;
  s.namespace = namespace;
  s.name = name;
  s.version = version;
  n = find_llist_match(symbols, &match_symbol, &s);
  if (n)
    return llist_node_data(n);
  else
    return NULL;
}

static int
match_imported_symbol(const void *d1, const void *d2)
{
  const struct imported_symbol *is = d1;
  const struct module *m = d2;
  return is->module == m;
}

static struct imported_symbol *
find_imported_symbol(struct symbol *s, struct module *m)
{
  struct llist_node *n;
  n = find_llist_match(s->imports, &match_imported_symbol, m);
  if (n)
    return llist_node_data(n);
  else
    return NULL;
}

void
unimport_all_symbols(struct module *m)
{
  struct llist_node *n;

  while ((n = llist_head(m->imported_symbols)))
    {
      struct imported_symbol *is = llist_node_data(n);
      unimport_symbol(m, is->sym->namespace, is->sym->version, is->sym->name);
    }
}

void
unexport_all_symbols(struct module *m)
{
  struct llist_node *n;

  while ((n = llist_head(m->exported_symbols)))
    {
      struct symbol *s = llist_node_data(n);
      unexport_symbol(m, s->namespace, s->version, s->name);
    }

}

static void
unexport_imported_symbol(void *d1, void *d2 unused_attribute)
{
  struct imported_symbol *is = d1;
  *is->sym_pointer = NULL;
}

/** \brief Process the symbol table for a newly loaded module
 * \param old_module module being replaced, or NULL
 * \param new_module module being loaded
 * \param symbol_table the symbol table from module m
 * \return true if all symbols were processed successfully
 *
 * import_symbol and export_symbol will be called for every symbol in
 * the table, as appropriate.
 *
 * If the function returns false, then some symbols may have been
 * imported or exported anyway. These should probably be revoked right
 * away with unimport_all_symbols and unexport_all_symbols.
 *
 * If old_module is specified, then symbols will be unexported from it
 * as necessary.
 */
int
process_symbol_table(struct module *old_module, struct module *new_module, const struct pending_symbol **symbol_table)
{
  for (; *symbol_table; symbol_table++)
    {
      const struct pending_symbol *symbol = *symbol_table;
      if (symbol->import)
        {
          if (!import_symbol(new_module,
                             symbol->namespace, symbol->version, symbol->name,
                             symbol->sym_pointer, symbol->required))
            return 0;
        }
      else
        {
          struct symbol *s;
          if (old_module && (s = find_symbol(symbol->namespace, symbol->version, symbol->name)))
            {
              if (s->exporter == old_module)
                {
                  /* We can't call unexport_symbol, something might
                   * require this symbol.  We can, however, force it
                   * here; we're about to export a replacement.
                   */
                  foreach_llist(s->imports, &unexport_imported_symbol, NULL);
                  delete_llist_data(s->exporter->exported_symbols, s);

                  log("core", LOG_PRIORITY_DEBUG, "Unexport symbol %s.%s::%s from %s",
                      symbol->namespace, symbol->version, symbol->name, s->exporter->name);
                  s->exporter = NULL;
                  s->ptr = NULL;
                }
            }
          if (!export_symbol(new_module,
                             symbol->namespace, symbol->version, symbol->name,
                             symbol->ptr))
            {
              /* This cannot fail, we unexported it above. Bad things would happen if it did */
              if (old_module)
                abort();
              return 0;
            }
        }
    }
  return 1;
}

/** \brief Checks the symbol table to see if it exports the given symbol
 * \param symbol_table table to search
 * \param namespace namespace to check for
 * \param version version of namespace to check for
 * \param name symbol to check for
 * \return True if namespace.version::name is in symbol_table, false otherwise
 */
static inline int
exports_symbol(const struct pending_symbol **symbol_table, const char *namespace, const char *version, const char *name)
{
  for (; *symbol_table; symbol_table++)
    {
      const struct pending_symbol *symbol = *symbol_table;
      if (!symbol->import)
        {
          if ((strcmp(namespace, symbol->namespace) == 0) &&
              (strcmp(version, symbol->version) == 0) &&
              (strcmp(name, symbol->name) == 0))
            return 1;
        }
    }
  return 0;
}

/** \brief Check to see whether a given module has any required symbols
 * \param m module that is potentially being unloaded
 * \param log_warning true if warnings should be logged
 * \param symbol_table symbols from the new module being loaded
 *
 * If symbol_table is non-NULL, then it will be treated as a symbol
 * table intended to replace the current one used by module m.
 *
 * Module m will be checked to see whether it has exported any symbols
 * which are required by other modules, and which are not also
 * exported by symbol_table.
 *
 * symbol_table will be checked to see whether it exports any symbols
 * that are already exported by any module other than m (ie, if it
 * would conflict with any currently loaded modules), and whether it
 * imports any symbols that are not present, or that are exported by
 * m.
 *
 * If any of the above conditions are found, the function will return
 * 1. Warnings will be emitted for those it finds, if log_warning is
 * true.
 */
int
check_busy_symbols(struct module *m, int log_warning, const struct pending_symbol **symbol_table)
{
  struct llist_node *n, *n2;
  int busy = 0;

  for (n2 = llist_head(m->exported_symbols); n2; n2 = llist_node_next(n2))
    {
      struct symbol *s = llist_node_data(n2);
      for (n = llist_head(s->imports); n; n = llist_node_next(n))
        {
          struct imported_symbol *is = llist_node_data(n);
          if (is->required && (!symbol_table || !exports_symbol(symbol_table, s->namespace, s->version, s->name)))
            {
              if (log_warning)
                log("core", LOG_PRIORITY_WARN, "Symbol %s.%s::%s in module %s required by module %s",
                    s->namespace, s->version, s->name, m->name, is->module->name);
              busy = 1;
            }
        }
    }

  if (symbol_table)
    /* Now check imports in the new table */
    for (; *symbol_table; symbol_table++)
      {
        const struct pending_symbol *symbol = *symbol_table;
        struct symbol *s;
        s = find_symbol(symbol->namespace, symbol->version, symbol->name);
        if (!symbol->import)
          {
            if (s && s->exporter && (s->exporter != m))
              {
                if (log_warning)
                  log("core", LOG_PRIORITY_WARN, "Symbol %s.%s::%s in new module already exported by module %s",
                      s->namespace, s->version, s->name, s->exporter->name);
                busy = 1;
              }
          }
        else
          {
            if (!s || !s->exporter)
              {
                if (log_warning)
                  log("core", LOG_PRIORITY_WARN, "Symbol %s.%s::%s in new module is not exported anywhere",
                      symbol->namespace, symbol->version, symbol->name);
                busy = 1;
              }
            else if (s->exporter == m)
              {
                if (log_warning)
                  log("core", LOG_PRIORITY_WARN, "Symbol %s.%s::%s in new module is exported by current module %s",
                      s->namespace, s->version, s->name, s->exporter->name);
                busy = 1;
              }
          }
      }

  return busy;
}

static void
export_imported_symbol(void *d1, void *d2)
{
  const struct imported_symbol *is = d1;
  const struct symbol *s = d2;
  *is->sym_pointer = s->ptr;
}

int
export_symbol(struct module *m, const char *namespace, const char *version,
              const char *name, void *ptr)
{
  struct symbol *s;
  if (!m || !namespace || !version || !name || !ptr)
    {
      log("core", LOG_PRIORITY_WARN, "Cannot export symbol %s.%s::%s from %s: invalid argument",
	  namespace, version, name, m->name);
      return 0;
    }

  if ((s = find_symbol(namespace, version, name)))
    {
      if (s->exporter)
        {
          log("core", LOG_PRIORITY_WARN, "Symbol %s.%s::%s is already exported from %s",
              namespace, version, name, s->exporter->name);
          return 0;
        }
    }
  else
    {
      s = xmalloc(sizeof(struct symbol));
      s->namespace = xstrdup(namespace);
      s->version = xstrdup(version);
      s->name = xstrdup(name);
      s->imports = new_llist();
      add_llist_head(symbols, s);
    }
  s->exporter = m;
  s->ptr = ptr;
  foreach_llist(s->imports, &export_imported_symbol, s);
  add_llist_head(m->exported_symbols, s);
  log("core", LOG_PRIORITY_DEBUG, "Export symbol %s.%s::%s from %s",
      namespace, version, name, m->name);
  return 1;
}

int
unexport_symbol(struct module *m, const char *namespace, const char *version, const char *name)
{
  struct symbol *s;
  struct llist_node *n;

  if (!namespace || !version || !name)
    return 0;

  if (!(s = find_symbol(namespace, version, name)))
    return 1;

  if (!s->exporter || (s->exporter != m))
    return 0;

  /* OK, we have to scan the list twice */
  /* First see if we can unexport at present */
  for (n = llist_head(s->imports); n; n = llist_node_next(n))
    {
      struct imported_symbol *is = llist_node_data(n);
      if (is->required)
	{
	  log("core", LOG_PRIORITY_WARN, "Cannot unexport symbol %s.%s::%s from module %s: required by module %s",
	      namespace, version, name, s->exporter->name, is->module->name);
	  return 0;
	}
    }
  /* Then do it */
  foreach_llist(s->imports, &unexport_imported_symbol, NULL);
  delete_llist_data(s->exporter->exported_symbols, s);

  s->exporter = NULL;
  s->ptr = NULL;

  log("core", LOG_PRIORITY_DEBUG, "Unexport symbol %s.%s::%s from %s",
      namespace, version, name, m->name);

  if (!llist_head(s->imports))
    {
      /* Nothing is using this symbol. Throw it away */
      delete_llist_data(symbols, s);
      xfree(s->version);
      xfree(s->namespace);
      xfree(s->name);
      xfree(s);
    }

  return 1;
}

int
import_symbol(struct module *m, const char *namespace, const char *version,
              const char *name, void **sym_pointer, int required)
{
  struct symbol *s;
  struct imported_symbol *is;

  if (!m || !namespace || !version || !name || !sym_pointer)
    {
      log("core", LOG_PRIORITY_WARN, "Cannot import symbol %s.%s::%s into %s: invalid argument",
	  namespace, version, name, m->name);
      return 0;
    }

  if (!(s = find_symbol(namespace, version, name)))
    {
      if (required)
	{
	  log("core", LOG_PRIORITY_WARN, "Cannot import required symbol %s.%s::%s into %s: symbol not found",
	      namespace, version, name, m->name);
	  return 0;
	}
      s = xmalloc(sizeof(struct symbol));
      s->exporter = NULL;
      s->namespace = xstrdup(namespace);
      s->version = xstrdup(version);
      s->name = xstrdup(name);
      s->imports = new_llist();
      s->ptr = NULL;
      add_llist_head(symbols, s);
    }
  else
    {
      if (find_imported_symbol(s, m))
	{
	  log("core", LOG_PRIORITY_WARN, "Cannot import symbol %s.%s::%s into %s: symbol already imported",
	      namespace, version, name, m->name);
	  return 0;
	}
      if (!s->exporter)
        {
	  log("core", LOG_PRIORITY_WARN, "Cannot import required symbol %s.%s::%s into %s: symbol not found",
	      namespace, version, name, m->name);
	  return 0;
        }
    }
  is = xmalloc(sizeof(struct imported_symbol));
  is->sym = s;
  is->module = m;
  is->required = required;
  is->sym_pointer = sym_pointer;
  *is->sym_pointer = s->ptr;
  add_llist_head(m->imported_symbols, is);
  add_llist_head(s->imports, is);
  log("core", LOG_PRIORITY_DEBUG, "Import symbol %s.%s::%s into %s",
      namespace, version, name, m->name);
  return 1;
}

void
unimport_symbol(struct module *m, const char *namespace, const char *version, const char *name)
{
  struct symbol *s;
  struct imported_symbol *is;

  if (!m || !namespace || !version || !name)
    return;

  if (!(s = find_symbol(namespace, version, name)))
    {
      log("core", LOG_PRIORITY_WARN, "Cannot unimport symbol %s.%s::%s from %s: symbol not found",
	  namespace, version, name, m->name);
      return;
    }

  if (!(is = find_imported_symbol(s, m)))
    {
      log("core", LOG_PRIORITY_WARN, "Cannot unimport symbol %s.%s::%s from %s: symbol not imported",
	  namespace, version, name, m->name);
      return;
    }

  delete_llist_data(s->imports, is);
  delete_llist_data(m->imported_symbols, is);
  *is->sym_pointer = NULL;
  xfree(is);

  log("core", LOG_PRIORITY_DEBUG, "Unimport symbol %s.%s::%s from %s",
      namespace, version, name, m->name);

  if (!s->exporter && !llist_head(s->imports))
    {
      /* This symbol no longer has anything referencing it. Free it */
      delete_llist_data(symbols, s);
      xfree(s->version);
      xfree(s->namespace);
      xfree(s->name);
      xfree(s);
    }
}
