/************************************************************************
 *   This is part of Dancer, a text relay daemon
 *   Copyright (C) 2002 Andrew Suffield <asuffield@users.sourceforge.net>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

#include <dancerd/config.h>
#include <dancerd/pidfile.h>
#include <dancerd/paths.h>
#include <dancerd/ctor.h>
#include <dancerd/gettext.h>

#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <signal.h>

/* This returns false if the pidfile does not point to an already-running ircd */

static int
check_pidfile(char *argv[])
{
  int fd = open(pidfile, O_RDONLY, 0);
  ssize_t len;
  char buf[64];
  pid_t pid;

  if (fd == -1)
    {
      if (argv)
	fprintf(stderr, _("%s: couldn't open %s for reading: %s\n"), argv[0], pidfile, strerror(errno));
      return 1;
    }

  /* Read the pid from the file */
  while (((len = read(fd, buf, 64)) == -1) && (errno == EINTR))
    ;
  if (len == -1)
    {
      if (argv)
	fprintf(stderr, _("%s: couldn't read from %s: %s\n"), argv[0], pidfile, strerror(errno));
      return 1;
    }
  buf[len] = '\0';
  close(fd);
  /* Convert to integer */
  pid = strtol(buf, NULL, 0);
  /* Catch these here so they don't screw up kill() */
  if (pid < 1)
    return 1;
  if (pid == getpid())
    /* That not only exists, it's us */
    return 2;
  if ((kill(pid, 0) == -1) && (errno == ESRCH))
    /* No such pid exists */
    return 0;
  else
    /* A process with that pid exists, better assume it's a dancer */
    return 1;
}

/* This one adds the current process to the pidfile... */

int write_pidfile(char *argv[])
{
  int fd = open(pidfile, O_CREAT | O_EXCL | O_WRONLY, S_IRUSR | S_IWUSR);

  /* If the file already exists... */
  if ((fd == -1) && (errno == EEXIST))
    {
      /* And it does not point to a running process... */
      if (!check_pidfile(argv))
	{
	  /* Then whack it and try again */
	  unlink(pidfile);
	  fd = open(pidfile, O_CREAT | O_EXCL | O_WRONLY, S_IRUSR | S_IWUSR);
	}
      else
	{
	  fprintf(stderr, _("%s: pidfile %s exists, is another copy already running?\n"), argv[0], pidfile);
	  return 0;
	}
    }

  if (fd == -1)
    {
      fprintf(stderr, _("%s: couldn't open %s for writing: %s\n"), argv[0], pidfile, strerror(errno));
      return 0;
    }
  else
    {
      char buf[16];
      snprintf(buf, 16, "%d\n", getpid());
      if (write(fd, buf, strlen(buf)) == -1)
	{
	  fprintf(stderr, _("%s: write to %s failed: %s\n"), argv[0], pidfile, strerror(errno));
	  return 0;
	}
      close(fd);
      return 1;
    }
}

/* ...and this one removes it */

void clear_pidfile(void)
{
  unlink(pidfile);
}

GLOBAL_DESTRUCTOR(main_pidfile, NULL)
{
  /* If the pidfile points to this process, whack it */
  if (check_pidfile(NULL) == 2)
    clear_pidfile();
}
