/************************************************************************
 *   This is part of Dancer, a text relay daemon
 *   Copyright (C) 2002 Andrew Suffield <asuffield@users.sourceforge.net>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

#include <dancerd/config.h>
#include <dancerd/llist.h>
#include <dancerd/log.h>
#include <dancerd/xmalloc.h>
#include <dancerd/signal_handler.h>
#include <dancerd/ctor.h>
#include <dancerd/event.h>
#include <dancerd/gettext.h>

#include <stdlib.h>
#include <signal.h>
#include <assert.h>

#ifndef NSIG
/* Hack!
 * This should be good enough, though
 */
#define NSIG 64
#endif

struct signal_handler
{
  signal_handler_func *sh;
  void *data;
};

static struct llist *signal_handlers[NSIG];

static void sig_exit_handler(int, void *x unused_attribute) noreturn_attribute;
static void
sig_exit_handler(int sig, void *x unused_attribute)
{
  assert((sig == SIGTERM) || (sig == SIGINT));
  do_global_destruction();
  exit(0);
}

static struct event *sighup_event;

static void
sig_hup_handler(int sig, void *x unused_attribute)
{
  assert(sig == SIGHUP);
  raise_event(sighup_event);
}

static void
call_signal_handler(void *h_data, void *sig_data)
{
  int *signum = sig_data;
  struct signal_handler *h = h_data;
  (*h->sh)(*signum, h->data);
}

static void signal_handler(int sig)
{
  foreach_llist(signal_handlers[sig], &call_signal_handler, &sig);
}

GLOBAL_CONSTRUCTOR(main_signal_handler, NULL)
{
  int i;
  sigset_t block_mask;
  for (i = 0; i < NSIG; i++)
    signal_handlers[i] = new_llist();

  /* Block all the signals, except the ones which we should not handle (fatal errors) */
  sigfillset(&block_mask);
  sigdelset(&block_mask, SIGSTOP);
  sigdelset(&block_mask, SIGKILL);
  sigdelset(&block_mask, SIGFPE);
  sigdelset(&block_mask, SIGSEGV);
  sigdelset(&block_mask, SIGILL);
  sigdelset(&block_mask, SIGBUS);
  sigdelset(&block_mask, SIGABRT);
  /* Leave these alone for debuggers, if they exist */
#ifdef SIGPROF
  sigdelset(&block_mask, SIGPROF);
#endif
#ifdef SIGTRAP
  sigdelset(&block_mask, SIGTRAP);
#endif
  sigprocmask(SIG_BLOCK, &block_mask, NULL);

  /* And register these ones to make sure we exit when we should */
  register_signal_handler(SIGTERM, &sig_exit_handler, NULL);
  register_signal_handler(SIGINT, &sig_exit_handler, NULL);

  /* Send out the "core.sighup" event on SIGHUP */
  sighup_event = event("core.sighup");
  register_signal_handler(SIGHUP, &sig_hup_handler, NULL);
}

void
poll_signals(void)
{
  sigset_t pending_signals;
  int i;
  sigpending(&pending_signals);
  for (i = 1; i < NSIG; i++)
    if (sigismember(&pending_signals, i))
      {
        struct sigaction sa;
        sigset_t block_mask;

        sa.sa_handler = &signal_handler;
        sigfillset(&sa.sa_mask);
        sa.sa_flags = SA_ONESHOT;
        sigaction(i, &sa, NULL);

        sigfillset(&block_mask);
        sigdelset(&block_mask, i);
        sigsuspend(&block_mask);
      }
}

void
register_signal_handler(int signum, signal_handler_func *sh, void *data)
{
  struct signal_handler *h;
  if (signum >= NSIG)
    {
      log("core", LOG_PRIORITY_CRITICAL,
	  N_("register_signal_handler: attempt to register signal %d which is larger than NSIG (%d)"),
	  signum, NSIG);
      exit(1);
    }
  h = xmalloc(sizeof(struct signal_handler));
  h->sh = sh;
  h->data = data;
  add_llist_head(signal_handlers[signum], h);
}

static int match_signal_handler(const void *h_data, const void *h_match_data)
{
  const struct signal_handler *h = h_data;
  const struct signal_handler *h_match = h_match_data;
  return (h->sh == h_match->sh) && (h->data == h_match->data);
}

void
deregister_signal_handler(int signum, signal_handler_func *sh, void *data)
{
  struct signal_handler h;
  struct llist_node *n;
  if (signum >= NSIG)
    return;
  h.sh = sh;
  h.data = data;
  n = find_llist_match(signal_handlers[signum], &match_signal_handler, &h);
  if (!n)
    return;
  xfree(llist_node_data(n));
  delete_llist_node(n);
}
