/************************************************************************
 *   This is part of Dancer, a text relay daemon
 *   Copyright (C) 2002 Andrew Suffield <asuffield@users.sourceforge.net>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

#define DANCER_CORE__XMALLOC_SOURCE__
#define _GNU_SOURCE

#include <dancerd/config.h>
#include <dancerd/xmalloc.h>
#include <dancerd/xstring.h>

#include <sys/types.h>
#include <string.h>
#include <ctype.h>
#include <stdarg.h>
#include <stdio.h>

void *
xstrndup(const char *s, size_t len)
{
  char *data;
  if (!s)
    return NULL;
  data = xmalloc(len + 1);
  strlcpy(data, s, len + 1);
  return data;
}

int
xasprintf(char **strp, const char *fmt, ...)
{
  va_list args;
  int ret;
  va_start(args, fmt);
  ret = vasprintf(strp, fmt, args);
  va_end(args);
  if (ret == -1)
    out_of_memory();
  return ret;
}

char *
strip_spaces(char *s)
{
  char *last_non_space = s, *s2;
  while (isspace(*s))
    s++;
  for (s2 = s; *s2; s2++)
    if (!isspace(*s2))
      last_non_space = s2;
  if (*last_non_space)
    last_non_space[1] = '\0';
  return s;
}

char *
strip_control(char *s)
{
  return s;
}

/* This is more or less the same as the algorithm used by dpkg, since
 * that seems to have worked pretty well
 *
 * aaa < bbb (lexicographic comparison)
 * a < aa (absence of a character sorts lowest)
 * 1 < 10 (numeric comparison)
 * a1 < a10 (composite)
 * a10 < aa1 (lexicographic by components)
 *
 * Basically, you start from the left, and break both strings into
 * components that consist entirely of letters or numbers. A component
 * of letters is compared lexicographically, a component of numbers is
 * compared numerically.
 *
 * Numeric comparisons are made in base 10 only.
 *
 * The two strings of components are then compared lexicographically.
 *
 * s1 < s2          xstrverscmp(s1, s2) < 0
 * s1 == s2         xstrverscmp(s1, s2) == 0
 * s1 > s2          xstrverscmp(s1, s2) > 0
 */
int
xstrverscmp(const char *s1, const char *s2)
{
  static const char empty[] = "";

  if (!s1)
    s1 = empty;
  if (!s2)
    s2 = empty;

  while (1)
    {
      char c1, c2;
      c1 = *s1;
      c2 = *s2;
      /* If one of them is NULL... */
      if (c1 == '\0')
        /* This is 0 if c2 is also null, and negative if c2 is non-null */
        return c1 - c2;

      if (isdigit(c1))
        {
          long int n1, n2;
          /* If c1 is a digit and c2 is not, then we have this case:
           * s1 == ...a1
           * s2 == ...aa1
           * Therefore s1 < s2
           */
          if (!isdigit(c2))
            return -1;
          /* Both are digits */
          /* Note that this updates s1 and s2 to point to the next non-digit */
          for (n1 = 0; isdigit(*s1); n1 = (n1 * 10) + (*s1++ - '0'));
          for (n2 = 0; isdigit(*s2); n2 = (n2 * 10) + (*s2++ - '0'));
          if (n1 != n2)
            return n1 - n2; /* >0 if n1 is greater than n2 */
        }
      /* If c1 is not a digit and c2 is, then we have this case:
       * s1 == ...aa1
       * s2 == ...a1
       * Therefore s1 > s2
       */
      else if (isdigit(c2))
        return 1;
      else
        {
          /* Neither c1 nor c2 are digits. Compare them. */
          if (c1 == c2)
            {
              /* Carry on */
              s1++;
              s2++;
              continue;
            }
          /* They differ; we know the answer now */
          return c1 - c2; /* >0 if c1 is larger than c2 */
        }
    }
}
