/************************************************************************
 *   This is part of Dancer, a text relay daemon
 *   Copyright (C) 2002 Andrew Suffield <asuffield@users.sourceforge.net>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

#define _GNU_SOURCE

#include <dancerd/config.h>
#include <dancerd/config/items.h>
#include <dancerd/modules.h>
#include <dancerd/auth/auth.h>
#include <dancerd/log.h>
#include <dancerd/xmalloc.h>
#include <dancerd/xstring.h>
#include <dancerd/hashtable.h>
#include <dancerd/gettext.h>
#include <dancerd/database/database.h>
#include <dancerd/clock.h>

#include <string.h>

#include "auth_internal.h"

SETUP_MODULE("auth", MODVERSION);

#include "symtab"

struct hashtable *method_hash = NULL;
struct database *credentials_db = NULL, *credentials_index = NULL,
  *authorisation_db = NULL, *authorisation_entity_index = NULL, *authorisation_object_index = NULL,
  *authorisation_object_priv_index = NULL,
  *entity_attribute_db = NULL, *entity_attribute_index = NULL,
  *object_attribute_db = NULL, *object_attribute_index = NULL;
static char invalid_arguments[] = "Invalid arguments";
static char invalid_method[] = "Invalid method name";
static char method_unloaded[] = "Authentication method unloaded";
static char operation_not_supported[] = "Operation not supported";
static char database_failure[] = "Database failure";
static char invalid_stored_value[] = "Value stored in database is not valid";

static inline void
remove_session_method(struct auth_handle *handle)
{
  handle->method = NULL;
}

static void
unload_method_module(void *d)
{
  struct auth_method *method = d;
  method->aih = NULL;
  method->ach = NULL;
  method->apa = NULL;
  method->ard = NULL;
  method->agd = NULL;
  destroy_llist(method->handles, &remove_session_method);
  hashtable_delete(method_hash, method->name);
  xfree(method->name);
  xfree(method);
}

/** \brief Register a new authentication method
 * \param name Name of the method
 * \param module Handle of the module that contains this method
 * \param aih Function to call to initialise a new handle with this method
 * \param ach Function to call to cancel an existing handle with this method
 * \param apa Function to call when peer_accept is called
 * \param ard Function to call when read_data is called
 * \param agd Function to call when get_data is called
 * \return Handle for the new auth method, or NULL on error
 *
 * The method will be unregistered when module is unloaded.
 */
DEFINE_SYMBOL(struct auth_method *, auth, register_method, (const char *name, struct module *module,
                                                            auth_init_handle *aih, auth_cancel_handle *ach,
                                                            auth_peer_accept *apa,
                                                            auth_read_data *ard, auth_get_data *agd,
                                                            auth_set_credentials *asc))
{
  struct auth_method *method;

  if (!name || !aih || !ach)
    return NULL;

  if ((method = hashtable_lookup(method_hash, name)))
    {
      log(MYNAME, LOG_PRIORITY_ERROR, N_("Attempt to re-register auth method %s"), name);
      return NULL;
    }

  method = xmalloc(sizeof(struct auth_method));
  method->name = xstrdup(name);
  method->module = module;
  method->handles = new_llist();
  method->aih = aih;
  method->ach = ach;
  method->apa = apa;
  method->ard = ard;
  method->agd = agd;
  method->asc = asc;

  hashtable_insert(method_hash, method->name, method);
  register_module_unload_handler(MYHANDLE, method->module, &unload_method_module, method);

  return method;
}

static inline void
cancel_session(struct auth_handle *handle)
{
  (*handle->method->ach)(handle, handle->data);
}

/** \brief Unregister an authentication method
 * \param method Handle for the method to unregister
 */
DEFINE_SYMBOL(void, auth, unregister_method, (struct auth_method *method))
{
  deregister_module_unload_handler(MYHANDLE, method->module, &unload_method_module, method);
  destroy_llist(method->handles, &cancel_session);
  hashtable_delete(method_hash, method->name);
  xfree(method->name);
  xfree(method);
}

/** \brief Create a new authentication handle
 * \param method Name of the method to use
 * \param name Name to authenticate as
 * \param err Return variable for error string
 * \return New authentication handle, or NULL if error
 *
 * A new authentication session is started, using the given method and
 * name. On success, *err is set to NULL. On error, *err is set to a
 * statically allocated error string that will last until the next
 * call on this auth handle.
 */
DEFINE_SYMBOL(struct auth_handle *, auth, new_handle, (const char *method_name, const char *entity, char **err))
{
  struct auth_method *method;
  struct auth_handle *handle;
  char *key_name;
  uint32_t key_len;
  void *credentials;
  uint32_t credentials_len;

  if (err)
    *err = NULL;

  if (!method_name || !entity)
    {
      if (err)
        *err = invalid_arguments;
      return NULL;
    }

  method = hashtable_lookup(method_hash, method_name);
  if (!method)
    {
      if (err)
        *err = invalid_method;
      return NULL;
    }

  key_len = xasprintf(&key_name, "%s\01%s\01", entity, method_name);
  switch(SYM(database, get_value)(credentials_db, NULL, key_name, key_len, &credentials, &credentials_len))
    {
    case 1:
      xfree(key_name);
      break;
    case 0:
      xfree(key_name);
      credentials = NULL;
      credentials_len = 0;
      break;
    default:
      if (err)
        *err = database_failure;
      return 0;
    }

  handle = xmalloc(sizeof(struct auth_handle));
  handle->entity = xstrdup(entity);
  handle->method = method;
  handle->data = (*method->aih)(handle, entity, credentials, credentials_len);
  add_llist_tail(method->handles, handle);
  handle->handles_node = llist_tail(method->handles);

  return handle;
}

/** \brief Cancel an authentication session and free the handle
 * \param handle Handle to cancel and free
 *
 * The authentication session is immediately terminated, and all
 * resources are freed
 */
DEFINE_SYMBOL(void, auth, cancel_handle, (struct auth_handle *handle))
{
  if (!handle)
    return;
  if (handle->method)
    (*handle->method->ach)(handle, handle->data);
  delete_llist_node(handle->handles_node);
  xfree(handle->entity);
  xfree(handle);
}

/** \brief Signals acceptance of the session by the peer
 * \param handle Handle being accepted
 * \return Result code
 *
 * Calling this function indicates that the peer has accepted
 * authentication. The semantics of this operation are defined by the
 * method. Not all methods require acceptance by the peer.
 *
 * The return value is one of:
 *  AUTH_ACCEPT
 *  AUTH_CANCEL
 *  AUTH_CONTINUE
 *
 * If the return value is ACCEPT or CANCEL, then the handle is freed
 * before the function returns.
 *
 * If the return value is CONTINUE, then the caller should probably
 * probe for pending data right away.
 *
 * If the return value is CANCEL, *err will point to an error string
 * on return, otherwise it will point to NULL.
 */
DEFINE_SYMBOL(int, auth, peer_accept, (struct auth_handle *handle, char **err))
{
  if (err)
    *err = NULL;
  if (!handle)
    {
      if (err)
        *err = invalid_arguments;
      return AUTH_CANCEL;
    }
  if (handle->method)
    {
      int ret;
      if (handle->method->apa)
        switch (ret = (*handle->method->apa)(handle, handle->data, err))
          {
          case AUTH_ACCEPT:
          case AUTH_CANCEL:
            delete_llist_node(handle->handles_node);
            xfree(handle->entity);
            xfree(handle);
            return ret;
          case AUTH_CONTINUE:
            return AUTH_CONTINUE;
          default:
            /* Cancel if the return is weird */
            log(MYNAME, LOG_PRIORITY_ERROR, N_("Got invalid return %d from auth method %s in %s"),
                ret, handle->method->name, "peer_accept");
            (*handle->method->ach)(handle, handle->data);
            delete_llist_node(handle->handles_node);
            xfree(handle->entity);
            xfree(handle);
            return AUTH_CANCEL;
          }
      else
        return AUTH_CONTINUE;
    }
  else
    {
      delete_llist_node(handle->handles_node);
      xfree(handle->entity);
      xfree(handle);
      if (err)
        *err = method_unloaded;
      return AUTH_CANCEL;
    }
}

/** \brief Pass an array of data from the peer
 * \param handle Handle to read data on
 * \param argc Count of arguments in array
 * \param argv Array of strings from the peer
 * \param err Return variable for error string
 * \return Result code
 *
 * The return value is one of:
 *  AUTH_ACCEPT
 *  AUTH_CANCEL
 *  AUTH_CONTINUE
 *
 * If the return value is ACCEPT or CANCEL, then the handle is freed
 * before the function returns.
 *
 * If the return value is CONTINUE, then the caller should probably
 * probe for pending data right away.
 *
 * If the return value is CANCEL, *err will point to an error string
 * on return, otherwise it will point to NULL.
 */
DEFINE_SYMBOL(int, auth, read_data, (struct auth_handle *handle, int argc, char **argv, char **err))
{
  if (err)
    *err = NULL;
  if (!handle)
    {
      if (err)
        *err = invalid_arguments;
      return AUTH_CANCEL;
    }
  if (handle->method)
    {
      int ret;
      if (handle->method->apa)
        switch (ret = (*handle->method->ard)(handle, handle->data, argc, argv, err))
          {
          case AUTH_ACCEPT:
          case AUTH_CANCEL:
            delete_llist_node(handle->handles_node);
            xfree(handle->entity);
            xfree(handle);
            return ret;
          case AUTH_CONTINUE:
            return AUTH_CONTINUE;
          default:
            /* Cancel if the return is weird */
            log(MYNAME, LOG_PRIORITY_ERROR, N_("Got invalid return %d from auth method %s in %s"),
                ret, handle->method->name, "read_data");
            (*handle->method->ach)(handle, handle->data);
            delete_llist_node(handle->handles_node);
            xfree(handle->entity);
            xfree(handle);
            return AUTH_CANCEL;
          }
      else
        return AUTH_CONTINUE;
    }
  else
    {
      delete_llist_node(handle->handles_node);
      xfree(handle->entity);
      xfree(handle);
      if (err)
        *err = method_unloaded;
      return AUTH_CANCEL;
    }
}

/** \brief Get an array of data for the peer
 * \param handle Handle to get data from
 * \param argc Return variable for count of arguments
 * \param argv Return variable for array of arguments
 * \param err Return variable for error string
 * \return Result code
 *
 * The return value is one of:
 *  AUTH_ACCEPT
 *  AUTH_CANCEL
 *  AUTH_CONTINUE
 *
 * If the return value is ACCEPT or CANCEL, then the handle is freed
 * before the function returns.
 *
 * If the return value is CANCEL, *err will point to an error string
 * on return, otherwise it will point to NULL.
 *
 * The returned array will be valid until the next call to this
 * function on the same handle. The error string will be valid until
 * the next call to any function on this handle.
 */
DEFINE_SYMBOL(int, auth, get_data, (struct auth_handle *handle, int *argc, char ***argv, char **err))
{
  if (err)
    *err = NULL;
  if (!handle)
    {
      if (err)
        *err = invalid_arguments;
      return AUTH_CANCEL;
    }
  if (handle->method)
    {
      int ret;
      if (handle->method->apa)
        switch (ret = (*handle->method->agd)(handle, handle->data, argc, argv, err))
          {
          case AUTH_ACCEPT:
          case AUTH_CANCEL:
            delete_llist_node(handle->handles_node);
            xfree(handle->entity);
            xfree(handle);
            *argc = 0;
            *argv = NULL;
            return ret;
          case AUTH_CONTINUE:
            return AUTH_CONTINUE;
          default:
            /* Cancel if the return is weird */
            log(MYNAME, LOG_PRIORITY_ERROR, N_("Got invalid return %d from auth method %s in %s"),
                ret, handle->method->name, "get_data");
            (*handle->method->ach)(handle, handle->data);
            delete_llist_node(handle->handles_node);
            xfree(handle->entity);
            xfree(handle);
            return AUTH_CANCEL;
          }
      else
        return AUTH_CONTINUE;
    }
  else
    {
      delete_llist_node(handle->handles_node);
      xfree(handle->entity);
      xfree(handle);
      if (err)
        *err = method_unloaded;
      return AUTH_CANCEL;
    }
}

struct delete_keys_args
{
  struct database *database;
  char *entity;
  uint32_t entity_len;
  char **err;
};

static inline int
delete_keys__(struct database_transaction *dbtxn, struct database *database,
              char *entity, uint32_t entity_len, char **err)
{
  struct database_cursor *cursor;
  void *data;
  uint32_t datalen;
  if (!(cursor = SYM(database, get_cursor)(database, dbtxn, entity, entity_len, &data, &datalen)))
    {
      *err = database_failure;
      return 0;
    }
  while (data)
    {
      switch (SYM(database, cursor_delete)(cursor))
        {
        case 1:
          break;
        case -1:
          *err = database_failure;
          return 0;
        }
      switch (SYM(database, cursor_next_dup)(cursor, &data, &datalen))
        {
        case 1:
          continue;
        case 0:
          data = NULL;
          break;
        case -1:
          *err = database_failure;
          SYM(database, free_cursor)(cursor);
          return 0;
        }
    }
  SYM(database, free_cursor)(cursor);
  return 1;
}

static int
delete_keys_(struct database_transaction *dbtxn, void *data)
{
  struct delete_keys_args *args = data;
  return delete_keys__(dbtxn, args->database, args->entity, args->entity_len, args->err);
}

static inline int
delete_keys(struct database *database, char *entity, uint32_t entity_len, char **err)
{
  return SYM(database, run_transaction)(&delete_keys_,
                                        (struct delete_keys_args[]) {{
                                          .database = database,
                                             .entity = entity, 
                                             .entity_len = entity_len,
                                             .err = err}}, 1);
}

static inline int
drop_entity(char *entity, uint32_t entity_len, char **err)
{
  int ret = 1;
  char *lerr;

  struct database_transaction *dbtxn = SYM(database, begin_transaction)();
  if (!delete_keys__(dbtxn, credentials_index, entity, entity_len, &lerr))
    {
      *err = lerr;
      ret = 0;
    }
  if (!delete_keys__(dbtxn, authorisation_entity_index, entity, entity_len, &lerr))
    {
      *err = lerr;
      ret = 0;
    }
  if (!delete_keys__(dbtxn, entity_attribute_index, entity, entity_len, &lerr))
    {
      *err = lerr;
      ret = 0;
    }

  if (ret)
    return SYM(database, commit_transaction)(dbtxn);
  else
    SYM(database, abort_transaction)(dbtxn);

  return ret;
}

/** \brief Remove an entity from the database, and all references to it
 * \param entity Name of the entity to remove
 * \param err Return value for an error string
 * \return boolean true on success, false on error
 *
 * The named entity will be removed from the database, along with all
 * references and credentials relating to it. If it does not already
 * exist, the function will silently succeed.
 *
 * If the function fails, *err will be set to an error string,
 * otherwise it will be set to NULL.
 */
DEFINE_SYMBOL(int, auth, drop_entity, (const char *entity, char **err))
{
  char *key_name;
  uint32_t key_len;
  int ret;
  char *errnull;

  if (err)
    *err = NULL;
  if (!entity)
    {
      if (err)
        *err = invalid_arguments;
      return 0;
    }

  key_len = xasprintf(&key_name, "%s\01", entity);
  if (!err)
    err = &errnull;
  ret = drop_entity(key_name, key_len, err);
  xfree(key_name);
  return ret;
}

static inline int
drop_object(char *object, uint32_t object_len, char **err)
{
  int ret = 1;
  char *lerr;

  struct database_transaction *dbtxn = SYM(database, begin_transaction)();
  if (!delete_keys__(dbtxn, authorisation_object_index, object, object_len, &lerr))
    {
      *err = lerr;
      ret = 0;
    }
  if (!delete_keys__(dbtxn, object_attribute_index, object, object_len, &lerr))
    {
      *err = lerr;
      ret = 0;
    }

  if (ret)
    return SYM(database, commit_transaction)(dbtxn);
  else
    SYM(database, abort_transaction)(dbtxn);

  return ret;
}

/** \brief Remove an object from the database, and all references to it
 * \param object_type Type of the object to remove
 * \param object_name Name of the object to remove
 * \param err Return value for an error string
 * \return boolean true on success, false on error
 *
 * The named object will be removed from the database, along with all
 * references relating to it. If it does not already exist, the
 * function will silently succeed.
 *
 * If the function fails, *err will be set to an error string,
 * otherwise it will be set to NULL.
 */
DEFINE_SYMBOL(int, auth, drop_object, (const char *object_type, const char *object_name, char **err))
{
  char *key_name;
  uint32_t key_len;
  int ret;
  char *errnull;

  if (err)
    *err = NULL;
  if (!object_type || !object_name || strchr(object_type, ':'))
    {
      if (err)
        *err = invalid_arguments;
      return 0;
    }

  key_len = xasprintf(&key_name, "%s:%s\01", object_type, object_name);
  if (!err)
    err = &errnull;
  ret = drop_object(key_name, key_len, err);
  xfree(key_name);
  return ret;
}

/** \brief Set the authentication credentials for an entity via a given method
 * \param entity Name of the entity to operate on
 * \param method Name of the authentication method
 * \param argc Count of arguments
 * \param argv Array of arguments
 * \param err Return value for an error string
 * \return boolean true on success, false on error
 *
 * The precise semantics of this function are defined by the
 * authentication method used.
 *
 * If the function fails, *err will be set to an error string,
 * otherwise it will be set to NULL.
 */
DEFINE_SYMBOL(int, auth, set_credentials, (const char *entity, const char *method_name, int argc, char **argv, char **err))
{
  struct auth_method *method;
  if (err)
    *err = NULL;
  if (!entity || !method_name)
    {
      if (err)
        *err = invalid_arguments;
      return 0;
    }
  if (!(method = hashtable_lookup(method_hash, method_name)))
    {
      if (err)
        *err = invalid_method;
      return 0;
    }
  if (method->asc)
    {
      char *data;
      size_t datalen;
      if ((*method->asc)(entity, argc, argv, &data, &datalen, err))
        {
          char *key_name;
          uint32_t key_len = xasprintf(&key_name, "%s\01%s\01", entity, method_name);
          switch(SYM(database, put_value)(credentials_db, NULL, key_name, key_len, data, datalen))
            {
            case 1:
              xfree(key_name);
              return 1;
            default:
              xfree(key_name);
              if (err)
                *err = database_failure;
              return 0;
            }
        }
      else
        {
          return 0;
        }
    }
  else
    {
      if (err)
        *err = operation_not_supported;
      return 0;
    }
}

/** \brief Grant authorisation for an entity
 * \param entity Name of the entity
 * \param object Name of the object
 * \param priv Name of the privilege
 * \param err Return value for an error string
 * \return boolean true on success, false on error
 *
 * If the function fails, *err will be set to an error string,
 * otherwise it will be set to NULL.
 */
DEFINE_SYMBOL(int, auth, grant_authorisation, (const char *entity, const char *object_type, const char *object_name,
                                               const char *priv, int32_t value, char **err))
{
  char *key_name;
  uint32_t key_len;

  if (err)
    *err = NULL;
  if (!entity || !object_type || !object_name || !priv || strchr(object_type, ':'))
    {
      if (err)
        *err = invalid_arguments;
      return 0;
    }

  key_len = xasprintf(&key_name, "%s\01%s:%s\01%s\01", entity, object_type, object_name, priv);
  switch (SYM(database, put_value)(authorisation_db, NULL, key_name, key_len, &value, sizeof(value)))
    {
    case 1:
      xfree(key_name);
      return 1;
    default:
      xfree(key_name);
      if (err)
        *err = database_failure;
      return 0;
    }
}

/** \brief Revoke authorisation for an entity
 * \param entity Name of the entity
 * \param object Name of the object
 * \param priv Name of the privilege
 * \param err Return value for an error string
 * \return boolean true on success, false on error
 *
 * If the function fails, *err will be set to an error string,
 * otherwise it will be set to NULL.
 */
DEFINE_SYMBOL(int, auth, revoke_authorisation, (const char *entity, const char *object_type, const char *object_name,
                                                const char *priv, char **err))
{
  char *key_name;
  uint32_t key_len;

  if (err)
    *err = NULL;
  if (!entity || !object_type || !object_name || !priv || strchr(object_type, ':'))
    {
      if (err)
        *err = invalid_arguments;
      return 0;
    }

  key_len = xasprintf(&key_name, "%s\01%s:%s\01%s\01", entity, object_type, object_name, priv);
  switch (SYM(database, delete_value)(authorisation_db, NULL, key_name, key_len))
    {
    case 1:
      xfree(key_name);
      return 1;
    default:
      xfree(key_name);
      if (err)
        *err = database_failure;
      return 0;
    }
}

/** \brief Check authorisation for an entity
 * \param entity Name of the entity
 * \param object_type Type of the object
 * \param object_name Name of the object
 * \param priv Name of the privilege
 * \param need_value Minimum value for the function to return true
 * \param err Return value for an error string
 * \return boolean true if the entity is authorised, false if not
 *
 * If the function fails, *err will be set to an error string,
 * otherwise it will be set to NULL.
 */
DEFINE_SYMBOL(int, auth, check_authorisation, (const char *entity, const char *object_type, const char *object_name,
                                               const char *priv, int32_t need_value, char **err))
{
  char *key_name;
  uint32_t key_len;
  void *value;
  uint32_t value_len;

  if (err)
    *err = NULL;
  if (!entity || !object_type || !object_name || !priv || strchr(object_type, ':'))
    {
      if (err)
        *err = invalid_arguments;
      return 0;
    }

  key_len = xasprintf(&key_name, "%s\01%s:%s\01%s\01", entity, object_type, object_name, priv);
  switch (SYM(database, get_value)(authorisation_db, NULL, key_name, key_len, &value, &value_len))
    {
    case 1:
      xfree(key_name);
      if (value_len != sizeof(int32_t))
        {
          if (err)
            *err = invalid_stored_value;
          return 0;
        }
      return (*(int32_t *)value >= need_value);
    case 0:
      xfree(key_name);
      return 0;
    default:
      xfree(key_name);
      if (err)
        *err = database_failure;
      return 0;
    }
}

/** \brief Get authorisation value for an entity
 * \param entity Name of the entity
 * \param object_type Type of the object
 * \param object_name Name of the object
 * \param priv Name of the privilege
 * \param return_value Return value
 * \param err Return value for an error string
 * \return boolean true if the entity is authorised, false if not
 *
 * If the function fails, *err will be set to an error string,
 * otherwise it will be set to NULL.
 */
DEFINE_SYMBOL(int, auth, get_authorisation, (const char *entity, const char *object_type, const char *object_name,
                                             const char *priv, int32_t *return_value, char **err))
{
  char *key_name;
  uint32_t key_len;
  void *value;
  uint32_t value_len;

  *return_value = 0;

  if (err)
    *err = NULL;
  if (!entity || !object_type || !object_name || !priv || strchr(object_type, ':'))
    {
      if (err)
        *err = invalid_arguments;
      return 0;
    }

  key_len = xasprintf(&key_name, "%s\01%s:%s\01%s\01", entity, object_type, object_name, priv);
  switch (SYM(database, get_value)(authorisation_db, NULL, key_name, key_len, &value, &value_len))
    {
    case 1:
      xfree(key_name);
      if (value)
        {
          *return_value = *(int32_t *)value;
          return 1;
        }
      else
        return 0;
    case 0:
      xfree(key_name);
      return 0;
    default:
      xfree(key_name);
      if (err)
        *err = database_failure;
      return 0;
    }
}

/** \brief Set an attribute for an entity
 * \param entity Name of the entity
 * \param attribute Name of the attribute
 * \param data Pointer to the value to store for the attribute
 * \param len Length of the value
 * \param err Return value for an error string
 * \return boolean true on success, false on error
 *
 * The given value is stored, and will be returned when the attribute
 * it retrieved.
 *
 * \note
 * NULL values are "stored" as non-existant nodes; internally, setting
 * the data parameter to NULL will delete the attribute from the
 * database.
 */
DEFINE_SYMBOL(int, auth, set_entity_attribute, (const char *entity, const char *attribute,
                                                void *data, uint32_t len, char **err))
{
  char *key_name;
  uint32_t key_len;

  if (err)
    *err = NULL;
  if (!entity || !attribute)
    {
      if (err)
        *err = invalid_arguments;
      return 0;
    }

  key_len = xasprintf(&key_name, "%s\01%s\01", entity, attribute);
  if (data)
    {
      switch (SYM(database, put_value)(entity_attribute_db, NULL, key_name, key_len, data, len))
        {
        case 1:
          xfree(key_name);
          return 1;
        default:
          xfree(key_name);
          if (err)
            *err = database_failure;
          return 0;
        }
    }
  else
    {
      switch (SYM(database, delete_value)(entity_attribute_db, NULL, key_name, key_len))
        {
        case 1:
          xfree(key_name);
          return 1;
        default:
          xfree(key_name);
          if (err)
            *err = database_failure;
          return 0;
        }
    }
}

/** \brief Get an attribute for an entity
 * \param entity Name of the entity
 * \param attribute Name of the attribute
 * \param data Return value for the stored data
 * \param len Return value for the length of the stored data
 * \param err Return value for an error string
 * \return boolean true on success, false on error
 *
 * If the attribute is not present, it is returned as NULL/0 and the
 * function suceeds.
 */
DEFINE_SYMBOL(int, auth, get_entity_attribute, (const char *entity, const char *attribute,
                                                void **data, uint32_t *len, char **err))
{
  char *key_name;
  uint32_t key_len;

  if (err)
    *err = NULL;
  if (!entity || !attribute)
    {
      if (err)
        *err = invalid_arguments;
      return 0;
    }

  if (data)
    *data = NULL;
  if (len)
    *len = 0;

  key_len = xasprintf(&key_name, "%s\01%s\01", entity, attribute);
  switch (SYM(database, get_value)(entity_attribute_db, NULL, key_name, key_len, data, len))
    {
    case 1:
    case 0:
      xfree(key_name);
      return 1;
    default:
      xfree(key_name);
      if (err)
        *err = database_failure;
      return 0;
    }
}

/** \brief Set an attribute for an object
 * \param object_type Type of the object
 * \param object_name Name of the object
 * \param attribute Name of the attribute
 * \param data Pointer to the value to store for the attribute
 * \param len Length of the value
 * \param err Return value for an error string
 * \return boolean true on success, false on error
 *
 * The given value is stored, and will be returned when the attribute
 * it retrieved.
 *
 * \note
 * NULL values are "stored" as non-existant nodes; internally, setting
 * the data parameter to NULL will delete the attribute from the
 * database.
 */
DEFINE_SYMBOL(int, auth, set_object_attribute, (const char *object_type, const char *object_name, const char *attribute,
                                                void *data, uint32_t len, char **err))
{
  char *key_name;
  uint32_t key_len;

  if (err)
    *err = NULL;
  if (!object_type || !object_name || !attribute || strchr(object_type, ':'))
    {
      if (err)
        *err = invalid_arguments;
      return 0;
    }

  key_len = xasprintf(&key_name, "%s:%s\01%s\01", object_type, object_name, attribute);
  if (data)
    {
      switch (SYM(database, put_value)(object_attribute_db, NULL, key_name, key_len, data, len))
        {
        case 1:
          xfree(key_name);
          return 1;
        default:
          xfree(key_name);
          if (err)
            *err = database_failure;
          return 0;
        }
    }
  else
    {
      switch (SYM(database, delete_value)(object_attribute_db, NULL, key_name, key_len))
        {
        case 1:
          xfree(key_name);
          return 1;
        default:
          xfree(key_name);
          if (err)
            *err = database_failure;
          return 0;
        }
    }
}

/** \brief Get an attribute for an object
 * \param object_type Type of the object
 * \param object_name Name of the object
 * \param attribute Name of the attribute
 * \param data Return value for the stored data
 * \param len Return value for the length of the stored data
 * \param err Return value for an error string
 * \return boolean true on success, false on error
 *
 * If the attribute is not present, it is returned as NULL/0 and the
 * function suceeds.
 */
DEFINE_SYMBOL(int, auth, get_object_attribute, (const char *object_type, const char *object_name, const char *attribute,
                                                void **data, uint32_t *len, char **err))
{
  char *key_name;
  uint32_t key_len;

  if (err)
    *err = NULL;
  if (!object_type || !object_name || !attribute || strchr(object_type, ':'))
    {
      if (err)
        *err = invalid_arguments;
      return 0;
    }

  if (data)
    *data = NULL;
  if (len)
    *len = 0;

  key_len = xasprintf(&key_name, "%s:%s\01%s\01", object_type, object_name, attribute);
  switch (SYM(database, get_value)(object_attribute_db, NULL, key_name, key_len, data, len))
    {
    case 1:
    case 0:
      xfree(key_name);
      return 1;
    default:
      xfree(key_name);
      if (err)
        *err = database_failure;
      return 0;
    }
}

struct auth_iterator
{
  struct database_cursor *cursor;
};

/** \brief Get the next value in an iterated list
 * \param it The iterator
 * \param err Return value for error
 * \return Next value in the list
 *
 * Iterators work like this:
 * \code
 * {
 *   char *value, *err;
 *   struct auth_iterator *it;
 *   if (!(it = SYM(auth, get_entity_methods)("foo", &value, &err)))
 *     {
 *       // Report error
 *       return;
 *     }
 *   while ((value = SYM(auth, iterator_next)(it, &err)))
 *     {
 *       // Do something
 *     }
 *   if (err)
 *     {
 *       // Report error
 *     }
 *   SYM(auth, free_iterator)(it);
 * }
 * \endcode
 */
DEFINE_SYMBOL(const char *, auth, iterator_next, (struct auth_iterator *it, char **err))
{
  void *value;
  uint32_t len;
  if (err)
    *err = NULL;
  if (!it)
    {
      if (err)
        *err = invalid_arguments;
      return NULL;
    }

  switch (SYM(database, cursor_primary_next_dup)(it->cursor, &value, &len))
    {
    case 1:
      /* We stored one extra \01 at the end of the string; change it to \00 */
      if (value && len)
        ((char *)value)[len - 1] = '\0';
      return value;
    case 0:
      return NULL;
    default:
      if (err)
        *err = database_failure;
      return 0;
    }
}

/** \brief Free an iterator when it's no longer needed
 * \param it Iterator to free
 */
DEFINE_SYMBOL(void, auth, free_iterator, (struct auth_iterator *it))
{
  if (!it)
    return;
  SYM(database, free_cursor)(it->cursor);
  xfree(it);
}

static struct auth_iterator *
new_iterator(struct database *db, char *name, uint32_t name_len, char **value, char **err)
{
  struct auth_iterator *it;
  struct database_cursor *cursor;
  void *data;
  uint32_t len;

  if (err)
    *err = NULL;

  if (!db || !name)
    {
      if (err)
        *err = invalid_arguments;
      return NULL;
    }

  if (!(cursor = SYM(database, get_primary_cursor)(db, NULL, name, name_len, &data, &len)))
    {
      if (err)
        *err = database_failure;
      return NULL;      
    }

  /* We stored one extra \01 at the end of the string; change it to \00 */
  if (data && len)
    ((char *)data)[len - 1] = '\0';

  if (value)
    *value = data;

  it = xmalloc(sizeof(struct auth_iterator));
  it->cursor = cursor;
  return it;
}

/** \brief Get the methods that this entity has credentials for
 * \param name Name of the entity
 * \param value Return value
 * \param err Return error
 * \return Iterator to find the rest of the list
 *
 * Returns keys of the form "entity\01method"
 */
DEFINE_SYMBOL(struct auth_iterator *, auth, get_entity_methods, (const char *name, char **value, char **err))
{
  char *key_name;
  uint32_t key_len = xasprintf(&key_name, "%s\01", name);
  struct auth_iterator *it = new_iterator(credentials_index, key_name, key_len, value, err);
  xfree(key_name);
  return it;
}

/** \brief Get the attributes that this entity has
 * \param name Name of the entity
 * \param value Return value
 * \param err Return error
 * \return Iterator to find the rest of the list
 *
 * Returns keys of the form "entity\01attribute"
 */
DEFINE_SYMBOL(struct auth_iterator *, auth, get_entity_attributes, (const char *name, char **value, char **err))
{
  char *key_name;
  uint32_t key_len = xasprintf(&key_name, "%s\01", name);
  struct auth_iterator *it = new_iterator(entity_attribute_index, key_name, key_len, value, err);
  xfree(key_name);
  return it;
}

/** \brief Get the attributes that this object has
 * \param type Type of the entity
 * \param name Name of the entity
 * \param value Return value
 * \param err Return error
 * \return Iterator to find the rest of the list
 *
 * Returns keys of the form "type:name\01attribute"
 */
DEFINE_SYMBOL(struct auth_iterator *, auth, get_object_attributes, (const char *type, const char *name,
                                                                    char **value, char **err))
{
  char *key_name;
  uint32_t key_len = xasprintf(&key_name, "%s:%s\01", type, name);
  struct auth_iterator *it = new_iterator(object_attribute_index, key_name, key_len, value, err);
  xfree(key_name);
  return it;
}

/** \brief Get the authorisation keys associated with this entity
 * \param type Type of the entity
 * \param name Name of the entity
 * \param value Return value
 * \param err Return error
 * \return Iterator to find the rest of the list
 *
 * Returns keys of the form "entity\01object_type:object_name\01attribute"
 */
DEFINE_SYMBOL(struct auth_iterator *, auth, get_entity_authorisation, (const char *name, char **value, char **err))
{
  char *key_name;
  uint32_t key_len = xasprintf(&key_name, "%s\01", name);
  struct auth_iterator *it = new_iterator(authorisation_entity_index, key_name, key_len, value, err);
  xfree(key_name);
  return it;
}

/** \brief Get the authorisation keys associated with this object
 * \param type Type of the entity
 * \param name Name of the entity
 * \param value Return value
 * \param err Return error
 * \return Iterator to find the rest of the list
 *
 * Returns keys of the form "entity\01object_type:object_name\01attribute"
 */
DEFINE_SYMBOL(struct auth_iterator *, auth, get_object_authorisation, (const char *type, const char *name,
                                                                       char **value, char **err))
{
  char *key_name;
  uint32_t key_len = xasprintf(&key_name, "%s:%s\01", type, name);
  struct auth_iterator *it = new_iterator(authorisation_object_index, key_name, key_len, value, err);
  xfree(key_name);
  return it;
}

/** \brief Get the authorisation keys associated with this object and privilege
 * \param type Type of the entity
 * \param name Name of the entity
 * \param priv Name of the privilege
 * \param value Return value
 * \param err Return error
 * \return Iterator to find the rest of the list
 *
 * Returns keys of the form "entity\01object_type:object_name\01attribute"
 */
DEFINE_SYMBOL(struct auth_iterator *, auth, get_object_priv_authorisation, (const char *type, const char *name,
                                                                            const char *priv,
                                                                            char **value, char **err))
{
  char *key_name;
  uint32_t key_len = xasprintf(&key_name, "%s:%s\01%s\01", type, name, priv);
  struct auth_iterator *it = new_iterator(authorisation_object_priv_index, key_name, key_len, value, err);
  xfree(key_name);
  return it;
}

static int
credentials_key(void *key, uint32_t keylen,
                void *data unused_attribute, uint32_t datalen unused_attribute,
                void **newkey, uint32_t *newkeylen)
{
  void *end = memchr(key, '\01', keylen);
  if (!end)
    return ASSOCIATED_KEY_IGNORE;
  if (((char *)end + 1) <= ((char *)key + keylen))
    {
      *newkey = key;
      *newkeylen = ((char *)end - (char *)key) + 1;
      return ASSOCIATED_KEY_STORE;
    }
  else
    return ASSOCIATED_KEY_IGNORE;
}

static int
attribute_key(void *key, uint32_t keylen,
              void *data unused_attribute, uint32_t datalen unused_attribute,
              void **newkey, uint32_t *newkeylen)
{
  void *end = memchr(key, '\01', keylen);
  if (!end)
    return ASSOCIATED_KEY_IGNORE;
  if (((char *)end + 1) <= ((char *)key + keylen))
    {
      *newkey = key;
      *newkeylen = ((char *)end - (char *)key) + 1;
      return ASSOCIATED_KEY_STORE;
    }
  else
    return ASSOCIATED_KEY_IGNORE;
}

static int
authorisation_entity_key(void *key, uint32_t keylen,
                         void *data unused_attribute, uint32_t datalen unused_attribute,
                         void **newkey, uint32_t *newkeylen)
{
  void *end = memchr(key, '\01', keylen);
  if (((char *)end + 1) <= ((char *)key + keylen))
    {
      *newkey = key;
      *newkeylen = ((char *)end - (char *)key) + 1;
      return ASSOCIATED_KEY_STORE;
    }
  else
    return ASSOCIATED_KEY_IGNORE;
}

static int
authorisation_object_key(void *key, uint32_t keylen,
                         void *data unused_attribute, uint32_t datalen unused_attribute,
                         void **newkey, uint32_t *newkeylen)
{
  void *end = memchr(key, '\01', keylen);
  ptrdiff_t chunk2len;
  void *end2;
  if (!end)
    return ASSOCIATED_KEY_IGNORE;
  chunk2len = keylen - (((char *)end - (char *)key) + 1);
  end2 = memchr((char *)end + 1, '\01', chunk2len);
  if (!end2)
    return ASSOCIATED_KEY_IGNORE;
  *newkey = (char *)end + 1;
  *newkeylen = ((char *)end2 - ((char *)end + 1)) + 1;
  return ASSOCIATED_KEY_STORE;
}

static int
authorisation_object_priv_key(void *key, uint32_t keylen,
                              void *data unused_attribute, uint32_t datalen unused_attribute,
                              void **newkey, uint32_t *newkeylen)
{
  void *end = memchr(key, '\01', keylen);
  if (!end)
    return ASSOCIATED_KEY_IGNORE;
  *newkey = (char *)end + 1;
  *newkeylen = keylen - (((char *)end - (char *)key) + 1);
  return ASSOCIATED_KEY_STORE;
}

static int
open_dbs(struct database_transaction *dbtxn, void *data unused_attribute)
{
  if (!(credentials_db = SYM(database, open_database)("auth", "credentials", dbtxn, 0)))
    {
      log(MYNAME, LOG_PRIORITY_ERROR, N_("Failed to open credentials database"));
      return 0;
    }
  if (!(credentials_index = SYM(database, open_database)("auth", "credentials_index", dbtxn, 1)))
    {
      log(MYNAME, LOG_PRIORITY_ERROR, N_("Failed to open credentials index"));
      return 0;
    }
  if (SYM(database, associate_database)(credentials_db, credentials_index, dbtxn, &credentials_key) != 1)
    {
      log(MYNAME, LOG_PRIORITY_ERROR, N_("Failed to associate credentials index with database"));
      return 0;
    }
  if (!(authorisation_db = SYM(database, open_database)("auth", "authorisation", dbtxn, 0)))
    {
      log(MYNAME, LOG_PRIORITY_ERROR, N_("Failed to open authorisation database"));
      return 0;
    }
  if (!(authorisation_entity_index = SYM(database, open_database)("auth", "authorisation_entity_index",
                                                                  dbtxn, 1)))
    {
      log(MYNAME, LOG_PRIORITY_ERROR, N_("Failed to open authorisation entity index"));
      return 0;
    }
  if (!(authorisation_object_index = SYM(database, open_database)("auth", "authorisation_object_index",
                                                                  dbtxn, 1)))
    {
      log(MYNAME, LOG_PRIORITY_ERROR, N_("Failed to open authorisation object index"));
      return 0;
    }
  if (!(authorisation_object_priv_index = SYM(database, open_database)("auth", "authorisation_object_priv_index",
                                                                       dbtxn, 1)))
    {
      log(MYNAME, LOG_PRIORITY_ERROR, N_("Failed to open authorisation object privilege index"));
      return 0;
    }
  if (SYM(database, associate_database)(authorisation_db, authorisation_entity_index,
                                        dbtxn, &authorisation_entity_key) != 1)
    {
      log(MYNAME, LOG_PRIORITY_ERROR, N_("Failed to associate authorisation entity index with database"));
      return 0;
    }
  if (SYM(database, associate_database)(authorisation_db, authorisation_object_index,
                                        dbtxn, &authorisation_object_key) != 1)
    {
      log(MYNAME, LOG_PRIORITY_ERROR, N_("Failed to associate authorisation object index with database"));
      return 0;
    }
  if (SYM(database, associate_database)(authorisation_db, authorisation_object_priv_index,
                                        dbtxn, &authorisation_object_priv_key) != 1)
    {
      log(MYNAME, LOG_PRIORITY_ERROR, N_("Failed to associate authorisation object privilege index with database"));
      return 0;
    }
  if (!(entity_attribute_db = SYM(database, open_database)("auth", "entity_attribute", dbtxn, 0)))
    {
      log(MYNAME, LOG_PRIORITY_ERROR, N_("Failed to open entity attribute database"));
      return 0;
    }
  if (!(entity_attribute_index = SYM(database, open_database)("auth", "entity_attribute_index", dbtxn, 1)))
    {
      log(MYNAME, LOG_PRIORITY_ERROR, N_("Failed to open entity attribute index"));
      return 0;
    }
  if (!(object_attribute_db = SYM(database, open_database)("auth", "object_attribute", dbtxn, 0)))
    {
      log(MYNAME, LOG_PRIORITY_ERROR, N_("Failed to open object attribute database"));
      return 0;
    }
  if (!(object_attribute_index = SYM(database, open_database)("auth", "object_attribute_index", dbtxn, 1)))
    {
      log(MYNAME, LOG_PRIORITY_ERROR, N_("Failed to open object attribute index"));
      return 0;
    }
  if (SYM(database, associate_database)(entity_attribute_db, entity_attribute_index, dbtxn, &attribute_key) != 1)
    {
      log(MYNAME, LOG_PRIORITY_ERROR, N_("Failed to associate entity attribute index with database"));
      return 0;
    }
  if (SYM(database, associate_database)(object_attribute_db, object_attribute_index, dbtxn, &attribute_key) != 1)
    {
      log(MYNAME, LOG_PRIORITY_ERROR, N_("Failed to associate object attribute index with database"));
      return 0;
    }
  return 1;
}

static void
close_dbs(void)
{
  if (credentials_index)
    SYM(database, close_database)(credentials_index);
  if (credentials_db)
    SYM(database, close_database)(credentials_db);
  if (authorisation_entity_index)
    SYM(database, close_database)(authorisation_entity_index);
  if (authorisation_object_index)
    SYM(database, close_database)(authorisation_object_index);
  if (authorisation_object_priv_index)
    SYM(database, close_database)(authorisation_object_priv_index);
  if (authorisation_db)
    SYM(database, close_database)(authorisation_db);
  if (entity_attribute_index)
    SYM(database, close_database)(entity_attribute_index);
  if (entity_attribute_db)
    SYM(database, close_database)(entity_attribute_db);
  if (object_attribute_index)
    SYM(database, close_database)(object_attribute_index);
  if (object_attribute_db)
    SYM(database, close_database)(object_attribute_db);
  credentials_index = credentials_db =
    authorisation_entity_index = authorisation_object_index = authorisation_object_priv_index = authorisation_db = 
    entity_attribute_index = entity_attribute_db = object_attribute_index = object_attribute_db = NULL;
}

int
MODULE_INIT(void *data unused_attribute)
{
  method_hash = new_hashtable(&hash_ascii, 19);
  if (!SYM(database, run_transaction)(&open_dbs, NULL, 2))
    {
      close_dbs();
      return MODULE_LOAD_FAILED;
    }

  return MODULE_LOAD_OK;
}

int
MODULE_DEINIT(const char *new_version, module_init_func *init)
{
  int ret = MODULE_UNLOAD_OK;
  if (new_version)
    {
      switch ((*init)(NULL))
        {
        case MODULE_LOAD_OK:
          ret = MODULE_LOAD_OK;
          break;
        case MODULE_LOAD_FAILED:
          return MODULE_LOAD_FAILED;
        default:
          log(MYNAME, LOG_PRIORITY_ERROR, N_("Unknown return from new module version %s, aborting"), new_version);
          return MODULE_LOAD_FAILED;
        }
    }

  close_dbs();
  free_hashtable(method_hash);

  return ret;
}
