/************************************************************************
 *   This is part of Dancer, a text relay daemon
 *   Copyright (C) 2002 Andrew Suffield <asuffield@users.sourceforge.net>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

#ifdef AUTH_H
struct auth_handle;
struct auth_method;
#else
#define AUTH_H
#include <dancerd/config.h>
#include <dancerd/modules.h>

#include <inttypes.h>

/** \file auth/auth.h
 * \brief Authentication layer
 */

/** \defgroup auth Authentication and authoristaion layer
 *
 * The auth layer handles the actual process of authentication, and
 * provides an API to update and query the auth database.
 *
 * Values returned from these functions, either by direct return or by
 * filling in a passed pointer, unless otherwise specified, do not
 * persist beyond the next call to the auth module. Thusly, they
 * should be saved by the caller if they are needed for longer.
 *
 * Generally, the return value is an error code, and the data is
 * returned by modifying the target of a pointer parameter. Notable
 * exceptions to this rule include new_handle, register_method,
 * check_authorisation, and iterator_next.
 *
 * A handle is a reference to a current authentication session in
 * progress.
 *
 * An entity can authenticate.
 *
 * An object cannot authenticate.
 *
 * Authorisation for a given privilege is granted on a pair (entity,
 * object).
 *
 * An object is represented by a pair (type, name).
 *
 * A method is an authentication algorithm.
 *
 * A credential is an authentication secret.
 *
 * An attribute is arbitrary data associated with an entity or object.
 *
 * All keys are C strings. Authorisation values are 32-bit
 * integers. Credential and attribute values are arbitrary chunks of
 * memory.
 *
 * @{
 */

struct auth_handle;
struct auth_method;
struct auth_iterator;

#define AUTH_ACCEPT 1
#define AUTH_CANCEL 2
#define AUTH_CONTINUE 3

typedef void *auth_init_handle(struct auth_handle *, const char *, void *, uint32_t);
typedef void auth_cancel_handle(struct auth_handle *, void *);
typedef int auth_peer_accept(struct auth_handle *, void *, char **);
typedef int auth_read_data(struct auth_handle *, void *, int, char **, char **);
typedef int auth_get_data(struct auth_handle *, void *, int *, char ***, char **);
typedef int auth_set_credentials(const char *, int, char **, char **, uint32_t *, char **);

DEFINE_NAMESPACE(auth, 0);

/* These are used from the router, to authenticate nodes */
DECLARE_SYMBOL(struct auth_handle *, auth, new_handle, (const char *, const char *, char **));
DECLARE_SYMBOL(void, auth, cancel_handle, (struct auth_handle *));
DECLARE_SYMBOL(int, auth, peer_accept, (struct auth_handle *, char **));
DECLARE_SYMBOL(int, auth, read_data, (struct auth_handle *, int, char **, char **));
DECLARE_SYMBOL(int, auth, get_data, (struct auth_handle *, int *, char ***, char **));

/* These are used to update the database */
DECLARE_SYMBOL(int, auth, drop_entity, (const char *, char **));
DECLARE_SYMBOL(int, auth, drop_object, (const char *, const char *, char **));
DECLARE_SYMBOL(int, auth, set_credentials, (const char *, const char *, int, char **, char **));
DECLARE_SYMBOL(int, auth, grant_authorisation, (const char *, const char *, const char *, const char *, int32_t, char **));
DECLARE_SYMBOL(int, auth, revoke_authorisation, (const char *, const char *, const char *, const char *, char **));
DECLARE_SYMBOL(int, auth, set_entity_attribute, (const char *, const char *, void *, uint32_t, char **));
DECLARE_SYMBOL(int, auth, set_object_attribute, (const char *, const char *, const char *, void *, uint32_t, char **));

/* These are used to query the database */
DECLARE_SYMBOL(int, auth, get_entity_attribute, (const char *, const char *, void **, uint32_t *, char **));
DECLARE_SYMBOL(int, auth, get_object_attribute, (const char *, const char *, const char *, void **, uint32_t *, char **));
DECLARE_SYMBOL(int, auth, get_authorisation, (const char *, const char *, const char *, const char *, int32_t *, char **));
DECLARE_SYMBOL(int, auth, check_authorisation, (const char *, const char *, const char *, const char *, int32_t, char **));
DECLARE_SYMBOL(struct auth_iterator *, auth, get_entity_methods, (const char *, char **, char **));
DECLARE_SYMBOL(struct auth_iterator *, auth, get_entity_attributes, (const char *, char **, char **));
DECLARE_SYMBOL(struct auth_iterator *, auth, get_object_attributes, (const char *, const char *, char **, char **));
DECLARE_SYMBOL(struct auth_iterator *, auth, get_entity_authorisation, (const char *, char **, char **));
DECLARE_SYMBOL(struct auth_iterator *, auth, get_object_authorisation, (const char *, const char *, char **, char **));
DECLARE_SYMBOL(struct auth_iterator *, auth, get_object_priv_authorisation, (const char *, const char *, const char *,
                                                                             char **, char **));
DECLARE_SYMBOL(const char *, auth, iterator_next, (struct auth_iterator *, char **));
DECLARE_SYMBOL(void, auth, free_iterator, (struct auth_iterator *));

/* These are used by authentication methods */
DECLARE_SYMBOL(struct auth_method *, auth, register_method, (const char *, struct module *,
                                                             auth_init_handle *, auth_cancel_handle *,
                                                             auth_peer_accept *,
                                                             auth_read_data *, auth_get_data *,
                                                             auth_set_credentials *));
DECLARE_SYMBOL(void, auth, unregister_method, (struct auth_method *));

END_NAMESPACE(auth);

/** @}
 */

#endif
