/************************************************************************
 *   This is part of Dancer, a text relay daemon
 *   Copyright (C) 2002 Andrew Suffield <asuffield@users.sourceforge.net>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

#include <dancerd/config.h>
#include <dancerd/modules.h>
#include <dancerd/log.h>
#include <dancerd/auth/auth.h>
#include <dancerd/xmalloc.h>
#include <dancerd/xstring.h>

#include <stddef.h>

SETUP_MODULE("auth.password", MODVERSION);
RELOAD_INTERFACE(0);

#include "symtab"

static struct auth_method *method;

static auth_init_handle init_handle;
static void *
init_handle(struct auth_handle *handle unused_attribute, const char *authname unused_attribute,
            void *credentials, uint32_t len unused_attribute)
{
  char *data;
  if (!handle || !authname || !credentials)
    return NULL;
  /* Use the credentials as a flat string; they're all we
   * need. read_data will do a strcmp() on this and immediately either
   * suceed or fail.
   */
  data = xmalloc(len + 1);
  memcpy(data, credentials, len);
  data[len] = '\0';
  return data;
}

static auth_cancel_handle cancel_handle;
static void
cancel_handle(struct auth_handle *handle unused_attribute, void *data)
{
  if (data)
    xfree(data);
}

static auth_peer_accept peer_accept;
static int
peer_accept(struct auth_handle *handle unused_attribute, void *data unused_attribute, char **err unused_attribute)
{
  return AUTH_CONTINUE;
}

static auth_read_data read_data;
static int
read_data(struct auth_handle *handle unused_attribute, void *data, int argc, char **argv, char **err unused_attribute)
{
  if (!data)
    return AUTH_CANCEL;
  if (argc >= 1)
    {
      int ret;
      if (strcmp(data, argv[0]) == 0)
        ret = AUTH_ACCEPT;
      else
        {
          static char incorrect_password[] = "Incorrect password";
          *err = incorrect_password;
          ret = AUTH_CANCEL;
        }
      xfree(data);
      return ret;
    }
  else
    return AUTH_CONTINUE;
}

static auth_read_data read_data;
static int
get_data(struct auth_handle *handle unused_attribute, void *data unused_attribute,
         int *argc unused_attribute, char ***argv unused_attribute, char **err unused_attribute)
{
  return AUTH_CONTINUE;
}

static auth_set_credentials set_credentials;
static int
set_credentials(const char *entity unused_attribute, int argc, char **argv,
                char **data, uint32_t *len, char **err unused_attribute)
{
  if (argc >= 1)
    {
      *data = argv[0];
      *len = strlen(*data);
      return 1;
    }
  else
    return 0;
}

int
MODULE_INIT(void *data)
{
  method = SYM(auth, register_method)("password", MYHANDLE,
                                      &init_handle, &cancel_handle,
                                      &peer_accept, &read_data, &get_data,
                                      &set_credentials);
  if (!method)
    return MODULE_LOAD_FAILED;
  if (data)
    {
      log(MYNAME, LOG_PRIORITY_ERROR, "Passed unknown data when initialising, aborting");
      return MODULE_LOAD_FAILED;
    }
  else
    return MODULE_LOAD_OK;
}

int
MODULE_DEINIT(const char *new_version, module_init_func *init)
{
  int ret = MODULE_UNLOAD_OK;

  if (new_version)
    {
      switch ((*init)(NULL))
        {
        case MODULE_LOAD_OK:
          ret = MODULE_LOAD_OK;
          break;
        case MODULE_LOAD_FAILED:
          return MODULE_LOAD_FAILED;
        default:
          log(MYNAME, LOG_PRIORITY_ERROR, "Unknown return from new module version %s, aborting", new_version);
          return MODULE_LOAD_FAILED;
        }
    }

  SYM(auth, unregister_method)(method);

  return ret;
}
