/************************************************************************
 *   This is part of Dancer, a text relay daemon
 *   Copyright (C) 2002 Andrew Suffield <asuffield@users.sourceforge.net>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

#include <dancerd/config.h>
#include <dancerd/modules.h>
#include <dancerd/log.h>
#include <dancerd/clock.h>
#include <dancerd/paths.h>
#include <dancerd/xmalloc.h>
#include <dancerd/gettext.h>

#include <stddef.h>
#include <string.h>
#include <unistd.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>

SETUP_MODULE("basic.logfile", MODVERSION);
RELOAD_INTERFACE(0);

static char *logfile_name = NULL;

struct logfile_data
{
  int fd;
};

struct logger *basic_logger;

static int
logfile_writer(const char *facility, int priority, void *data, const char *format, va_list args)
{
  struct logfile_data *d = data;
  char buf[1024];
  if (d->fd < 0)
    return 0;
  snprintf(buf, 1024, "%s %s [%s]: ", get_real_time_str(), facility, priority_name(priority));
  if (write(d->fd, buf, strlen(buf)) == -1)
    return 0;
  vsnprintf(buf, 1024, gettext(format), args);
  if (write(d->fd, buf, strlen(buf)) == -1)
    return 0;
  if (write(d->fd, "\n", 1) == -1)
    return 0;
  else
    return 1;
}

static int
logfile_filter(const char *facility unused_attribute, int priority unused_attribute,
               void *data unused_attribute, const char *msg unused_attribute, va_list args unused_attribute)
{
  return 1;
}

static int
logfile_reopen(void *data)
{
  struct logfile_data *d = data;
  close(d->fd);
  d->fd = open(logfile_name, O_CREAT | O_WRONLY | O_APPEND, S_IRUSR | S_IWUSR);
  if (d->fd == -1)
    return 0;
  else
    return 1;
}

static void
logfile_close(void *data)
{
  struct logfile_data *d = data;
  close(d->fd);
  xfree(d);
}

int
MODULE_INIT(void *data unused_attribute)
{
  struct logfile_data *d;

  logfile_name = xmalloc(strlen(logdir) + strlen("basic.log") + 1);
  strcpy(logfile_name, logdir);
  strcat(logfile_name, "basic.log");

  d = xmalloc(sizeof(struct logfile_data));
  if ((d->fd = open(logfile_name, O_CREAT | O_WRONLY | O_APPEND, S_IRUSR | S_IWUSR)) == -1)
    {
      log(MYNAME, LOG_PRIORITY_ERROR, "Failed to open basic logfile %s: %s", logfile_name, strerror(errno));
      return MODULE_LOAD_FAILED;
    }

  if (!(basic_logger = register_logger("Basic error log", &logfile_writer, &logfile_filter,
				       &logfile_reopen, &logfile_close, d)))
    {
      close(d->fd);
      xfree(d);
      return MODULE_LOAD_FAILED;
    }

  log(MYNAME, LOG_PRIORITY_INFO, "basic.log opened");

  return MODULE_LOAD_OK;
}

int
MODULE_DEINIT(const char *new_version, module_init_func *init)
{
  if (new_version)
    {
      switch((*init)(NULL))
        {
        case MODULE_LOAD_OK:
          close_logger(basic_logger);
          basic_logger = NULL;
          xfree(logfile_name);
          logfile_name = NULL;
          return MODULE_LOAD_OK;
        case MODULE_LOAD_FAILED:
          return MODULE_LOAD_FAILED;
        default:
          log(MYNAME, LOG_PRIORITY_ERROR, "Unknown return from new module version %s, aborting", new_version);
          return MODULE_LOAD_FAILED;
        }
    }

  log(MYNAME, LOG_PRIORITY_INFO, "basic.log closed");
  close_logger(basic_logger);
  basic_logger = NULL;
  xfree(logfile_name);
  logfile_name = NULL;
  return MODULE_UNLOAD_OK;
}
