/************************************************************************
 *   This is part of Dancer, a text relay daemon
 *   Copyright (C) 2002 Andrew Suffield <asuffield@users.sourceforge.net>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

#include <dancerd/config.h>
#include <dancerd/modules.h>
#include <dancerd/log.h>
#include <dancerd/paths.h>
#include <dancerd/signal_handler.h>
#include <dancerd/xmalloc.h>
#include <dancerd/xstring.h>

#include <stddef.h>
#include <string.h>
#include <assert.h>
#include <errno.h>
#include <ctype.h>
#include <stdio.h>

SETUP_MODULE("basic.modconf", MODVERSION);
RELOAD_INTERFACE(0);

static void
process_modconf(int hup)
{
  char *filename;
  FILE *f;
  char buf[256];

  filename = xmalloc(strlen(confdir) + strlen("modules.conf") + 1);
  strcpy(filename, confdir);
  strcat(filename, "modules.conf");

  if (!(f = fopen(filename, "r")))
    {
      log(MYNAME, LOG_PRIORITY_ERROR, "Failed to open %s: %s", filename, strerror(errno));
      xfree(filename);
      return;
    }

  while (fgets(buf, 256, f))
    {
      char *c = strip_spaces(buf);
      if (*c == '#')
	continue;
      /* "load" gives a module to load on startup */
      if (strncasecmp(c, "load ", strlen("load ")) == 0)
	{
	  c += strlen("load ");
	  c = strip_spaces(c);
	  if (!hup)
	    {
	      log(MYNAME, LOG_PRIORITY_TRACE, "Loading module %s", c);
	      load_module(c, NULL);
	    }
	}
      /* "force-load" gives a module to load on startup or HUP */
      else if (strncasecmp(c, "force-load ", strlen("force-load ")) == 0)
	{
	  c += strlen("force-load ");
	  c = strip_spaces(c);
	  log(MYNAME, LOG_PRIORITY_TRACE, "Loading module %s", c);
	  load_module(c, NULL);
	}
      /* "unload" gives a module to unload on HUP */
      else if (strncasecmp(c, "unload ", strlen("unload ")) == 0)
	{
	  c += strlen("unload ");
	  c = strip_spaces(c);
	  if (hup)
	    {
	      log(MYNAME, LOG_PRIORITY_TRACE, "Unloading module %s", c);
	      unload_module(c);
	    }
	}
      else
	{
	  log(MYNAME, LOG_PRIORITY_WARN, "Unrecognised line in %s: %s", filename, buf);
	}
    }
  xfree(filename);
}

static void
sighup_handler(int sig, void *data unused_attribute)
{
  assert(sig == SIGHUP);
  process_modconf(1);
}

int
MODULE_INIT(void *data)
{
  process_modconf(data ? 1 : 0);
  register_signal_handler(SIGHUP, &sighup_handler, NULL);

  return 1;
}

int
MODULE_DEINIT(const char *new_version, module_init_func *init)
{
  if (new_version)
    {
      switch((*init)((void *)1))
        {
        case MODULE_LOAD_OK:
          deregister_signal_handler(SIGHUP, &sighup_handler, NULL);
          return MODULE_LOAD_OK;
        case MODULE_LOAD_FAILED:
          return MODULE_LOAD_FAILED;
        default:
          log(MYNAME, LOG_PRIORITY_ERROR, "Unknown return from new module version %s, aborting", new_version);
          return MODULE_LOAD_FAILED;
        }
    }

  deregister_signal_handler(SIGHUP, &sighup_handler, NULL);
  return MODULE_UNLOAD_OK;
}
