/************************************************************************
 *   This is part of Dancer, a text relay daemon
 *   Copyright (C) 2002 Andrew Suffield <asuffield@users.sourceforge.net>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

#include <dancerd/config.h>
#include <dancerd/modules.h>
#include <dancerd/log.h>
#include <dancerd/xmalloc.h>

#include <dancerd/database/database.h>
#include <dancerd/config/items.h>

#include <stddef.h>
#include <stdint.h>

SETUP_MODULE("config.database", MODVERSION);
RELOAD_INTERFACE(0);

IMPORTED_SYMBOL(database, open_database);
IMPORTED_SYMBOL(database, close_database);
IMPORTED_SYMBOL(database, get_value);
IMPORTED_SYMBOL(database, put_value);
IMPORTED_SYMBOL(database, delete_value);

#include "symtab"

static struct database *config_db = NULL;

DEFINE_SYMBOL(int, config, get_item, (char *key, struct config_item *item))
{
  void *d;
  char *data;
  uint32_t len;

  if (!key || !item)
    return 0;

  switch (SYM(database, get_value)(config_db, NULL, key, strlen(key), &d, &len))
    {
    case 0:
      item->type = citem_null;
      return 1;
    case -1:
      return 0;
    }

  if (len < 2)
    return 0;

  data = d;

  item->key = key;
  switch (data[0])
    {
    case 1:
      item->type = citem_bool;
      item->boolean = data[1];
      return 1;
    case 2:
      item->type = citem_string;
      item->str = data + 1;
      return 1;
    case 3:
      if (len < (sizeof(uint32_t) * 2))
	return 0;
      item->type = citem_uint32;
      item->uint32 = ((uint32_t *)data)[1];
      return 1;
    case 4:
      if (len < (sizeof(int32_t) * 2))
	return 0;
      item->type = citem_sint32;
      item->sint32 = ((int32_t *)data)[1];
      return 1;
    default:
      log(MYNAME, LOG_PRIORITY_WARN, "Unrecognised field type %c in config item %s", data[0], key);
      return 0;
    }
}

DEFINE_SYMBOL(int, config, put_item, (struct config_item *item))
{
  void *d;
  uint32_t len;
  int d_malloced = 0;
  int ret;

  if (!item)
    return 0;

  switch (item->type)
    {
    case citem_bool:
      {
	static char data[2];
	len = 2;
	data[0] = 1;
	data[1] = item->boolean;
	d = data;
      }
      break;
    case citem_string:
      {
	char *data;
	len = strlen(item->str) + 2;
	data = xmalloc(len);
	data[0] = 2;
	memcpy(data + 1, item->str, len - 1);
	d = data;
	d_malloced = 1;
      }
      break;
    case citem_uint32:
      {
	static char data[sizeof(uint32_t) * 2];
	len = sizeof(uint32_t) * 2;
	data[0] = 3;
	((uint32_t *)data)[1] = item->uint32;
	d = data;
      }
      break;
    case citem_sint32:
      {
	static char data[sizeof(int32_t) * 2];
	len = sizeof(int32_t) * 2;
	data[0] = 4;
	((int32_t *)data)[1] = item->sint32;
	d = data;
      }
      break;
    default:
      log(MYNAME, LOG_PRIORITY_WARN, "Unrecognised value for type of config item %s", item->key);
      return 0;
    }

  ret = SYM(database, put_value)(config_db, NULL, item->key, strlen(item->key), d, len);

  if (d_malloced)
    xfree(d);

  return (ret == 1);
}

DEFINE_SYMBOL(int, config, delete_item, (char *key))
{
  if (!key)
    return 0;

  return (SYM(database, delete_value)(config_db, NULL, key, strlen(key)) > -1);
}

int
MODULE_INIT(void *data)
{
  if (data)
    {
      config_db = data;
    }
  else
    {
      if (!(config_db = SYM(database, open_database)("config", "items", NULL, 0)))
        {
          log(MYNAME, LOG_PRIORITY_ERROR, "Failed to open config item database");
          return MODULE_LOAD_FAILED;
        }
    }

  return MODULE_LOAD_OK;
}

int
MODULE_DEINIT(const char *new_version, module_init_func *init)
{
  if (new_version)
    {
      switch ((*init)(config_db))
        {
        case MODULE_LOAD_OK:
          config_db = NULL;
          return MODULE_LOAD_OK;
        case MODULE_LOAD_FAILED:
          return MODULE_LOAD_FAILED;
        default:
          log(MYNAME, LOG_PRIORITY_ERROR, "Unknown return from new module version %s, aborting", new_version);
          return MODULE_LOAD_FAILED;
        }
    }

  SYM(database,close_database)(config_db);
  config_db = NULL;
  return MODULE_UNLOAD_OK;
}
