/************************************************************************
 *   This is part of Dancer, a text relay daemon
 *   Copyright (C) 2002 Andrew Suffield <asuffield@users.sourceforge.net>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

#include <dancerd/config.h>
#include <dancerd/modules.h>
#include <dancerd/log.h>
#include <dancerd/xmalloc.h>
#include <dancerd/paths.h>
#include "db_internal.h"
#include <dancerd/database/database.h>
#include <dancerd/xstring.h>

#include <stddef.h>
#include <string.h>
#include <db.h>

#if DB_VERSION_MAJOR != 4
#error Berkeley libdb version is not 4
#endif

SETUP_MODULE("database.db", MODVERSION);
RELOAD_INTERFACE(0);

#include "symtab"

static DB_ENV *dbenv = NULL;
static struct llist *databases = NULL;
static struct llist *transactions = NULL;

static void
db_errcall(const char *prefix, char *msg)
{
  log(MYNAME, LOG_PRIORITY_ERROR, "libdb error in %s: %s", prefix, msg);
}

void
set_errpfx(const char *prefix)
{
  if (dbenv)
    dbenv->set_errpfx(dbenv, prefix);
}

static void
close_dbenv(void)
{
  int err;

  if (!dbenv)
    return;

  set_errpfx("close_dbenv");

  if ((err = dbenv->close(dbenv, 0)) != 0)
    log(MYNAME, LOG_PRIORITY_ERROR, "Failed to cleanly close db environment: %s", db_strerror(err));

  dbenv = NULL;
}

static int
open_dbenv(void)
{
  int err;

  db_env_set_func_malloc(&xmalloc);
  db_env_set_func_realloc(&xrealloc);
  db_env_set_func_free(&xfree);

  if ((err = db_env_create(&dbenv, 0)) != 0)
    {
      log(MYNAME, LOG_PRIORITY_ERROR, "Failed to create db environment: %s", db_strerror(err));
      return 0;
    }

  dbenv->set_errcall(dbenv, &db_errcall);

  set_errpfx("open_dbenv");

  char dbpath[strlen(statedir) + strlen("db") + 1];
  strcpy(dbpath, statedir);
  strcat(dbpath, "db");

  if ((err = dbenv->open(dbenv, dbpath,
			 DB_INIT_LOCK | DB_INIT_MPOOL | DB_INIT_TXN |
			 DB_RECOVER | DB_CREATE, 0)) != 0)
    {
      log(MYNAME, LOG_PRIORITY_ERROR, "Failed to open db environment %s: %s", dbpath, db_strerror(err));
      close_dbenv();
      set_errpfx(NULL);
      return 0;
    }

  if ((err = dbenv->set_flags(dbenv, DB_AUTO_COMMIT | DB_DIRECT_DB | DB_DIRECT_LOG, 1)) != 0)
    {
      log(MYNAME, LOG_PRIORITY_ERROR, "Failed to set flags on db environment %s: %s", dbpath, db_strerror(err));
      close_dbenv();
      set_errpfx(NULL);
      return 0;
    }

  set_errpfx(NULL);
  return 1;
}

static inline DB *
open_db_(const char *file, const char *database, DB_TXN *txn, int dup,
         compare_func *dc, compare_func *bc, prefix_func *bp)
{
  int err;
  DB *db;

  if ((err = db_create(&db, dbenv, 0)) != 0)
    {
      log(MYNAME, LOG_PRIORITY_ERROR, "Failed to create database handle: %s", db_strerror(err));
      return NULL;
    }

  if (bc)
    {
      if ((err = db->set_bt_compare(db, bc)) != 0)
	{
	  log(MYNAME, LOG_PRIORITY_ERROR, "Failed to set btree key comparison function: %s", db_strerror(err));
	  db->close(db, 0);
	  return NULL;
	}
    }

  if (bp)
    {
      if ((err = db->set_bt_prefix(db, bp)) != 0)
	{
	  log(MYNAME, LOG_PRIORITY_ERROR, "Failed to set btree key prefix function: %s", db_strerror(err));
	  db->close(db, 0);
	  return NULL;
	}
    }

  if (dup && !dc)
    {
      if ((err = db->set_flags(db, DB_DUP)) != 0)
	{
	  log(MYNAME, LOG_PRIORITY_ERROR, "Failed to set DB_DUP flag: %s", db_strerror(err));
	  db->close(db, 0);
	  return NULL;
	}
    }
  else if (dup && dc)
    {
      if ((err = db->set_flags(db, DB_DUPSORT)) != 0)
	{
	  log(MYNAME, LOG_PRIORITY_ERROR, "Failed to set DB_DUPSORT flag: %s", db_strerror(err));
	  db->close(db, 0);
	  return NULL;
	}
      if ((err = db->set_dup_compare(db, dc)) != 0)
	{
	  log(MYNAME, LOG_PRIORITY_ERROR, "Failed to set duplicate comparison function: %s", db_strerror(err));
	  db->close(db, 0);
	  return NULL;
	}
    }

  if ((err = db->open(db, txn, file, database, DB_BTREE, DB_CREATE, 0)) != 0)
    {
      log(MYNAME, LOG_PRIORITY_ERROR, "Failed to open database %s (in file %s): %s", database, file, db_strerror(err));
      db->close(db, 0);
      return NULL;
    }

  return db;
}

DB *
open_db(const char *file, const char *database, DB_TXN *txn, int dup,
        compare_func *dc, compare_func *bc, prefix_func *bp)
{
  /* Database opens must be transaction-protected, so we'll do it
     ourselves if we have to */
  if (txn)
    {
      DB *ret;
      set_errpfx("open_db");
      ret = open_db_(file, database, txn, dup, dc, bc, bp);
      set_errpfx(NULL);
      return ret;
    }
  else
    {
      txn = begin_transaction();
      DB *ret;
      set_errpfx("open_db");
      ret = open_db_(file, database, txn, dup, dc, bc, bp);
      set_errpfx(NULL);
      if (!ret)
        {
          abort_transaction(txn);
          return NULL;
        }
      if (!commit_transaction(txn))
        return NULL;
      return ret;
    }
}

void
close_db(DB *db)
{
  int err;

  set_errpfx("close_db");
  if ((err = db->close(db, 0)) != 0)
    log(MYNAME, LOG_PRIORITY_ERROR, "Failed to close database: %s", db_strerror(err));
  set_errpfx(NULL);
}

DB_TXN *
begin_transaction(void)
{
  DB_TXN *txn;
  int err;

  set_errpfx("begin_transaction");
  if ((err = dbenv->txn_begin(dbenv, NULL, &txn, DB_TXN_NOWAIT)) != 0)
    {
      log(MYNAME, LOG_PRIORITY_ERROR, "Failed to start transaction: %s", db_strerror(err));
      txn = NULL;
    }
  set_errpfx(NULL);

  return txn;
}

void
abort_transaction(DB_TXN *txn)
{
  set_errpfx("abort_transaction");
  if (!txn)
    return;
  txn->abort(txn);
  set_errpfx(NULL);
}

int
commit_transaction(DB_TXN *txn)
{
  int err = 1;

  set_errpfx("commit_transaction");
  if (txn->commit(txn, 0) != 0)
    err = 0;
  set_errpfx(NULL);

  return err;
}

DEFINE_SYMBOL(const char *, database, driver_name, (void))
{
  static const char *name = "db";
  return name;
}

DEFINE_SYMBOL(struct database *, database, open_database, (const char *namespace, const char *name,
                                                           struct database_transaction *dbtxn,
                                                           int dup))
{
  struct database *database;
  DB *db;

  if (!namespace || !name)
    return NULL;

  if (!(db = open_db(namespace, name, dbtxn ? dbtxn->txn : NULL, dup, NULL, NULL, NULL)))
    return NULL;

  database = xmalloc(sizeof(struct database));
  database->namespace = xstrdup(namespace);
  database->name = xstrdup(name);
  database->db = db;
  database->temp_db = NULL;
  database->dup = dup;
  database->associated_key_callback = NULL;
  db->app_private = database;
  add_llist_tail(databases, database);
  return database;
}

static int
associate_callback(DB *db, const DBT *keyt, const DBT *datat, DBT *newkeyt)
{
  struct database *secondary = db->app_private;
  associated_key_callback *callback = secondary->associated_key_callback;
  int err;

  if (!callback)
    return -1;

  switch (err = (*callback)(keyt->data, keyt->size, datat->data, datat->size, &newkeyt->data, &newkeyt->size))
    {
    case ASSOCIATED_KEY_STORE_MALLOCED:
      newkeyt->flags = DB_DBT_APPMALLOC;
      return 0;
    case ASSOCIATED_KEY_STORE:
      return 0;
    case ASSOCIATED_KEY_IGNORE:
      return DB_DONOTINDEX;
    default:
      return err;
    }
}

DEFINE_SYMBOL(int, database, associate_database, (struct database *primary, struct database *secondary,
                                                  struct database_transaction *dbtxn,
                                                  associated_key_callback *callback))
{
  if (!primary || !secondary || !callback)
    return 0;

  int err;
  if (secondary->associated_key_callback)
    return 0;

  secondary->associated_key_callback = callback;
  if ((err = primary->db->associate(primary->db, dbtxn ? dbtxn->txn : NULL, secondary->db, &associate_callback,
                                    DB_CREATE)) != 0)
    {
      secondary->associated_key_callback = NULL;
      switch (err)
	{
	default:
	  log(MYNAME, LOG_PRIORITY_ERROR, "Failed to associate %s/%s with %s/%s: %s",
	      secondary->namespace, secondary->name, primary->namespace, primary->name, db_strerror(err));
	  return -1;
	}
    }
  return 1;
}

static int
match_database(const void *d1, const void *d2)
{
  const struct database *db1 = d1, *db2 = d2;
  return (strcmp(db1->namespace, db2->namespace) == 0) && (strcmp(db1->name, db2->name) == 0);
}

static inline void
free_database(struct database *db)
{
  struct llist_node *n;
  if (!db)
    return;
  if ((n = find_llist_match(databases, &match_database, db)))
    delete_llist_node(n);
  close_db(db->db);
  xfree(db->namespace);
  xfree(db->name);
  xfree(db);
}

DEFINE_SYMBOL(void, database, close_database, (struct database *db))
{
  free_database(db);
}

DEFINE_SYMBOL(int, database, get_value, (struct database *db, struct database_transaction *dbtxn,
                                         void *key, uint32_t keylen,
                                         void **data, uint32_t *datalen))
{
  int err;
  DBT keyt, datat;

  if (!db || !key || !keylen)
    return -1;

  memset(&keyt, 0, sizeof(DBT));
  memset(&datat, 0, sizeof(DBT));

  keyt.data = key;
  keyt.size = keylen;

  set_errpfx("get_value");
  if ((err = db->db->get(db->db, dbtxn ? dbtxn->txn : NULL, &keyt, &datat, 0)) != 0)
    {
      set_errpfx(NULL);
      switch (err)
	{
	case DB_NOTFOUND:
	  return 0;
	  break;
	default:
	  log(MYNAME, LOG_PRIORITY_ERROR, "Failed to get key from %s/%s: %s",
	      db->namespace, db->name, db_strerror(err));
	  return -1;
	}
    }
  set_errpfx(NULL);

  if (data)
    *data = datat.data;
  if (datalen)
    *datalen = datat.size;
  return 1;
}

DEFINE_SYMBOL(int, database, put_value, (struct database *db, struct database_transaction *dbtxn,
                                         void *key, uint32_t keylen,
                                         void *data, uint32_t datalen))
{
  int err;

  DBT keyt, datat;

  if (!db || !key)
    return -1;

  memset(&keyt, 0, sizeof(DBT));
  memset(&datat, 0, sizeof(DBT));

  keyt.data = key;
  keyt.size = keylen;
  datat.data = data;
  datat.size = datalen;

  set_errpfx("put_value");
  if ((err = db->db->put(db->db, dbtxn ? dbtxn->txn : NULL, &keyt, &datat, 0)) != 0)
    {
      set_errpfx(NULL);
      log(MYNAME, LOG_PRIORITY_ERROR, "Failed to put key to %s/%s: %s",
	  db->namespace, db->name, db_strerror(err));
      return -1;
    }
  set_errpfx(NULL);

  return 1;
}

DEFINE_SYMBOL(int, database, delete_value, (struct database *db, struct database_transaction *dbtxn,
                                            void *key, uint32_t keylen))
{
  int err;

  DBT keyt;

  if (!db || !key)
    return -1;
  
  memset(&keyt, 0, sizeof(DBT));

  keyt.data = key;
  keyt.size = keylen;

  set_errpfx("delete_value");
  if ((err = db->db->del(db->db, dbtxn ? dbtxn->txn : NULL, &keyt, 0)) != 0)
    {
      set_errpfx(NULL);
      switch (err)
	{
	case DB_NOTFOUND:
	  return 0;
	default:
	  log(MYNAME, LOG_PRIORITY_ERROR, "Failed to delete key from %s/%s: %s",
	      db->namespace, db->name, db_strerror(err));
	  return -1;
	}
    }
  set_errpfx(NULL);

  return 1;
}

DEFINE_SYMBOL(int, database, check_sanity, (struct database *db unused_attribute))
{
  return 1;
}

DEFINE_SYMBOL(struct database_cursor *, database, get_cursor, (struct database *db,
                                                               struct database_transaction *dbtxn,
                                                               void *key, uint32_t keylen,
                                                               void **data, uint32_t *datalen))
{
  struct database_cursor *cursor;
  DBT keyt, datat;
  int err;

  if (!db || !key || !keylen || !data || !datalen)
    return NULL;

  cursor = xmalloc(sizeof(struct database_cursor));
  cursor->db = db;
  cursor->cursor = NULL;
  set_errpfx("get_cursor");
  if ((err = db->db->cursor(db->db, dbtxn ? dbtxn->txn : NULL, &cursor->cursor, 0)) != 0)
    {
      set_errpfx(NULL);
      xfree(cursor);
      switch (err)
	{
	default:
	  log(MYNAME, LOG_PRIORITY_ERROR, "Failed to get cursor from %s/%s: %s",
	      db->namespace, db->name, db_strerror(err));
	  return NULL;
	}
    }

  memset(&keyt, 0, sizeof(DBT));
  memset(&datat, 0, sizeof(DBT));

  keyt.data = key;
  keyt.size = keylen;

  if ((err = cursor->cursor->c_get(cursor->cursor, &keyt, &datat, key ? DB_SET : DB_FIRST)) != 0)
    {
      set_errpfx(NULL);
      switch (err)
	{
        case DB_NOTFOUND:
          *data = NULL;
          *datalen = 0;
          cursor->cursor->c_close(cursor->cursor);
          cursor->cursor = NULL;
          return cursor;
	default:
          cursor->cursor->c_close(cursor->cursor);
          xfree(cursor);
	  log(MYNAME, LOG_PRIORITY_ERROR, "Failed to get cursor from %s/%s: %s",
	      db->namespace, db->name, db_strerror(err));
	  return NULL;
	}
    }
  set_errpfx(NULL);

  *data = datat.data;
  *datalen = datat.size;
  return cursor;
}

DEFINE_SYMBOL(struct database_cursor *, database, get_primary_cursor, (struct database *db,
                                                                       struct database_transaction *dbtxn,
                                                                       void *key, uint32_t keylen,
                                                                       void **data, uint32_t *datalen))
{
  struct database_cursor *cursor;
  DBT keyt, pkeyt, datat;
  int err;

  if (!db || !key || !keylen || !data || !datalen)
    return NULL;

  cursor = xmalloc(sizeof(struct database_cursor));
  cursor->db = db;
  cursor->cursor = NULL;
  set_errpfx("get_cursor");
  if ((err = db->db->cursor(db->db, dbtxn ? dbtxn->txn : NULL, &cursor->cursor, 0)) != 0)
    {
      set_errpfx(NULL);
      xfree(cursor);
      switch (err)
	{
	default:
	  log(MYNAME, LOG_PRIORITY_ERROR, "Failed to get cursor from %s/%s: %s",
	      db->namespace, db->name, db_strerror(err));
	  return NULL;
	}
    }

  memset(&keyt, 0, sizeof(DBT));
  memset(&pkeyt, 0, sizeof(DBT));
  memset(&datat, 0, sizeof(DBT));

  keyt.data = key;
  keyt.size = keylen;

  if ((err = cursor->cursor->c_pget(cursor->cursor, &keyt, &pkeyt, &datat, key ? DB_SET : DB_FIRST)) != 0)
    {
      set_errpfx(NULL);
      switch (err)
	{
        case DB_NOTFOUND:
          *data = NULL;
          *datalen = 0;
          cursor->cursor->c_close(cursor->cursor);
          cursor->cursor = NULL;
          return cursor;
	default:
          cursor->cursor->c_close(cursor->cursor);
          xfree(cursor);
	  log(MYNAME, LOG_PRIORITY_ERROR, "Failed to get cursor from %s/%s: %s",
	      db->namespace, db->name, db_strerror(err));
	  return NULL;
	}
    }
  set_errpfx(NULL);

  *data = pkeyt.data;
  *datalen = pkeyt.size;
  return cursor;
}

DEFINE_SYMBOL(void, database, free_cursor, (struct database_cursor *cursor))
{
  if (!cursor)
    return;
  if (cursor->cursor)
    cursor->cursor->c_close(cursor->cursor);
  xfree(cursor);
}

static int
cursor_get(struct database_cursor *cursor, void **data, uint32_t *datalen, int flags)
{
  DBT keyt, datat;
  int err;

  *data = NULL;
  *datalen = 0;
  if (!cursor)
    return -1;

  if (!cursor->cursor)
    return 0;

  memset(&keyt, 0, sizeof(DBT));
  memset(&datat, 0, sizeof(DBT));

  set_errpfx("cursor_get");
  if ((err = cursor->cursor->c_get(cursor->cursor, &keyt, &datat, flags)) != 0)
    {
      set_errpfx(NULL);
      switch (err)
	{
        case DB_NOTFOUND:
          *data = NULL;
          *datalen = 0;
          return 0;
	default:
	  log(MYNAME, LOG_PRIORITY_ERROR, "Failed to get from cursor on %s/%s: %s",
	      cursor->db->namespace, cursor->db->name, db_strerror(err));
	  return -1;
	}
    }
  
  set_errpfx(NULL);
  *data = datat.data;
  *datalen = datat.size;
  return 1;
}

static int
cursor_pget(struct database_cursor *cursor, void **data, uint32_t *datalen, int flags)
{
  DBT keyt, pkeyt, datat;
  int err;

  *data = NULL;
  *datalen = 0;
  if (!cursor)
    return -1;

  if (!cursor->cursor)
    return 0;

  memset(&keyt, 0, sizeof(DBT));
  memset(&pkeyt, 0, sizeof(DBT));
  memset(&datat, 0, sizeof(DBT));

  set_errpfx("cursor_get");
  if ((err = cursor->cursor->c_pget(cursor->cursor, &keyt, &pkeyt, &datat, flags)) != 0)
    {
      set_errpfx(NULL);
      switch (err)
	{
        case DB_NOTFOUND:
          *data = NULL;
          *datalen = 0;
          return 0;
	default:
	  log(MYNAME, LOG_PRIORITY_ERROR, "Failed to get from cursor on %s/%s: %s",
	      cursor->db->namespace, cursor->db->name, db_strerror(err));
	  return -1;
	}
    }
  
  set_errpfx(NULL);
  *data = pkeyt.data;
  *datalen = pkeyt.size;
  return 1;
}

DEFINE_SYMBOL(int, database, cursor_next, (struct database_cursor *cursor, void **data, uint32_t *datalen))
{
  return cursor_get(cursor, data, datalen, DB_NEXT);
}

DEFINE_SYMBOL(int, database, cursor_primary_next, (struct database_cursor *cursor, void **data, uint32_t *datalen))
{
  return cursor_pget(cursor, data, datalen, DB_NEXT);
}

DEFINE_SYMBOL(int, database, cursor_prev, (struct database_cursor *cursor, void **data, uint32_t *datalen))
{
  return cursor_get(cursor, data, datalen, DB_PREV);
}

DEFINE_SYMBOL(int, database, cursor_primary_prev, (struct database_cursor *cursor, void **data, uint32_t *datalen))
{
  return cursor_pget(cursor, data, datalen, DB_PREV);
}

DEFINE_SYMBOL(int, database, cursor_get, (struct database_cursor *cursor, void **data, uint32_t *datalen))
{
  return cursor_get(cursor, data, datalen, DB_CURRENT);
}

DEFINE_SYMBOL(int, database, cursor_primary_get, (struct database_cursor *cursor, void **data, uint32_t *datalen))
{
  return cursor_pget(cursor, data, datalen, DB_CURRENT);
}

DEFINE_SYMBOL(int, database, cursor_next_dup, (struct database_cursor *cursor, void **data, uint32_t *datalen))
{
  return cursor_get(cursor, data, datalen, DB_NEXT_DUP);
}

DEFINE_SYMBOL(int, database, cursor_primary_next_dup, (struct database_cursor *cursor, void **data, uint32_t *datalen))
{
  return cursor_pget(cursor, data, datalen, DB_NEXT_DUP);
}

DEFINE_SYMBOL(int, database, cursor_delete, (struct database_cursor *cursor))
{
  int err;
  if (!cursor)
    return -1;
  set_errpfx("cursor_delete");
  if ((err = cursor->cursor->c_del(cursor->cursor, 0)) != 0)
    {
      set_errpfx(NULL);
      switch (err)
	{
        case DB_KEYEMPTY:
        case DB_NOTFOUND:
          return 1;
	default:
	  log(MYNAME, LOG_PRIORITY_ERROR, "Failed to get cursor from %s/%s: %s",
	      cursor->db->namespace, cursor->db->name, db_strerror(err));
          cursor->cursor->c_close(cursor->cursor);
          xfree(cursor);
	  return -1;
	}
    }
  set_errpfx(NULL);
  return 1;
}

static struct database_cursor *
do_join(struct database *db, DB_TXN *txn, int args, void **data, uint32_t *len, DBC **carray)
{
  struct database_cursor *cursor;
  int i, err;
  DBT keyt, datat;

  memset(&keyt, 0, sizeof(DBT));
  memset(&datat, 0, sizeof(DBT));

  for (i = 0; i < args; i++)
    {
      if ((err = db->db->cursor(db->db, txn, &carray[i], 0) != 0))
        {
          switch (err)
            {
            default:
              log(MYNAME, LOG_PRIORITY_ERROR, "Failed to get cursor from %s/%s: %s",
                  db->namespace, db->name, db_strerror(err));
              return NULL;
            }
        }
      keyt.data = data[i];
      keyt.size = len[i];
      if ((err = carray[i]->c_get(carray[i], &keyt, &datat, DB_SET)) != 0)
        {
          switch (err)
            {
            case DB_NOTFOUND:
              return NULL;
            default:
              log(MYNAME, LOG_PRIORITY_ERROR, "Failed to get cursor from %s/%s: %s",
                  db->namespace, db->name, db_strerror(err));
              return NULL;
            }
        }
    }

  cursor = xmalloc(sizeof(struct database_cursor));
  cursor->db = db;
  cursor->cursor = NULL;
  if ((err = db->db->join(db->db, carray, &cursor->cursor, 0)) != 0)
    {
      switch (err)
        {
        default:
          log(MYNAME, LOG_PRIORITY_ERROR, "Failed to get cursor from %s/%s: %s",
              cursor->db->namespace, cursor->db->name, db_strerror(err));
          return NULL;
        }
    }
  return cursor;
}

DEFINE_SYMBOL(struct database_cursor *, database, join, (struct database *db, struct database_transaction *dbtxn,
                                                         int args, void **data, uint32_t *len))
{
  DBC **carray;
  struct database_cursor *cursor;
  int i;

  if (!db || !args || !data || !len)
    return NULL;

  for (i = 0; i < args; i++)
    if (!data[i])
      return NULL;

  carray = xmalloc(sizeof(DBC *) * args);
  memset(carray, 0, sizeof(DBC *) * args);
  set_errpfx("join");
  if ((cursor = do_join(db, dbtxn ? dbtxn->txn : NULL, args, data, len, carray)))
    {
      set_errpfx(NULL);
      xfree(carray);
      return cursor;
    }
  else
    {
      set_errpfx(NULL);
      xfree(carray);
      for (i = 0; i < args; i++)
        if (carray[i])
          carray[i]->c_close(carray[i]);
      return NULL;
    }
}

DEFINE_SYMBOL(struct database_transaction *, database, begin_transaction, (void))
{
  DB_TXN *txn = begin_transaction();
  if (!txn)
    return NULL;
  struct database_transaction *dbtxn = xmalloc(sizeof(*dbtxn));
  dbtxn->txn = txn;
  return dbtxn;
}

DEFINE_SYMBOL(int, database, commit_transaction, (struct database_transaction *dbtxn))
{
  if (!dbtxn)
    return 1;
  int ret = commit_transaction(dbtxn->txn);
  xfree(dbtxn);
  return ret;
}

DEFINE_SYMBOL(void, database, abort_transaction, (struct database_transaction *dbtxn))
{
  if (!dbtxn)
    return;
  abort_transaction(dbtxn->txn);
  xfree(dbtxn);
}

/** \brief Run a given function as an ACID transaction
 * \param func Pointer to the function to run
 * \param data Data pointer for function
 * \param retries Number of times to retry if the transaction aborts
 * \return non-zero if the transaction commits, zero if the transaction aborts
 *
 * The function supplied will be called with a new tranaction and the
 * data pointer. If it returns zero, or if the transaction cannot be
 * committed, then the transaction will be aborted. If retries is
 * non-zero then the function will be rerun until either it commits,
 * or the retry limit is reached. If retries is zero, the function
 * will not be rerun.
 *
 * This means that the function itself must be ACID in nature - it
 * must not perform any operations that are not protected by the
 * transaction handle that it is called with, including updating any
 * external variables or modifying the structure pointed to by the
 * data pointer.
 */
DEFINE_SYMBOL(int, database, run_transaction, (int (*func)(struct database_transaction *, void *),
                                               void *data,
                                               int retries))
{
  do
    {
      struct database_transaction *txn = LSYM(database, begin_transaction)();
      if (!txn)
        return 0;
      if ((*func)(txn, data))
        {
          if (LSYM(database, commit_transaction)(txn))
            return 1;
        }
      else
        LSYM(database, abort_transaction)(txn);
    } while (retries-- > 0);
  return 0;
}

struct reload_data
{
  struct llist *databases;
};

int
MODULE_INIT(void *data)
{
  transactions = new_llist();
  if (data)
    {
      struct reload_data *rd = data;
      struct llist_node *n;
      int fail = 0;

      /* Open a new dbenv using whatever libdb we are linked with, and
       * reopen all the databases in that
       */
      open_dbenv();
      databases = rd->databases;

      for (n = llist_head(databases); n; n = llist_node_next(n))
        {
          struct database *db = llist_node_data(n);
          if (!(db->temp_db = open_db(db->namespace, db->name, NULL, 0, NULL, NULL, NULL)))
            {
              fail = 1;
              break;
            }
        }

      if (fail)
        {
          for (n = llist_head(databases); n; n = llist_node_next(n))
            {
              struct database *db = llist_node_data(n);
              close_db(db->temp_db);
              db->temp_db = NULL;
            }
          return MODULE_LOAD_FAILED;
        }
      else
        {
          for (n = llist_head(databases); n; n = llist_node_next(n))
            {
              struct database *db = llist_node_data(n);
              close_db(db->db);
              db->db = db->temp_db;
              db->temp_db = NULL;
            }
          return MODULE_LOAD_OK;
        }
    }
  else
    {
      databases = new_llist();
      if (!open_dbenv())
        return MODULE_LOAD_FAILED;
    }

  return MODULE_LOAD_OK;
}

int
MODULE_DEINIT(const char *new_version, module_init_func *init)
{
  /* There is no sane way to preserve a transaction across a module
   * reload, so abort them all. There shouldn't really be any
   * outstanding transactions at this point anyway - letting
   * transactions span multiple messages would probably cause more
   * trouble than it is worth
   */
  destroy_llist(transactions, LSYM(database, abort_transaction));

  if (new_version)
    {
      struct reload_data rd = {databases};
      switch ((*init)(&rd))
        {
        case MODULE_LOAD_OK:
          close_dbenv();
          databases = NULL;
          dbenv = NULL;
          return MODULE_LOAD_OK;
        case MODULE_LOAD_FAILED:
          return MODULE_LOAD_FAILED;
        default:
          log(MYNAME, LOG_PRIORITY_ERROR, "Unknown return from new module version %s, aborting", new_version);
          return MODULE_LOAD_FAILED;
        }
    }

  destroy_llist(databases, free_database);
  close_dbenv();

  return MODULE_UNLOAD_OK;
}
