/************************************************************************
 *   This is part of Dancer, a text relay daemon
 *   Copyright (C) 2002 Andrew Suffield <asuffield@users.sourceforge.net>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

#include <dancerd/config.h>
#include <dancerd/main.h>
#include <dancerd/modules.h>
#include <dancerd/log.h>
#include <dancerd/fdbuffer.h>
#include <dancerd/fd.h>
#include <dancerd/config/items.h>
#include <dancerd/xstring.h>

#include <assert.h>
#include <stddef.h>
#include <string.h>

SETUP_MODULE("edit.config", MODVERSION);

static struct fdbuffer *stdin_fdb = NULL, *stdout_fdb = NULL;

IMPORTED_SYMBOL(config, get_item);
IMPORTED_SYMBOL(config, put_item);
IMPORTED_SYMBOL(config, delete_item);

#include "symtab"

static void
do_get_config(struct dbuffer *buf, ssize_t line_end)
{
  char *key_buf, *key;
  struct config_item item;

  assert(line_end >= 0);

  key_buf = xmalloc((size_t)(line_end + 1));
  dbuffer_extract(buf, key_buf, (size_t)line_end);
  /* Remove the trailing \n */
  dbuffer_remove(buf, 1);
  key_buf[line_end] = '\0';

  key = strip_spaces(key_buf);
  if (SYM(config, get_item)(key, &item))
    switch(item.type)
      {
      case citem_bool:
        if (item.boolean)
          printf_fdbuffer(stdout_fdb, "GET %s: true\n", key);
        else
          printf_fdbuffer(stdout_fdb, "GET %s: false\n", key);
        break;
      case citem_string:
        printf_fdbuffer(stdout_fdb, "GET %s: \"%s\"\n", key, item.str);
        break;
      case citem_uint32:
        {
          long unsigned int i = item.uint32;
          printf_fdbuffer(stdout_fdb, "GET %s: %lu\n", key, i);
          break;
        }
      case citem_sint32:
        {
          long signed int i = item.sint32;
          printf_fdbuffer(stdout_fdb, "GET %s: %+ld\n", key, i);
          break;
        }
      case citem_null:
        printf_fdbuffer(stdout_fdb, "GET %s: null\n", key);
        break;
      default:
        printf_fdbuffer(stdout_fdb, "ERROR: Unknown item type was retrieved\n");
        break;
      }
  else
    printf_fdbuffer(stdout_fdb, "ERROR: Couldn't retrieve config item\n");

  xfree(key_buf);
}

static void
do_put_config(struct dbuffer *buf, ssize_t line_end)
{
  ssize_t key_end = dbuffer_find_char(buf, ' ');
  char *key_buf, *value_buf, *value;
  struct config_item item;

  /* If we didn't find a space, set the end of the key to be the end of the line */
  if ((key_end == -1) || (key_end >= line_end))
    {
      printf_fdbuffer(stdout_fdb, "ERROR: no value given for PUT\n");
      dbuffer_remove(buf, (size_t)(line_end + 1));
      return;
    }

  assert(key_end >= 0);
  key_buf = xmalloc((size_t)(key_end + 1));
  dbuffer_extract(buf, key_buf, (size_t)key_end);
  /* And remove the trailing space */
  dbuffer_remove(buf, 1);
  line_end -= key_end + 1;
  key_buf[key_end] = '\0';

  value_buf = xmalloc((size_t)(line_end + 1));
  dbuffer_extract(buf, value_buf, (size_t)line_end);
  /* ...and the trailing \n */
  dbuffer_remove(buf, 1);
  value_buf[line_end] = '\0';

  item.key = strip_spaces(key_buf);
  value = strip_spaces(value_buf);
  switch(value[0])
    {
    case 't':
      item.type = citem_bool;
      item.boolean = 1;
      break;
    case 'f':
      item.type = citem_bool;
      item.boolean = 0;
      break;
    case '"':
      item.type = citem_string;
      item.str = value + 1;
      if ((value = strchr(item.str, '"')))
        *value = '\0';
      else
        {
          printf_fdbuffer(stdout_fdb, "ERROR: Unterminated string value");
          xfree(key_buf);
          xfree(value_buf);
          return;
        }
      break;
    case '+':
    case '-':
      item.type = citem_sint32;
      item.sint32 = strtol(value, NULL, 0);
      break;
    case '0': case '1': case '2': case '3': case '4':
    case '5': case '6': case '7': case '8': case '9':
      item.type = citem_uint32;
      item.uint32 = strtoul(value, NULL, 0);
      break;
    default:
      printf_fdbuffer(stdout_fdb, "ERROR: Unrecognised value type\n");
      xfree(key_buf);
      xfree(value_buf);
      return;
    }

  if (SYM(config, put_item)(&item))
    printf_fdbuffer(stdout_fdb, "PUT %s\n", item.key);
  else
    printf_fdbuffer(stdout_fdb, "ERROR: failed to put config item\n");
}

static void
do_delete_config(struct dbuffer *buf, ssize_t line_end)
{
  char *key_buf, *key;

  assert(line_end >= 0);
  key_buf = xmalloc((size_t)(line_end + 1));
  dbuffer_extract(buf, key_buf, (size_t)line_end);
  dbuffer_remove(buf, 1);
  key_buf[line_end] = '\0';

  key = strip_spaces(key_buf);
  if (SYM(config, delete_item)(key))
    printf_fdbuffer(stdout_fdb, "DELETE %s: ok\n", key);
  else
    printf_fdbuffer(stdout_fdb, "ERROR: Couldn't delete config item");
  xfree(key_buf);
}

static void
stdin_read_handler(struct fdbuffer *fdb unused_attribute, struct dbuffer *buf, const char *data, size_t len)
{
  ssize_t line_end;

  if (!data || !len)
    {
      log(MYNAME, LOG_PRIORITY_INFO, "stdin was closed, unloading self");
      unload_module(MYNAME);
      return;
    }

  if (strchr(data, '\n'))
    while ((line_end = dbuffer_find_char(buf, '\n')) != -1)
      {
	ssize_t command_end = dbuffer_find_char(buf, ' ');
	char command_buf[32];
	/* If we didn't find a space, set the end of the command to be the end of the line */
	if ((command_end == -1) || (command_end > line_end))
	  command_end = line_end;

	assert(command_end >= 0);

	/* Extract the first 31 characters of the command */
	dbuffer_get(buf, command_buf, (31 < command_end) ? 31 : (size_t)command_end);
	/* And NULL-terminate the string */
	command_buf[(31 < command_end) ? 31 : command_end] = '\0';
	/* Eat the space, if one exists */
	if (command_end < line_end)
	  command_end++;
	/* Delete the command and space from the dbuffer */
	dbuffer_remove(buf, (size_t)command_end);
	line_end -= command_end;
	/* OK, let's go parse */
	if (strcasecmp(command_buf, "GET") == 0)
	  do_get_config(buf, line_end);
	else if (strcasecmp(command_buf, "PUT") == 0)
	  do_put_config(buf, line_end);
	else if (strcasecmp(command_buf, "DELETE") == 0)
	  do_delete_config(buf, line_end);
	else
	  {
	    assert(line_end >= 0);

	    /* Remove anything else in this line, plus the trailing \n */
	    dbuffer_remove(buf, (size_t)(line_end + 1));
	    printf_fdbuffer(stdout_fdb, "ERROR: unknown command\n");
	  }
      }
}

static void 
stdout_read_handler(struct fdbuffer *fdb unused_attribute, struct dbuffer *buf unused_attribute,
		    const char *data, size_t len)
{
  if (data || len)
    {
      log(MYNAME, LOG_PRIORITY_WARN, "WTF: Data for reading from stdout?");
      return;
    }
  /* Actually I don't care if stdout gets closed */
}

struct reload_data
{
  struct fdbuffer *stdin, *stdout;
};

int
MODULE_INIT(void *data)
{
  if (data)
    {
      struct reload_data *rd = data;
      stdin_fdb = rd->stdin;
      stdout_fdb = rd->stdout;
      attach_fdbuffer_read_handler(MYHANDLE, stdin_fdb, &stdin_read_handler);
      attach_fdbuffer_read_handler(MYHANDLE, stdout_fdb, &stdout_read_handler);

      return MODULE_LOAD_OK;
    }
  else
    {
      if (!have_stdio)
        {
          log(MYNAME, LOG_PRIORITY_ERROR, "stdin/stdout are not available");
          return MODULE_LOAD_FAILED;
        }

      have_stdio = 0;

      stdin_fdb = setup_fdbuffer(MYHANDLE, 0, FD_CLASS_LOCAL, &stdin_read_handler);
      block_fd_writes(0);
      stdout_fdb = setup_fdbuffer(MYHANDLE, 1, FD_CLASS_LOCAL, &stdout_read_handler);
      block_fd_reads(1);
      printf_fdbuffer(stdout_fdb, "%s ready\n", MYNAME);

      return MODULE_LOAD_OK;
    }
}

int
MODULE_DEINIT(const char *new_version, module_init_func *init)
{
  if (new_version)
    {
      struct reload_data rd = {stdin_fdb, stdout_fdb};
      switch ((*init)(&rd))
        {
        case MODULE_LOAD_OK:
          stdin_fdb = NULL;
          stdout_fdb = NULL;
          return MODULE_LOAD_OK;
        case MODULE_LOAD_FAILED:
          return MODULE_LOAD_FAILED;
        default:
          log(MYNAME, LOG_PRIORITY_ERROR, "Unknown return from new module version %s, aborting", new_version);
          return MODULE_LOAD_FAILED;
        }
    }

  destroy_fdbuffer(stdin_fdb);
  stdin_fdb = NULL;
  destroy_fdbuffer(stdout_fdb);
  stdout_fdb = NULL;

  have_stdio = 1;

  return MODULE_UNLOAD_OK;
}
