/************************************************************************
 *   This is part of Dancer, a text relay daemon
 *   Copyright (C) 2002 Andrew Suffield <asuffield@users.sourceforge.net>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

#include <dancerd/config.h>
#include <dancerd/modules.h>
#include <dancerd/protocol/protocol.h>
#include <dancerd/protocol/message.h>
#include <dancerd/log.h>
#include <dancerd/xmalloc.h>
#include <dancerd/xstring.h>

#include <dgmtp/dgmtp.h>
#include <dgmtp/message.h>
#include <dgmtp/errors.h>

#include <assert.h>
#include <stddef.h>
#include <string.h>

SETUP_MODULE("protocol.dgmtp", MODVERSION);

IMPORTED_SYMBOL(protocol, register_protocol);
IMPORTED_SYMBOL(protocol, unregister_protocol);
IMPORTED_SYMBOL(protocol, write_data);
IMPORTED_SYMBOL(message, get_message_type_name);

#include "symtab"

static struct protocol *my_protocol = NULL;
struct dgmtp_client
{
  struct session *session;
  struct dgmtp_connection *conn;
};

static int protocol_error;

static protocol_identify_func identify;
static int
identify(struct dbuffer *buf)
{
  /* We identify a DGMTP connection by this byte string (taken
   * directly from doc/protocol.txt in the dgmtp source tree)
   */
  static const char dgmtp_handshake[] =
    {
      0,
      0, 0, 0, 11,
      1, 0, 0, 0, 6, 'D', 'A', 'N', 'C', 'E', 'R'};

  /* Grab all we can from the buffer */
  char cbuf[sizeof(dgmtp_handshake)];
  size_t len = dbuffer_get(buf, cbuf, sizeof(dgmtp_handshake));

  /* All the bytes we get must match, or else it isn't DGMTP */
  if (memcmp(cbuf, dgmtp_handshake, len) != 0)
    return -1;

  /* We don't know it's valid until we get all the bytes */
  if (len == sizeof(dgmtp_handshake))
    return 1;
  return 0;
}

static protocol_init_session_func init_session;
static void *
init_session(struct session *session, void *d)
{
  struct dgmtp_client *client = d;

  if (client)
    return client;

  client = xmalloc(sizeof(*client));
  client->session = session;
  client->conn = dgmtp_new_connection(0);

  return client;
}

static protocol_connection_ready_func connection_ready;
static void *
connection_ready(struct session *session)
{
  struct dgmtp_client *client;
  client = xmalloc(sizeof(*client));
  client->session = session;
  client->conn = dgmtp_new_connection(1);
  return client;
}

/* Returns false if there was an error that should cause the
 * connection to be dropped, true otherwise
 */
static int
process_dgmtp_message(struct session *session, struct dgmtp_message *dmsg)
{
  if (dmsg->type)
    {
      int type = SYM(message, get_message_type)(dmsg->type);

      if (type == -1)
        return 0;

      int strc = 0, uint32c = 0, sint32c = 0;
      for (int i = 0; i < dmsg->members; i++)
        switch (dmsg->member[i].type)
          {
          case dgmtp_string:
            strc++;
            break;
          case dgmtp_uint32:
            uint32c++;
            break;
          case dgmtp_int32:
            sint32c++;
            break;
          }
      const char *strv[strc];
      uint32_t uint32v[uint32c];
      int32_t sint32v[sint32c];
      int strc2 = 0, uint32c2 = 0, sint32c2 = 0;
      for (int i = 0; i < dmsg->members; i++)
        switch (dmsg->member[i].type)
          {
          case dgmtp_string:
            strv[strc2++] = dmsg->member[i].string;
            break;
          case dgmtp_uint32:
            uint32v[uint32c2++] = dmsg->member[i].uint32;
            break;
          case dgmtp_int32:
            sint32v[sint32c2++] = dmsg->member[i].int32;
            break;
          }
      struct message *msg = SYM(message, build_message)(type,
                                                        dmsg->origin, dmsg->dest,
                                                        strc, strv,
                                                        uint32c, uint32v,
                                                        sint32c, sint32v);
      SYM(protocol, read_message)(session, msg);
      return 1;
    }
  else
    {
      /* This is a basic_message */
      if (dmsg->members == 0)
        {
          /* Handshake complete */
          /* Currently we have no use for this. Ignore it. */
          return 1;
        }
      if (dmsg->member[0].type != dgmtp_string)
        {
          /* The heck is this? Close the connection */
          return 0;
        }
      if (!strcmp(dmsg->member[0].string, "ERROR"))
        {
          struct message *msg;
          assert(dmsg->members >= 2);
          msg = SYM(message, build_message)(protocol_error,
                                            NULL, NULL,
                                            1, (const char *[]){dmsg->member[1].string},
                                            0, NULL,
                                            0, NULL);
          SYM(protocol, read_message)(session, msg);
          /* Note, it is not our job to close the connection. libdgmtp
           * will queue up a NULL message behind this one which will
           * do that
           */
          return 1;
        }
      /* If we get this far, we don't know what that message is; drop
       * the connection
       */
      return 0;
    }
}

static protocol_read_data_func read_data;
static void
read_data(struct session *session, void *d,
          struct dbuffer *buf unused_attribute, const void *data, size_t len)
{
  struct dgmtp_client *client = d;

  assert(session == client->session);

  if (len > 0)
    {
      dgmtp_read_data(client->conn, data, len);
      dbuffer_remove(buf, len);
    }
  else
    {
      char cbuf[dbuffer_len(buf)];
      size_t clen = dbuffer_extract(buf, cbuf, dbuffer_len(buf));
      /* We might read zero bytes from an empty buffer immediately
       * after connection_ready was called. This makes things easier,
       * since we need to write some data, and we can't do that from
       * connection_ready (session->data hasn't been initialised at
       * that point)
       */
      if (clen > 0)
        dgmtp_read_data(client->conn, cbuf, clen);
    }

  /* Dequeue anything we might have received during that */
  while (dgmtp_message_queue_size(client->conn) > 0)
    {
      struct dgmtp_message *dmsg = dgmtp_get_message(client->conn);

      if (!dmsg)
        {
          dgmtp_close_connection(client->conn);
          client->conn = NULL;
          SYM(protocol, shutdown_session)(session);
          return;
        }

      if (!process_dgmtp_message(session, dmsg))
        {
          dgmtp_free_message(dmsg);
          dgmtp_close_connection(client->conn);
          client->conn = NULL;
          SYM(protocol, shutdown_session)(session);
          return;
        }

      dgmtp_free_message(dmsg);
    }

  /* Dequeue anything we can from libdgmtp, and pass it on up */
  while (client->conn && dgmtp_write_queue_size(client->conn) > 0)
    {
      static char cbuf[4096];
      size_t clen = dgmtp_get_write_data(client->conn, cbuf, 4096);
      SYM(protocol, write_data)(client->session, cbuf, clen);
      dgmtp_remove_write_data(client->conn, clen);
    }
}

static protocol_send_message_func send_message;
static void
send_message(struct session *session, void *d, const struct message *msg)
{
  struct dgmtp_client *client = d;

  assert(session == client->session);

  if (!msg)
    return;

  /* Build a struct dgmtp_message from the struct message */
  /* Use a C99-style stack allocation */
  char dmsg_data[sizeof(struct dgmtp_message) + (sizeof(struct dgmtp_member) *
                                                 (msg->strc + msg->uint32c + msg->sint32c))];
  memset(dmsg_data, 0, sizeof(dmsg_data));
  struct dgmtp_message *dmsg = (struct dgmtp_message *)dmsg_data;
  dmsg->origin = msg->origin;
  dmsg->dest = msg->dest;

  /* This is strdup(), on the stack */
  const char *type_name = SYM(message, get_message_type_name)(msg->type);
  size_t type_name_len = strlen(type_name);
  char dmsg_type[type_name_len + 1];
  memcpy(dmsg_type, type_name, type_name_len + 1);
  dmsg->type = dmsg_type;

  /* Fill in the body */
  dmsg->members = msg->strc + msg->uint32c + msg->sint32c;
  int i = 0;
  for (int j = 0; j < msg->strc; i++, j++)
    {
      dmsg->member[i].type = dgmtp_string;
      dmsg->member[i].string = msg->strv[j];
    }
  for (int j = 0; j < msg->uint32c; i++, j++)
    {
      dmsg->member[i].type = dgmtp_uint32;
      dmsg->member[i].uint32 = msg->uint32v[j];
    }
  for (int j = 0; j < msg->sint32c; i++, j++)
    {
      dmsg->member[i].type = dgmtp_int32;
      dmsg->member[i].int32 = msg->sint32v[j];
    }
  assert(i == dmsg->members);
  /* Pass the dgmtp_message to libdgmtp for processing */
  dgmtp_send_message(client->conn, dmsg);

  /* Dequeue anything we can from libdgmtp, and pass it on up */
  while (dgmtp_write_queue_size(client->conn) > 0)
    {
      static char cbuf[4096];
      size_t clen = dgmtp_get_write_data(client->conn, cbuf, 4096);
      SYM(protocol, write_data)(client->session, cbuf, clen);
      dgmtp_remove_write_data(client->conn, clen);
    }
}

static protocol_close_session_func close_session;
static void
close_session(struct session *session unused_attribute, void *d)
{
  struct dgmtp_client *client = d;
  if (client->conn)
    dgmtp_close_connection(client->conn);
  client->conn = NULL;
  xfree(client);
}

int
MODULE_INIT(void *data unused_attribute)
{
  my_protocol = SYM(protocol, register_protocol)("dgmtp", MYHANDLE,
                                                 &identify, &init_session, &connection_ready,
                                                 &read_data, &send_message,
                                                 &close_session);
  protocol_error = SYM(message, register_message_type)("protocol error");

  if (!my_protocol)
    return MODULE_LOAD_FAILED;

  return MODULE_LOAD_OK;
}

int
MODULE_DEINIT(const char *new_version, module_init_func *init)
{
  /* This deinit cannot be done safely - we must unregister our own
   * protocol handle before loading the new one, and we don't know if
   * we'll be able to get it back afterwards.
   *
   * If the load fails and we can't get back our handle, that's a
   * FUBAR.
   */
  if (my_protocol)
    SYM(protocol, unregister_protocol)(my_protocol);
  my_protocol = NULL;

  if (new_version)
    {
      switch ((*init)(NULL))
        {
        case MODULE_LOAD_OK:
          return MODULE_LOAD_OK;
        case MODULE_LOAD_FAILED:
          my_protocol = SYM(protocol, register_protocol)("dgmtp", MYHANDLE,
                                                         &identify, &init_session, &connection_ready,
                                                         &read_data, &send_message,
                                                         &close_session);
          if (!my_protocol)
            return MODULE_LOAD_FUBAR;
          return MODULE_LOAD_FAILED;
        default:
          log(MYNAME, LOG_PRIORITY_ERROR, "Unknown return from new module version %s, aborting", new_version);
          my_protocol = SYM(protocol, register_protocol)("dgmtp", MYHANDLE,
                                                         &identify, &init_session, &connection_ready,
                                                         &read_data, &send_message,
                                                         &close_session);
          if (!my_protocol)
            return MODULE_LOAD_FUBAR;
          return MODULE_LOAD_FAILED;
        }
    }

  return MODULE_UNLOAD_OK;
}
