/************************************************************************
 *   This is part of Dancer, a text relay daemon
 *   Copyright (C) 2002 Andrew Suffield <asuffield@users.sourceforge.net>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

/** \file protocol/message.c
 *
 * The protocol layer encodes/decodes messages from sessions; messages
 * are the only form of communication that the higher layers will see
 */

#include <dancerd/config.h>
#include <dancerd/hashtable.h>
#include <dancerd/llist.h>
#include <dancerd/protocol/message.h>
#include <dancerd/modules.h>
#include <dancerd/protocol/protocol.h>
#include <dancerd/protocol/protocol_internal.h>
#include <dancerd/xmalloc.h>
#include <dancerd/xstring.h>

#include <assert.h>
#include <stdint.h>

EXTERN_MODULE();

struct message_type
{
  char *name;
  int id;
  struct llist *handlers;
};

static struct hashtable *message_type_hash = NULL;
static struct message_type *message_type_array = NULL;
static struct message_type message_type_array_negative[1] = {{NULL, -2, NULL}};
static int message_types_allocated = 0, message_types_registered = 0;
#define MESSAGE_TYPE_ALLOCATE_CHUNK 64

/** \brief Duplicate a message
 * \param old_msg message to duplicate
 * \return a newly allocated message equal to old_msg
 *
 * A new message is allocated, and all fields from old_msg are
 * deep-copied, creating a new and independant message. The new
 * message returned must be freed with free_message once it is no
 * longer needed.
 *
 * There is no relationship between the old and new message once this
 * function returns, other than the equality of their contents. Each
 * can be used and freed without regard to the state of the other.
 */
DEFINE_SYMBOL(struct message *, message, dup_message, (const struct message *old_msg))
{
  if (!old_msg)
    return NULL;

  return LSYM(message, build_message)(old_msg->type, old_msg->origin, old_msg->dest,
                                      old_msg->strc, (const char **)old_msg->strv,
                                      old_msg->uint32c, old_msg->uint32v,
                                      old_msg->sint32c, old_msg->sint32v);
}

/** \brief Deallocate a message
 * \param msg message to deallocate
 *
 * All memory used by the message is freed. After this function
 * returns, the struct is invalid and should not be referred to
 * again.
 */
DEFINE_SYMBOL(void, message, free_message, (struct message *msg))
{
  int i;
  if (!msg)
    return;

  xfree(msg->origin);
  xfree(msg->dest);
  for (i = 0; i < msg->strc; i++)
    xfree(msg->strv[i]);
  xfree(msg->strv);
  xfree(msg->uint32v);
  xfree(msg->sint32v);
  
  xfree(msg);
}

/** \brief Translate message names to types
 * \param name message type to look up
 * \return integer value of message type, or -1 if invalid
 *
 * All message types will be positive integers. A given type name will
 * always have the same value while the protocol interface module is
 * loaded.
 *
 * \note Message type names will be forgotten if the protocol
 * interface module is unloaded
 */
DEFINE_SYMBOL(int, message, get_message_type, (const char *name))
{
  struct message_type *type;

  if (!name || !*name)
    return -1;

  if (!message_type_hash)
    return -1;

  if ((type = hashtable_lookup(message_type_hash, name)))
    {
      return type->id;
    }
  return -1;
}

/** \brief Register a new message type name
 * \param name message type to register
 * \return integer value of message type
 *
 * This behaves the same as get_message_type, but will attempt to
 * create a new message type if the name is not known. It will return
 * -1 if it was unable to create the message type.
 *
 * All message types are guaranteed to be positive and fit into an
 * int. Message types are invariant while the protocol module is
 * loaded, but are not preserved when it is unloaded. A module may
 * assume message types do not change if it depends on the protocol
 * interface module.
 */
DEFINE_SYMBOL(int, message, register_message_type, (const char *name))
{
  struct message_type *type;

  if (!name || !*name)
    return -1;

  if (!message_type_array)
    {
      int i;
      message_type_array = xmalloc(MESSAGE_TYPE_ALLOCATE_CHUNK * sizeof(struct message_type));
      message_types_allocated = MESSAGE_TYPE_ALLOCATE_CHUNK;
      for (i = 0; i < message_types_allocated; i++)
        {
          message_type_array[i].id = -1;
          message_type_array[i].name = NULL;
          message_type_array[i].handlers = NULL;
        }
    }
  if (!message_type_hash)
    message_type_hash = new_hashtable(&hash_ascii, 107);

  if ((type = hashtable_lookup(message_type_hash, name)))
    return type->id;

  if (message_types_allocated == message_types_registered)
    {
      int allocate_chunk = MESSAGE_TYPE_ALLOCATE_CHUNK;
      struct message_type *new_array;
      int i;

      new_array = xmalloc(sizeof(struct message_type) * (message_types_allocated + allocate_chunk));
      memcpy(new_array, message_type_array, sizeof(struct message_type) * (message_types_allocated));
      xfree(message_type_array);
      message_type_array = new_array;
      /* Update the hash table to point at the new structure */
      for (i = 0; i < message_types_registered; i++)
        {
          hashtable_insert(message_type_hash, message_type_array[i].name, &message_type_array[i]);
        }
      for (i = message_types_allocated; i < message_types_allocated + allocate_chunk; i++)
        {
          message_type_array[i].id = -1;
          message_type_array[i].name = NULL;
          message_type_array[i].handlers = NULL;
        }
      message_types_allocated += allocate_chunk;
    }

  message_type_array[message_types_registered].id = message_types_registered;
  message_type_array[message_types_registered].name = xstrdup(name);
  hashtable_insert(message_type_hash, name, &message_type_array[message_types_registered]);
  return message_types_registered++;
}

/** \brief Get the name of a message type
 * \param type integer value of message type
 * \return name of message type, or NULL if invalid
 */
DEFINE_SYMBOL(const char *, message, get_message_type_name, (int type))
{
  if (!message_type_array)
    return NULL;
  if ((type < 0) || (type >= message_types_registered))
    return NULL;
  if (message_type_array[type].id == -1)
    return NULL;
  return message_type_array[type].name;
}

/** \brief Allocate and populate a message
 * \param type message type, as returned by get_message_type
 * \param origin pointer to a single string that is the origin of the message
 * \param dest pointer to a single string that is the target of the message
 * \param strc size of strv array
 * \param strv array of string components
 * \param uint32c size of uint32v array
 * \param uint32v array of unsigned 32-bit integer components
 * \param sint32c size of sint32v array
 * \param sint32v array of signed 32-bit integer components
 * \return newly allocated message populated with the values given as parameters, or NULL on error
 *
 * A new message is allocated and filled with the given data. This is
 * the only way to allocate a new message.
 *
 * String members are deep-copied. Cacheable string members are
 * shallow-copied, since they are allocated by this module anyway.
 */
DEFINE_SYMBOL(struct message *, message, build_message, (int type,
                                                         const char *origin, const char *dest,
                                                         int strc, const char *const *strv,
                                                         int uint32c, uint32_t *uint32v,
                                                         int sint32c, int32_t *sint32v))
{
  struct message *msg;
  int i;

  if (type < 0)
    return NULL;

  /* These are of the form:
   *
   * if (strc > 0 && !strv)
   *   return NULL;
   */
#define CHECK_ARG(NAME) if (NAME ## c > 0 && !NAME ## v) return NULL
  CHECK_ARG(str);
  CHECK_ARG(uint32);
  CHECK_ARG(sint32);
#undef CHECK_ARG

  msg = xmalloc(sizeof(struct message));
  msg->type = type;
  msg->origin = xstrdup(origin);
  msg->dest = xstrdup(dest);
  msg->strv = strc ? xmalloc(strc * sizeof(char *)) : NULL;
  msg->uint32v = uint32c ? xmalloc(uint32c * sizeof(uint32_t)) : NULL;
  msg->sint32v = sint32c ? xmalloc(sint32c * sizeof(int32_t)) : NULL;

  if ((msg->strc = strc))
    for (i = 0; i < strc; i++)
      msg->strv[i] = xstrdup(strv[i]);

  if ((msg->uint32c = uint32c))
    memcpy(msg->uint32v, uint32v, uint32c * sizeof(uint32_t));

  if ((msg->sint32c = sint32c))
    memcpy(msg->sint32v, sint32v, sint32c * sizeof(int32_t));

  return msg;
}

struct message_handler
{
  int type;
  message_handler_func *f;
  void *data;
  struct module *module;
  struct llist_node *node;
};

static int
match_message_handler(const void *d1, const void *d2)
{
  const struct message_handler *mh1 = d1, *mh2 = d2;
  return (mh1->f == mh2->f) && (mh1->data == mh2->data) && (mh1->type == mh2->type) && (mh1->module == mh2->module);
}

static void
deregister_handler(void *d)
{
  struct message_handler *mh = d;
  delete_llist_node(mh->node);
  xfree(mh);
}

static void
free_message_handlers(struct session *session)
{
  struct llist_node *n, *next;
  if (!session || !session->handlers)
    return;
  for (n = llist_head(session->handlers); n; n = next)
    {
      struct message_handler *mh = llist_node_data(n);
      next = llist_node_next(n);
      deregister_module_unload_handler(MYHANDLE, mh->module, &deregister_handler, mh);
      xfree(mh);
      delete_llist_node(n);
    }
  free_llist(session->handlers);
  session->handlers = NULL;
}

static struct llist_node **next_message_handler;

static void
process_message(struct session *session, struct message *msg)
{
  int type;
  if (!session)
    return;

  if (msg)
    type = msg->type;
  else
    type = MESSAGE_TYPE_SHUTDOWN;

  if (type > message_types_registered)
    {
      log(MYNAME, LOG_PRIORITY_ERROR, "Attempt to accept message with invalid type %d (registered %d)",
          type, message_types_registered);
      return;
    }

  /* Now, this is a little complicated. We iterate over all the
   * message handlers, and call the ones that match.  Unfortunately, a
   * message handler might call protocol::shutdown_session, which can
   * cause session->buf and session itself to be immediately freed.
   *
   * In order to prevent this happening, there is code in
   * shutdown_session and close_session to delay them when
   * session->refcount > 0.
   *
   * However, we must also stop calling message handlers as soon as
   * shutdown_session is called. The exception is the NULL message,
   * which will always be handled - it's sent by close_session itself,
   * to signal that the session really is going away now and
   * everything should forget about it.
   *
   * If one of the ignore_*_from_* functions is called, it will check
   * next_message_handler to see if that is the one it is deleting -
   * if it is, then it will update next_message_handler to point to
   * the following node.
   */
  session->refcount++;
  if ((type >= 0) && message_type_array[type].handlers)
    {
      struct llist_node *n, *next = NULL;
      next_message_handler = &next;
      for (n = llist_head(message_type_array[type].handlers); n && (!msg || !session->shutting_down);
           n = next)
        {
          next = llist_node_next(n);
          struct message_handler *mh = llist_node_data(n);
          (*mh->f)(session, msg, mh->data);
        }
    }
  next_message_handler = NULL;

  if ((type < -1) && message_type_array_negative[(-type) - 2].handlers)
    {
      struct llist_node *n, *next = NULL;
      next_message_handler = &next;
      for (n = llist_head(message_type_array_negative[(-type) - 2].handlers); n && (!msg || !session->shutting_down);
           n = next)
        {
          next = llist_node_next(n);
          struct message_handler *mh = llist_node_data(n);
          (*mh->f)(session, msg, mh->data);
        }
    }
  next_message_handler = NULL;

  if (session->handlers)
    {
      struct llist_node *n, *next = NULL;
      next_message_handler = &next;
      for (n = llist_head(session->handlers); n && (!msg || !session->shutting_down); n = next)
        {
          next = llist_node_next(n);
          struct message_handler *mh = llist_node_data(n);
          if ((mh->type == -1) || (mh->type == type))
            (*mh->f)(session, msg, mh->data);
        }
    }
  next_message_handler = NULL;

  session->refcount--;
  if (!msg)
    free_message_handlers(session);
  
  if (session->refcount == 0 && session->shutting_down)
    LSYM(protocol, shutdown_session)(session);
}

static int processing_message = 0;
static struct llist *pending_messages = NULL;

struct pending_message
{
  struct session *session;
  struct message *msg;
};

DEFINE_SYMBOL(void, message, accept_message, (struct session *session, struct message *msg))
{
  /* Don't accept any messages for a session once it has started
   * closing, other than the final NULL
   */
  if (session->closed && msg != NULL)
    return;
  /* Don't let messages recurse, queue them up and let the current one
   * unwind before firing off the next. This simplifies a lot of code.
   */
  if (processing_message)
    {
      struct pending_message *pm = xmalloc(sizeof(struct pending_message));
      if (pending_messages == NULL)
        pending_messages = new_llist();
      pm->session = session;
      pm->msg = msg;
      /* Hang on to this session, we still have messages for it */
      session->refcount++;
      add_llist_tail(pending_messages, pm);
    }
  else
    {
      processing_message = 1;
      process_message(session, msg);
      while (pending_messages && llist_head(pending_messages))
        {
          struct pending_message *pm = llist_node_data(llist_head(pending_messages));
          delete_llist_node(llist_head(pending_messages));
          /* We no longer need to hold this session open; if this
           * reduces refcount to zero, process_message will shut it
           * down when it's done
           */
          pm->session->refcount--;
          process_message(pm->session, pm->msg);
        }
      processing_message = 0;
    }
}

DEFINE_SYMBOL(void, message, request_all_from_session, (struct module *module, struct session *session,
                                                        message_handler_func *f, void *data))
{
  struct message_handler *mh;
  if (!session || !f)
    return;

  mh = xmalloc(sizeof(struct message_handler));
  mh->type = -1;
  mh->f = f;
  mh->data = data;
  mh->module = module;

  if (!session->handlers)
    session->handlers = new_llist();

  add_llist_head(session->handlers, mh);
  mh->node = llist_head(session->handlers);
  if (module)
    register_module_unload_handler(MYHANDLE, module, &deregister_handler, mh);
}

DEFINE_SYMBOL(void, message, ignore_all_from_session, (struct module *module, struct session *session,
                                                         message_handler_func *f, void *data))
{
  struct message_handler mh;
  struct llist_node *n;

  if (!session || !session->handlers || !f)
    return;

  mh.type = -1;
  mh.f = f;
  mh.data = data;
  mh.module = module;

  if ((n = find_llist_match(session->handlers, &match_message_handler, &mh)))
    {
      struct message_handler *mh2 = llist_node_data(n);
      if (module)
        deregister_module_unload_handler(MYHANDLE, mh2->module, &deregister_handler, mh2);

      /* If necessary, poke the loop in process_message() to skip this
       * item, since we just deleted it
       */
      if (next_message_handler && n == *next_message_handler)
        *next_message_handler = llist_node_next(n);

      delete_llist_node(n);
    }
}

DEFINE_SYMBOL(void, message, request_type_from_session, (struct module *module, struct session *session, int type,
                                                           message_handler_func *f, void *data))
{
  struct message_handler *mh;
  if (!session || !f || type == -1)
    return;

  mh = xmalloc(sizeof(struct message_handler));
  mh->type = type;
  mh->f = f;
  mh->data = data;
  mh->module = module;

  if (!session->handlers)
    session->handlers = new_llist();

  add_llist_head(session->handlers, mh);
  mh->node = llist_head(session->handlers);
  if (module)
    register_module_unload_handler(MYHANDLE, module, &deregister_handler, mh);
}

DEFINE_SYMBOL(void, message, ignore_type_from_session, (struct module *module, struct session *session, int type,
                                                          message_handler_func *f, void *data))
{
  struct message_handler mh;
  struct llist_node *n;

  if (!session || !session->handlers || !f || type == -1)
    return;

  mh.type = type;
  mh.f = f;
  mh.data = data;
  mh.module = module;

  if ((n = find_llist_match(session->handlers, &match_message_handler, &mh)))
    {
      struct message_handler *mh2 = llist_node_data(n);
      if (module)
        deregister_module_unload_handler(MYHANDLE, mh2->module, &deregister_handler, mh2);

      /* If necessary, poke the loop in process_message() to skip this
       * item, since we just deleted it
       */
      if (next_message_handler && n == *next_message_handler)
        *next_message_handler = llist_node_next(n);

      delete_llist_node(n);
    }
}

DEFINE_SYMBOL(void, message, request_type_from_all, (struct module *module, int type,
                                                     message_handler_func *f, void *data))
{
  struct message_handler *mh;
  struct message_type *mt;

  if (!f || type == -1 || type >= message_types_registered)
    return;

  mt = (type < 0) ? (&message_type_array_negative[(-type) - 2]) : &message_type_array[type];

  mh = xmalloc(sizeof(struct message_handler));
  mh->type = type;
  mh->f = f;
  mh->data = data;
  mh->module = module;

  if (!mt->handlers)
    mt->handlers = new_llist();

  add_llist_head(mt->handlers, mh);
  mh->node = llist_head(mt->handlers);
  if (module)
    register_module_unload_handler(MYHANDLE, module, &deregister_handler, mh);
}

DEFINE_SYMBOL(void, message, ignore_type_from_all, (struct module *module, int type,
                                                    message_handler_func *f, void *data))
{
  struct message_handler mh;
  struct llist_node *n;
  struct message_type *mt;

  if (!f || type == -1 || type >= message_types_registered)
    return;

  mt = (type < 0) ? (&message_type_array_negative[(-type) - 2]) : &message_type_array[type];
  if (!mt->handlers)
    return;

  mh.type = type;
  mh.f = f;
  mh.data = data;
  mh.module = module;

  if ((n = find_llist_match(mt->handlers, &match_message_handler, &mh)))
    {
      struct message_handler *mh2 = llist_node_data(n);
      if (module)
        deregister_module_unload_handler(MYHANDLE, mh2->module, &deregister_handler, mh2);

      /* If necessary, poke the loop in process_message() to skip this
       * item, since we just deleted it
       */
      if (next_message_handler && n == *next_message_handler)
        *next_message_handler = llist_node_next(n);

      delete_llist_node(n);
    }
}
