/************************************************************************
 *   This is part of Dancer, a text relay daemon
 *   Copyright (C) 2002 Andrew Suffield <asuffield@users.sourceforge.net>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

#ifdef MESSAGE_H

struct message;

#else
#define MESSAGE_H
#include <dancerd/config.h>
#include <dancerd/modules.h>
#include <dancerd/protocol/protocol.h>

#include <stdint.h>

/** \file protocol/message.h
 * \brief Protocol messages
 */

/** \defgroup message Protocol messages
 * \ingroup protocol
 * The protocol layer encodes/decodes messages from sessions; messages
 * are the only form of communication that the higher layers will see
 *
 * @{
 */

/** \brief Message structure
 *
 * This is the exported form of messages that other modules can work
 * with. It consists of a message type, and a series of
 * components. Sequence is preserved between components of the same
 * type; there is no ordering between two components of different
 * types, and they may be transmitted in a different sequence as long
 * as they reassemble into an equivalent struct on the remote side.
 *
 * Each group of same-typed components is stored in a count/array
 * pair, like argc/argv; indexes into the array equal to or greater
 * than the count are undefined and will probably segfault if
 * accessed.
 */
struct message
{
  /** Type number for this message (from protocol/get_message_type()) */
  int type;
  /** Name of the origin of this message */
  char *origin;
  /** Name of the destination of this message */
  char *dest;
  /** Number of elements in strv */
  int strc;
  /** Array of string components of this message, in order */
  char **strv;
  /** Number of elements in uint32v */
  int uint32c;
  /** Array of unsigned 32-bit integer components of this message, in order */
  uint32_t *uint32v;
  /** Number of elements in sint32v */
  int sint32c;
  /** Array of signed 32-bit integer components of this message, in order */
  int32_t *sint32v;
};

/** A message handler is called with the session that the message came
 * from, the message itself, and the arbitrary void* given when the
 * handler was registered
 */
typedef void message_handler_func(struct session *, const struct message *, void *);

/** A handler for MESSAGE_TYPE_SHUTDOWN is invoked with a NULL message
 * argument, when the session is being shut down
 */
#define MESSAGE_TYPE_SHUTDOWN -2

DEFINE_NAMESPACE(message, 0);

/* These functions do stuff to struct message */
DECLARE_SYMBOL(struct message *, message, dup_message, (const struct message *));
DECLARE_SYMBOL(void, message, free_message, (struct message *));
DECLARE_SYMBOL(int, message, get_message_type, (const char *));
DECLARE_SYMBOL(int, message, register_message_type, (const char *));
DECLARE_SYMBOL(const char *, message, get_message_type_name, (int));
DECLARE_SYMBOL(struct message *, message, build_message, (int, const char *, const char *,
                                                          int, const char *const *,
                                                          int, uint32_t *,
                                                          int, int32_t *));

/* This is called when a message is to be passed up to the message handlers */
DECLARE_SYMBOL(void, message, accept_message, (struct session *, struct message *));

/* These are used to (de)register message handlers */
DECLARE_SYMBOL(void, message, request_all_from_session, (struct module *, struct session *, message_handler_func *, void *));
DECLARE_SYMBOL(void, message, ignore_all_from_session, (struct module *, struct session *, message_handler_func *, void *));
DECLARE_SYMBOL(void, message, request_type_from_session, (struct module *, struct session *, int, message_handler_func *, void *));
DECLARE_SYMBOL(void, message, ignore_type_from_session, (struct module *, struct session *, int, message_handler_func *, void *));
DECLARE_SYMBOL(void, message, request_type_from_all, (struct module *, int, message_handler_func *, void *));
DECLARE_SYMBOL(void, message, ignore_type_from_all, (struct module *, int, message_handler_func *, void *));

END_NAMESPACE(message);

/** @}
 */

#endif
