/************************************************************************
 *   This is part of Dancer, a text relay daemon
 *   Copyright (C) 2002 Andrew Suffield <asuffield@users.sourceforge.net>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

#include <dancerd/config.h>
#include <dancerd/config/items.h>
#include <dancerd/event.h>
#include <dancerd/log.h>
#include <dancerd/protocol/message.h>
#include <dancerd/modules.h>
#include <dancerd/protocol/protocol.h>
#include <dancerd/protocol/protocol_internal.h>
#include <dancerd/protocol/scheduler.h>
#include <dancerd/transport/transport.h>
#include <dancerd/xmalloc.h>
#include <dancerd/xstring.h>

#include <assert.h>
#include <stddef.h>
#include <string.h>

SETUP_MODULE("protocol", MODVERSION);

static int max_unknown_send_queue = 20;

IMPORTED_SYMBOL(transport, shutdown_connection);
IMPORTED_SYMBOL(transport, abort_connection);
IMPORTED_SYMBOL(transport, write_data);
IMPORTED_SYMBOL(transport, get_connection_meta);
IMPORTED_SYMBOL(config, get_item);

#include "symtab"

static struct llist *protocols = NULL;
static struct llist *sessions = NULL;
static struct llist *new_session_handlers = NULL;

struct message_queue
{
  struct llist *queue;
  unsigned int length, limit;
};

struct new_session_handler
{
  new_session_handler *nsh;
  void *data;
  struct module *module;
};

static int
match_protocol(const void *d1, const void *d2)
{
  const struct protocol *p = d1;
  const char *s = d2;
  return strcmp(p->name, s) == 0;
}

static struct protocol *
find_protocol(const char *name)
{
  struct llist_node *n;
  if (!name)
    return NULL;
  n = find_llist_match(protocols, &match_protocol, name);
  return n ? llist_node_data(n) : NULL;
}

static void
free_protocol(struct protocol *proto)
{
  struct llist_node *n, *next;
  if (!proto)
    return;

  if (proto->sessions)
    for (n = llist_head(proto->sessions); n; n = next)
      {
        struct session *session = llist_node_data(n);
        next = llist_node_next(n);
        SYM(transport, abort_connection)(session->conn);
      }
  free_llist(proto->sessions);
  xfree(proto->name);
  xfree(proto);
}

static struct protocol *
get_protocol(const char *name)
{
  struct protocol *proto = find_protocol(name);
  if (proto)
    return proto;
  proto = xmalloc(sizeof(struct protocol));
  proto->name = xstrdup(name);
  proto->pi = NULL;
  proto->pis = NULL;
  proto->prd = NULL;
  proto->pcs = NULL;
  proto->sessions = new_llist();
  proto->active = 0;
  add_llist_head(protocols, proto);
  return proto;
}

static inline struct message_queue *
new_message_queue(unsigned int limit)
{
  struct message_queue *mq;
  if (!limit)
    return NULL;
  mq = xmalloc(sizeof(struct message_queue));
  mq->queue = new_llist();
  mq->length = 0;
  mq->limit = limit;
  return mq;
}

static inline int
message_queue_add(struct message_queue *mq, struct message *msg)
{
  if (mq->length >= mq->limit)
    return 0;
  add_llist_tail(mq->queue, msg);
  mq->length++;
  return 1;
}

static inline struct message *
message_queue_extract(struct message_queue *mq)
{
  struct message *msg;
  if (!mq->length)
    return NULL;
  msg = llist_node_data(llist_head(mq->queue));
  delete_llist_node(llist_head(mq->queue));
  mq->length--;
  return msg;
}

static inline void
destroy_message_queue(struct message_queue *mq)
{
  if (!mq)
    return;
  destroy_llist(mq->queue, LSYM(message, free_message));
  xfree(mq);
}

static void
dequeue_messages(struct session *session)
{
  struct message *msg;
  if (!session->proto || !session->proto->active || !session->proto->psm)
    return;
  while ((msg = message_queue_extract(session->send_queue)))
    {
      (*session->proto->psm)(session, session->data, msg);
      LSYM(message, free_message)(msg);
    }
}

static void
invoke_init_session_handlers(void *d1, void *d2)
{
  struct session *session = d1;
  struct protocol *proto = d2;
  /* If the session is already established, reinitialise it */
  if (session->established)
    {
      if (proto->pis)
        session->data = (*proto->pis)(session, session->data);
      dequeue_messages(session);
      if (dbuffer_len(session->buf) && proto->prd)
        (*session->proto->prd)(session, session->data, session->buf, NULL, 0);
    }
  else
    {
      /* Otherwise it's a new outgoing connection, so start it as such */
      session->established = 1;
      if (proto->pcr)
        session->data = (*proto->pcr)(session);
      dequeue_messages(session);
    }
}

static void
unregister_protocol_(void *d)
{
  struct protocol *proto = d;
  /* Don't call the handlers, because the module might not be loaded */  
  proto->active = 0;
  proto->pi = NULL;
  proto->pis = NULL;
  proto->prd = NULL;
  proto->psm = NULL;
  proto->pcs = NULL;
}

DEFINE_SYMBOL(struct protocol *, protocol, register_protocol, (const char *name, struct module *module,
                                                               protocol_identify_func *pi,
                                                               protocol_init_session_func *pis,
                                                               protocol_connection_ready_func *pcr,
                                                               protocol_read_data_func *prd,
                                                               protocol_send_message_func *psm,
                                                               protocol_close_session_func *pcs))
{
  struct protocol *proto = get_protocol(name);

  assert(proto);

  if (proto->active)
    {
      log(MYNAME, LOG_PRIORITY_ERROR, "Attempt to re-register active protocol %s", name);
      return NULL;
    }

  proto->module = module;
  proto->pi = pi;
  proto->pis = pis;
  proto->pcr = pcr;
  proto->prd = prd;
  proto->psm = psm;
  proto->pcs = pcs;
  proto->active = 1;

  if (module)
    register_module_unload_handler(MYHANDLE, module, &unregister_protocol_, proto);

  foreach_llist(proto->sessions, &invoke_init_session_handlers, (void*)proto);

  return proto;
}

DEFINE_SYMBOL(void, protocol, unregister_protocol, (struct protocol *proto))
{
  if (!proto)
    return;

  if (!proto->active)
    {
      log(MYNAME, LOG_PRIORITY_ERROR, "Attempt to de-register inactive protocol %s", proto->name);
      return;
    }

  if (proto->module)
    deregister_module_unload_handler(MYHANDLE, proto->module, &unregister_protocol_, proto);
  proto->module = NULL;

  proto->active = 0;
  proto->pi = NULL;
  proto->pis = NULL;
  proto->prd = NULL;
  proto->psm = NULL;
  proto->pcs = NULL;
}

/** \brief Calls any new session handlers on this session
 * \param session The new session
 * \param outgoing True if this is an outgoing connection
 * \return true if successful, false if the session was terminated
 */
static int
call_new_session_handlers(struct session *session, int outgoing)
{
  struct llist_node *n2, *next2;
  for (n2 = llist_head(new_session_handlers); n2; n2 = next2)
    {
      struct new_session_handler *nsh = llist_node_data(n2);
      next2 = llist_node_next(n2);
      if (!(*nsh->nsh)(session, nsh->data, outgoing))
        {
          /* This means the handler wants us to kill off the session */
          LSYM(protocol, shutdown_session)(session);
          return 0;
        }
    }
  return 1;
}

static void
try_identify_session(struct session *session)
{
  struct llist_node *n, *next;
  for (n = llist_head(session->candidate_protocols); n; n = next)
    {
      struct protocol *proto = llist_node_data(n);
      next = llist_node_next(n);
      if (!proto->active)
        {
          delete_llist_node(n);
          continue;
        }
      switch ((*proto->pi)(session->buf))
        {
        case 1:
          /* This protocol accepts the session. We're done identifying */
          free_llist(session->candidate_protocols);
          session->candidate_protocols = NULL;
          if (!call_new_session_handlers(session, 0))
            return;

          session->proto = proto;
          session->established = 1;
          add_llist_head(proto->sessions, session);
          session->proto_sessions_node = llist_head(proto->sessions);
          if (proto->pis)
            session->data = (*proto->pis)(session, session->data);
          dequeue_messages(session);
          if (proto->prd)
            (*proto->prd)(session, session->data, session->buf, NULL, 0);
          /* Don't drop through to the llist_head(session->candidate_protocols) call below */
          return;
        case 0:
          /* This protocol is not yet sure, skip it for now */
          break;
        case -1:
          /* This protocol does not match, drop it from the list */
          delete_llist_node(n);
          break;
        default:
          log(MYNAME, LOG_PRIORITY_ERROR, "Protocol %s returned invalid value from identity function", proto->name);
          break;
        }
    }
  if (!llist_head(session->candidate_protocols))
    {
      log(MYNAME, LOG_PRIORITY_INFO, "Dropping session %s, unrecognised protocol", "MOOOOOOOOOOOOOOOO");
      SYM(transport, abort_connection)(session->conn);
    }
}

static struct session *
new_session(void)
{
  struct session *session;
  session = xmalloc(sizeof(struct session));
  session->conn = NULL;
  session->buf = NULL;
  session->data = NULL;
  session->proto = NULL;
  session->candidate_protocols = dup_llist(protocols);
  session->send_queue = new_message_queue(max_unknown_send_queue);
  session->sched = NULL;
  session->proto_sessions_node = NULL;
  session->handlers = NULL;
  session->metadata = new_hashtable(&hash_ascii, 17);
  session->attributes = NULL;
  session->shutting_down = 0;
  session->established = 0;
  session->attributes_allocated = 0;
  session->closed = 0;
  session->refcount = 0;
  add_llist_tail(sessions, session);
  session->sessions_entry = llist_tail(sessions);
  return session;
}

DEFINE_SYMBOL(struct session *, protocol, new_session, (struct connection *conn, struct dbuffer *buf))
{
  struct session *session;

  if (!conn)
    return NULL;

  session = new_session();
  session->conn = conn;
  session->buf = buf;

  return session;
}

DEFINE_SYMBOL(void, protocol, data_read, (struct session *session, const void *data, size_t len))
{
  if (len <= 0)
    {
      len = 0;
      data = NULL;
    }

  session->refcount++;
  if (!session->proto)
    try_identify_session(session);
  else
    {
      if (session->proto->active && session->proto->prd)
        (*session->proto->prd)(session, session->data, session->buf, data, len);
    }
  session->refcount--;
  if (!session->refcount && session->shutting_down)
    LSYM(protocol, shutdown_session)(session);
}

DEFINE_SYMBOL(void, protocol, close_session, (struct session *session))
{
  if (!session)
    return;
  session->conn = NULL;
  session->buf = NULL;
  /* Similar to shutdown_session, we can't actually close the session
   * while we're inside a handler that might want to operate on it.
   *
   * This case only happens when the connection is aborted; if it is
   * shutdown cleanly, then the code in shutdown_session will catch
   * it.
   *
   * Note that once the connection has been aborted, session->buf is
   * invalid. Also note that after calling this function, no further
   * messages will be accepted for this session.
   */
  if (!session->closed)
    {
      /* Set this so that message::accept_message passes control to
       * protocol::shutdown_session, which will come back to us here
       */
      session->shutting_down = 1;
      session->closed = 1;
      /* We don't want the trailing NULL message to trigger a
       * re-entrance into this function - that would cause a
       * segfault. So we claim a reference on it here; if refcount is
       * 0, then we'll delete the session right afterwards
       */
      session->refcount++;
      LSYM(message, accept_message)(session, NULL);
      session->refcount--;
    }
  if (session->refcount > 0)
    return;
  if (session->proto && session->proto->active && session->proto->pcs)
    (*session->proto->pcs)(session, session->data);
  if (session->proto_sessions_node)
    delete_llist_node(session->proto_sessions_node);
  free_llist(session->candidate_protocols);
  destroy_message_queue(session->send_queue);
  free_hashtable(session->metadata);
  delete_llist_node(session->sessions_entry);
  xfree(session->attributes);
  xfree(session);
}

DEFINE_SYMBOL(void, protocol, shutdown_session, (struct session *session))
{
  if (!session)
    return;

  /* If we got here but the session is tagged as closed, then we came
   * from message::accept_message and should just hand over to
   * close_session.
   *
   * This sequence occurs when the connection has been aborted
   */
  if (session->closed)
    {
      LSYM(protocol, close_session)(session);
      return;
    }

  /* Note that we want to shut down the session */
  session->shutting_down = 1;
  /* But don't actually do it if the session is currently in use.
   *
   * This occurs when the session is shut down from within a message
   * or data_read handler. We must delay the shutdown until the
   * handler returns, so we'll stop here and check shutting_down at
   * the end of those two routines
   */
  if (session->refcount > 0)
    return;
  SYM(transport, shutdown_connection)(session->conn);
}

DEFINE_SYMBOL(void *, protocol, get_session_meta, (struct session *session, const char *key))
{
  void *data;
  if (!session || !key)
    return NULL;

  data = session->metadata ? hashtable_lookup(session->metadata, key) : NULL;
  return data ? data : SYM(transport, get_connection_meta)(session->conn, key);
}

DEFINE_SYMBOL(void, protocol, set_session_meta, (struct session *session, const char *key, void *data))
{
  if (!session || !key || !session->metadata)
    return;

  hashtable_insert(session->metadata, key, data);
}

/** \brief Pass a received message into the protocol layer
 * \param session the session that received the message
 * \param msg the message that was received
 *
 * A protocol implementation module should generate messages with
 * build_message, then pass them here. The module should then forget
 * about the message and not refer to it again; it will be managed and
 * freed by the protocol interface once it is no longer needed
 *
 * \internal
 * The message is handed off to a scheduler if one is present,
 * otherwise it is immediately accepted and freed.
 */
DEFINE_SYMBOL(void, protocol, read_message, (struct session *session,
					     struct message *msg))
{
  if (session->shutting_down)
    {
      LSYM(message, free_message)(msg);
    }
  else if (session->sched)
    {
      LSYM(scheduler, queue_message)(session, session->sched, msg);
    }
  else
    {
      LSYM(message, accept_message)(session, msg);
    }
}

/** \brief Pass a message to send into the protocol layer
 * \param session the session to send the message to
 * \param msg the message to send
 *
 * A message should be generated with build_message and then passed
 * here; this function will then take control of the message and free
 * it when appropriate. The caller should not refer to the message
 * again after passing it here.
 */
DEFINE_SYMBOL(void, protocol, send_message, (struct session *session,
                                             struct message *msg))
{
  if (session->shutting_down)
    return;

  /* If we can't send yet, queue the message */
  if (!session->proto || !session->proto->active)
    {
      if (!message_queue_add(session->send_queue, msg))
        LSYM(protocol, shutdown_session)(session);
      return;
    }

  if (session->proto->psm)
    (*session->proto->psm)(session, session->data, msg);
  LSYM(message, free_message)(msg);
}

static void
unregister_new_session_handler_(void *d)
{
  struct new_session_handler *nsh = d;
  /* Unset this so it doesn't try to unregister the module unload handler */
  nsh->module = NULL;
  LSYM(protocol, unregister_new_session_handler)(NULL, nsh->nsh, nsh->data);
}

DEFINE_SYMBOL(void, protocol, register_new_session_handler, (struct module *module,
                                                             new_session_handler *nsh, void *data))
{
  struct new_session_handler *n;
  if (!nsh)
    return;
  n = xmalloc(sizeof(struct new_session_handler));
  n->nsh = nsh;
  n->data = data;
  n->module = module;
  add_llist_head(new_session_handlers, n);
  if (module)
    register_module_unload_handler(MYHANDLE, module, &unregister_new_session_handler_, n);
}

static int
match_new_session_handler(const void *d1, const void *d2)
{
  const struct new_session_handler *n1 = d1, *n2 = d2;
  return (n1->nsh == n2->nsh) && (n1->data == n2->data) && (n1->module == n2->module);
}

DEFINE_SYMBOL(void, protocol, unregister_new_session_handler, (struct module *module,
                                                               new_session_handler *nsh, void *data))
{
  struct new_session_handler n;
  struct llist_node *node;
  n.nsh = nsh;
  n.data = data;
  n.module = module;
  node = find_llist_match(new_session_handlers, &match_new_session_handler, &n);
  if (node)
    {
      struct new_session_handler *n2 = llist_node_data(node);
      if (n2->module)
        deregister_module_unload_handler(MYHANDLE, module, &unregister_new_session_handler_, n2);
      xfree(n2);
      delete_llist_node(node);
    }
}

DEFINE_SYMBOL(void, protocol, write_data, (struct session *session, const void *data, size_t len))
{
  SYM(transport, write_data)(session->conn, data, len);
}

DEFINE_SYMBOL(struct session *, protocol, begin_connection, (const char *protocol,
                                                             const char *transport, const char *class,
                                                             const char *target, void **argv, int argc))
{
  struct session *session;
  struct connection *conn;
  struct protocol *proto;

  if (!protocol || !transport || !class)
    return NULL;

  if (!(proto = find_protocol(protocol)))
    {
      log(MYNAME, LOG_PRIORITY_ERROR, "Invalid protocol name '%s' when trying to establish connection",
          protocol);
      return NULL;
    }

  if (!proto->pcr)
    {
      log(MYNAME, LOG_PRIORITY_ERROR, "Protocol '%s' does not support outgoing connections",
          protocol);
      return NULL;
    }

  session = new_session();
  session->proto = proto;

  if (!(conn = SYM(transport, begin_connection)(session, transport, class, target, argv, argc)))
    return NULL;

  session->conn = conn;

  return session;
}

DEFINE_SYMBOL(void, protocol, connection_ready, (struct session *session, struct dbuffer *buf))
{
  /* We can't set up a new connection when the protocol layer isn't
   * active, so don't bother trying.
   *
   * FIXME: This should probably be buffered until the protocol
   * is active again, deal with that later
   */
  if (!session->proto->active)
    {
      LSYM(protocol, shutdown_session)(session);
      return;
    }
  session->buf = buf;
  add_llist_head(session->proto->sessions, session);
  session->proto_sessions_node = llist_head(session->proto->sessions);
  session->refcount++;
  session->established = 1;
  session->data = (*session->proto->pcr)(session);
  dequeue_messages(session);
  call_new_session_handlers(session, 1);
  (*session->proto->prd)(session, session->data, session->buf, NULL, 0);
  session->refcount--;
  if (!session->refcount && session->shutting_down)
    LSYM(protocol, shutdown_session)(session);
}

static int attributes = 0;
static const char **attribute_keys = NULL;

DEFINE_SYMBOL(int, protocol, register_attribute, (const char *key))
{
  if (attribute_keys)
    {
      int i;
      for (i = 0; i < attributes; i++)
        if (strcmp(attribute_keys[i], key) == 0)
          return i;
    }
  attribute_keys = xrealloc(attribute_keys, sizeof(char *) * (attributes + 1));
  attribute_keys[attributes] = xstrdup(key);
  return attributes++;
}

DEFINE_SYMBOL(void *, protocol, get_attribute, (struct session *session, int i))
{
  if (!session || i < 0)
    return NULL;

  if (i >= session->attributes_allocated)
    return NULL;

  return session->attributes[i];
}

DEFINE_SYMBOL(void, protocol, set_attribute, (struct session *session, int i, void *data))
{
  if (!session || i < 0)
    return;

  if (i >= session->attributes_allocated)
    {
      session->attributes = xrealloc(session->attributes, sizeof(void *) * (i + 1));
      session->attributes_allocated = i + 1;
    }

  session->attributes[i] = data;
}

static void
read_config(void)
{
  struct config_item item;
  static char key_name[] = "protocol.max-unknown-send-queue";

  SYM(config, get_item)(key_name, &item);
  if (item.type != citem_uint32)
    {
      log(MYNAME, LOG_PRIORITY_ERROR, "config key %s is not an unsigned integer type", key_name);
      max_unknown_send_queue = 4;
    }
  else
    max_unknown_send_queue = item.uint32;
}

static void
read_config_event(struct event *e unused_attribute)
{
  read_config();
}

struct reload_data
{
  struct llist *protocols, *sessions, *new_session_handlers;
};

int
MODULE_INIT(void *data)
{
  if (data)
    {
      struct reload_data *rd = data;
      protocols = rd->protocols;
      sessions = rd->sessions;
      new_session_handlers = rd->new_session_handlers;
    }
  else
    {
      protocols = new_llist();
      sessions = new_llist();
      new_session_handlers = new_llist();
    }

  read_config();

  add_event_handler(event("protocol.reload-config"), &read_config_event);
  add_event_handler(event("core.sighup"), &read_config_event);
  raise_event(event("protocol.load"));

  return MODULE_LOAD_OK;
}

int
MODULE_DEINIT(const char *new_version, module_init_func *init)
{
  delete_event_handler(event("protocol.reload-config"), &read_config_event);
  delete_event_handler(event("core.sighup"), &read_config_event);
  if (new_version)
    {
      struct reload_data rd = {protocols, sessions, new_session_handlers};
      switch ((*init)(&rd))
        {
        case MODULE_LOAD_OK:
          protocols = NULL;
          sessions = NULL;
          new_session_handlers = NULL;
          return MODULE_LOAD_OK;
        case MODULE_LOAD_FAILED:
          add_event_handler(event("protocol.reload-config"), &read_config_event);
          add_event_handler(event("core.sighup"), &read_config_event);
          return MODULE_LOAD_FAILED;
        default:
          log(MYNAME, LOG_PRIORITY_ERROR, "Unknown return from new module version %s, aborting", new_version);
          add_event_handler(event("protocol.reload-config"), &read_config_event);
          add_event_handler(event("core.sighup"), &read_config_event);
          return MODULE_LOAD_FAILED;
        }
    }

  destroy_llist(sessions, LSYM(protocol, close_session));
  destroy_llist(protocols, free_protocol);
  destroy_llist(new_session_handlers, xfree);

  return MODULE_UNLOAD_OK;
}
