/************************************************************************
 *   This is part of Dancer, a text relay daemon
 *   Copyright (C) 2002 Andrew Suffield <asuffield@users.sourceforge.net>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

#ifdef PROTOCOL_H

struct protocol;
struct session;

#else
#define PROTOCOL_H
#include <dancerd/config.h>
#include <dancerd/modules.h>
#include <dancerd/dbuffer.h>
#include <dancerd/transport/transport.h>
#include <dancerd/protocol/message.h>

#include <sys/types.h>

struct protocol;
struct session;

/** \file protocol/protocol.h
 * \brief Protocol layer
 */

/** \defgroup protocol Protocol layer
 *
 * The protocol layer encodes/decodes messages from sessions, and
 * passes them up to interested modules.
 *
 * @{
 */

/* This is passed the current buffer to try and identify if it is handled by this protocol.
 * Return 1 for "definitely mine", 0 for "inconclusive", and -1 for "not mine"
 */
typedef int protocol_identify_func(struct dbuffer *);
/* Called when a session is assigned to the protocol (ie, after the identity function returns 1) */
typedef void *protocol_init_session_func(struct session *, void *);
/* Called when a new outgoing connection is ready, so this protocol should begin handshaking */
typedef void *protocol_connection_ready_func(struct session *);
/* Called whenever data is ready on a session */
typedef void protocol_read_data_func(struct session *, void *, struct dbuffer *, const void *, size_t);
/* Called whenever data is to be sent to a session */
typedef void protocol_send_message_func(struct session *, void *, const struct message *);
/* Called when the transport layer closes the session */
typedef void protocol_close_session_func(struct session *, void *);

/* Returns false if the new session is to be closed */
typedef int new_session_handler(struct session *, void *, int);

DEFINE_NAMESPACE(protocol, 0);

DECLARE_SYMBOL(struct protocol *, protocol, register_protocol, (const char *, struct module *,
                                                                protocol_identify_func *,
                                                                protocol_init_session_func *,
                                                                protocol_connection_ready_func *,
                                                                protocol_read_data_func *,
                                                                protocol_send_message_func *,
                                                                protocol_close_session_func *));
DECLARE_SYMBOL(void, protocol, unregister_protocol, (struct protocol *));

/* These are called by protocol modules */
DECLARE_SYMBOL(void, protocol, read_message, (struct session *, struct message *));
DECLARE_SYMBOL(void, protocol, write_data, (struct session *, const void *, size_t));

/* These are called by protocol modules or higher layers */
DECLARE_SYMBOL(void, protocol, shutdown_session, (struct session *));
DECLARE_SYMBOL(void *, protocol, get_session_meta, (struct session *, const char *));
DECLARE_SYMBOL(void, protocol, set_session_meta, (struct session *, const char *, void *));

/* This is the interface to the higher layer */
DECLARE_SYMBOL(void, protocol, send_message, (struct session *, struct message *));
DECLARE_SYMBOL(void, protocol, register_new_session_handler, (struct module *, new_session_handler *, void *));
DECLARE_SYMBOL(void, protocol, unregister_new_session_handler, (struct module *, new_session_handler *, void *));

DECLARE_SYMBOL(struct session *, protocol, begin_connection, (const char *, const char *,
                                                              const char *, const char *,
                                                              void **, int));

DECLARE_SYMBOL(int, protocol, register_attribute, (const char *));
DECLARE_SYMBOL(void *, protocol, get_attribute, (struct session *, int));
DECLARE_SYMBOL(void, protocol, set_attribute, (struct session *, int, void *));

/* These are called by the transport layer */
DECLARE_SYMBOL(struct session *, protocol, new_session, (struct connection *, struct dbuffer *));
DECLARE_SYMBOL(void, protocol, data_read, (struct session *, const void *, size_t));
DECLARE_SYMBOL(void, protocol, close_session, (struct session *));
DECLARE_SYMBOL(void, protocol, connection_ready, (struct session *, struct dbuffer *));

END_NAMESPACE(protocol);

/** @}
 */

#endif
