/************************************************************************
 *   This is part of Dancer, a text relay daemon
 *   Copyright (C) 2002 Andrew Suffield <asuffield@users.sourceforge.net>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

#ifdef PROTOCOL_INTERNAL_H

struct protocol;
struct session;

#else
#define PROTOCOL_INTERNAL_H
#include <dancerd/config.h>
#include <dancerd/modules.h>
#include <dancerd/protocol/protocol.h>
#include <dancerd/hashtable.h>
#include <dancerd/llist.h>

/** \file protocol_internal.h
 * \internal
 * Shared structures within the protocol interface module
 */

/** \brief Buffering and association between transports, protocols, and message handlers
 */
struct session
{
  /** Connection (to the transport layer) used by this session */
  struct connection *conn;
  /** Arbitrary pointer passed to all protocol handlers */
  void *data;
  /** Buffer that the transport layer writes to, and the protocol layer reads from */
  struct dbuffer *buf;
  /** Protocol used by this session, if known */
  struct protocol *proto;
  /** Node in the llist proto->sessions that points to this session */
  struct llist_node *proto_sessions_node;
  /** Protocols which could be used by this session, if proto == NULL */
  struct llist *candidate_protocols;
  /** Queue of messages awaiting sending */
  struct message_queue *send_queue;
  /** Scheduler used by this session, or NULL if packets are to be
   * processed directly
   */
  struct scheduler *sched;
  /** Message handlers attached to this session */
  struct llist *handlers;
  /** Metadata associated with this session */
  struct hashtable *metadata;
  /** When true, no further messages will be accepted in either direction on this session */
  bool shutting_down;
  /** When true, the session has been established
   *
   * An established session is one which has been passed to the
   * protocol implementation for initialisation.
   *
   * If established is false and proto is null, this is a new incoming
   * connection.
   *
   * If established is false and protocol is valid, this is a new
   * outgoing connection that should be finished when the protocol
   * becomes active again.
   */
  bool established;
  /** Number of attributes allocated in this session */
  int attributes_allocated;
  /** Array of attributes */
  void **attributes;
  /** Count of the number of things that are currently using this
   * session; while non-zero, it won't be freed
   */
  int refcount;
  /** True if this session is closed; when refcount goes to zero,
   * close_session should be called */
  bool closed;
  /** \brief Node that points to this session in the sessions llist
   */
  struct llist_node *sessions_entry;
};

/** \brief A registered protocol
 */
struct protocol
{
  /** Name of the protocol */
  char *name;
  /** Module this protocol is implemented in (will automatically
   * unregister when this module unloads)
   */
  struct module *module;
  /** Called for new connections to try and identify if they use this protocol */
  protocol_identify_func *pi;
  /** Called when the protocol used by a session has been identified and it is ready to go */
  protocol_init_session_func *pis;
  /** Called when a new outgoing connection is ready, so the protocol should start handshaking */
  protocol_connection_ready_func *pcr;
  /** Called when new data is read by the transport */
  protocol_read_data_func *prd;
  /** Called when a message is to be sent out on the transport */
  protocol_send_message_func *psm;
  /** Called when the session is to be closed, cleanly if possible */
  protocol_close_session_func *pcs;
  /** List of sessions attached to this protocol */
  struct llist *sessions;
  /** True if something has currently registered this protocol */
  bool active;
};


#endif
