/************************************************************************
 *   This is part of Dancer, a text relay daemon
 *   Copyright (C) 2002 Andrew Suffield <asuffield@users.sourceforge.net>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

#include <dancerd/config.h>
#include <dancerd/modules.h>
#include <dancerd/xmalloc.h>
#include <dancerd/xstring.h>
#include <dancerd/hashtable.h>
#include <dancerd/llist.h>
#include <iconv.h>
#include <errno.h>
#include <assert.h>

#include "client.h"
#include "charset.h"

#define ICONV_BUFFER_MALLOC_CHUNK 100

static struct hashtable *charset_hash;
static struct llist *charset_list;

struct rfc1459_charset *new_iconv(const char *name)
{
  struct rfc1459_charset *charset = hashtable_lookup(charset_hash, name);
  iconv_t cd_to_utf8;
  iconv_t cd_from_utf8;
  if (charset)
    {
      charset->refs++;
      return charset;
    }

  cd_to_utf8 = iconv_open("UTF-8", name ? name : "ASCII");
  if (cd_to_utf8 == (iconv_t)(-1))
    {
      return NULL;
    }
  cd_from_utf8 = iconv_open(name ? name : "ASCII", "UTF-8");
  if (cd_from_utf8 == (iconv_t)(-1))
    {
      /* Preserve errno */
      int e = errno;
      iconv_close(cd_to_utf8);
      errno = e;
      return NULL;
    }

  charset = xmalloc(sizeof(struct rfc1459_charset));
  charset->name = xstrdup(name);
  charset->refs = 1;
  charset->cd_to_utf8 = cd_to_utf8;
  charset->cd_from_utf8 = cd_from_utf8;
  hashtable_insert(charset_hash, charset->name, charset);
  add_llist_tail(charset_list, charset);
  charset->node = llist_tail(charset_list);
  return charset;
}

static void
free_iconv(struct rfc1459_charset *charset)
{
  iconv_close(charset->cd_from_utf8);
  iconv_close(charset->cd_to_utf8);
  xfree(charset->name);
  xfree(charset);
}

void release_iconv(struct rfc1459_charset *charset)
{
  if (!charset)
    return;

  charset->refs--;
  if (charset->refs == 0)
    {
      hashtable_delete(charset_hash, charset->name);
      delete_llist_node(charset->node);
      free_iconv(charset);
    }
}

static char *xiconv(iconv_t cd, char *inbuf, size_t inbytesleft, int from_utf8)
{
  char *buf = xmalloc(ICONV_BUFFER_MALLOC_CHUNK);
  char *outbuf = buf;
  size_t outbufsize = ICONV_BUFFER_MALLOC_CHUNK;
  size_t outbytesleft = outbufsize;
  outbuf[0] = '\0';
  while (inbuf && (iconv(cd,
                         &inbuf, &inbytesleft,
                         &outbuf, &outbytesleft) == (size_t)(-1)))
    {
      switch(errno)
        {
        case EILSEQ:
          if (*inbuf)
            {
              /* Heuristic: change one character to '?' and try to continue
               *
               * Note that we have to change to '?' on the UTF-8 side,
               * since we don't know how to represent a '?' in the
               * other character set
               */
              if (from_utf8)
                {
                  assert(inbytesleft);
                  /** \todo Make this UTF-8 aware, so it only
                   * generates one ? even for multibyte characters
                   */
                  *inbuf = '?';
                }
              else
                {
                  /* There's no way to avoid generating too many '?' 
                   * characters here; the problem is an invalid input
                   * string, not an inability to represent the input
                   * string in the output character set
                   */
                  if (!outbytesleft)
                    {
                      outbufsize += ICONV_BUFFER_MALLOC_CHUNK;
                      outbytesleft += ICONV_BUFFER_MALLOC_CHUNK;
                      buf = xrealloc(buf, outbufsize);
                      outbuf = buf + outbufsize - outbytesleft;
                    }
                  inbuf++;
                  inbytesleft--;
                  *outbuf++ = '?';
                  outbytesleft--;
                }
            }
          break;
        case E2BIG:
          outbufsize += ICONV_BUFFER_MALLOC_CHUNK;
          outbytesleft += ICONV_BUFFER_MALLOC_CHUNK;
          buf = xrealloc(buf, outbufsize);
          outbuf = buf + outbufsize - outbytesleft;
          break;
        case EINVAL:
        default:
          xfree(buf);
          return NULL;
        }
    }

  buf = xrealloc(buf, (outbufsize + 1) - outbytesleft);
  buf[outbufsize - outbytesleft] = '\0';
  return buf;
}

char *convert_to_utf8(struct rfc1459_charset *charset, const char *indata)
{
  char *inbuf;
  char *outbuf;
  size_t len;

  if (!indata || !charset)
    return NULL;

  len = strlen(indata);
  inbuf = xmalloc(len + 1);
  memcpy(inbuf, indata, len + 1);
  outbuf = xiconv(charset->cd_to_utf8, inbuf, len, 0);
  xfree(inbuf);
  return outbuf;
}

char *convert_from_utf8(struct rfc1459_charset *charset, const char *indata)
{
  char *inbuf;
  char *outbuf;
  size_t len;

  if (!indata || !charset)
    return NULL;

  len = strlen(indata);
  inbuf = xmalloc(len + 1);
  memcpy(inbuf, indata, len + 1);
  outbuf = xiconv(charset->cd_from_utf8, inbuf, len, 1);
  xfree(inbuf);
  return outbuf;
}

void
init_charset(void)
{
  charset_hash = new_hashtable(&hash_ascii, 149);
  charset_list = new_llist();
}

void
reinit_charset(void)
{
  struct llist_node *n;
  for (n = llist_head(charset_list); n; n = llist_node_next(n))
    {
      struct rfc1459_charset *charset = llist_node_data(n);
      charset->cd_to_utf8 = iconv_open("UTF-8", charset->name ? charset->name : "ASCII");
      charset->cd_from_utf8 = iconv_open(charset->name ? charset->name : "ASCII", "UTF-8");
    }
}

void
deinit_charset(int reload)
{
  if (reload)
    {
      struct llist_node *n;
      for (n = llist_head(charset_list); n; n = llist_node_next(n))
        {
          struct rfc1459_charset *charset = llist_node_data(n);
          iconv_close(charset->cd_to_utf8);
          iconv_close(charset->cd_from_utf8);
          charset->cd_to_utf8 = (iconv_t)-1;
          charset->cd_from_utf8 = (iconv_t)-1;
        }
    }
  else
    {
      destroy_llist(charset_list, free_iconv);
      charset_list = NULL;
      free_hashtable(charset_hash);
      charset_hash = NULL;
    }
}
