/************************************************************************
 *   This is part of Dancer, a text relay daemon
 *   Copyright (C) 2002 Andrew Suffield <asuffield@users.sourceforge.net>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

#include <dancerd/config.h>
#include <dancerd/modules.h>
#include <dancerd/protocol/protocol.h>
#include <dancerd/log.h>
#include <dancerd/xmalloc.h>
#include <dancerd/xstring.h>
#include <dancerd/gettext.h>

#include <assert.h>
#include <locale.h>
#include <stddef.h>
#include <string.h>
#include <stdio.h>

#include "client.h"
#include "charset.h"
#include "handlers.h"
#include "messages.h"

SETUP_MODULE("protocol.rfc1459.client", MODVERSION);

IMPORTED_SYMBOL(protocol, register_protocol);
IMPORTED_SYMBOL(protocol, unregister_protocol);
IMPORTED_SYMBOL(protocol, write_data);
IMPORTED_SYMBOL(message, get_message_type_name);

#include "symtab"

static struct protocol *my_protocol = NULL;
static struct rfc1459_charset *ascii_charset = NULL;

const struct rfc1459_command *rfc1459_get_command(const char *, unsigned int);
/* gperf has an irritating habit of using "index", which is also a
   function. Work around it. */
#ifdef index
#undef index
#endif
#define index gperf_index
#include "commands.tab"
#undef index

const char *
rfc1459_client_gettext(struct rfc1459_client *client, const char *msgid)
{
  const char *translated;
  char *saved_language = NULL;
  if (client->language)
    {
      saved_language = xstrdup(setlocale(LC_MESSAGES, NULL));
      setlocale(LC_MESSAGES, client->language);
    }
  translated = gettext(msgid);
  if (saved_language)
    {
      setlocale(LC_MESSAGES, saved_language);
      xfree(saved_language);
    }
  return translated;
}

const char *
rfc1459_client_ngettext(struct rfc1459_client *client, const char *msgid, const char *msgid_plural, unsigned long int n)
{
  const char *translated;
  char *saved_language = NULL;
  if (client->language)
    {
      saved_language = xstrdup(setlocale(LC_MESSAGES, NULL));
      setlocale(LC_MESSAGES, client->language);
    }
  translated = ngettext(msgid, msgid_plural, n);
  if (saved_language)
    {
      setlocale(LC_MESSAGES, saved_language);
      xfree(saved_language);
    }
  return translated;
}

/* Okay, here's the heuristic we're going to use to identify rfc1459 clients:
 *
 * We look at up to the first 2048 bytes of the session.
 *
 * Include any bytes outside the printable range (32-126) and it's not ours.
 * Fail to include \r or \n for 512 bytes and it's not ours.
 * Include the sequences "\nSERVER " or "\rSERVER " and it's not ours
 * Include the sequences "\nUSER ", "\rUSER ", "\nNICK " or "\rNICK " and it's ours
 */
static protocol_identify_func identify;
static int
identify(struct dbuffer *buf)
{
  char cbuf[2049];
  size_t len = dbuffer_get(buf, cbuf, 2048);
  cbuf[2048] = '\0';
  size_t i, message_start = 0;

  for (i = 0; i < len; i++)
    {
      /* If the line has more than 512 characters total, or more than
       * 510 [^\r\n] characters, it's too long
       */
      if ((message_start + 510) < i)
        return -1;
      /* Catch the case where this character is a line end */
      if (((cbuf[i] == '\r') || (cbuf[i] == '\n')) && (i != message_start))
        {
          if (strncasecmp(cbuf + message_start, "SERVER ", min(7, i - message_start)) == 0)
            return -1;
          if (strncasecmp(cbuf + message_start, "USER ", min(5, i - message_start)) == 0)
            return 1; 
          if (strncasecmp(cbuf + message_start, "NICK ", min(5, i - message_start)) == 0)
            return 1;
          if (strncasecmp(cbuf + message_start, "QUIT", min(4, i - message_start)) == 0)
            return 1;
          message_start = i + 1;
       }
      /* Any other non-printables are invalid
       * Note that this one catches NULLs
       */
      else if ((cbuf[i] < 32) || (cbuf[i] > 126))
        return -1;
    }
  return 0;
}

static protocol_init_session_func init_session;
static void *
init_session(struct session *session, void *d)
{
  struct rfc1459_client *client = d;

  if (client)
    return client;

  client = xmalloc(sizeof(struct rfc1459_client));
  client->done_handshaking = 0;
  client->session = session;
  client->name = NULL;
  client->myname = NULL;
  client->done_user = 0;
  client->done_nick = 0;
  client->charset = new_iconv("ASCII");
  client->language = NULL;

  return client;
}

static void
parse_line(struct rfc1459_client *client unused_attribute, char *line)
{
  const char *command_end;
  const struct rfc1459_command *command;
  unsigned int command_length;

  if (line[0] == ':')
    {
      /* This message has a prefix. Prefixes are ignored from clients,
       * so skip over it
       */
      char *c = strchr(line, ' ');
      if (!c)
        /* There is nothing but a prefix, so this line is empty. Ignore it */
        return;

      while (*c == ' ')
        c++;
      /* Skip the prefix */
      line = c;
    }

  command_end = strchr(line, ' ');
  if (command_end)
    command_length = command_end - line;
  else
    command_length = strlen(line);
  if (command_length == 0)
    /* No command, ignore this message */
    return;
  command = rfc1459_get_command(line, command_length);

  if (!command || !command->handler)
    {
      static char buf[512];
      line[command_length] = '\0';
      size_t len = snprintf(buf, 510, "421 %s %s :%s",
                            client->name ? client->name : "", line, _c(client, "Unknown command"));
      len = min(len, 510);
      buf[len] = '\r';
      buf[len + 1] = '\n';
      SYM(protocol, write_data)(client->session, buf, len + 2);
      return;
    }

  /* Advance the line to the first word after the command */
  line += command_length;
  /* Skip any spaces */
  while (*line == ' ')
    line++;

  /* We happen to never need more than 16 words */
  {
    char *argv[16];
    unsigned int argc = 0;
    while ((argc < 16) && *line)
      {
        char *c;
        /* Stop if we hit the quote
         *
         * If we have reached the maximum number of parameters,
         * silently quote what is left
         */
        if ((line[0] == ':') || ((argc + 1) == command->max_params))
          {
            if (line[0] == ':')
              line++;
            argv[argc++] = line;
            break;
          }

        /* Find the next space */
        c = strchr(line, ' ');

        /* Null terminate it, and point to the start of the next word
         * (or NULL if this is the last word
         */
        if (c)
          {
            *c++ = '\0';

            /* Skip any extra spaces here
             */
            while (*c == ' ')
              c++;
          }

        argv[argc++] = line;

        /* c is NULL if last word, otherwise pointer to next word */
        if (c)
          line = c;
        else
          break;
      }
    if (argc < command->min_params)
      {
        static char buf[512];
        size_t len = snprintf(buf, 510, "461 %s %s :%s",
                              client->name ? client->name : "", command->name, _c(client, "Not enough parameters"));
        len = min(len, 510);
        buf[len] = '\r';
        buf[len + 1] = '\n';
        SYM(protocol, write_data)(client->session, buf, len + 2);
        return;
      }
    else
      (*command->handler)(client, command, argc, argv);
  }
}

static protocol_read_data_func read_data;
static void
read_data(struct session *session, void *d,
          struct dbuffer *buf, const void *data, size_t len)
{
  struct rfc1459_client *client = d;

  assert(session == client->session);

  /* If we have a 'data' parameter, use it to reduce the number of
   * slower dbuffer scans
   */
  if (!data || memchr(data, '\r', len) || memchr(data, '\n', len))
    while (1)
      {
        ssize_t cr_pos, nl_pos, line_end;
        static char cbuf[513];

        cr_pos = dbuffer_find_char(buf, '\r');
        nl_pos = dbuffer_find_char(buf, '\n');

        /* We are done when there are no more complete lines */
        if ((cr_pos == -1) && (nl_pos == -1))
          break;

        if ((cr_pos != -1) && (nl_pos != -1) && ((cr_pos + 1) == nl_pos))
          {
            /* Line ends in a CRLF. Extract up to 510, then remove the
             * lot
             */
            dbuffer_get(buf, cbuf, min(cr_pos, 510));
            cbuf[min(cr_pos, 510)] = '\0';
            dbuffer_remove(buf, nl_pos + 1);
          }
        else
          {
            /* Line ends in a CR or LF */
            if (cr_pos == -1)
              line_end = nl_pos;
            else if (nl_pos == -1)
              line_end = cr_pos;
            else
              line_end = min(cr_pos, nl_pos);
            /* line_end is the index from zero of the first character
             * past the end. This can be no greater than 510, for 510
             * interesting or 511 total characters.
             */
            if (line_end > 510)
              {
                /* Take out the 510 interesting characters */
                dbuffer_extract(buf, cbuf, 510);
                cbuf[510] = '\0';
                /* Throw away anything else, in an effort to reduce
                 * spurious "unknown command" messages
                 */
                dbuffer_remove(buf, line_end - 510);
              }
            else
              {
                dbuffer_extract(buf, cbuf, line_end + 1);
                cbuf[line_end] = '\0';
              }
          }

        parse_line(client, cbuf);
      }
}

static protocol_send_message_func send_message;
static void
send_message(struct session *session, void *d, const struct message *msg)
{
  struct rfc1459_client *client = d;
  struct rfc1459_message *m;

  assert(session == client->session);

  if (!msg)
    return;

  if ((msg->type > max_rfc1459_message_type) || !rfc1459_message_types[msg->type])
    {
      log(MYNAME, LOG_PRIORITY_DEBUG, "Can't send message type %d (%s), no handler", msg->type,
          SYM(message, get_message_type_name)(msg->type));
      return;
    }

  m = rfc1459_message_types[msg->type];

  /* Is everything we need satisfied? */
  if ((m->required_components & RMC_ORIGIN) && (!msg->origin))
    {
      log(MYNAME, LOG_PRIORITY_ERROR, "Can't send '%s' message: missing origin", m->message_type_name);
      return;
    }

  if ((m->required_components & RMC_DEST) && (!msg->dest))
    {
      log(MYNAME, LOG_PRIORITY_ERROR, "Can't send '%s' message: missing dest", m->message_type_name);
      return;
    }

  if (m->min_strc > msg->strc)
    {
      log(MYNAME, LOG_PRIORITY_ERROR, "Can't send '%s' message: insufficient string components (%d < %d)",
          m->message_type_name, msg->strc, m->min_strc);
      return;
    }

  if (m->min_uint32c > msg->uint32c)
    {
      log(MYNAME, LOG_PRIORITY_ERROR, "Can't send '%s' message: insufficient uint32 components (%d < %d)",
          m->message_type_name, msg->uint32c, m->min_uint32c);
      return;
    }

  if (m->min_sint32c > msg->sint32c)
    {
      log(MYNAME, LOG_PRIORITY_ERROR, "Can't send '%s' message: insufficient sint32 components (%d < %d)",
          m->message_type_name, msg->sint32c, m->min_sint32c);
      return;
    }

  /* If there is a special handler, call it, otherwise ignore the message */
  if (m->handler)
    {
      (*m->handler)(client, msg);
      return;
    }
}

static protocol_close_session_func close_session;
static void
close_session(struct session *session unused_attribute, void *d)
{
  struct rfc1459_client *client = d;
  release_iconv(client->charset);
  xfree(client->name);
  xfree(client->myname);
  xfree(client);
}

int
MODULE_INIT(void *data unused_attribute)
{
  my_protocol = SYM(protocol, register_protocol)("rfc1459.client", MYHANDLE,
                                                 &identify, &init_session, NULL,
                                                 &read_data, &send_message,
                                                 &close_session);

  if (!my_protocol)
    return MODULE_LOAD_FAILED;

  init_charset();
  init_rfc1459_messages();
  init_rfc1459_handlers();

  /* Just to make sure this one is always open */
  ascii_charset = new_iconv("ASCII");

  return MODULE_LOAD_OK;
}

int
MODULE_DEINIT(const char *new_version, module_init_func *init)
{
  /* This deinit cannot be done safely - we must unregister our own
   * protocol handle before loading the new one, and we don't know if
   * we'll be able to get it back afterwards.
   *
   * If the load fails and we can't get back our handle, that's a
   * FUBAR.
   */
  if (my_protocol)
    SYM(protocol, unregister_protocol)(my_protocol);
  my_protocol = NULL;

  release_iconv(ascii_charset);
  ascii_charset = NULL;
  deinit_charset(0);

  if (new_version)
    {
      switch ((*init)(NULL))
        {
        case MODULE_LOAD_OK:
          return MODULE_LOAD_OK;
        case MODULE_LOAD_FAILED:
          my_protocol = SYM(protocol, register_protocol)("rfc1459.client", MYHANDLE,
                                                         &identify, &init_session, NULL,
                                                         &read_data, &send_message,
                                                         &close_session);
          if (!my_protocol)
            return MODULE_LOAD_FUBAR;
          return MODULE_LOAD_FAILED;
        default:
          log(MYNAME, LOG_PRIORITY_ERROR, "Unknown return from new module version %s, aborting", new_version);
          my_protocol = SYM(protocol, register_protocol)("rfc1459.client", MYHANDLE,
                                                         &identify, &init_session, NULL,
                                                         &read_data, &send_message,
                                                         &close_session);
          if (!my_protocol)
            return MODULE_LOAD_FUBAR;
          return MODULE_LOAD_FAILED;
        }
    }

  return MODULE_UNLOAD_OK;
}
