/************************************************************************
 *   This is part of Dancer, a text relay daemon
 *   Copyright (C) 2002 Andrew Suffield <asuffield@users.sourceforge.net>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

#include <dancerd/config.h>
#include <dancerd/modules.h>
#include <dancerd/protocol/protocol.h>
#include <dancerd/protocol/message.h>
#include <dancerd/xmalloc.h>

#include <assert.h>
#include <stddef.h>
#include <stdio.h>
#include <errno.h>

#include "charset.h"
#include "client.h"
#include "handlers.h"

EXTERN_MODULE();

IMPORTED_SYMBOL(protocol, read_message);
IMPORTED_SYMBOL(message, build_message);
IMPORTED_SYMBOL(message, register_message_type);
IMPORTED_SYMBOL(protocol, write_data);

/* handlers.tab is automatically generated from this file by
 * make-handlers (in the current source directory); it finds every
 * call to message::build_message in this file, extracts the first
 * parameter (which should be a variable name), and emits code into
 * handlers.tab that
 *  a) defines that variable as a static int
 *  b) assigns it the return value from message::register_message_type
 *     with a parameter (message name) generated by applying s/_/ /g to
 *     the variable name
 *
 * So, if there is an intruction in this file like this:
 *
 *  SYM(protocol, read_message)(client->session,
 *                              SYM(message, build_message)(set_real_name, origin, NULL,
 *
 * Then set_real_name will be defined, and assigned the message type
 * "set real name".
 */

#include "handlers.tab"

static char buf[512];

void
p_user(struct rfc1459_client *client, const struct rfc1459_command *command unused_attribute,
       unsigned int argc, char **argv)
{
  const char *origin;
  const char *args[1];
  assert(argc == 4);

  origin = client->name;
  args[0] = argv[3];

  SYM(protocol, read_message)(client->session,
                              SYM(message, build_message)(set_real_name, origin, NULL,
                                                          1, args,
                                                          0, NULL,
                                                          0, NULL));

  /* Generate a done_handshake message after we get the last of a pair
   * of valid NICK/USER messages
   *
   * Here's the effect:
   * Receive NICK
   * Reject NICK
   * Receive USER
   * Receive NICK
   * Generate done_handshake
   * Reject NICK
   * Receive NICK
   * Generate done_handshake
   *
   * Which happens to be what we want.
   */
  if (!client->done_user)
    {
      client->done_user = 1;
      if (client->done_nick)
        SYM(protocol, read_message)(client->session,
                                    SYM(message, build_message)(done_handshake, origin, NULL,
                                                                0, NULL,
                                                                0, NULL,
                                                                0, NULL));
    }
}

void
p_nick(struct rfc1459_client *client, const struct rfc1459_command *command unused_attribute,
       unsigned int argc, char **argv)
{
  const char *origin;
  const char *args[1];
  assert(argc >= 1);

  origin = client->name;
  args[0] = argv[0];

  SYM(protocol, read_message)(client->session,
                              SYM(message, build_message)(set_name, origin, NULL,
                                                          1, args,
                                                          0, NULL,
                                                          0, NULL));

  if (!client->done_nick)
    {
      client->done_nick = 1;
      if (client->done_user)
        SYM(protocol, read_message)(client->session,
                                    SYM(message, build_message)(done_handshake, origin, NULL,
                                                                0, NULL,
                                                                0, NULL,
                                                                0, NULL));
    }
}

void
p_whois(struct rfc1459_client *client, const struct rfc1459_command *command unused_attribute,
        unsigned int argc, char **argv)
{
  const char *origin, *target;
  const char *args[1];

  origin = client->name;
  if (argc > 1)
    {
      target = argv[0];
      args[0] = argv[1];
    }
  else
    {
      target = NULL;
      args[0] = argv[0];
    }

  SYM(protocol, read_message)(client->session,
                              SYM(message, build_message)(query_node_information, origin, target,
                                                          1, args,
                                                          0, NULL,
                                                          0, NULL));
}

void
p_quit(struct rfc1459_client *client, const struct rfc1459_command *command unused_attribute,
       unsigned int argc, char **argv)
{
  const char *origin;

  origin = client->name;

  SYM(protocol, read_message)(client->session,
                              SYM(message, build_message)(close_session, origin, NULL,
                                                          (argc >= 1) ? 2 : 1, (const char *[]) {client->name, argc >= 1 ? argv[0] : NULL},
                                                          0, NULL,
                                                          0, NULL));
}

void
p_version(struct rfc1459_client *client, const struct rfc1459_command *command unused_attribute,
          unsigned int argc unused_attribute, char **argv unused_attribute)
{
  const char *origin;

  origin = client->name;

  SYM(protocol, read_message)(client->session,
                              SYM(message, build_message)(query_node_software_information, origin, NULL,
                                                          0, NULL,
                                                          0, NULL,
                                                          0, NULL));
}

void
p_license(struct rfc1459_client *client, const struct rfc1459_command *command unused_attribute,
          unsigned int argc unused_attribute, char **argv unused_attribute)
{
  const char *origin;

  origin = client->name;

  SYM(protocol, read_message)(client->session,
                              SYM(message, build_message)(query_node_license_information, origin, NULL,
                                                          0, NULL,
                                                          0, NULL,
                                                          0, NULL));
}

void
p_info(struct rfc1459_client *client, const struct rfc1459_command *command unused_attribute,
          unsigned int argc unused_attribute, char **argv unused_attribute)
{
  const char *origin;

  origin = client->name;

  SYM(protocol, read_message)(client->session,
                              SYM(message, build_message)(query_node_author_information, origin, NULL,
                                                          0, NULL,
                                                          0, NULL,
                                                          0, NULL));
}

void
p_charset(struct rfc1459_client *client, const struct rfc1459_command *command unused_attribute,
          unsigned int argc, char **argv)
{
  size_t len;
  const char *from = client->myname ? client->myname : "";
  const char *to = client->name ? client->name : "CHARSET";

  assert(argc >= 1);
  release_iconv(client->charset);
  client->charset = new_iconv(argv[0]);

  if (client->charset)
    {
      len = snprintf(buf, 510, _c(client, ":%s NOTICE %s :Your character set is now '%s'"),
                     from, to, client->charset->name);
    }
  else
    {
      if (errno == EINVAL)
        len = snprintf(buf, 510, _c(client, ":%s NOTICE %s :The character set '%s' is not supported"),
                       from, to, argv[0]);
      else
        len = snprintf(buf, 510, _c(client, ":%s NOTICE %s :Couldn't select character set '%s': %s"),
                       from, to, argv[0], strerror(errno));
    }
  len = min(len,510);
  buf[len] = '\r';
  buf[len+1] = '\n';
  SYM(protocol, write_data)(client->session, buf, len + 2);
}

void
p_privmsg(struct rfc1459_client *client, const struct rfc1459_command *command unused_attribute,
          unsigned int argc unused_attribute, char **argv)
{
  const char *origin;

  origin = client->name;

  if (*argv[0] == '#')
    {
      /* Channel message */
      const char *strv[2];
      char *msg = argv[1];
      char *utf8_message;
      size_t len = strlen(msg);
      if ((msg[0] == '\01') && (msg[len - 1] == '\01'))
        {
          msg[len - 1] = '\0';
          msg++;
          utf8_message = convert_to_utf8(client->charset, msg);
          strv[0] = argv[0] + 1;
          strv[1] = utf8_message;
          SYM(protocol, read_message)(client->session,
                                      SYM(message, build_message)(channel_ctcp_query, origin, NULL,
                                                                  2, strv,
                                                                  0, NULL,
                                                                  0, NULL));
        }
      else
        {
          utf8_message = convert_to_utf8(client->charset, msg);
          strv[0] = argv[0] + 1;
          strv[1] = utf8_message;
          SYM(protocol, read_message)(client->session,
                                      SYM(message, build_message)(channel, origin, NULL,
                                                                  2, strv,
                                                                  0, NULL,
                                                                  0, NULL));
        }
      xfree(utf8_message);
    }
  else
    {
      /* Private message */
      const char *strv[1];
      char *msg = argv[1];
      char *utf8_message;
      size_t len = strlen(msg);
      if ((msg[0] == '\01') && (msg[len - 1] == '\01'))
        {
          msg[len - 1] = '\0';
          msg++;
          utf8_message = convert_to_utf8(client->charset, msg);
          strv[0] = utf8_message;
          SYM(protocol, read_message)(client->session,
                                      SYM(message, build_message)(ctcp_query, origin, argv[0],
                                                                  1, strv,
                                                                  0, NULL,
                                                                  0, NULL));
        }
      else
        {
          utf8_message = convert_to_utf8(client->charset, msg);
          strv[0] = utf8_message;
          SYM(protocol, read_message)(client->session,
                                      SYM(message, build_message)(private, origin, argv[0],
                                                                  1, strv,
                                                                  0, NULL,
                                                                  0, NULL));
        }
      xfree(utf8_message);
    }
}

void
p_beginauth(struct rfc1459_client *client, const struct rfc1459_command *command unused_attribute,
            unsigned int argc, char **argv)
{
  const char *origin;
  const char *args[2];

  assert(argc == 2);

  origin = client->name;
  args[0] = argv[0];
  args[1] = argv[1];

  SYM(protocol, read_message)(client->session,
                              SYM(message, build_message)(begin_auth, origin, NULL,
                                                          2, args,
                                                          0, NULL,
                                                          0, NULL));
}

void
p_authdata(struct rfc1459_client *client, const struct rfc1459_command *command unused_attribute,
           unsigned int argc, char **argv)
{
  const char *origin;
  const char *args[16];
  unsigned int i;

  assert(argc < 16);
  origin = client->name;
  for (i = 0; i < argc; i++)
    args[i] = argv[i];

  SYM(protocol, read_message)(client->session,
                              SYM(message, build_message)(auth_data, origin, NULL,
                                                          argc, args,
                                                          0, NULL,
                                                          0, NULL));
}

void
p_cancelauth(struct rfc1459_client *client, const struct rfc1459_command *command unused_attribute,
             unsigned int argc, char **argv)
{
  const char *origin;
  const char *args[1];

  origin = client->name;
  if (argc > 0)
    args[0] = argv[0];

  SYM(protocol, read_message)(client->session,
                              SYM(message, build_message)(cancel_auth, origin, NULL,
                                                          (argc && argv[0]) ? 1 : 0, args,
                                                          0, NULL,
                                                          0, NULL));
}

void
p_acceptauth(struct rfc1459_client *client, const struct rfc1459_command *command unused_attribute,
             unsigned int argc unused_attribute, char **argv unused_attribute)
{
  const char *origin;

  origin = client->name;

  SYM(protocol, read_message)(client->session,
                              SYM(message, build_message)(accept_auth, origin, NULL,
                                                          0, NULL,
                                                          0, NULL,
                                                          0, NULL));
}

void
p_setcred(struct rfc1459_client *client, const struct rfc1459_command *command unused_attribute,
          unsigned int argc, char **argv)
{
  const char *origin;
  const char *args[16];
  unsigned int i;

  assert(argc < 16);
  origin = client->name;
  for (i = 0; i < argc; i++)
    args[i] = argv[i];

  SYM(protocol, read_message)(client->session,
                              SYM(message, build_message)(set_credentials, origin, NULL,
                                                          argc, args,
                                                          0, NULL,
                                                          0, NULL));
}

void
p_register(struct rfc1459_client *client, const struct rfc1459_command *command unused_attribute,
           unsigned int argc, char **argv)
{
  const char *origin;

  assert(argc == 1);

  origin = client->name;

  SYM(protocol, read_message)(client->session,
                              SYM(message, build_message)(register_user, origin, NULL,
                                                          1, (const char *[]) {argv[0]},
                                                          0, NULL,
                                                          0, NULL));
}

void
p_connect(struct rfc1459_client *client, const struct rfc1459_command *command unused_attribute,
          unsigned int argc, char **argv)
{
  const char *origin;

  assert(argc == 4);

  origin = client->name;

  SYM(protocol, read_message)(client->session,
                              SYM(message, build_message)(begin_connection, origin, NULL,
                                                          4, (const char *[]) {argv[0], argv[1], argv[2], argv[3]},
                                                          0, NULL,
                                                          0, NULL));
}

void
p_ping(struct rfc1459_client *client, const struct rfc1459_command *command unused_attribute,
       unsigned int argc, char **argv)
{
  const char *origin;

  assert(argc >= 1);

  origin = client->name;

  SYM(protocol, read_message)(client->session,
                              SYM(message, build_message)(echo_request, origin, argv[0],
                                                          argc >= 2 ? 1 : 0, (const char *[]) {argc >= 2 ? argv[1] : NULL},
                                                          0, NULL,
                                                          0, NULL));
}


void
p_pong(struct rfc1459_client *client, const struct rfc1459_command *command unused_attribute,
       unsigned int argc, char **argv)
{
  const char *origin;

  assert(argc >= 1);

  origin = client->name;

  SYM(protocol, read_message)(client->session,
                              SYM(message, build_message)(echo_reply, origin, argv[0],
                                                          argc >= 2 ? 1 : 0, (const char *[]) {argc >= 2 ? argv[1] : NULL},
                                                          0, NULL,
                                                          0, NULL));
}
