/************************************************************************
 *   This is part of Dancer, a text relay daemon
 *   Copyright (C) 2002 Andrew Suffield <asuffield@users.sourceforge.net>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

#include <dancerd/config.h>
#include <dancerd/modules.h>
#include <dancerd/protocol/message.h>
#include <dancerd/version.h>
#include <dancerd/xmalloc.h>
#include <dancerd/xstring.h>

#include <stddef.h>
#include <stdio.h>

#include "charset.h"
#include "messages.h"

EXTERN_MODULE();

IMPORTED_SYMBOL(protocol, write_data);
IMPORTED_SYMBOL(protocol, shutdown_session);
IMPORTED_SYMBOL(message, register_message_type);
IMPORTED_SYMBOL(message, get_message_type_name);
IMPORTED_SYMBOL(protocol, get_session_meta);
IMPORTED_SYMBOL(protocol, read_message);
IMPORTED_SYMBOL(message, build_message);

static char buf[512];

static int accept_handshake;

static void
s_accept_handshake(struct rfc1459_client *client, const struct message *msg)
{
  size_t len;
  const char *listen_host;
  int *port;

  client->done_handshaking = 1;
  xfree(client->myname);
  xfree(client->name);
  client->myname = xstrdup(msg->origin);
  client->name = xstrdup(msg->dest);

  len = snprintf(buf, 511, _c(client, ":%s 001 %s :Welcome to the Internet Relay Network %s"),
                 msg->origin, msg->dest, msg->dest);
  len = min(len,510);
  buf[len] = '\r';
  buf[len+1] = '\n';
  SYM(protocol, write_data)(client->session, buf, len + 2);

  listen_host = SYM(protocol, get_session_meta)(client->session, "transport.listen-host");
  port = SYM(protocol, get_session_meta)(client->session, "transport.listen-port");
  
  len = snprintf(buf, 511, _c(client, ":%s 002 %s :Your host is %s[%s/%d], running version %s/protocol.rfc1459.client-%s"),
                 msg->origin, msg->dest,
                 listen_host ? listen_host : "*", listen_host ? listen_host : "*",
                 port ? *port : 0,
                 dancer_version, RFC1459_CLIENT_VERSION);
  len = min(len, 510);
  buf[len] = '\r';
  buf[len+1] = '\n';
  SYM(protocol, write_data)(client->session, buf, len + 2);

  /* Simulate the acknowledgement, since rfc1459 doesn't have one */
  SYM(protocol, read_message)(client->session,
                              SYM(message, build_message)(accept_handshake, msg->dest, msg->origin,
                                                          0, NULL,
                                                          0, NULL,
                                                          0, NULL));
}

static void
s_reject_set_name(struct rfc1459_client *client, const struct message *msg)
{
  size_t len;
  const char *oldtargetname = msg->dest ? msg->dest : "*";
  const char *newtargetname = msg->strv[0];

  client->done_nick = 0;

  /* If there is a reason, we send out an "errnoeous nickname"
   * message, otherwise it's "nick in use"
   */
  if (msg->strc > 1)
    {
      const char *reason = msg->strv[1] ? msg->strv[1] : _c(client, "Erroneous nickname");
      len = snprintf(buf, 511, ":%s 432 %s %s :%s", msg->origin, oldtargetname, newtargetname, reason);
    }
  else
    {
      len = snprintf(buf, 511, ":%s 433 %s %s :%s", msg->origin, oldtargetname, newtargetname,
                     _c(client, "Nickname is already in use"));
    }
  len = min(len, 510);
  buf[len] = '\r';
  buf[len+1] = '\n';
  SYM(protocol, write_data)(client->session, buf, len + 2);
}

static void
s_reject_set_real_name(struct rfc1459_client *client, const struct message *msg)
{
  size_t len;
  const char *targetname = msg->dest ? msg->dest : "*";
  const char *newrealname = msg->strv[0];

  client->done_user = 0;

  /* There's no way for rfc1459 clients to handle this. ERROR and
   * close.
   */
  if ((msg->strc > 1) && msg->strv[1])
    {
      const char *reason = msg->strv[1];
      len = snprintf(buf, 511, "ERROR :%s: %s (%s [%s])", _c(client, "Closing link"), targetname, reason, newrealname);
    }
  else
    {
      len = snprintf(buf, 511, "ERROR :%s: %s (%s [%s])", _c(client, "Closing link"),
                     targetname, _c(client, "Invalid username"), newrealname);
    }
  len = min(len, 510);
  buf[len] = '\r';
  buf[len+1] = '\n';
  SYM(protocol, write_data)(client->session, buf, len + 2);
  SYM(protocol, shutdown_session)(client->session);
}

static void
s_wrong_parameters(struct rfc1459_client *client, const struct message *msg)
{
  size_t len;
  const char *targetname = msg->dest ? msg->dest : "*";

  if ((msg->strc > 0) && msg->strv[0])
    {
      const char *reason = msg->strv[0];
      len = snprintf(buf, 511, ":%s 461 %s * :%s", msg->origin, targetname, reason);
    }
  else
    {
      len = snprintf(buf, 511, ":%s 461 %s * :%s", msg->origin, targetname, _c(client, "Not enough parameters"));
    }
  len = min(len, 510);
  buf[len] = '\r';
  buf[len+1] = '\n';
  SYM(protocol, write_data)(client->session, buf, len + 2);
}

static void
s_handshake_required(struct rfc1459_client *client, const struct message *msg)
{
  size_t len;
  const char *targetname = msg->dest ? msg->dest : "*";

  len = snprintf(buf, 511, ":%s 451 %s * :%s", msg->origin, targetname, _c(client, "Register first"));
  len = min(len, 510);
  buf[len] = '\r';
  buf[len+1] = '\n';
  SYM(protocol, write_data)(client->session, buf, len + 2);
}

static void
s_node_information(struct rfc1459_client *client, const struct message *msg)
{
  size_t len;

  if (msg->uint32v[0])
    {
      const char *name = msg->strv[0] ? msg->strv[0] : "*";
      const char *hostname = msg->strv[1] ? msg->strv[1] : "*";
      const char *realname = msg->strv[2] ? msg->strv[2] : "";
      if (msg->strc < 3)
        {
          log(MYNAME, LOG_PRIORITY_ERROR, "Can't send '%s' message: insufficient string components (%d < %d)",
              SYM(message, get_message_type_name)(msg->type), msg->strc, 3);
          return;
        }

      len = snprintf(buf, 511, ":%s 311 %s %s %s %s * :%s", msg->origin, msg->dest, name, name, hostname, realname);
      len = min(len, 510);
      buf[len] = '\r';
      buf[len+1] = '\n';
      SYM(protocol, write_data)(client->session, buf, len + 2);

      len = snprintf(buf, 511, ":%s 318 %s %s :%s", msg->origin, msg->dest, name, _c(client, "End of /WHOIS list"));
      len = min(len, 510);
      buf[len] = '\r';
      buf[len+1] = '\n';
      SYM(protocol, write_data)(client->session, buf, len + 2);
    }
  else
    {
      if (msg->strc > 1)
        {
          len = snprintf(buf, 511, ":%s 401 %s %s :%s", msg->origin, msg->dest, msg->strv[0], msg->strv[1]);
        }
      else
        {
          len = snprintf(buf, 511, ":%s 401 %s %s :%s", msg->origin, msg->dest, msg->strv[0],
                         _c(client, "No such nick/channel"));
        }
      len = min(len, 510);
      buf[len] = '\r';
      buf[len+1] = '\n';
      SYM(protocol, write_data)(client->session, buf, len + 2);
    }
}

static void
s_relaying_denied(struct rfc1459_client *client, const struct message *msg)
{
  size_t len;

  len = snprintf(buf, 511, ":%s 481 %s :%s", msg->origin, msg->dest, _c(client, "You can't do that, you're no deity"));
  len = min(len, 510);
  buf[len] = '\r';
  buf[len+1] = '\n';
  SYM(protocol, write_data)(client->session, buf, len + 2);
}

static void
s_node_software_information(struct rfc1459_client *client, const struct message *msg)
{
  size_t len;

  len = snprintf(buf, 511, ":%s 351 %s %s-%s(%s) %s :%s", msg->origin, msg->dest,
                 msg->strv[0], msg->strv[1], msg->strv[2], msg->origin, msg->strv[4]);
  len = min(len, 510);
  buf[len] = '\r';
  buf[len+1] = '\n';
  SYM(protocol, write_data)(client->session, buf, len + 2);
}

static void
s_node_license_information(struct rfc1459_client *client, const struct message *msg)
{
  size_t len;
  int i;

  for (i = 0; i < msg->strc; i++)
    {    
      len = snprintf(buf, 511, ":%s 371 %s :%s", msg->origin, msg->dest, msg->strv[i]);
      len = min(len, 510);
      buf[len] = '\r';
      buf[len+1] = '\n';
      SYM(protocol, write_data)(client->session, buf, len + 2);
    }
  len = snprintf(buf, 511, ":%s 374 %s :%s", msg->origin, msg->dest, _c(client, "End of /LICENSE list"));
  len = min(len, 510);
  buf[len] = '\r';
  buf[len+1] = '\n';
  SYM(protocol, write_data)(client->session, buf, len + 2);
}

static void
s_node_author_information(struct rfc1459_client *client, const struct message *msg)
{
  size_t len;
  int i;

  for (i = 0; i < msg->strc; i++)
    {    
      len = snprintf(buf, 511, ":%s 371 %s :%s", msg->origin, msg->dest, msg->strv[i]);
      len = min(len, 510);
      buf[len] = '\r';
      buf[len+1] = '\n';
      SYM(protocol, write_data)(client->session, buf, len + 2);
    }
  len = snprintf(buf, 511, ":%s 374 %s :%s", msg->origin, msg->dest, _c(client, "End of /INFO list"));
  len = min(len, 510);
  buf[len] = '\r';
  buf[len+1] = '\n';
  SYM(protocol, write_data)(client->session, buf, len + 2);
}

static void
s_private(struct rfc1459_client *client, const struct message *msg)
{
  size_t len;

  char *native_string = convert_from_utf8(client->charset, msg->strv[0]);
  len = snprintf(buf, 511, ":%s PRIVMSG %s :%s", msg->origin, msg->dest, native_string);
  len = min(len, 510);
  buf[len] = '\r';
  buf[len+1] = '\n';
  SYM(protocol, write_data)(client->session, buf, len + 2);
  xfree(native_string);
}

static void
s_ctcp_query(struct rfc1459_client *client, const struct message *msg)
{
  size_t len;

  char *native_string = convert_from_utf8(client->charset, msg->strv[0]);
  len = snprintf(buf, 510, ":%s PRIVMSG %s :\01%s", msg->origin, msg->dest, native_string);
  len = min(len, 509);
  /* Ensure that the trailing \01 is always added */
  buf[len] = '\01';
  buf[len+1] = '\r';
  buf[len+2] = '\n';
  SYM(protocol, write_data)(client->session, buf, len + 3);
  xfree(native_string);
}

static void
s_ctcp_response(struct rfc1459_client *client, const struct message *msg)
{
  size_t len;

  char *native_string = convert_from_utf8(client->charset, msg->strv[0]);
  len = snprintf(buf, 510, ":%s NOTICE %s :\01%s", msg->origin, msg->dest, native_string);
  len = min(len, 509);
  /* Ensure that the trailing \01 is always added */
  buf[len] = '\01';
  buf[len+1] = '\r';
  buf[len+2] = '\n';
  SYM(protocol, write_data)(client->session, buf, len + 3);
  xfree(native_string);
}

static void
s_channel(struct rfc1459_client *client, const struct message *msg)
{
  size_t len;

  char *native_string = convert_from_utf8(client->charset, msg->strv[1]);
  len = snprintf(buf, 511, ":%s PRIVMSG #%s :%s", msg->origin, msg->strv[0], native_string);
  len = min(len, 510);
  buf[len] = '\r';
  buf[len+1] = '\n';
  SYM(protocol, write_data)(client->session, buf, len + 2);
  xfree(native_string);
}

static void
s_channel_ctcp_query(struct rfc1459_client *client, const struct message *msg)
{
  size_t len;

  char *native_string = convert_from_utf8(client->charset, msg->strv[1]);
  len = snprintf(buf, 510, ":%s PRIVMSG #%s :\01%s", msg->origin, msg->strv[0], native_string);
  len = min(len, 509);
  /* Ensure that the trailing \01 is always added */
  buf[len] = '\01';
  buf[len+1] = '\r';
  buf[len+2] = '\n';
  SYM(protocol, write_data)(client->session, buf, len + 3);
  xfree(native_string);
}

static void
s_accept_auth(struct rfc1459_client *client, const struct message *msg)
{
  size_t len;

  len = snprintf(buf, 511, ":%s 051 %s :%s", msg->origin, msg->dest, _c(client, "Authentication accepted"));
  len = min(len, 510);
  buf[len] = '\r';
  buf[len+1] = '\n';
  SYM(protocol, write_data)(client->session, buf, len + 2);
}

static void
s_cancel_auth(struct rfc1459_client *client, const struct message *msg)
{
  size_t len;

  if (msg->strc > 0)
    len = snprintf(buf, 511, ":%s 052 %s :%s", msg->origin, msg->dest, msg->strv[0]);
  else
    len = snprintf(buf, 511, ":%s 052 %s :%s", msg->origin, msg->dest, _c(client, "Authentication cancelled"));
  len = min(len, 510);
  buf[len] = '\r';
  buf[len+1] = '\n';
  SYM(protocol, write_data)(client->session, buf, len + 2);
}

static void
s_auth_data(struct rfc1459_client *client, const struct message *msg)
{
  size_t len;
  char *p;
  int i;

  len = snprintf(buf, 511, ":%s 050 %s", msg->origin, msg->dest);
  len = min(len, 510);

  p = buf + len;
  for (i = 0; i < msg->strc; i++)
    {
      size_t l;
      /* We need at least two free bytes, or else there's no point even trying */
      if (len >= 509)
        break;
      l = snprintf(p, 511 - len, " %s", msg->strv[i]);
      len += l;
      p += l;
    }

  len = min(len, 510);
  buf[len] = '\r';
  buf[len+1] = '\n';
  SYM(protocol, write_data)(client->session, buf, len + 2);
}

static void
s_not_authenticated(struct rfc1459_client *client, const struct message *msg)
{
  size_t len;

  len = snprintf(buf, 511, ":%s 399 %s :%s", msg->origin, msg->dest, _c(client, "Authenticate first"));
  len = min(len, 510);
  buf[len] = '\r';
  buf[len+1] = '\n';
  SYM(protocol, write_data)(client->session, buf, len + 2);
}

static void
s_not_authorised(struct rfc1459_client *client, const struct message *msg)
{
  size_t len;

  len = snprintf(buf, 511, ":%s 399 %s :%s", msg->origin, msg->dest, _c(client, "You are not authorised to perform that operation"));
  len = min(len, 510);
  buf[len] = '\r';
  buf[len+1] = '\n';
  SYM(protocol, write_data)(client->session, buf, len + 2);
}

static void
s_accept_set_credentials(struct rfc1459_client *client, const struct message *msg)
{
  size_t len;

  len = snprintf(buf, 511, ":%s 399 %s :%s", msg->origin, msg->dest, _c(client, "Credentials updated"));
  len = min(len, 510);
  buf[len] = '\r';
  buf[len+1] = '\n';
  SYM(protocol, write_data)(client->session, buf, len + 2);
}

static void
s_reject_set_credentials(struct rfc1459_client *client, const struct message *msg)
{
  size_t len;

  if (msg->strc >= 2)
    len = snprintf(buf, 511, ":%s 399 %s :%s%s", msg->origin, msg->dest,
                   _c(client, "Credentials not updated: "), msg->strv[1]);
  else    
    len = snprintf(buf, 511, ":%s 399 %s :%s", msg->origin, msg->dest,
                   _c(client, "Credentials not updated: invalid authentication method specified"));
  len = min(len, 510);
  buf[len] = '\r';
  buf[len+1] = '\n';
  SYM(protocol, write_data)(client->session, buf, len + 2);
}


static void
s_accept_register_user(struct rfc1459_client *client, const struct message *msg)
{
  size_t len;

  len = snprintf(buf, 511, ":%s 399 %s :%s", msg->origin, msg->dest,
                 _c(client, "Your username has been registered"));
  len = min(len, 510);
  buf[len] = '\r';
  buf[len+1] = '\n';
  SYM(protocol, write_data)(client->session, buf, len + 2);
}

static void
s_reject_register_user(struct rfc1459_client *client, const struct message *msg)
{
  size_t len;

  if (msg->strc >= 1)
    len = snprintf(buf, 511, ":%s 399 %s :%s%s", msg->origin, msg->dest,
                   _c(client, "User not registered: "), msg->strv[0]);
  else    
    len = snprintf(buf, 511, ":%s 399 %s :%s", msg->origin, msg->dest,
                   _c(client, "User not registered"));
  len = min(len, 510);
  buf[len] = '\r';
  buf[len+1] = '\n';
  SYM(protocol, write_data)(client->session, buf, len + 2);
}

static void
s_registered_user(struct rfc1459_client *client, const struct message *msg)
{
  size_t len;

  len = snprintf(buf, 511, ":%s 399 %s :%s", msg->origin, msg->dest,
                 _c(client, "This username has been registered. Please authenticate."));
  len = min(len, 510);
  buf[len] = '\r';
  buf[len+1] = '\n';
  SYM(protocol, write_data)(client->session, buf, len + 2);
}

static void
s_unregistered_user(struct rfc1459_client *client, const struct message *msg)
{
  size_t len;

  len = snprintf(buf, 511, ":%s 399 %s :%s", msg->origin, msg->dest,
                 _c(client, "This username has not been registered. Please register."));
  len = min(len, 510);
  buf[len] = '\r';
  buf[len+1] = '\n';
  SYM(protocol, write_data)(client->session, buf, len + 2);
}

static void
s_invalid_destination(struct rfc1459_client *client, const struct message *msg)
{
  size_t len;

  len = snprintf(buf, 511, ":%s 401 %s %s :%s", msg->origin, msg->dest, msg->strv[0],
                 _c(client, "No such nick/channel"));
  len = min(len, 510);
  buf[len] = '\r';
  buf[len+1] = '\n';
  SYM(protocol, write_data)(client->session, buf, len + 2);
}

static void
s_echo_request(struct rfc1459_client *client, const struct message *msg)
{
  size_t len;

  if (msg->strc >= 1)
    len = snprintf(buf, 511, ":%s PING %s :%s", msg->origin, msg->dest, msg->strv[0]);
  else
    len = snprintf(buf, 511, ":%s PING %s", msg->origin, msg->dest);
  len = min(len, 510);
  buf[len] = '\r';
  buf[len+1] = '\n';
  SYM(protocol, write_data)(client->session, buf, len + 2);
}

static void
s_echo_reply(struct rfc1459_client *client, const struct message *msg)
{
  size_t len;

  if (msg->strc >= 1)
    len = snprintf(buf, 511, ":%s PONG %s :%s", msg->origin, msg->dest, msg->strv[0]);
  else
    len = snprintf(buf, 511, ":%s PONG %s", msg->origin, msg->dest);
  len = min(len, 510);
  buf[len] = '\r';
  buf[len+1] = '\n';
  SYM(protocol, write_data)(client->session, buf, len + 2);
}

static struct rfc1459_message rfc1459_messages_[] = 
  {
    {"accept handshake", 0, &s_accept_handshake, RMC_ORIGIN | RMC_DEST, 0, 0, 0},
    {"reject set name", 0, &s_reject_set_name, RMC_ORIGIN, 1, 0, 0},
    {"reject set real name", 0, &s_reject_set_real_name, 0, 1, 0, 0},
    {"wrong parameters", 0, &s_wrong_parameters, RMC_ORIGIN, 0, 0, 0},
    {"relaying denied", 0, &s_relaying_denied, RMC_ORIGIN | RMC_DEST, 1, 0, 0},
    {"handshake required", 0, &s_handshake_required, RMC_ORIGIN, 0, 0, 0},
    {"node information", 0, &s_node_information, RMC_ORIGIN | RMC_DEST, 1, 1, 0},
    {"node software information", 0, &s_node_software_information, RMC_ORIGIN | RMC_DEST, 5, 0, 0},
    {"node license information", 0, &s_node_license_information, RMC_ORIGIN | RMC_DEST, 0, 0, 0},
    {"node author information", 0, &s_node_author_information, RMC_ORIGIN | RMC_DEST, 0, 0, 0},
    {"private", 0, &s_private, RMC_ORIGIN | RMC_DEST, 1, 0, 0},
    {"ctcp query", 0, &s_ctcp_query, RMC_ORIGIN | RMC_DEST, 1, 0, 0},
    {"ctcp response", 0, &s_ctcp_response, RMC_ORIGIN | RMC_DEST, 1, 0, 0},
    {"channel", 0, &s_channel, RMC_ORIGIN, 2, 0, 0},
    {"channel ctcp query", 0, &s_channel_ctcp_query, RMC_ORIGIN, 2, 0, 0},
    {"done handshake", 0, NULL, RMC_ORIGIN | RMC_DEST, 0, 0, 0},
    {"accept auth", 0, &s_accept_auth, RMC_ORIGIN | RMC_DEST, 0, 0, 0},
    {"cancel auth", 0, &s_cancel_auth, RMC_ORIGIN | RMC_DEST, 0, 0, 0},
    {"auth data", 0, &s_auth_data, RMC_ORIGIN | RMC_DEST, 0, 0, 0},
    {"not authenticated", 0, &s_not_authenticated, RMC_ORIGIN | RMC_DEST, 0, 1, 0},
    {"not authorised", 0, &s_not_authorised, RMC_ORIGIN | RMC_DEST, 0, 1, 0},
    {"accept set credentials", 0, &s_accept_set_credentials, RMC_ORIGIN | RMC_DEST, 1, 0, 0},
    {"reject set credentials", 0, &s_reject_set_credentials, RMC_ORIGIN | RMC_DEST, 1, 0, 0},
    {"accept register user", 0, &s_accept_register_user, RMC_ORIGIN | RMC_DEST, 0, 0, 0},
    {"reject register user", 0, &s_reject_register_user, RMC_ORIGIN | RMC_DEST, 0, 0, 0},
    {"registered user", 0, &s_registered_user, RMC_ORIGIN | RMC_DEST, 0, 0, 0},
    {"unregistered user", 0, &s_unregistered_user, RMC_ORIGIN | RMC_DEST, 0, 0, 0},
    {"invalid destination", 0, &s_invalid_destination, RMC_ORIGIN | RMC_DEST, 1, 0, 0},
    {"echo request", 0, &s_echo_request, RMC_ORIGIN | RMC_DEST, 0, 0, 0},
    {"echo reply", 0, &s_echo_reply, RMC_ORIGIN | RMC_DEST, 0, 0, 0},
    {NULL, 0, NULL, 0, 0, 0, 0},
  };

struct rfc1459_message *rfc1459_messages = rfc1459_messages_;

int max_rfc1459_message_type;
struct rfc1459_message **rfc1459_message_types;

void
init_rfc1459_messages(void)
{
  struct rfc1459_message *m;
  max_rfc1459_message_type = -1;
  for (m = rfc1459_messages; m->message_type_name; m++)
    {
      m->message_type = SYM(message, register_message_type)(m->message_type_name);
      max_rfc1459_message_type =
        (max_rfc1459_message_type > m->message_type) ? max_rfc1459_message_type : m->message_type;
    }
  rfc1459_message_types = xmalloc(sizeof(struct rfc1459_message *) * (max_rfc1459_message_type + 1));
  for (m = rfc1459_messages; m->message_type_name; m++)
    rfc1459_message_types[m->message_type] = m;

  accept_handshake = SYM(message, register_message_type)("accept handshake");
}
