/************************************************************************
 *   This is part of Dancer, a text relay daemon
 *   Copyright (C) 2002 Andrew Suffield <asuffield@users.sourceforge.net>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

/** \file scheduler.c
 *
 * A scheduler decides when and in what order messages will be processed.
 */

#include <dancerd/config.h>
#include <dancerd/hashtable.h>
#include <dancerd/protocol/message.h>
#include <dancerd/modules.h>
#include <dancerd/protocol/scheduler.h>
#include <dancerd/llist.h>
#include <dancerd/xmalloc.h>
#include <dancerd/xstring.h>

EXTERN_MODULE();

struct scheduler
{
  char *name;
  queue_packet_handler *qph;
  add_session_handler *ash;
  remove_session_handler *rsh;
  int active;
  struct llist *sessions;
};

static struct hashtable *scheduler_hash = NULL;

DEFINE_SYMBOL(struct scheduler *, scheduler, register_scheduler, (const char *name, queue_packet_handler *qph,
                                                                  add_session_handler *ash, remove_session_handler *rsh))
{
  struct scheduler *sched;

  if (!name || !qph)
    return NULL;

  if (!scheduler_hash)
    scheduler_hash = new_hashtable(&hash_ascii, 61);

  if (!(sched = hashtable_lookup(scheduler_hash, name)))
    {
      sched = xmalloc(sizeof(struct scheduler));
      sched->name = xstrdup(name);
      sched->sessions = new_llist();
    }
  else
    {
      if (sched->active)
        {
          log(MYNAME, LOG_PRIORITY_ERROR, "Attempt to reregister active scheduler %s", name);
          return NULL;
        }
    }

  sched->qph = qph;
  sched->ash = ash;
  sched->rsh = rsh;
  sched->active = 1;

  hashtable_insert(scheduler_hash, name, sched);

  if (ash)
    {
      struct llist_node *n;
      for (n = llist_head(sched->sessions); n; n = llist_node_next(n))
        (*ash)(llist_node_data(n));
    }

  return sched;
}

DEFINE_SYMBOL(void, scheduler, unregister_scheduler, (struct scheduler *sched))
{
  if (!sched || !scheduler_hash)
    return;

  sched->active = 0;
  sched->qph = NULL;
  sched->ash = NULL;
  sched->rsh = NULL;
}

DEFINE_SYMBOL(struct scheduler *, scheduler, get_scheduler, (const char *name))
{
  struct scheduler *sched;

  if (!name || !scheduler_hash)
    return NULL;

  sched = hashtable_lookup(scheduler_hash, name);
  if (sched->active)
    return sched;

  return NULL;
}

DEFINE_SYMBOL(void, scheduler, add_session, (struct session *session, struct scheduler *sched))
{
  if (!sched || !session)
    return;

  if (sched->active && sched->ash)
    (*sched->ash)(session);

  add_llist_tail(sched->sessions, session);
}

DEFINE_SYMBOL(void, scheduler, remove_session, (struct session *session, struct scheduler *sched))
{
  if (!sched || !session)
    return;

  if (sched->active && sched->rsh)
    (*sched->rsh)(session);

  delete_llist_data(sched->sessions, session);
}

DEFINE_SYMBOL(void, scheduler, queue_message, (struct session *session, struct scheduler *sched, struct message *msg))
{
  if (!session || !sched || !msg)
    return;

  /* We do not queue messages when the scheduler is not loaded. That
   * is what the scheduler is *for*.
   */
  if (!sched->active)
    {
      LSYM(message, free_message)(msg);
      LSYM(protocol, shutdown_session)(session);
      return;
    }

  (*sched->qph)(session, msg);
}
