/************************************************************************
 *   This is part of Dancer, a text relay daemon
 *   Copyright (C) 2002 Andrew Suffield <asuffield@users.sourceforge.net>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

#include <dancerd/config.h>
#include <dancerd/xmalloc.h>
#include <dancerd/auth/auth.h>
#include <dancerd/clock.h>
#include <dancerd/xstring.h>

#include <assert.h>

#include "node.h"
#include "handlers.h"
#include "send.h"

EXTERN_MODULE();

#include "auth.c.imptab"

void
m_begin_auth(struct peer *peer unused_attribute, struct node *origin, const struct message *msg)
{
  char *err;
  assert(msg->strc >= 2);
  SYM(auth, cancel_handle)(origin->auth);
  origin->next_authenticated_name = xstrdup(msg->strv[0]);
  origin->next_authenticated = 0;
  origin->auth = SYM(auth, new_handle)(msg->strv[0], msg->strv[1], &err);
  if (!origin->auth)
    send_cancel_auth(origin, err);
}

void
m_cancel_auth(struct peer *peer unused_attribute, struct node *origin, const struct message *msg unused_attribute)
{
  if (origin->auth)
    {
      SYM(auth, cancel_handle)(origin->auth);
    }
  send_cancel_auth(origin, NULL);
  origin->auth = NULL;
}

void
m_accept_auth(struct peer *peer unused_attribute, struct node *origin, const struct message *msg unused_attribute)
{
  char *err;
  if (origin->auth)
    {
      switch (SYM(auth, peer_accept)(origin->auth, &err))
        {
        case AUTH_ACCEPT:
          send_accept_auth(origin);
          origin->auth = NULL;
          if (strcmp(origin->next_authenticated_name, origin->name) == 0)
            {
              origin->authenticated = 1;
              origin->next_authenticated = 0;
              xfree(origin->next_authenticated_name);
              origin->next_authenticated_name = NULL;
            }
          else
            origin->next_authenticated = 1;
          break;
        case AUTH_CANCEL:
          send_cancel_auth(origin, err);
          origin->auth = NULL;
          break;
        case AUTH_CONTINUE:
          {
            int argc;
            char **argv;
            switch (SYM(auth, get_data)(origin->auth, &argc, &argv, &err))
              {
              case AUTH_ACCEPT:
                send_accept_auth(origin);
                origin->auth = NULL;
                if (strcmp(origin->next_authenticated_name, origin->name) == 0)
                  {
                    origin->authenticated = 1;
                    origin->next_authenticated = 0;
                    xfree(origin->next_authenticated_name);
                    origin->next_authenticated_name = NULL;
                  }
                else
                  origin->next_authenticated = 1;
                break;
              case AUTH_CANCEL:
                send_cancel_auth(origin, err);
                origin->auth = NULL;
                break;
              case AUTH_CONTINUE:
                send_auth_data(origin, argc, argv);
                break;
              default:
                abort();
              }
            break;
          }
        default:
          abort();
        }
    }
  else
    send_cancel_auth(origin, NULL);
}

void
m_auth_data(struct peer *peer unused_attribute, struct node *origin, const struct message *msg)
{
  char *err;
  if (origin->auth)
    {
      switch (SYM(auth, read_data)(origin->auth, msg->strc, msg->strv, &err))
        {
        case AUTH_ACCEPT:
          send_accept_auth(origin);
          origin->auth = NULL;
          if (strcmp(origin->next_authenticated_name, origin->name) == 0)
            {
              origin->authenticated = 1;
              origin->next_authenticated = 0;
              xfree(origin->next_authenticated_name);
              origin->next_authenticated_name = NULL;
            }
          else
            origin->next_authenticated = 1;
          break;
        case AUTH_CANCEL:
          send_cancel_auth(origin, err);
          origin->auth = NULL;
          break;
        case AUTH_CONTINUE:
          {
            int argc;
            char **argv;
            switch (SYM(auth, get_data)(origin->auth, &argc, &argv, &err))
              {
              case AUTH_ACCEPT:
                send_accept_auth(origin);
                origin->auth = NULL;
                break;
              case AUTH_CANCEL:
                send_cancel_auth(origin, err);
                origin->auth = NULL;
                break;
              case AUTH_CONTINUE:
                send_auth_data(origin, argc, argv);
                break;
              default:
                abort();
              }
            break;
          }
        default:
          abort();
        }
    }
  else
    send_cancel_auth(origin, NULL);
}

void
m_set_credentials(struct peer *peer unused_attribute, struct node *origin, const struct message *msg)
{
  char *err;
  if (SYM(auth, set_credentials)(origin->name, msg->strv[0], msg->strc - 1, msg->strv + 1, &err))
    {
      send_accept_set_credentials(origin, msg->strv[0]);
    }
  else
    {
      send_reject_set_credentials(origin, msg->strv[0], err);
    }
}

void
m_register_user(struct peer *peer unused_attribute, struct node *origin, const struct message *msg)
{
  char *err;
  void *registration_time;
  /* Is this user available? */
  if (SYM(auth, get_entity_attribute)(origin->name, "registration time", &registration_time, 0, &err))
    {
      if (registration_time)
        {
          send_reject_register_user(origin, "User is already registered");
          return;
        }
    }
  else
    {
      /* Error retrieving registration time */
      send_reject_register_user(origin, err);
      return;
    }
  /* Okay, we're allowed to regiser this one */
  /* First, store the creation time */
  if (!SYM(auth, set_entity_attribute)(origin->name, "registration time", &current_time, sizeof(current_time), &err))
    {
      send_reject_register_user(origin, err);
      return;
    }
  /* Next, store the real name */
  SYM(auth, set_entity_attribute)(origin->name, "real name", msg->strv[0], strlen(msg->strv[0]) + 1, NULL);
  /* Tag the user as authenticated */
  origin->authenticated = 1;
  /* Inform the user that registration was successful */
  send_accept_register_user(origin);
}
