/************************************************************************
 *   This is part of Dancer, a text relay daemon
 *   Copyright (C) 2002 Andrew Suffield <asuffield@users.sourceforge.net>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

#include <dancerd/config.h>
#include <dancerd/llist.h>
#include <dancerd/modules.h>
#include <dancerd/protocol/message.h>
#include <dancerd/protocol/protocol.h>
#include <dancerd/auth/auth.h>
#include <dancerd/xmalloc.h>
#include <dancerd/hashtable.h>

#include <assert.h>
#include <limits.h>

#include "router.h"
#include "node.h"
#include "handlers.h"
#include "send.h"

EXTERN_MODULE();

#include "handlers.c.imptab"

static router_message_filter_func reject_filter;
static int
accept_filter(struct peer *peer unused_attribute, struct node *origin unused_attribute,
              struct node *target unused_attribute, const struct message *msg unused_attribute)
{
  return 1;
}

static router_message_filter_func reject_filter;
static int
reject_filter(struct peer *peer unused_attribute, struct node *origin unused_attribute,
              struct node *target unused_attribute, const struct message *msg unused_attribute)
{
  return 0;
}

static struct message_handler message_handlers_[] =
  {
  /*  name,      0, handler,     filter,         required, */
    {"set name", 0, &m_set_name, &reject_filter, AON_HANDSHAKE,
  /* Ranges (min, max for each):
   * string,    uint32,    sint32
   */
     1, 1,      0, 0,      0, 0},

    {"set real name", 0, &m_set_real_name, &reject_filter, AON_HANDSHAKE,
     1, 1,      0, 0,      0, 0},

    {"query node information", 0, &m_query_node_information, &reject_filter, 0,
     1, 1,      0, 0,      0, 0},

    {"query node software information", 0, &m_query_node_software_information, &reject_filter, 0,
     0, 0,      0, 0,      0, 0},

    {"query node license information", 0, &m_query_node_license_information, &reject_filter, 0,
     0, 0,      0, 0,      0, 0},

    {"query node author information", 0, &m_query_node_author_information, &reject_filter, 0,
     0, 0,      0, 0,      0, 0},

    {"done handshake", 0, &m_done_handshake, &reject_filter, AON_HANDSHAKE,
     0, 0,      0, 0,      0, 0},

    {"accept handshake", 0, &m_accept_handshake, &reject_filter, RMC_ORIGIN | RMC_DEST | AON_HANDSHAKE,
     0, 0,      0, 0,      0, 0},

    {"reject set name", 0, &m_reject_name, &reject_filter, AON_HANDSHAKE,
     1, 2,      0, 0,      0, 0},

    {"reject set real name", 0, &m_reject_real_name, &reject_filter, AON_HANDSHAKE,
     1, 2,      0, 0,      0, 0},

    {"close session", 0, &m_close_session, &reject_filter, AON_HANDSHAKE,
     1, 2,      0, 0,      0, 0},

    {"private", 0, &m_private, &f_private, RMC_ORIGIN | RMC_DEST,
     1, 1,      0, 0,      0, 0},

    {"begin auth", 0, &m_begin_auth, &reject_filter, RMC_ORIGIN,
     2, 2,      0, 0,      0, 0},

    {"cancel auth", 0, &m_cancel_auth, &reject_filter, RMC_ORIGIN,
     0, 1,      0, 0,      0, 0},

    {"accept auth", 0, &m_accept_auth, &reject_filter, RMC_ORIGIN,
     0, 0,      0, 0,      0, 0},

    {"auth data", 0, &m_auth_data, &reject_filter, RMC_ORIGIN,
     0, INT_MAX,0, 0,      0, 0},

    {"set credentials", 0, &m_set_credentials, &reject_filter, RMC_ORIGIN | RON_AUTH,
     1, INT_MAX,0, 0,      0, 0},

    {"register user", 0, &m_register_user, &reject_filter, RMC_ORIGIN,
     1, 1,      0, 0,      0, 0},

    {"begin connection", 0, &m_begin_connection, &reject_filter, RMC_ORIGIN,
     4, INT_MAX,0, 0,      0, 0},

    {"query routable nodes", 0, &m_query_routable_nodes, &reject_filter, RMC_ORIGIN,
     0, 0,      0, 0,      0, 0},

    {"done routable nodes", 0, &m_done_routable_nodes, &reject_filter, RMC_ORIGIN | RON_ROUTER,
     0, 0,      0, 0,      0, 0},

    {"routable node", 0, &m_routable_node, &reject_filter, RMC_ORIGIN | RON_ROUTER,
     2, 3,      0, 0,      0, 0},

    {"invalid destination", 0, &m_invalid_destination, &error_filter, RMC_ORIGIN | RMC_DEST,
     1, 1,      1, 1,      0, 0},

    {"echo request", 0, &m_echo_request, &accept_filter, RMC_ORIGIN,
     0, 0,      0, 0,      0, 0},

    {"echo reply", 0, &m_echo_reply, &accept_filter, RMC_ORIGIN,
     0, 0,      0, 0,      0, 0},

    {NULL, 0, NULL, &reject_filter, 0, 0, 0, 0, 0, 0, 0}
  };

struct message_handler *message_handlers = message_handlers_;
static message_handler_func handle_close_session;

void
do_close_session(struct node *origin, struct node *target, const char *reason, int suppress_messages)
{
  /* The graph should be acyclic; cycles are bugs */
  assert(!target->is_closing);
  target->is_closing = true;

  /* If this is a local node, make sure the session is shut down, and that we won't get called again */
  if (target->adjacent == me)
    {
      assert(target->peer);
      SYM(message, ignore_type_from_session)(MYHANDLE, target->peer->session, MESSAGE_TYPE_SHUTDOWN, &handle_close_session, target);
      SYM(protocol, shutdown_session)(target->peer->session);
    }

  /* Recurse through far peers of this target, suppressing
   * notifications for them
   */
  struct llist_node *n;
  while ((n = llist_head(target->peers)))
    {
      struct node *current = llist_node_data(n);
      do_close_session(origin, current, NULL, 1);
      /* This assertion checks that do_close_session has removed the
       * offending node from target->peers
       */
      assert(n != llist_head(target->peers));
    }

  if (!suppress_messages)
    /* Tell any interested routers about this closed session */
    for (n = llist_head(routers); n; n = llist_node_next(n))
      {
        struct node *current = llist_node_data(n);
        
        /* If node is reached via the path to current, then we don't
         * want to send a notification to current - it already knows
         */
        if (target->peer == current->peer)
          continue;
        
        /* Otherwise send it out
         */
        send_close_session(origin, current, target, reason);
      }

  if (target->name)
    hashtable_delete(node_hash, target->name);

  SYM(auth, cancel_handle)(target->auth);
  free_node(target);
}

static void
handle_close_session(struct session *session unused_attribute, const struct message *msg unused_attribute, void *d)
{
  do_close_session(NULL, d, NULL, 0);
}

int
new_session(struct session *session, void *d unused_attribute, int outgoing)
{
  struct peer *peer = make_peer(session);
  struct node *node = make_node(peer);

  peer->node = node;
  peer->outgoing = outgoing;

  add_llist_tail(me->peers, node);
  node->adjacent_peers_entry = llist_tail(me->peers);

  node->adjacent = me;

  /* The node will be added to the routers list at the time when it is
   * sent a list of routable nodes - send_routable_nodes() in
   * handshake.c
   *
   * Doing it earlier could lead to duplicate 'routable node' messages
   * being sent.
   */

  SYM(message, request_type_from_session)(MYHANDLE, session, MESSAGE_TYPE_SHUTDOWN, &handle_close_session, node);

  if (outgoing)
    {
      send_set_name(peer->node, NULL, peer->me->name);
      send_set_real_name(peer->node, NULL, peer->me->realname);
      send_done_handshake(peer->node);
      peer->sent_my_handshake = true;
    }

  return 1;
}

/* This intermediate handler checks the parameters and calls the real
 * handler with more agreeable arguments
 */
static void
initial_message_handler(struct session *session, const struct message *msg, void *d)
{
  struct message_handler *mh = d;
  struct node *origin;
  struct peer *peer = SYM(protocol, get_attribute)(session, router_attribute);

  if (!peer->accepted_peer_handshake || !peer->accepted_my_handshake)
    {
      if (!(mh->required & AON_HANDSHAKE))
        {
          send_handshake_required(peer->node, msg->type);
          return;
        }
      /* The origin is always the peer before handshaking has completed */
      origin = peer->node;
    }
  else
    {
      /* If the peer is not a router, the origin *must* be the
       * peer. We'll ignore whatever the peer claimed in the origin
       * field in this case. We'll also allow the origin field to be
       * empty, although the protocol does _not_.
       */
      if (!peer->node->router)
        origin = peer->node;
      else
        {
          /* All messages must have an origin after handshaking */
          if (!msg->origin)
            {
              send_wrong_parameters(peer->node, msg->type, _n(peer->node, "No origin specified"));
              return;
            }
          origin = hashtable_lookup(node_hash, msg->origin);
          if (!origin)
            {
              send_wrong_parameters(peer->node, msg->type, _n(peer->node, "Origin not known"));
              return;
            }
        }
    }

  /* All messages with RMC_ORIGIN must always have an origin */
  if (!msg->origin && (mh->required & RMC_ORIGIN))
    {
      send_wrong_parameters(peer->node, msg->type, _n(peer->node, "No origin specified"));
      return;
    }

  if (!msg->dest && (mh->required & RMC_DEST))
    {
      send_wrong_parameters(peer->node, msg->type, _n(peer->node, "No destination specified"));
      return;
    }

  if ((mh->required & RON_AUTH) && (!origin || !origin->authenticated))
    {
      send_not_authenticated(peer->node, msg->type);
      return;
    }

  if ((mh->required & RON_ROUTER) && !origin->router)
    {
      /* \todo Should report router-only commands as non-existant to
       * non-routers, not drop their connections
       */
      send_fatal_error(peer->node, _n(peer->node, "Not a router"));
      return;
    }

  if (msg->strc < mh->min_strc)
    {
      send_wrong_parameters(peer->node, msg->type, _n(peer->node, "Not enough string parameters"));
      return;
    }
  if (msg->strc > mh->max_strc)
    {
      send_wrong_parameters(peer->node, msg->type, _n(peer->node, "Too many string parameters"));
      return;
    }
  if (msg->uint32c < mh->min_uint32c)
    {
      send_wrong_parameters(peer->node, msg->type, _n(peer->node, "Not enough unsigned 32-bit integer parameters"));
      return;
    }
  if (msg->uint32c > mh->max_uint32c)
    {
      send_wrong_parameters(peer->node, msg->type, _n(peer->node, "Too many unsigned 32-bit integer parameters"));
      return;
    }
  if (msg->sint32c < mh->min_sint32c)
    {
      send_wrong_parameters(peer->node, msg->type, _n(peer->node, "Not enough signed 32-bit integer parameters"));
      return;
    }
  if (msg->sint32c > mh->max_sint32c)
    {
      send_wrong_parameters(peer->node, msg->type, _n(peer->node, "Too many signed 32-bit integer parameters"));
      return;
    }

  if (msg->dest)
    {
      if (strcmp(msg->dest, peer->me->name) != 0)
        {
          /* This is a request for us to route the message
           * We don't route anything for an unregistered host
           */
          if (peer->accepted_peer_handshake && peer->accepted_my_handshake)
            {
              struct node *dest = hashtable_lookup(node_hash, msg->dest);
              if (!dest)
                {
                  /* We don't know who that dest is. Bounce the message */
                  send_invalid_destination(origin, msg->dest, msg->type);
                  return;
                }
              if (dest->peer == peer)
                {
                  /* Routing failure. We should not have received that
                   * message. For now, kill the connection
                   */
                  send_fatal_error(peer->node, _n(peer->node, "Received message travelling in the wrong direction"));
                  return;
                }
              if (!mh->filter || (*mh->filter)(peer, origin, dest, msg))
                {
                  SYM(protocol, send_message)(dest->peer->session, SYM(message, dup_message)(msg));
                }
              else
                {
                  send_relaying_denied(origin, dest);
                }
              return;
            }
        }
      /* Dest is us, so we carry on */
    }

  (*mh->handler)(peer, origin, msg);
}

void
init_message_handlers(void)
{
  struct message_handler *mh;

  for (mh = message_handlers; mh->type_name; mh++)
    {
      assert(mh->handler);
      mh->type = SYM(message, register_message_type)(mh->type_name);
      SYM(message, request_type_from_all)(MYHANDLE, mh->type, &initial_message_handler, mh);
    }
}

void
deinit_message_handlers(void)
{
  struct message_handler *mh;

  for (mh = message_handlers; mh->type_name; mh++)
    {
      SYM(message, ignore_type_from_all)(MYHANDLE, mh->type, &initial_message_handler, mh);
    }
}
