/************************************************************************
 *   This is part of Dancer, a text relay daemon
 *   Copyright (C) 2002 Andrew Suffield <asuffield@users.sourceforge.net>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

#include <dancerd/config.h>
#include <dancerd/xmalloc.h>
#include <dancerd/xstring.h>
#include <dancerd/protocol/message.h>

#include <string.h>

#include "router.h"
#include "node.h"
#include "handlers.h"
#include "send.h"

EXTERN_MODULE();

#include "handshake.c.imptab"

void m_set_name(struct peer *peer unused_attribute, struct node *origin, const struct message *msg)
{
  char buf[33];
  size_t len = strlen(msg->strv[0]);
  len = min(len, 32);
  memcpy(buf, msg->strv[0], len);
  buf[len] = '\0';
  if (origin->name)
    {
      if (strcmp(origin->name, buf) != 0)
        {
          /* If the handshake has not completed, this cancels any
           * previous 'accept handshake' message from the peer
           */
          if (!peer->accepted_peer_handshake || !peer->accepted_my_handshake)
            peer->accepted_my_handshake = false;
        }
    }

  if (hashtable_lookup(node_hash, buf))
    {
      /* NULL reason indicates that the name is already in use */
      send_reject_set_name(origin, buf, NULL);
      return;
    }

  /* If there was already a name set... */
  if (origin->name)
    {
      /* Acknowledge the change */
      send_set_name(origin, origin->name, buf);
      hashtable_delete(node_hash, origin->name);
    }

  /* Copy the name into the node struct */
  origin->name = xrealloc(origin->name, len + 1);
  memcpy(origin->name, buf, len);
  origin->name[len] = '\0';
  hashtable_insert(node_hash, origin->name, origin);

  /* Handle pre-authentication magic */
  if (origin->next_authenticated_name && origin->next_authenticated
      && strcmp(origin->name, origin->next_authenticated_name) == 0)
    {
      /* It has pre-authenticated for this name, so atomically
       * authenticate while changing name
       */
      xfree(origin->next_authenticated_name);
      origin->next_authenticated_name = NULL;
      origin->authenticated = true;
      origin->next_authenticated = false;
      send_accept_auth(origin);
    }
  else
    origin->authenticated = false;
}

void m_set_real_name(struct peer *peer unused_attribute, struct node *origin, const struct message *msg)
{
  /* We always accept real names, at present */

  /* Acknowledge the change, if there was already a realname set */
  if (origin->realname)
    send_set_real_name(origin, origin, origin->realname);

  /* Copy it into the node struct */
  size_t len = strlen(msg->strv[0]);
  len = min(len, 32);
  origin->realname = xrealloc(origin->realname, len + 1);
  memcpy(origin->realname, msg->strv[0], len);
  origin->realname[len] = '\0';
}

/** \brief Send a full list of nodes that we will route messages for
 * \param node Destination to send the list to
 *
 * Sends the list of all the nodes which are _not_ reached via
 * node->peer - ie, all the nodes which node should reach by routing
 * messages to us.
 */
static void
send_routable_nodes(struct node *node)
{
  struct llist_node *n;
  for (n = llist_head(global_nodes); n; n = llist_node_next(n))
    {
      struct node *current = llist_node_data(n);

      /* If current is reached via the path to node, then we don't
       * want to offer a route for it
       */
      if (current->peer == node->peer)
        continue;

      /* Otherwise, send it */
      send_routable_node(node, current);
    }

  send_done_routable_nodes(node);

  /* Add node to the list of routers, so that it receives messages for
   * any new routable nodes which are introduced later
   */
  add_llist_tail(routers, node);
  node->routers_entry = llist_tail(routers);
  node->router = true;
}

static void
finished_handshake(struct node *origin)
{
  /* If the user is not authenticated, then warn them about that */
  if (!origin->authenticated)
    {
      void *registration_time;
      /* FIXME: what if this is false? */
      if (SYM(auth, get_entity_attribute)(origin->name, "registration time", &registration_time, 0, NULL))
        {
          /* If the username is registered, warn the user to
           * authenticate, otherwise warn them to register
           */
          if (registration_time)
            send_registered_user(origin);
          else
            send_unregistered_user(origin);
        }
    }

  /* Tell any interested routers about this new node */
  struct llist_node *n;
  for (n = llist_head(routers); n; n = llist_node_next(n))
    {
      struct node *current = llist_node_data(n);

      /* At this stage, routers is precisely the set of things which
       * need to know about this node
       */
      send_routable_node(current, origin);
    }

  if (origin->peer->outgoing)
    {
      send_query_routable_nodes(origin);
      send_routable_nodes(origin);
    }
}

void m_done_handshake(struct peer *peer,
                      struct node *origin, const struct message *msg unused_attribute)
{
  if (peer->node->name && peer->node->realname)
    {
      peer->accepted_peer_handshake = true;
      if (!peer->sent_my_handshake)
        {
          send_set_name(peer->node, NULL, peer->me->name);
          send_set_real_name(peer->node, NULL, peer->me->realname);
          send_done_handshake(peer->node);
          peer->sent_my_handshake = true;
        }
      send_accept_handshake(peer->node);
      if (peer->accepted_my_handshake)
        finished_handshake(origin);
    }
}

void m_accept_handshake(struct peer *peer, struct node *origin, const struct message *msg)
{
  if (!peer->node->name)
    {
      send_fatal_error(peer->node, _n(peer->node, "Handshake failed (your name was not set)"));
      return;
    }
  if (strcmp(peer->node->name, msg->origin) != 0)
    {
      send_fatal_error(peer->node, _n(peer->node, "Handshake failed (your name does not match)"));
      return;
    }
  if (strcmp(peer->me->name, msg->dest) != 0)
    {
      send_fatal_error(peer->node, _n(peer->node, "Handshake failed (my name does not match)"));
      return;
    }
  peer->accepted_my_handshake = true;
  if (peer->accepted_peer_handshake)
    finished_handshake(origin);
}

void
m_begin_connection(struct peer *peer unused_attribute, struct node *origin unused_attribute,
                   const struct message *msg)
{
  SYM(protocol, begin_connection)(msg->strv[0], msg->strv[1], msg->strv[2], msg->strv[3],
                                  NULL, 0);
}

void
m_query_routable_nodes(struct peer *peer unused_attribute, struct node *origin,
                       const struct message *msg unused_attribute)
{
  send_routable_nodes(origin);
}

void
m_done_routable_nodes(struct peer *peer unused_attribute, struct node *origin,
                       const struct message *msg unused_attribute)
{
  log(MYNAME, LOG_PRIORITY_INFO, "Connection to router %s established", origin->name);
}

void
m_routable_node(struct peer *peer, struct node *origin,
                const struct message *msg)
{
  if (strlen(msg->strv[0]) > 32)
    {
      send_fatal_error(origin, "Name too long");
      return;
    }
  
  if (hashtable_lookup(node_hash, msg->strv[0]))
    {
      /* FIXME: We should usually be killing one of the clones here,
       * not the router. Exact logic will have to wait until the auth
       * layer is ready.
       */
      send_fatal_error(origin, "Duplicate node detected");
      return;
    }

  struct node *adjacent = NULL;
  if (msg->strc > 2 && !(adjacent = hashtable_lookup(node_hash, msg->strv[2])))
    {
      send_fatal_error(origin, "Adjacent node name is not known to me");
      return;
    }
  if (!adjacent)
    adjacent = origin;

  struct node *node = make_node(peer);

  node->name = xstrdup(msg->strv[0]);
  hashtable_insert(node_hash, node->name, node);

  node->realname = xstrdup(msg->strv[1]);
  node->adjacent = adjacent;
  add_llist_tail(adjacent->peers, node);
  node->adjacent_peers_entry = llist_tail(adjacent->peers);

  /* Tell any interested routers about this new node */
  struct llist_node *n;
  for (n = llist_head(routers); n; n = llist_node_next(n))
    {
      struct node *current = llist_node_data(n);

      /* If node is reached via the path to current, then we don't
       * want to send a notification to current - it has the right
       * route already
       */
      if (node->peer == current->peer)
        continue;

      /* Otherwise advertise the route
       */
      send_routable_node(current, node);
    }
}
