/************************************************************************
 *   This is part of Dancer, a text relay daemon
 *   Copyright (C) 2002 Andrew Suffield <asuffield@users.sourceforge.net>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

#ifdef ROUTER_PEER_H
#else
#define ROUTER_PEER_H

#include <dancerd/config.h>
#include <dancerd/protocol/protocol.h>
#include <dancerd/auth/auth.h>
#include <dancerd/hashtable.h>
#include <dancerd/gettext.h>

/** \file router/node.h
 * \brief Node structures for the message router
 */

/** \defgroup routernode Node structures for the message router
 * \ingroup router
 *
 * Every entity in the network that can send and receive messages is a
 * node.
 *
 * A peer is a node which is adjacent to this one.
 *
 * @{
 */

struct peer;

/** \brief A node I communicate with
 *
 * A node is an entity in the network that can send and receive messages.
 *
 * It contains connection-independant data associated with the node
 */
struct node
{
  /** \brief Name of the node */
  char *name;
  /** \brief True if the node has authenticated for its current name */
  int authenticated;
  /** \brief 'next' authenticated name for the node */
  char *next_authenticated_name;
  /** \brief true if next authentication has completed */
  int next_authenticated;
  /** \brief "Real name" of the node
   *
   * This is a descriptive, free-form field
   */
  char *realname;
  /** \brief The peer to send data to in order to find this node
   *
   * This node is locally connected iff node == node->peer->node
   *
   * This node is me iff peer == NULL
   */
  struct peer *peer;
  /** \brief All the nodes which are connected to this node, excluding
   * the one which is nearest to us.
   *
   * This provides us with a way to recurse through the tree.
   */
  struct llist *peers;
  /** \brief The node nearest us that this node is connected to.
   *
   * If this node is local, then adjacent == me.
   */
  struct node *adjacent;
  /** \brief The node in adjacent->peers which references this one
   */
  struct llist_node *adjacent_peers_entry;
  /** \brief The language used by this node
   */
  char *language;
  /** \brief The handle for the current authentication session, or NULL
   */
  struct auth_handle *auth;
  /** \brief The llist_node in global_nodes which references this node */
  struct llist_node *global_nodes_entry;
  /** \brief The llist_node in routers which references this node,
   * or NULL if this node is not a router
   */
  struct llist_node *routers_entry;
  /** \brief True if this node is a router.
   *
   * Used for basic, fast checks on whether message arguments are
   * valid or not.
   *
   * For example, only a router may send 'routable node' messages, or
   * send messages where the origin is not itself.
   */
  bool router;
  /** \brief True if this node is being processed by
   * do_close_session. This is used to trap re-entry.
   */
  bool is_closing;
};

/** \brief A peer I communicate with
 *
 * A peer is another node with an immediate connection to me; it has
 * an associated session.
 *
 * It contains data associated with this specific connection.
 */
struct peer
{
  /** \brief The node that this peer is a connection to
   */
  struct node *node;
  /** \brief The 'me' that this peer is talking to
   *
   * This is present to support virtual hosts, if they are ever
   * implemented.
   */
  struct node *me;
  /** \brief The session I have with this peer */
  struct session *session;
  /** \brief True if I have accepted the peer's handshake */
  bool accepted_peer_handshake;
  /** \brief True if the peer has accepted my handshake */
  bool accepted_my_handshake;
  /** \brief True if I have sent my handshake */
  bool sent_my_handshake;
  /** \brief True if I initiated this connection */
  bool outgoing;
};

extern struct hashtable *node_hash;

extern struct node *make_node(struct peer *) malloc_attribute;
extern struct peer *make_peer(struct session *) malloc_attribute;

extern void free_node(struct node *);
extern void free_peer(struct peer *);

/* do_close_session is defined in handlers.c */
extern void do_close_session(struct node *, struct node *, const char *, int);

/* Mirror the logic in gettext.h here; if ENABLE_NLS is not defined or
 * false, we inline the strings
 */
#if ENABLE_NLS
extern const char *node_gettext(struct node *, const char *);
extern const char *node_ngettext(struct node *, const char *, const char *, unsigned long int);

#define _n(C, S) node_gettext(C, S)
#define _nn(C, S, SP, N) node_ngettext(C, S, SP, N)
#else
#define _n(C, S) ((const char *)(C))
#define _nn(C, S, SP, N) ((N) == 1 ? (const char *) (S) : (const char *) (SP))
#endif

/** @}
 */

#endif
