/************************************************************************
 *   This is part of Dancer, a text relay daemon
 *   Copyright (C) 2002 Andrew Suffield <asuffield@users.sourceforge.net>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

#include <dancerd/config.h>
#include <dancerd/config/items.h>
#include <dancerd/modules.h>
#include <dancerd/log.h>
#include <dancerd/xstring.h>
#include <dancerd/gettext.h>
#include <dancerd/llist.h>

#include <stddef.h>
#include <string.h>
#include <locale.h>
#include <assert.h>

#include "handlers.h"
#include "router.h"
#include "send.h"
#include "node.h"

SETUP_MODULE("router", MODVERSION);

#include "symtab"

struct node *me = NULL;
struct llist *routers = NULL;
struct llist *global_nodes = NULL;
int router_attribute;

const char *
node_gettext(struct node *node, const char *msgid)
{
  const char *translated;
  char *saved_language = NULL;
  if (node->language)
    {
      saved_language = xstrdup(setlocale(LC_MESSAGES, NULL));
      setlocale(LC_MESSAGES, node->language);
    }
  translated = gettext(msgid);
  if (saved_language)
    {
      setlocale(LC_MESSAGES, saved_language);
      xfree(saved_language);
    }
  return translated;
}

const char *
node_ngettext(struct node *node, const char *msgid, const char *msgid_plural, unsigned long int n)
{
  const char *translated;
  char *saved_language = NULL;
  if (node->language)
    {
      saved_language = xstrdup(setlocale(LC_MESSAGES, NULL));
      setlocale(LC_MESSAGES, node->language);
    }
  translated = ngettext(msgid, msgid_plural, n);
  if (saved_language)
    {
      setlocale(LC_MESSAGES, saved_language);
      xfree(saved_language);
    }
  return translated;
}

int
MODULE_INIT(void *data unused_attribute)
{
  struct config_item name_item, real_name_item;
  static char name_key_name[] = "router.name";
  static char real_name_key_name[] = "router.real_name";

  if (!SYM(config, get_item)(name_key_name, &name_item))
    {
      log(MYNAME, LOG_PRIORITY_CRITICAL, "Couldn't get config item %s", name_key_name);
      return MODULE_LOAD_FAILED;
    }
  if (name_item.type != citem_string)
    {
      log(MYNAME, LOG_PRIORITY_CRITICAL, "config key %s is not a string type (%d)", name_key_name, real_name_item.type);
      return MODULE_LOAD_FAILED;
    }
  /* We need to preserve the value that get_item returned, because the
   * next call can overwrite it... but we don't want to malloc yet,
   * since the next one might fail. So we'll put it on the stack and
   * copy it twice.
   */
  size_t name_len = strlen(name_item.str);
  char name[name_len + 1];
  memcpy(name, name_item.str, name_len + 1);

  if (!SYM(config, get_item)(real_name_key_name, &real_name_item))
    {
      log(MYNAME, LOG_PRIORITY_CRITICAL, "Couldn't get config item %s", real_name_key_name);
      return MODULE_LOAD_FAILED;
    }
  if (real_name_item.type != citem_string)
    {
      log(MYNAME, LOG_PRIORITY_CRITICAL, "config key %s is not a string type (%d)", real_name_key_name, real_name_item.type);
      return MODULE_LOAD_FAILED;
    }

  router_attribute = SYM(protocol, register_attribute)("router");

  routers = new_llist();
  global_nodes = new_llist();  
  me = make_node(NULL);

  me->realname = xstrdup(real_name_item.str);
  me->name = xmemdup(name, name_len + 1);

  /* me shouldn't be in global_nodes - that is defined as the list of
   * things which are not me
   */
  delete_llist_node(me->global_nodes_entry);
  me->global_nodes_entry = NULL;

  node_hash = new_hashtable(&hash_ascii, 137);
  hashtable_insert(node_hash, me->name, me);

  init_send();
  init_message_handlers();

  SYM(protocol, register_new_session_handler)(MYHANDLE, &new_session, NULL);

  return MODULE_LOAD_OK;
}

int
MODULE_DEINIT(const char *new_version, module_init_func *init)
{
  SYM(protocol, unregister_new_session_handler)(MYHANDLE, &new_session, NULL);
  do_close_session(me, me, "Router module unloading", 0);
  deinit_message_handlers();
  assert(!llist_head(global_nodes));
  free_llist(routers);
  free_llist(global_nodes);
  free_hashtable(node_hash);

  if (new_version)
    {
      switch ((*init)(NULL))
        {
        case MODULE_LOAD_OK:
          return MODULE_LOAD_OK;
        case MODULE_LOAD_FAILED:
          return MODULE_LOAD_FAILED;
        default:
          log(MYNAME, LOG_PRIORITY_ERROR, "Unknown return from new module version %s, aborting", new_version);
          return MODULE_LOAD_FAILED;
        }
    }

  return MODULE_UNLOAD_OK;
}
