/************************************************************************
 *   This is part of Dancer, a text relay daemon
 *   Copyright (C) 2002 Andrew Suffield <asuffield@users.sourceforge.net>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

#include <dancerd/config.h>
#include <dancerd/modules.h>
#include <dancerd/protocol/message.h>
#include <dancerd/xmalloc.h>

#include <assert.h>
#include <string.h>

#include "node.h"
#include "send.h"

EXTERN_MODULE();

#include "send.c.imptab"

static int wrong_parameters, accept_handshake, done_handshake, close_session, reject_set_name, handshake_required,
  fatal_error, node_information, set_name, set_real_name, relaying_denied, node_software_information,
  node_license_information, node_author_information, invalid_destination,
  accept_auth, cancel_auth, auth_data, not_authorised, not_authenticated,
  accept_set_credentials, reject_set_credentials, accept_register_user, reject_register_user,
  node_authenticated, node_not_authenticated, registered_user, unregistered_user,
  query_routable_nodes, done_routable_nodes, routable_node, echo_request, echo_reply;

static struct message_type
{
  int *var;
  const char *name;
} message_types[] =
  {
    {&wrong_parameters, "wrong parameters"},
    {&reject_set_name, "reject set name"},
    {&set_name, "set name"},
    {&set_real_name, "set real name"},
    {&accept_handshake, "accept handshake"},
    {&done_handshake, "done handshake"},
    {&close_session, "close session"},
    {&handshake_required, "handshake required"},
    {&fatal_error, "fatal error"},
    {&node_information, "node information"},
    {&node_software_information, "node software information"},
    {&node_license_information, "node license information"},
    {&node_author_information, "node author information"},
    {&relaying_denied, "relaying denied"},
    {&invalid_destination, "invalid destination"},
    {&accept_auth, "accept auth"},
    {&cancel_auth, "cancel auth"},
    {&auth_data, "auth data"},
    {&not_authenticated, "not authenticated"},
    {&not_authorised, "not authorised"},
    {&accept_set_credentials, "accept set credentials"},
    {&reject_set_credentials, "reject set credentials"},
    {&accept_register_user, "accept register user"},
    {&reject_register_user, "reject register user"},
    {&node_authenticated, "node authenticated"},
    {&node_not_authenticated, "node not authenticated"},
    {&registered_user, "registered user"},
    {&unregistered_user, "unregistered user"},
    {&query_routable_nodes, "query routable nodes"},
    {&done_routable_nodes, "done routable nodes"},
    {&routable_node, "routable node"},
    {&echo_request, "echo request"},
    {&echo_reply, "echo reply"},
    {NULL, NULL}
  };

void
init_send(void)
{
  struct message_type *mt;
  for (mt = message_types; mt->var; mt++)
    {
      (*mt->var) = SYM(message, register_message_type)(mt->name);
    }
}

void
send_wrong_parameters(struct node *node, int type, const char *reason)
{
  const char *origin, *dest;

  origin = node->peer->me->name;
  dest = node->name;

  SYM(protocol, send_message)(node->peer->session,
                              SYM(message, build_message)(wrong_parameters, origin, dest,
                                                          reason ? 1 : 0, (const char*[]) {reason},
                                                          1, (uint32_t[]) {type},
                                                          0, NULL));
}

void
send_handshake_required(struct node *node, int type)
{
  const char *origin, *dest;

  origin = node->peer->me->name;
  dest = node->name;

  SYM(protocol, send_message)(node->peer->session,
                              SYM(message, build_message)(handshake_required, origin, dest,
                                                          0, NULL,
                                                          1, (uint32_t[]) {type},
                                                          0, NULL));
}

void
send_invalid_destination(struct node *node, const char *invalid_dest, int type)
{
  const char *origin, *dest;

  origin = node->peer->me->name;
  dest = node->name;

  SYM(protocol, send_message)(node->peer->session,
                              SYM(message, build_message)(invalid_destination, origin, dest,
                                                          1, (const char*[]) {invalid_dest},
                                                          1, (uint32_t[]) {type},
                                                          0, NULL));
}

void
send_done_handshake(struct node *node)
{
  const char *origin, *dest;

  origin = node->peer->me->name;
  dest = node->name;

  SYM(protocol, send_message)(node->peer->session,
                              SYM(message, build_message)(done_handshake, origin, dest,
                                                          0, NULL,
                                                          0, NULL,
                                                          0, NULL));
}

void
send_accept_handshake(struct node *node)
{
  const char *origin, *dest;

  origin = node->peer->me->name;
  dest = node->name;

  SYM(protocol, send_message)(node->peer->session,
                              SYM(message, build_message)(accept_handshake, origin, dest,
                                                          0, NULL,
                                                          0, NULL,
                                                          0, NULL));
}

void
send_query_routable_nodes(struct node *node)
{
  const char *origin, *dest;

  origin = node->peer->me->name;
  dest = node->name;

  SYM(protocol, send_message)(node->peer->session,
                              SYM(message, build_message)(query_routable_nodes, origin, dest,
                                                          0, NULL,
                                                          0, NULL,
                                                          0, NULL));
}

void
send_done_routable_nodes(struct node *node)
{
  const char *origin, *dest;

  origin = node->peer->me->name;
  dest = node->name;

  SYM(protocol, send_message)(node->peer->session,
                              SYM(message, build_message)(done_routable_nodes, origin, dest,
                                                          0, NULL,
                                                          0, NULL,
                                                          0, NULL));
}

void
send_routable_node(struct node *node, struct node *target)
{
  const char *origin, *dest;

  origin = node->peer->me->name;
  dest = node->name;

  int local = (target->peer->node == target);
  if (!local)
    assert(target->adjacent);

  SYM(protocol, send_message)(node->peer->session,
                              SYM(message, build_message)(routable_node, origin, dest,
                                                          local ? 2 : 3,
                                                          local ? (const char *[]) {target->name, target->realname}
                                                          : (const char *[]) {target->name, target->realname, target->adjacent->name},
                                                          0, NULL,
                                                          0, NULL));
}

void
send_close_session(struct node *origin, struct node *dest, struct node *target, const char *reason)
{
  SYM(protocol, send_message)(dest->peer->session,
                              SYM(message, build_message)(close_session, origin ? origin->name : dest->peer->me->name, dest->name,
                                                          reason ? 2 : 1, (const char*[]) {target->name, reason},
                                                          0, NULL,
                                                          0, NULL));
}

void
send_reject_set_name(struct node *node, const char *new_name, const char *reason)
{
  const char *origin, *dest;

  origin = node->peer->me->name;
  dest = node->name;

  SYM(protocol, send_message)(node->peer->session,
                              SYM(message, build_message)(reject_set_name, origin, dest,
                                                          reason ? 2 : 1, (const char*[]) {new_name, reason},
                                                          0, NULL,
                                                          0, NULL));
}

void
send_fatal_error(struct node *node, const char *reason)
{
  const char *origin, *dest;

  origin = node->peer->me->name;
  dest = node->name;

  SYM(protocol, send_message)(node->peer->session,
                              SYM(message, build_message)(fatal_error, origin, dest,
                                                          reason ? 1 : 0, (const char*[]) {reason},
                                                          0, NULL,
                                                          0, NULL));
  SYM(protocol, shutdown_session)(node->peer->session);
}

void
send_relaying_denied(struct node *node, struct node *original_dest)
{
  const char *origin, *dest;

  origin = node->peer->me->name;
  dest = node->name;

  SYM(protocol, send_message)(node->peer->session,
                              SYM(message, build_message)(relaying_denied, origin, dest,
                                                          1, (const char*[]) {original_dest->name}, 
                                                          0, NULL,
                                                          0, NULL));
}

void
send_node_information(struct node *node, const char *name, const char *hostname, const char *realname)
{
  const char *origin, *dest;

  origin = node->peer->me->name;
  dest = node->name;

  SYM(protocol, send_message)(node->peer->session,
                              SYM(message, build_message)(node_information, origin, dest,
                                                          3, (const char*[]) {name, hostname, realname}, 
                                                          1, (uint32_t []) {1},
                                                          0, NULL));
}

void
send_failed_node_information(struct node *node, const char *name, const char *error)
{
  const char *origin, *dest;

  origin = node->peer->me->name;
  dest = node->name;

  SYM(protocol, send_message)(node->peer->session,
                              SYM(message, build_message)(node_information, origin, dest,
                                                          error ? 2 : 1, (const char*[]) {name, error},
                                                          1, (uint32_t[]) {0},
                                                          0, NULL));
}

void
send_set_name(struct node *node, const char *old_name, const char *new_name)
{
  const char *origin, *dest;

  origin = old_name;
  dest = node->name;

  SYM(protocol, send_message)(node->peer->session,
                              SYM(message, build_message)(set_name, origin, dest,
                                                          1, (const char*[]) {new_name},
                                                          0, NULL,
                                                          0, NULL));
}

void
send_set_real_name(struct node *dest, struct node *target, const char *new_name)
{
  SYM(protocol, send_message)(dest->peer->session,
                              SYM(message, build_message)(set_real_name, target ? target->name : NULL, dest->name,
                                                          1, (const char*[]) {new_name},
                                                          0, NULL,
                                                          0, NULL));
}

void
send_node_software_information(struct node *node, const char *package_name, const char *package_version,
                               const char *build_id,
                               const char *build_description, const char *package_description)
{
  const char *origin, *dest;

  origin = node->peer->me->name;
  dest = node->name;

  SYM(protocol, send_message)(node->peer->session,
                              SYM(message, build_message)(node_software_information, origin, dest,
                                                          5, (const char*[]) {package_name, package_version,
                                                                              build_id, build_description,
                                                                              package_description},
                                                          0, NULL,
                                                          0, NULL));
}

void
send_node_license_information(struct node *node, int lines, const char **license)
{
  const char *origin, *dest;

  origin = node->peer->me->name;
  dest = node->name;

  SYM(protocol, send_message)(node->peer->session,
                              SYM(message, build_message)(node_license_information, origin, dest,
                                                          lines, license, 
                                                          0, NULL,
                                                          0, NULL));
}

void
send_node_author_information(struct node *node, int lines, const char **authors)
{
  const char *origin, *dest;

  origin = node->peer->me->name;
  dest = node->name;

  SYM(protocol, send_message)(node->peer->session,
                              SYM(message, build_message)(node_author_information, origin, dest,
                                                          lines, authors, 
                                                          0, NULL,
                                                          0, NULL));
}

void
send_accept_auth(struct node *node)
{
  const char *origin, *dest;

  origin = node->peer->me->name;
  dest = node->name;

  SYM(protocol, send_message)(node->peer->session,
                              SYM(message, build_message)(accept_auth, origin, dest,
                                                          0, NULL,
                                                          0, NULL,
                                                          0, NULL));
}

void
send_cancel_auth(struct node *node, const char *reason)
{
  const char *origin, *dest;

  origin = node->peer->me->name;
  dest = node->name;

  SYM(protocol, send_message)(node->peer->session,
                              SYM(message, build_message)(cancel_auth, origin, dest,
                                                          reason ? 1 : 0, (const char*[]){reason},
                                                          0, NULL,
                                                          0, NULL));
  xfree(node->next_authenticated_name);
  node->next_authenticated_name = NULL;
  node->next_authenticated = 0;
}

void
send_not_authenticated(struct node *node, uint32_t type)
{
  const char *origin, *dest;

  origin = node->peer->me->name;
  dest = node->name;

  SYM(protocol, send_message)(node->peer->session,
                              SYM(message, build_message)(not_authenticated, origin, dest,
                                                          0, NULL,
                                                          1, (uint32_t[]) {type},
                                                          0, NULL));
}

void
send_not_authorised(struct node *node, uint32_t type)
{
  const char *origin, *dest;

  origin = node->peer->me->name;
  dest = node->name;

  SYM(protocol, send_message)(node->peer->session,
                              SYM(message, build_message)(not_authorised, origin, dest,
                                                          0, NULL,
                                                          1, (uint32_t[]) {type},
                                                          0, NULL));
}

void
send_auth_data(struct node *node, int argc, char **argv)
{
  const char *origin, *dest;

  if (!argc)
    return;

  origin = node->peer->me->name;
  dest = node->name;

  SYM(protocol, send_message)(node->peer->session,
                              SYM(message, build_message)(auth_data, origin, dest,
                                                          argc, (const char**)argv,
                                                          0, NULL,
                                                          0, NULL));
}

void
send_accept_set_credentials(struct node *node, const char *method)
{
  const char *origin, *dest;

  origin = node->peer->me->name;
  dest = node->name;

  SYM(protocol, send_message)(node->peer->session,
                              SYM(message, build_message)(accept_set_credentials, origin, dest,
                                                          1, (const char*[]) {method},
                                                          0, NULL,
                                                          0, NULL));
}

void
send_reject_set_credentials(struct node *node, const char *method, const char *reason)
{
  const char *origin, *dest;

  origin = node->peer->me->name;
  dest = node->name;

  SYM(protocol, send_message)(node->peer->session,
                              SYM(message, build_message)(reject_set_credentials, origin, dest,
                                                          reason ? 2 : 1, (const char*[]) {method, reason},
                                                          0, NULL,
                                                          0, NULL));
}

void
send_accept_register_user(struct node *node)
{
  const char *origin, *dest;

  origin = node->peer->me->name;
  dest = node->name;

  SYM(protocol, send_message)(node->peer->session,
                              SYM(message, build_message)(accept_register_user, origin, dest,
                                                          0, NULL,
                                                          0, NULL,
                                                          0, NULL));
}

void
send_reject_register_user(struct node *node, const char *reason)
{
  const char *origin, *dest;

  origin = node->peer->me->name;
  dest = node->name;

  SYM(protocol, send_message)(node->peer->session,
                              SYM(message, build_message)(reject_register_user, origin, dest,
                                                          reason ? 1 : 0, (const char*[]) {reason},
                                                          0, NULL,
                                                          0, NULL));
}

void
send_unregistered_user(struct node *node)
{
  const char *origin, *dest;

  origin = node->peer->me->name;
  dest = node->name;

  SYM(protocol, send_message)(node->peer->session,
                              SYM(message, build_message)(unregistered_user, origin, dest,
                                                          0, NULL,
                                                          0, NULL,
                                                          0, NULL));
}

void
send_registered_user(struct node *node)
{
  const char *origin, *dest;

  origin = node->peer->me->name;
  dest = node->name;

  SYM(protocol, send_message)(node->peer->session,
                              SYM(message, build_message)(registered_user, origin, dest,
                                                          0, NULL,
                                                          0, NULL,
                                                          0, NULL));
}

void
send_node_authenticated(struct node *node, struct node *target)
{
  const char *origin, *dest;

  origin = node->peer->me->name;
  dest = node->name;

  SYM(protocol, send_message)(node->peer->session,
                              SYM(message, build_message)(node_authenticated, origin, dest,
                                                          1, (const char*[]) {target->name},
                                                          0, NULL,
                                                          0, NULL));
}

void
send_node_not_authenticated(struct node *node, struct node *target)
{
  const char *origin, *dest;

  origin = node->peer->me->name;
  dest = node->name;

  SYM(protocol, send_message)(node->peer->session,
                              SYM(message, build_message)(node_not_authenticated, origin, dest,
                                                          1, (const char*[]) {target->name},
                                                          0, NULL,
                                                          0, NULL));
}

void
send_echo_request(struct node *node, const char *id)
{
  const char *origin, *dest;

  origin = node->peer->me->name;
  dest = node->name;

  SYM(protocol, send_message)(node->peer->session,
                              SYM(message, build_message)(echo_request, origin, dest,
                                                          id ? 1 : 0, (const char *[]) {id},
                                                          0, NULL,
                                                          0, NULL));
}

void
send_echo_reply(struct node *node, const char *id)
{
  const char *origin, *dest;

  origin = node->peer->me->name;
  dest = node->name;

  SYM(protocol, send_message)(node->peer->session,
                              SYM(message, build_message)(echo_reply, origin, dest,
                                                          id ? 1 : 0, (const char *[]) {id},
                                                          0, NULL,
                                                          0, NULL));
}
