/************************************************************************
 *   This is part of Dancer, a text relay daemon
 *   Copyright (C) 2002 Andrew Suffield <asuffield@users.sourceforge.net>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

#include <dancerd/config.h>
#include <dancerd/dbuffer.h>
#include <dancerd/fd.h>
#include <dancerd/modules.h>
#include <dancerd/log.h>
#include <dancerd/transport/transport.h>
#include <dancerd/xmalloc.h>
#include <dancerd/config/items.h>

#include <stddef.h>
#include <string.h>
#include <stdio.h>
#include <errno.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <unistd.h>

SETUP_MODULE("transport.ipv4.tcp", MODVERSION);
RELOAD_INTERFACE(0);

static struct transport *my_transport = NULL;

IMPORTED_SYMBOL(transport, register_transport);
IMPORTED_SYMBOL(transport, unregister_transport);
IMPORTED_SYMBOL(transport, class_name);
IMPORTED_SYMBOL(transport, transport_connected);
IMPORTED_SYMBOL(transport, data_read);
IMPORTED_SYMBOL(transport, connection_closed);
IMPORTED_SYMBOL(transport, write_buffer);
IMPORTED_SYMBOL(config, get_item);

#include "symtab"

#define READ_BLOCK 2048
#define WRITE_BLOCK 2048

struct listener_class
{
  int fd;
  uint32_t address;
  uint16_t port;
  struct transport_class *tc;
};

struct tcp_connection
{
  int fd;
  int established;
  struct sockaddr_in saddr;
  struct transport_class *tc;
  struct connection *conn;
  struct dbuffer *wbuf;
};

/* bind_address is an unsigned 32-bit address in network order. It must be a valid IP address on this host, or zero.
 * bind_port is an unsigned 32-bit port number in network order. It must be in the range 1-65534.
 */
static transport_init_tclass_handler init_tclass;
static void *
init_tclass(struct transport_class *tc, void *data unused_attribute)
{
  struct listener_class *lc;
  const char *class_name = SYM(transport, class_name)(tc);
  char *key_name = xmalloc(strlen("transport.classes.") + strlen(class_name) + strlen(".bind_address") + 1);
  struct sockaddr_in saddr;
  struct config_item item;
  int fd;
  int sock_opt = 1;

  log(MYNAME, LOG_PRIORITY_DEBUG, "Init tclass %s", class_name);
  sprintf(key_name, "transport.classes.%s.bind_address", class_name);
  SYM(config, get_item)(key_name, &item);
  if (item.type != citem_uint32)
    {
      log(MYNAME, LOG_PRIORITY_ERROR, "config key %s is not an unsigned 32-bit integer type", key_name);
      xfree(key_name);
      return NULL;
    }

  saddr.sin_addr.s_addr = item.uint32;

  sprintf(key_name, "transport.classes.%s.bind_port", class_name);
  SYM(config, get_item)(key_name, &item);
  if (item.type != citem_uint32)
    {
      log(MYNAME, LOG_PRIORITY_ERROR, "config key %s is not an unsigned 32-bit integer type", key_name);
      xfree(key_name);
      return NULL;
    }

  if ((item.uint32 < 1) || (item.uint32 > 65534))
    {
      log(MYNAME, LOG_PRIORITY_ERROR, "config key %s (%d) is not a valid port address", key_name, item.uint32);
      xfree(key_name);
      return NULL;
    }
  xfree(key_name);

  saddr.sin_port = item.uint32;
  saddr.sin_family = AF_INET;

  fd = socket(PF_INET, SOCK_STREAM, 0);
  if (fd == -1)
    {
      log(MYNAME, LOG_PRIORITY_ERROR, "Failed to create tcp socket in class %s: %s", class_name, strerror(errno));
      return NULL;
    }

  if (setsockopt(fd, SOL_SOCKET, SO_REUSEADDR, (void *)&sock_opt, sizeof(sock_opt)))
    {
      log(MYNAME, LOG_PRIORITY_ERROR, "Failed to set SO_REUSEADDR for tcp socket in class %s: %s",
          class_name, strerror(errno));
      close(fd);
      return NULL;
    }

  if (bind(fd, (struct sockaddr *)&saddr, sizeof(struct sockaddr_in)) == -1)
    {
      log(MYNAME, LOG_PRIORITY_ERROR, "Failed to bind tcp socket in class %s to %s:%d : %s",
          class_name, inet_ntoa(saddr.sin_addr), ntohs(saddr.sin_port), strerror(errno));
      close(fd);
      return NULL;
    }

  if (listen(fd, SOMAXCONN) == -1)
    {
      log(MYNAME, LOG_PRIORITY_ERROR, "Failed to listen on tcp socket %s: %s", class_name, strerror(errno));
      close(fd);
      return NULL;
    }

  lc = xmalloc(sizeof(struct listener_class));
  lc->address = saddr.sin_addr.s_addr;
  lc->port = saddr.sin_port;
  lc->fd = fd;
  lc->tc = tc;

  log(MYNAME, LOG_PRIORITY_DEBUG, "Listening for class %s on %s:%d", class_name,
      inet_ntoa(saddr.sin_addr), ntohs(saddr.sin_port));

  register_fd_handler(fd, FD_CLASS_CLIENT, MYNAME, 0, lc);

  return lc;
}

static transport_deinit_tclass_handler deinit_tclass;
static void
deinit_tclass(struct transport_class *tc, void *data)
{
  struct listener_class *lc = data;
  const char *class_name = SYM(transport, class_name)(tc);
  log(MYNAME, LOG_PRIORITY_DEBUG, "Deinit tclass %s", class_name);
  if (lc->fd > -1)
    {
      deregister_fd_handler(lc->fd);
      close(lc->fd);
    }
  xfree(lc);
}

static fd_read_handler listen_fd_read;
static int
listen_fd_read(int fd, void *data)
{
  struct listener_class *lc = data;
  struct tcp_connection *tconn;
  socklen_t len;

  tconn = xcalloc(sizeof(struct tcp_connection));
  len = sizeof(tconn->saddr);
  tconn->established = 1;
  tconn->tc = lc->tc;

  int new_fd = accept(fd, (struct sockaddr *)&tconn->saddr, &len);
  if (new_fd == -1)
    {
      log(MYNAME, LOG_PRIORITY_WARN, "Listening socket %d (%d, %p) was ready, but accept() failed: %s (%d)",
          fd, lc->fd, (void *)lc, strerror(errno), errno);
      xfree(tconn);
      /* This is not fatal, it just means that the new connection failed */
      return 1;
    }
  tconn->fd = new_fd;

  log(MYNAME, LOG_PRIORITY_DEBUG, "Accepted connection from %d on %d (%s:%d)", fd, tconn->fd,
      inet_ntoa(tconn->saddr.sin_addr), ntohs(tconn->saddr.sin_port));

  tconn->conn = SYM(transport, transport_connected)(my_transport, tconn->tc, tconn);
  tconn->wbuf = SYM(transport, write_buffer)(tconn->conn);
  register_fd_handler(tconn->fd, FD_CLASS_CLIENT, MYNAME, 1, tconn);

  return 1;
}

static fd_close_handler listen_fd_close;
static void
listen_fd_close(int fd, void *data)
{
  struct listener_class *lc = data;
  const char *class_name = SYM(transport, class_name)(lc->tc);
  log(MYNAME, LOG_PRIORITY_WARN, "Listener fd for class %s has been closed", class_name);
  close(fd);
  lc->fd = -1;
}

static fd_read_handler connection_fd_read;
static int
connection_fd_read(int fd, void *data)
{
  struct tcp_connection *tconn = data;
  char buf[READ_BLOCK];
  ssize_t len;
  if ((len = read(fd, buf, READ_BLOCK)) == -1)
    {
      log(MYNAME, LOG_PRIORITY_WARN, "Read on socket failed, connection will be closed: %s", strerror(errno));
      return 0;
    }
  if (len == 0)
    {
      /* This is a normal termination, don't warn about it */
      return 0;
    }

  log(MYNAME, LOG_PRIORITY_DEBUG, "Read %d bytes from fd %d", len, fd);
  SYM(transport, data_read)(tconn->conn, buf, len);
  return 1;
}

static fd_write_handler connection_fd_write;
static int
connection_fd_write(int fd, void *data)
{
  struct tcp_connection *tconn = data;
  char buf[WRITE_BLOCK];
  ssize_t len;

  if (!tconn->established)
    {
      /* That was the connect() call completing. Send the signal back up. */
      tconn->established = 1;
      SYM(transport, finished_connection)(tconn->conn);
      return 1;
    }

  len = dbuffer_get(tconn->wbuf, buf, WRITE_BLOCK);
  if (len == 0)
    {
      /* Don't come here again until we have data to write */
      block_fd_writes(fd);
    }

  if ((len = write(fd, buf, len)) == -1)
    {
      log(MYNAME, LOG_PRIORITY_WARN, "Write failed, connection will be closed: %s", strerror(errno));
      return 0;
    }

  log(MYNAME, LOG_PRIORITY_DEBUG, "Wrote %d bytes to fd %d", len, fd);
  dbuffer_remove(tconn->wbuf, len);
  return 1;
}

static fd_close_handler connection_fd_close;
static void
connection_fd_close(int fd, void *data)
{
  struct tcp_connection *tconn = data;
  log(MYNAME, LOG_PRIORITY_DEBUG, "Closed fd %d", fd);
  close(fd);
  SYM(transport, connection_closed)(tconn->conn);
  xfree(tconn);
}

static transport_begin_connection_handler do_connect;
static void *
do_connect(struct connection *conn, const char *target,
           void **argv unused_attribute, int argc unused_attribute)
{
  struct sockaddr_in addr;
  char host_target[64];
  char *p = strchr(target, ':');
  int fd;
  struct tcp_connection *tconn;
  int sock_opt = 1;

  /* Parse the target address as an ipv4 dotted-quad-and-port string
   *
   * Example: 1.2.3.4:6667
   */
  if (!p || ((p - target) > 63))
    {
      log(MYNAME, LOG_PRIORITY_ERROR, "Attempt to connect to invalid target '%s'", target);
      return NULL;
    }
  memcpy(host_target, target, p - target);
  host_target[p - target] = '\0';

  addr.sin_family = AF_INET;

  if (inet_pton(AF_INET, host_target, &addr.sin_addr) <= 0)
    {
      log(MYNAME, LOG_PRIORITY_ERROR, "Couldn't convert '%s' to an ipv4 address: %s",
          host_target, strerror(errno));
      return NULL;
    }
  addr.sin_port = htons(strtoul(p + 1, NULL, 10));

  /* Create a new socket and kick off the connection */
  fd = socket(PF_INET, SOCK_STREAM, 0);
  if (fd == -1)
    {
      log(MYNAME, LOG_PRIORITY_ERROR, "Failed to create tcp socket: %s", strerror(errno));
      return NULL;
    }

  if (setsockopt(fd, SOL_SOCKET, SO_REUSEADDR, (void *)&sock_opt, sizeof(sock_opt)))
    {
      log(MYNAME, LOG_PRIORITY_ERROR, "Failed to set SO_REUSEADDR for tcp socket: %s", strerror(errno));
      close(fd);
      return NULL;
    }

  if (!set_nonblocking_fd(fd))
    {
      log(MYNAME, LOG_PRIORITY_ERROR, "Failed to set tcp socket non-blocking: %s", strerror(errno));
      close(fd);
      return NULL;
    }

  if (connect(fd, (struct sockaddr *)&addr, sizeof(addr)) == -1)
    {
      if (errno != EINPROGRESS)
        {
          log(MYNAME, LOG_PRIORITY_ERROR, "Failed to begin connection to %s: %s", target, strerror(errno));
          close(fd);
          return NULL;
        }
    }

  tconn = xmalloc(sizeof(struct tcp_connection));
  tconn->fd = fd;
  tconn->established = 0;
  tconn->conn = conn;
  tconn->wbuf = SYM(transport, write_buffer)(tconn->conn);
  memcpy(&tconn->saddr, &addr, sizeof(addr));
  register_fd_handler(tconn->fd, FD_CLASS_CLIENT, MYNAME, 1, tconn);

  return tconn;
}

static transport_write_handler do_write;
static void
do_write(struct connection *conn unused_attribute, void *data,
         struct dbuffer *wbuf unused_attribute,
         const void *buf unused_attribute, size_t len unused_attribute)
{
  struct tcp_connection *tconn = data;
  unblock_fd_writes(tconn->fd);
}

static transport_close_handler do_close;
static void
do_close(struct connection *conn unused_attribute, void *handle)
{
  struct tcp_connection *tconn = handle;
  deregister_fd_handler(tconn->fd);
  close(tconn->fd);
  SYM(transport, connection_closed)(conn);
  xfree(tconn);
}

static transport_abort_handler do_abort;
static void
do_abort(struct connection *conn, void *handle)
{
  struct tcp_connection *tconn = handle;
  if (!tconn)
    return;
  deregister_fd_handler(tconn->fd);
  close(tconn->fd);
  SYM(transport, connection_closed)(conn);
  xfree(tconn);
}

/* There really isn't any data to reload here - the transport
 * interface itself stores the necessary pointers
 *
 * Therefore, don't even think about changing the ABI here. Doing so
 * requires bumping the reload interface version.
 */
int
MODULE_INIT(void *data unused_attribute)
{
  register_fd_handler_group(MYHANDLE, MYNAME, 0, &listen_fd_read, NULL, &listen_fd_close);
  register_fd_handler_group(MYHANDLE, MYNAME, 1, &connection_fd_read,
                            &connection_fd_write, &connection_fd_close);

  my_transport = SYM(transport, register_transport)("ipv4.tcp", MYHANDLE,
                                                    &init_tclass, NULL, &deinit_tclass,
                                                    &do_connect, &do_write, &do_close, &do_abort);

  return MODULE_LOAD_OK;
}

int
MODULE_DEINIT(const char *new_version, module_init_func *init)
{
  SYM(transport, unregister_transport)(my_transport);
  my_transport = NULL;

  if (new_version)
    {
      switch ((*init)(NULL))
        {
        case MODULE_LOAD_OK:
          return MODULE_LOAD_OK;
        case MODULE_LOAD_FAILED:
          register_fd_handler_group(MYHANDLE, MYNAME, 0, &listen_fd_read, NULL, &listen_fd_close);
          register_fd_handler_group(MYHANDLE, MYNAME, 1, &connection_fd_read,
                                    &connection_fd_write, &connection_fd_close);
          my_transport = SYM(transport, register_transport)("ipv4.tcp", MYHANDLE,
                                                            &init_tclass, NULL, &deinit_tclass,
                                                            &do_connect, &do_write, &do_close, &do_abort);
          if (!my_transport)
            return MODULE_LOAD_FUBAR;
          return MODULE_LOAD_FAILED;
        default:
          log(MYNAME, LOG_PRIORITY_ERROR, "Unknown return from new module version %s, aborting", new_version);
          register_fd_handler_group(MYHANDLE, MYNAME, 0, &listen_fd_read, NULL, &listen_fd_close);
          register_fd_handler_group(MYHANDLE, MYNAME, 1, &connection_fd_read,
                                    &connection_fd_write, &connection_fd_close);
          my_transport = SYM(transport, register_transport)("ipv4.tcp", MYHANDLE,
                                                            &init_tclass, NULL, &deinit_tclass,
                                                            &do_connect, &do_write, &do_close, &do_abort);
          if (!my_transport)
            return MODULE_LOAD_FUBAR;
          return MODULE_LOAD_FAILED;
        }
    }

  return MODULE_UNLOAD_OK;
}
