/************************************************************************
 *   This is part of Dancer, a text relay daemon
 *   Copyright (C) 2002 Andrew Suffield <asuffield@users.sourceforge.net>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

#include <dancerd/config.h>
#include <dancerd/modules.h>
#include <dancerd/log.h>
#include <dancerd/xmalloc.h>
#include <dancerd/xstring.h>
#include <dancerd/llist.h>
#include <dancerd/config/items.h>

#include <dancerd/transport/tclass.h>

#include <stddef.h>
#include <unistd.h>
#include <stdio.h>

EXTERN_MODULE();

IMPORTED_SYMBOL(config, get_item);

struct llist *classes = NULL;

CONSTRUCTOR()
{
  classes = new_llist();
}

static int
match_transport_class(const void *d1, const void *d2)
{
  const struct transport_class *tc1 = d1;
  const char *s = d2;
  return strcmp(tc1->name, s) == 0;
}

struct transport_class *
find_transport_class(const char *name)
{
  if (!name)
    return NULL;
  return llist_node_data(find_llist_match(classes, &match_transport_class, name));
}

void
read_transport_class(const char *name)
{
  char *key_name = xmalloc(strlen("transport.classes.") + strlen(name) + strlen(".transport") + 1);
  struct config_item item;
  struct transport_class *tc;
  sprintf(key_name, "transport.classes.%s.transport", name);
  
  SYM(config, get_item)(key_name, &item);
  if (item.type != citem_string)
    {
      log(MYNAME, LOG_PRIORITY_ERROR, "config key %s is not a string type", key_name);
      xfree(key_name);
      return;
    }
  if (!*item.str)
    {
      log(MYNAME, LOG_PRIORITY_ERROR, "config key %s contains an empty string", key_name);
      xfree(key_name);
      return;
    }
  xfree(key_name);

  tc = find_transport_class(name);
  if (!tc)
    {
      tc = xmalloc(sizeof(struct transport_class));
      tc->name = xstrdup(name);
      tc->trans = NULL;
      tc->tdata = NULL;
      tc->connections = new_llist();
      tc->transport_classes_node = NULL;
      add_llist_tail(classes, tc);
    }
  tc->active = 1;

  if (!tc->trans)
    {
      tc->trans = get_transport(item.str);
      /* Insert into the new transport class list */
      add_llist_tail(tc->trans->classes, tc);
      tc->transport_classes_node = llist_tail(tc->trans->classes);
    }
  else
    {
      /* If this is a different transport to the one we currently have... */
      if (strcmp(item.str, tc->trans->name) != 0)
        {
          /* Detach from the old transport */
          if (tc->trans->active)
            (*tc->trans->tdic)(tc, tc->tdata);
          else
            xfree(tc->tdata);
          delete_llist_node(tc->transport_classes_node);
          tc->trans = get_transport(item.str);
          /* Insert into the new transport class list */
          add_llist_tail(tc->trans->classes, tc);
          tc->transport_classes_node = llist_tail(tc->trans->classes);
        }
    }

  if (tc->trans->active)
    tc->tdata = (*tc->trans->tic)(tc, tc->tdata);
}

void
check_transport_class(struct transport_class *tc)
{
  /* Call deactivate on the transport if appropriate */
  if (!tc->active && tc->trans && tc->trans->tdac)
    (*tc->trans->tdac)(tc, tc->tdata);
  /* We can free a tclass if:
   *  It has no connections
   *  It is inactive
   *  Either:
   *   It has no transport
   *  Or:
   *   Its transport is active so we can disassociate it from the class
   */
  if (!llist_head(tc->connections) && !tc->active && (!tc->trans || tc->trans->active))
    {
      /* This class is no longer needed. Free it */
      log(MYNAME, LOG_PRIORITY_DEBUG, "Freeing tclass %s", tc->name);
      if (tc->trans)
        (*tc->trans->tdic)(tc, tc->tdata);
      else
        xfree(tc->tdata);
      delete_llist_data(classes, tc);
      if (tc->trans)
        delete_llist_node(tc->transport_classes_node);
      free_llist(tc->connections);
      xfree(tc->name);
      xfree(tc);
    }
}
