/************************************************************************
 *   This is part of Dancer, a text relay daemon
 *   Copyright (C) 2002 Andrew Suffield <asuffield@users.sourceforge.net>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

#include <dancerd/config.h>
#include <dancerd/modules.h>
#include <dancerd/log.h>
#include <dancerd/protocol/protocol.h>
#include <dancerd/transport/transport.h>
#include <dancerd/xmalloc.h>
#include <dancerd/xstring.h>
#include <dancerd/llist.h>
#include <dancerd/config/items.h>
#include <dancerd/event.h>
#include <dancerd/dbuffer.h>

#include <dancerd/transport/transport_internal.h>
#include <dancerd/transport/tclass.h>

#include <assert.h>
#include <stddef.h>
#include <string.h>
#include <unistd.h>

SETUP_MODULE("transport", MODVERSION);
RELOAD_INTERFACE(0);

struct llist *transports = NULL, *connections = NULL;

IMPORTED_SYMBOL(config, get_item);
IMPORTED_WEAK_SYMBOL(protocol, new_session);
IMPORTED_WEAK_SYMBOL(protocol, data_read);
IMPORTED_WEAK_SYMBOL(protocol, close_session);

#include "symtab"

static int
match_transport(const void *d1, const void *d2)
{
  const struct transport *t = d1;
  const char *s = d2;
  return strcmp(t->name, s) == 0;
}

static struct transport *
find_transport(const char *name)
{
  struct llist_node *n;
  if (!name)
    return NULL;
  n = find_llist_match(transports, &match_transport, name);
  return n ? llist_node_data(n) : NULL;
}

struct transport *
get_transport(const char *name)
{
  struct transport *trans = find_transport(name);
  if (trans)
    return trans;
  trans = xmalloc(sizeof(struct transport));
  trans->name = xstrdup(name);
  trans->module = NULL;
  trans->tic = NULL;
  trans->tdac = NULL;
  trans->tdic = NULL;
  trans->tbc = NULL;
  trans->tw = NULL;
  trans->tc = NULL;
  trans->ta = NULL;
  trans->connections = new_llist();
  trans->classes = new_llist();
  trans->destroyed_connections = new_llist();
  trans->shutdown_connections = new_llist();
  trans->active = 0;
  add_llist_head(transports, trans);
  return trans;
}

static int
read_config(void)
{
  struct config_item item;
  char *c, *str;
  struct llist_node *n, *next;
  static char key_name[] = "transport.classes";

  SYM(config, get_item)(key_name, &item);
  if (item.type != citem_string)
    {
      log(MYNAME, LOG_PRIORITY_ERROR, "config key transport.classes is not a string type (%d)", item.type);
      return 0;
    }

  for (n = llist_head(classes); n; n = llist_node_next(n))
    {
      struct transport_class *tc = llist_node_data(n);
      tc->active = 0;
    }

  str = xstrdup(item.str);
  /* Break str up on commas, and pass each element to read_transport_class() */
  if (*str)
    {
      char *p = str;
      while ((c = strchr(p, ',')))
        {
          *c = '\0';
          
          read_transport_class(p);
          
          p = c + 1;
        }
      
      read_transport_class(p);
    }
  xfree(str);

  for (n = llist_head(classes); n; n = next)
    {
      struct transport_class *tc = llist_node_data(n);
      next = llist_node_next(n);
      check_transport_class(tc);
    }

  return 1;
}

static void
invoke_init_tclass_handlers(void *d1, void *d2)
{
  struct transport_class *tc = d1;
  transport_init_tclass_handler *tic = (transport_init_tclass_handler *)d2;
  tc->tdata = (*tic)(tc, tc->tdata);
}

static void
invoke_deactivate_tclass_handlers(void *d1, void *d2)
{
  struct transport_class *tc = d1;
  transport_deactivate_tclass_handler *tdac = (transport_deactivate_tclass_handler *)d2;
  (*tdac)(tc, tc->tdata);
}

static void
unregister_transport_(void *d)
{
  struct transport *trans = d;
  /* The module might not be loaded now; we won't call the handlers
   * when automatically unregistering a transport. If the handlers
   * need to be called, the transport should unregister formally,
   * before it unloads.
   */
  trans->module = NULL;
  trans->tic = NULL;
  trans->tdac = NULL;
  trans->tdic = NULL;
  trans->tbc = NULL;
  trans->tw = NULL;
  trans->tc = NULL;
  trans->ta = NULL;
  trans->active = 0;
}

DEFINE_SYMBOL(struct transport *, transport, register_transport, (const char *name, struct module *module,
                                                                  transport_init_tclass_handler *tic,
                                                                  transport_deactivate_tclass_handler *tdac,
                                                                  transport_deinit_tclass_handler *tdic,
                                                                  transport_begin_connection_handler *tbc,
                                                                  transport_write_handler *tw,
                                                                  transport_close_handler *tc,
                                                                  transport_abort_handler *ta))
{
  struct transport *trans = get_transport(name);
  struct llist_node *n, *next;

  assert(trans);

  if (trans->active)
    {
      log(MYNAME, LOG_PRIORITY_ERROR, "Attempt to re-register active transport %s", name);
      return NULL;
    }

  trans->module = module;
  trans->tic = tic;
  trans->tdac = tdac;
  trans->tdic = tdic;
  trans->tbc = tbc;
  trans->tw = tw;
  trans->tc = tc;
  trans->ta = ta;
  trans->active = 1;

  /* Register a handler to unregister the transport if it is unloaded */
  if (trans->module)
    register_module_unload_handler(MYHANDLE, module, &unregister_transport_, trans);

  /* Properly clear any classes which became redundant while the transport was not active */
  for (n = llist_head(trans->classes); n; n = next)
    {
      struct transport_class *tclass = llist_node_data(n);
      next = llist_node_next(n);
      check_transport_class(tclass);
    }

  /* Call the init handler for all classes which were waiting for this transport to load */
  foreach_llist(trans->classes, &invoke_init_tclass_handlers, (void*)tic);

  for (n = llist_head(trans->destroyed_connections); n; n = next)
    {
      struct connection *conn = llist_node_data(n);
      next = llist_node_next(n);
      (*trans->ta)(conn, conn->data);
      delete_llist_node(n);
    }

  for (n = llist_head(trans->shutdown_connections); n; n = next)
    {
      struct connection *conn = llist_node_data(n);
      next = llist_node_next(n);
      (*trans->tc)(conn, conn->data);
      delete_llist_node(n);
    }

  if (tw)
    for (n = llist_head(trans->connections); n; n = next)
      {
        struct connection *conn = llist_node_data(n);
        next = llist_node_next(n);
        if (dbuffer_len(conn->wbuf) && conn->trans->tw)
          (*conn->trans->tw)(conn, conn->data, conn->wbuf, NULL, 0);
      }

  return trans;
}

static void
unregister_transport(struct transport *trans)
{
  if (trans->module)
    deregister_module_unload_handler(MYHANDLE, trans->module, &unregister_transport_, trans);
  trans->module = NULL;

  if (trans->tdac)
    foreach_llist(trans->classes, &invoke_deactivate_tclass_handlers, (void*)trans->tdac);

  trans->active = 0;
  trans->tic = NULL;
  trans->tdac = NULL;
  trans->tdic = NULL;
  trans->tw = NULL;
  trans->tc = NULL;
  trans->ta = NULL;
}

DEFINE_SYMBOL(void, transport, unregister_transport, (struct transport *trans))
{
  if (!trans)
    return;

  if (!trans->active)
    {
      log(MYNAME, LOG_PRIORITY_ERROR, "Attempt to de-register inactive transport %s", trans->name);
      return;
    }

  unregister_transport(trans);
}

DEFINE_SYMBOL(struct connection *, transport, transport_connected, (struct transport *trans, struct transport_class *tc, void *data))
{
  struct connection *conn;

  if (!trans || !tc)
    return NULL;

  conn = xmalloc(sizeof(struct connection));
  conn->class = tc;
  conn->trans = trans;
  conn->data = data;
  conn->rbuf = new_dbuffer();
  conn->wbuf = new_dbuffer();
  conn->metadata = new_hashtable(&hash_ascii, 17);
  conn->shutting_down = 0;
  conn->fds = new_llist();
  conn->mallocs = new_llist();
  conn->session = NULL;
  conn->protocol_pending = PENDING_NONE;

  add_llist_tail(connections, conn);

  add_llist_head(trans->connections, conn);
  conn->transport_connection_node = llist_head(trans->connections);

  add_llist_head(tc->connections, conn);
  conn->class_connection_node = llist_head(tc->connections);

  if (!tc->active)
    (*trans->ta)(conn, data);
  else
    {
      if (SYM_PRESENT(protocol, new_session))
        conn->session = SYM(protocol, new_session)(conn, conn->rbuf);
      else
        conn->protocol_pending |= PENDING_INIT;
    }

  return conn;
}

/** \brief Begin the process of establishing a new connection
 * \param transport_name Name of the transport to use to connect
 * \param class_name Name of the transport class to put this connection in
 * \param target Name of the target to connect to
 * \param argc Array of arguments
 * \param argv Length of argc
 * \return The newly created connection
 *
 * The precise meaning of target, argc, and argv are defined by the
 * transport in question. Transports should use target where possible
 * for clarity and simplicity; argv/argc are present for those cases
 * where it is simply not sensible to use a string.
 */
DEFINE_SYMBOL(struct connection *, transport, begin_connection, (struct session *session,
                                                                 const char *transport_name,
                                                                 const char *class_name,
                                                                 const char *target,
                                                                 void **argv, int argc))
{
  struct transport *trans = find_transport(transport_name);
  struct transport_class *tclass = find_transport_class(class_name);
  struct connection *conn;

  if (!trans)
    {
      log(MYNAME, LOG_PRIORITY_ERROR, "Attempt to use transport '%s', which is not registered", transport_name);
      return NULL;
    }

  if (!tclass)
    {
      log(MYNAME, LOG_PRIORITY_ERROR, "Attempt to use transport class '%s', which is not registered",
          class_name);
      return NULL;
    }

  if (!trans->tbc)
    {
      log(MYNAME, LOG_PRIORITY_ERROR,
          "Attempt to begin connection with transport '%s', which does not support outgoing connections",
          transport_name);
      return NULL;
    }

  if (!tclass->active)
    return NULL;

  conn = xmalloc(sizeof(struct connection));
  conn->class = tclass;
  conn->trans = trans;
  conn->data = NULL;
  conn->rbuf = new_dbuffer();
  conn->wbuf = new_dbuffer();
  conn->metadata = new_hashtable(&hash_ascii, 17);
  conn->shutting_down = 0;
  conn->fds = new_llist();
  conn->mallocs = new_llist();
  conn->session = session;
  conn->protocol_pending = PENDING_NONE;
  conn->data = NULL;

  add_llist_tail(connections, conn);

  add_llist_head(trans->connections, conn);
  conn->transport_connection_node = llist_head(trans->connections);

  add_llist_head(tclass->connections, conn);
  conn->class_connection_node = llist_head(tclass->connections);

  conn->data = (*trans->tbc)(conn, target, argv, argc);

  if (!conn->data)
    {
      (*trans->ta)(conn, NULL);
      return NULL;
    }

  return conn;
}

DEFINE_SYMBOL(void, transport, finished_connection, (struct connection *conn))
{
  if (SYM_PRESENT(protocol, connection_ready))
    SYM(protocol, connection_ready)(conn->session, conn->rbuf);
  else
    conn->protocol_pending |= PENDING_FINISH;
}

static void
close_(void *d)
{
  int *fd = d;
  close(*fd);
  xfree(fd);
}

static void
xfree_(void *d)
{
  xfree(d);
}

DEFINE_SYMBOL(void, transport, attach_fd, (struct connection *conn, int fd))
{
  int *d;
  if (!conn)
    return;

  d = xmalloc(sizeof(int));
  *d = fd;
  add_llist_head(conn->fds, d);
}

DEFINE_SYMBOL(void, transport, attach_mem, (struct connection *conn, void *data))
{
  if (!conn)
    return;
  add_llist_head(conn->mallocs, data);
}

static void
check_transport(struct transport *trans)
{
  if (!trans || trans->active)
    return;

  if (!llist_head(trans->connections) && !llist_head(trans->classes) &&
      !llist_head(trans->destroyed_connections) && !llist_head(trans->shutdown_connections))
    {
      delete_llist_data(transports, trans);
      free_llist(trans->connections);
      free_llist(trans->classes);
      free_llist(trans->destroyed_connections);
      free_llist(trans->shutdown_connections);
      xfree(trans->name);
      xfree(trans);
    }
}

DEFINE_SYMBOL(void, transport, connection_closed, (struct connection *conn))
{
  if (!conn)
    return;
  delete_llist_node(conn->transport_connection_node);
  conn->transport_connection_node = NULL;
  delete_llist_node(conn->class_connection_node);
  conn->class_connection_node = NULL;
  destroy_llist(conn->fds, &close_);
  destroy_llist(conn->mallocs, &xfree_);
  conn->fds = NULL;
  conn->mallocs = NULL;
  if (!conn->session || SYM_PRESENT(protocol, close_session))
    {
      delete_llist_data(connections, conn);
      free_dbuffer(conn->rbuf);
      free_dbuffer(conn->wbuf);
      free_hashtable(conn->metadata);
      check_transport(conn->trans);
      if (conn->session)
        SYM(protocol, close_session)(conn->session);
      xfree(conn);
    }
  else
    {
      free_dbuffer(conn->rbuf);
      free_dbuffer(conn->wbuf);
      conn->rbuf = NULL;
      conn->wbuf = NULL;
      free_hashtable(conn->metadata);
      conn->metadata = NULL;
      conn->transport_connection_node = NULL;
      conn->class_connection_node = NULL;
      conn->fds = NULL;
      conn->mallocs = NULL;
      conn->shutting_down = 1;
      conn->protocol_pending |= PENDING_CLOSE;
    }
}

static void
destroy_connection(struct connection *conn)
{
  if (!conn)
    return;
  delete_llist_node(conn->transport_connection_node);
  delete_llist_node(conn->class_connection_node);
  destroy_llist(conn->fds, &close_);
  destroy_llist(conn->mallocs, &xfree_);
  free_dbuffer(conn->rbuf);
  free_dbuffer(conn->wbuf);
  free_hashtable(conn->metadata);
  delete_llist_data(connections, conn);
  check_transport(conn->trans);
  xfree(conn);
}

DEFINE_SYMBOL(void, transport, abort_connection, (struct connection *conn))
{
  if (!conn)
    return;
  if (conn->trans->active)
    (*conn->trans->ta)(conn, conn->data);
  else
    {
      /* Destroy these now, but keep the connection around so the transport can clean up later */
      if (conn->session && SYM_PRESENT(protocol, close_session))
        SYM(protocol, close_session)(conn->session);
      conn->session = NULL;
      conn->class = NULL;
      free_dbuffer(conn->rbuf);
      free_dbuffer(conn->wbuf);
      conn->rbuf = NULL;
      conn->wbuf = NULL;
      free_hashtable(conn->metadata);
      conn->metadata = NULL;
      destroy_llist(conn->fds, &close_);
      destroy_llist(conn->mallocs, &xfree_);
      conn->fds = NULL;
      conn->mallocs = NULL;
      add_llist_head(conn->trans->destroyed_connections, conn);
    }
}

DEFINE_SYMBOL(void, transport, shutdown_connection, (struct connection *conn))
{
  if (!conn)
    return;
  /* Don't call trans->tc twice */
  if (conn->shutting_down)
    return;
  conn->shutting_down = 1;
  if (conn->trans->active)
    (*conn->trans->tc)(conn, conn->data);
  else
    add_llist_head(conn->trans->shutdown_connections, conn);
}

DEFINE_SYMBOL(void, transport, data_read, (struct connection *conn, const void *data, size_t len))
{
  if (conn->shutting_down)
    return;

  dbuffer_add(conn->rbuf, data, len);

  if (SYM_PRESENT(protocol, data_read))
    SYM(protocol, data_read)(conn->session, data, len);
  else
    conn->protocol_pending |= PENDING_READ;
}

DEFINE_SYMBOL(void, transport, write_data, (struct connection *conn, const void *data, size_t len))
{
  if (conn->shutting_down)
    return;

  dbuffer_add(conn->wbuf, data, len);

  if (conn->trans->active && conn->trans->tw)
    (*conn->trans->tw)(conn, conn->data, conn->wbuf, data, len);
}

DEFINE_SYMBOL(struct dbuffer *, transport, read_buffer, (struct connection *conn))
{
  return conn->rbuf;
}

DEFINE_SYMBOL(struct dbuffer *, transport, write_buffer, (struct connection *conn))
{
  return conn->wbuf;
}

DEFINE_SYMBOL(const char *, transport, class_name, (struct transport_class *tc))
{
  return tc->name;
}

DEFINE_SYMBOL(void *, transport, get_connection_meta, (struct connection *conn, const char *key))
{
  if (!conn || !key || !conn->metadata)
    return NULL;

  return hashtable_lookup(conn->metadata, key);
}

DEFINE_SYMBOL(void, transport, set_connection_meta, (struct connection *conn, const char *key, void *data))
{
  if (!conn || !key || !conn->metadata)
    return;

  hashtable_insert(conn->metadata, key, data);
}

static void
protocol_loaded_event(struct event *e unused_attribute)
{
  struct llist_node *n, *next;
  for (n = llist_head(connections); n; n = next)
    {
      struct connection *conn = llist_node_data(n);
      next = llist_node_next(n);
      if ((conn->protocol_pending & PENDING_CLOSE) && SYM_PRESENT(protocol, close_session))
        {
          LSYM(transport, connection_closed)(conn);
          continue;
        }
      if ((conn->protocol_pending & PENDING_INIT) && SYM_PRESENT(protocol, new_session))
        {
          conn->session = SYM(protocol, new_session)(conn, conn->rbuf);
          conn->protocol_pending &= ~PENDING_INIT;
        }
      if ((conn->protocol_pending & PENDING_FINISH) && SYM_PRESENT(protocol, connection_ready))
        {
          SYM(protocol, connection_ready)(conn->session, conn->rbuf);
          conn->protocol_pending &= ~PENDING_FINISH;
        }
      if ((conn->protocol_pending & PENDING_READ) && SYM_PRESENT(protocol, data_read))
        {
          SYM(protocol, data_read)(conn->session, NULL, 0);
          conn->protocol_pending &= ~PENDING_READ;
        }
    }
}

static void
read_config_event(struct event *e unused_attribute)
{
  read_config();
}

struct reload_data
{
  struct llist *transports, *connections, *classes;
};

int
MODULE_INIT(void *data)
{
  if (data)
    {
      struct reload_data *rd = data;
      transports = rd->transports;
      connections = rd->connections;
      classes = rd->classes;
    }
  else
    {
      transports = new_llist();
      connections = new_llist();
    }

  add_event_handler(event("protocol.load"), &protocol_loaded_event);
  add_event_handler(event("transport.reload-config"), &read_config_event);
  add_event_handler(event("core.sighup"), &read_config_event);

  if (read_config())
    {
      return MODULE_LOAD_OK;
    }
  else
    {
      return MODULE_LOAD_FAILED;
    }
}

int
MODULE_DEINIT(const char *new_version, module_init_func *init)
{
  int ret = MODULE_UNLOAD_OK;

  if (new_version)
    {
      struct reload_data rd = {transports, connections, classes};
      switch ((*init)(&rd))
        {
        case MODULE_LOAD_OK:
          transports = NULL;
          connections = NULL;
          classes = NULL;
          ret = MODULE_LOAD_OK;
          break;
        case MODULE_LOAD_FAILED:
          return MODULE_LOAD_FAILED;
        default:
          log(MYNAME, LOG_PRIORITY_ERROR, "Unknown return from new module version %s, aborting", new_version);
          return MODULE_LOAD_FAILED;
        }
    }
  else
    {
      struct llist_node *n, *next;
      for (n = llist_head(connections); n; n = next)
        {
          struct connection *conn = llist_node_data(n);
          next = llist_node_next(n);
          LSYM(transport, abort_connection)(conn);
        }
      destroy_llist(connections, destroy_connection);
      connections = NULL;
      destroy_llist(transports, unregister_transport);
      transports = NULL;
      for (n = llist_head(classes); n; n = next)
        {
          struct transport_class *tc = llist_node_data(n);
          next = llist_node_next(n);
          /* Deactivate each class, then check it (should trigger cleanup and disposal) */
          tc->active = 0;
          /* Disassociate the class from its transport, since we already ditched all the transports */
          tc->trans = NULL;
          check_transport_class(tc);
        }
      assert(!llist_head(classes));
      free_llist(classes);
    }

  delete_event_handler(event("protocol.load"), &protocol_loaded_event);
  delete_event_handler(event("transport.reload-config"), &read_config_event);
  delete_event_handler(event("core.sighup"), &read_config_event);

  return ret;
}
