/************************************************************************
 *   This is part of Dancer, a text relay daemon
 *   Copyright (C) 2002 Andrew Suffield <asuffield@users.sourceforge.net>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

#ifdef TRANSPORT_H

struct transport;
struct connection;
struct transport_class;
#else
#define TRANSPORT_H
#include <dancerd/config.h>
#include <dancerd/modules.h>
#include <dancerd/dbuffer.h>
#include <dancerd/protocol/protocol.h>

struct transport;
struct connection;
struct transport_class;

/** \file transport/transport.h
 * \brief Transport layer
 */

/** \defgroup transport Transport layer
 *
 * The transport layer establishes connections with other processes,
 * and communicates with them.
 *
 * @{
 */

/* The init class handler is called once for every new class this
 * transport is attached to
 *
 * The deinit class handler is called just before the class is
 * destroyed (no connections will remain)
 *
 * The deactivate handler is called if the transport is to stop
 * accepting connections in that class (optional, they will be
 * rejected in transport_connected if necessary)
 *
 * A deactivated class is reactivated by calling the init handler
 * again.
 *
 * The void* returned from the init handler is passed to the
 * deactivate and deinit handlers. It must be a single return from
 * xmalloc(), and will be xfree()ed if the deinit handler cannot be
 * called because the transport is not loaded. It might be passed in
 * to the init handler if the transport is reloaded, whatever value
 * this might have; it might also be destroyed. Modules should
 * probably use the module-level reload data instead
 *
 * Thusly, the transport should:
 * a) allocate a new void* when the init handler is called with a NULL
 * paraneter
 * b) *not* deallocate it in the deinit handler
 */
typedef void *transport_init_tclass_handler(struct transport_class *, void *);
typedef void transport_deactivate_tclass_handler(struct transport_class *, void *);
typedef void transport_deinit_tclass_handler(struct transport_class *, void *);
/* The begin connection handler is called to make a new outgoing
 * connection. If it returns NULL, the connection is aborted
 * immediately (the abort handler will be called)
 */
typedef void *transport_begin_connection_handler(struct connection *, const char *, void **, int);
/* The write handler is called when the protocol layer has data it wants writing */
typedef void transport_write_handler(struct connection *, void *, struct dbuffer *, const void *, size_t);
/* The close handler is called to begin an orderly shutdown of the
 * connection; it should call transport/connection_closed when it has
 * finished
 */
typedef void transport_close_handler(struct connection *, void *);
/* The abort handler indicates the connection should be simply
 * released and aborted immediately, severing the connection.
 */
typedef void transport_abort_handler(struct connection *, void *);

DEFINE_NAMESPACE(transport, 0);

DECLARE_SYMBOL(struct transport *, transport, register_transport, (const char *, struct module *,
                                                                   transport_init_tclass_handler *,
                                                                   transport_deactivate_tclass_handler *,
                                                                   transport_deinit_tclass_handler *,
                                                                   transport_begin_connection_handler *,
                                                                   transport_write_handler *,
                                                                   transport_close_handler *,
                                                                   transport_abort_handler *));
DECLARE_SYMBOL(void, transport, unregister_transport, (struct transport *));
/* This is called by the transport every time a new connection is
 * ready
 *
 * It is passed: the transport which is connecting, the class this
 * connection is in, and the data handle which the
 * transport_connection should use
 */
DECLARE_SYMBOL(struct connection *, transport, transport_connected,
		(struct transport *, struct transport_class *, void *));
/* Attach an fd that will be close()d automatically when the connection is */
DECLARE_SYMBOL(void, transport, attach_fd, (struct connection *, int));
/* Same as attach_fd, but for memory and xfree */
DECLARE_SYMBOL(void, transport, attach_mem, (struct connection *, void *));
/* This is called by the transport every time new data is read */
DECLARE_SYMBOL(void, transport, data_read, (struct connection *, const void *, size_t));
/* These return the dbuffer to work on */
DECLARE_SYMBOL(struct dbuffer *, transport, read_buffer, (struct connection *));
DECLARE_SYMBOL(struct dbuffer *, transport, write_buffer, (struct connection *));
/* Write data to the connection (called from above, calls the transport) */
DECLARE_SYMBOL(void, transport, write_data, (struct connection *, const void *, size_t));

/* Start the process of opening a new connection */
DECLARE_SYMBOL(struct connection *, transport, begin_connection, (struct session *,
                                                                  const char *, const char *, const char *,
                                                                  void **, int));
/* Called by the transport when it's finished; passes the session back up */
DECLARE_SYMBOL(void, transport, finished_connection, (struct connection *));

/* This causes an immediate abort of the connection; it will be
 * terminated when this function returns
 */
DECLARE_SYMBOL(void, transport, abort_connection, (struct connection *));
/* This causes an orderly shutdown of the connection; no further data
 * may be transmitted but it may take some time to be torn down
 */
DECLARE_SYMBOL(void, transport, shutdown_connection, (struct connection *));

/* This is called when a transport has finished detaching from a connection
 */
DECLARE_SYMBOL(void, transport, connection_closed, (struct connection *));

/* Get/set arbitrary metadata fields from the connection
 */
DECLARE_SYMBOL(void *, transport, get_connection_meta, (struct connection *, const char *));
DECLARE_SYMBOL(void, transport, set_connection_meta, (struct connection *, const char *, void *));

/* Get the name of a tclass */
DECLARE_SYMBOL(const char *, transport, class_name, (struct transport_class *));

END_NAMESPACE(transport);

/** @}
 */

#endif
