/************************************************************************
 *   This is part of Dancer, a text relay daemon
 *   Copyright (C) 2002 Andrew Suffield <asuffield@users.sourceforge.net>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

#ifdef TRANSPORT_INTERNAL_H

struct transport;
struct connection;

#else
#define TRANSPORT_INTERNAL_H
#include <dancerd/config.h>
#include <dancerd/transport/transport.h>
#include <dancerd/modules.h>
#include <dancerd/llist.h>
#include <dancerd/hashtable.h>
#include <dancerd/protocol/protocol.h>

#include <dancerd/transport/tclass.h>

/** \file transport_internal.h
 * \brief Internals of the transport layer interface
 */

/** \defgroup transportguts Internals of the transport layer interface
 * \ingroup transport
 * This group collects the stuff used internally by the transport
 * layer interface (the transport module itself).
 *
 * @{
 */ 

#define PENDING_NONE   0x0000
#define PENDING_INIT   0x0001
#define PENDING_READ   0x0002
#define PENDING_CLOSE  0x0004
#define PENDING_FINISH 0x0008

/** \brief A registered transport
 */
struct transport
{
  /** Name of the transport */
  char *name;
  /** Module this transport is implemented in (will automatically
   * unregister when this module unloads)
   */
  struct module *module;
  /** Called when a new transport class is attached to the transport */
  transport_init_tclass_handler *tic;
  /** Called when a transport class is to stop accepting connections */
  transport_deactivate_tclass_handler *tdac;
  /** Called when a transport class is to be released entirely */
  transport_deinit_tclass_handler *tdic;
  /** Called to set up a new connection */
  transport_begin_connection_handler *tbc;
  /** Called when data is to be written */
  transport_write_handler *tw;
  /** Called when the connection is to be cleanly closed */
  transport_close_handler *tc;
  /** Called when the connection is to be immediately closed, abnormally if necessary */
  transport_abort_handler *ta;
  /** List of all the connections using this transport */
  struct llist *connections;
  /** List of all the classes using this transport */
  struct llist *classes;
  /** List of all the destroyed connections that should be aborted when the transport is reregistered */
  struct llist *destroyed_connections;
  /** List of all the connections that should be shut down when the transport is reregistered */
  struct llist *shutdown_connections;
  /** True if this transport is currently registered */
  int active;
};

/** \brief Connection to a remote client over a transport
 */
struct connection
{
  /** The class this connection is in */
  struct transport_class *class;
  /** The transport this connection uses */
  struct transport *trans;
  /** Buffer of data that has been read from this connection */
  struct dbuffer *rbuf;
  /** Buffer of data that is to be written to this connection */
  struct dbuffer *wbuf;
  /** Arbitrary handle that is passed to the transport on all calls */
  void *data;
  /** Node that contains this connection in trans->connections */
  struct llist_node *transport_connection_node;
  /** Node that contains this connection in class->connections */
  struct llist_node *class_connection_node;
  /** True if this connection is being shut down */
  int shutting_down;
  /** List of fds to close when this connection is freed */
  struct llist *fds;
  /** List of memory to free when this connection is freed */
  struct llist *mallocs;
  /** Metadata associated with this connection */
  struct hashtable *metadata;
  /** Session (at the protocol layer) attached to this connection */
  struct session *session;
  /** Bitfield of pending events to tell the protocol layer about */
  uint16_t protocol_pending;
};

extern struct transport *get_transport(const char *name);

/** @}
 */

#endif
