# Copyright (C) 2002 Andrew Suffield <asuffield@users.sourceforge.net>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA

use strict;
use 5.006;
use warnings;
use DejaPerl qw/:framework/;
use Dancer::CmdPipe;
use Dancer::Spawn;

my $dancer = new Dancer::Spawn "auth";
$dancer->install();
$dancer->start(modules => [qw/basic.logfile database.db auth test.auth/]);

my $pipe = new Dancer::CmdPipe $dancer, "auth", "test.auth", sanity_checks => 0;

verbose "Starting test run...";

$pipe->do("REGISTER", "register auth method");
$pipe->do("UNREGISTER", "unregister auth method");
$pipe->do("REGISTER", "register auth method");
$pipe->do("UNREGISTER", "unregister auth method");

$pipe->do("REGISTER", "register auth method");

$pipe->check_cmd_value("NEW_HANDLE test alpha", "INIT_HANDLE: alpha (null)", "new handle alpha");
$pipe->check_cmd_value("NEW_HANDLE test beta", "INIT_HANDLE: beta (null)", "new handle beta");

$pipe->check_cmd_value("CANCEL_HANDLE beta", "CANCEL_HANDLE: beta", "cancel handle beta");
$pipe->check_cmd_value("NEW_HANDLE test beta", "INIT_HANDLE: beta (null)", "new handle beta");

$pipe->do("STATUS_RESULT 3", "set status result to AUTH_CONTINUE");
$pipe->check_cmd_value("PEER_ACCEPT alpha", "PEER_ACCEPT: alpha\nCONTINUE", "peer accept alpha");
$pipe->do("STATUS_RESULT 1", "set status result to AUTH_ACCEPT");
$pipe->check_cmd_value("PEER_ACCEPT alpha", "PEER_ACCEPT: alpha\nACCEPT", "peer accept alpha");
$pipe->check_cmd_value("NEW_HANDLE test alpha", "INIT_HANDLE: alpha (null)", "new handle alpha");
$pipe->do("STATUS_RESULT 2", "set status result to AUTH_CANCEL");
$pipe->check_cmd_value("PEER_ACCEPT alpha", "PEER_ACCEPT: alpha\nCANCEL (null)", "peer accept alpha");
$pipe->check_cmd_value("NEW_HANDLE test alpha", "INIT_HANDLE: alpha (null)", "new handle alpha");
$pipe->do("STATUS_ERROR gamma", "set status error to 'gamma'");
$pipe->check_cmd_value("PEER_ACCEPT alpha", "PEER_ACCEPT: alpha\nCANCEL gamma", "peer accept alpha");
$pipe->check_cmd_value("NEW_HANDLE test alpha", "INIT_HANDLE: alpha (null)", "new handle alpha");

$pipe->do("STATUS_RESULT 3", "set status result to AUTH_CONTINUE");
$pipe->check_cmd_value("READ_DATA alpha epsilon eta theta", "READ_DATA: alpha epsilon eta theta\nCONTINUE",
                       "read data for alpha");
$pipe->do("STATUS_RESULT 1", "set status result to AUTH_ACCEPT");
$pipe->check_cmd_value("READ_DATA alpha epsilon eta theta", "READ_DATA: alpha epsilon eta theta\nACCEPT",
                       "read data for alpha");
$pipe->check_cmd_value("NEW_HANDLE test alpha", "INIT_HANDLE: alpha (null)", "new handle alpha");
$pipe->do("STATUS_RESULT 2", "set status result to AUTH_CANCEL");
$pipe->check_cmd_value("READ_DATA alpha epsilon eta theta", "READ_DATA: alpha epsilon eta theta\nCANCEL gamma",
                       "read data for alpha");
$pipe->check_cmd_value("NEW_HANDLE test alpha", "INIT_HANDLE: alpha (null)", "new handle alpha");

$pipe->do("STATUS_RESULT 3", "set status result to AUTH_CONTINUE");
$pipe->do("RESULT_ARGV iti ni san", "set result data");
$pipe->check_cmd_value("GET_DATA alpha", "GET_DATA: alpha\nCONTINUE\nDATA: iti ni san",
                       "get data for alpha");
$pipe->do("STATUS_RESULT 1", "set status result to AUTH_ACCEPT");
$pipe->check_cmd_value("GET_DATA alpha", "GET_DATA: alpha\nACCEPT\n",
                       "get data for alpha");
$pipe->check_cmd_value("NEW_HANDLE test alpha", "INIT_HANDLE: alpha (null)", "new handle alpha");
$pipe->do("STATUS_RESULT 2", "set status result to AUTH_CANCEL");
$pipe->check_cmd_value("GET_DATA alpha", "GET_DATA: alpha\nCANCEL gamma\n",
                       "get data for alpha");

$pipe->do("STATUS_RESULT 0", "set status result to 0 (false)");
$pipe->do("STATUS_ERROR delta", "set status error to 'delta'");
$pipe->do("RESULT_DATA omega", "set result data to omega");
$pipe->check_cmd_value("SET_CREDENTIALS alpha test beta gamma", "SET_CREDENTIALS: alpha test beta gamma\ndelta",
                       "set credentials for alpha");
$pipe->check_cmd_value("NEW_HANDLE test alpha", "INIT_HANDLE: alpha (null)", "new handle alpha");
$pipe->check_cmd_value("CANCEL_HANDLE alpha", "CANCEL_HANDLE: alpha", "cancel handle alpha");
$pipe->do("STATUS_RESULT 1", "set status result to 1 (true)");
$pipe->check_cmd_value("SET_CREDENTIALS alpha test beta gamma", "SET_CREDENTIALS: alpha test beta gamma\nOK",
                       "set credentials for alpha");
$pipe->check_cmd_value("NEW_HANDLE test alpha", "INIT_HANDLE: alpha omega", "new handle alpha");
$pipe->check_cmd_value("CANCEL_HANDLE alpha", "CANCEL_HANDLE: alpha", "cancel handle alpha");

# Drop the entity, and all its data should go away
$pipe->do("DROP_ENTITY alpha", "drop entity alpha");
$pipe->check_cmd_value("NEW_HANDLE test alpha", "INIT_HANDLE: alpha (null)", "new handle alpha");
$pipe->check_cmd_value("CANCEL_HANDLE alpha", "CANCEL_HANDLE: alpha", "cancel handle alpha");

# Dropping it again should be safe
$pipe->do("DROP_ENTITY alpha", "drop entity alpha");

$pipe->check_cmd_value("GET_ENTITY_METHODS alpha", "(null)", "get methods for entity alpha");
$pipe->check_cmd_value("ITERATOR_NEXT", "(null)", "get non-existant member from iterator");
$pipe->check_cmd_value("ITERATOR_NEXT", "(null)", "get non-existant member from iterator");
$pipe->check_cmd_value("ITERATOR_NEXT", "(null)", "get non-existant member from iterator");
$pipe->do("FREE_ITERATOR", "free iterator");

# Now we just want to use these functions to test the other operations
sub set_credentials($$)
  {
    my ($entity, $credentials) = @_;
    $pipe->do("RESULT_DATA $credentials", "set result data to $credentials", undef, [caller]);
    $pipe->check_cmd_value("SET_CREDENTIALS $entity test alpha", "SET_CREDENTIALS: $entity test alpha\nOK",
                           "set credentials for $entity", undef, [caller]);
  }

sub try_credentials($$)
  {
    my ($entity, $credentials) = @_;
    $pipe->check_cmd_value("NEW_HANDLE test $entity", "INIT_HANDLE: $entity $credentials", "new handle $entity");
    $pipe->check_cmd_value("CANCEL_HANDLE $entity", "CANCEL_HANDLE: $entity", "cancel handle $entity");
  }

sub set_authorisation($$$$)
  {
    my ($entity, $object, $priv, $value) = @_;
    $pipe->do("GRANT_AUTHORISATION $entity $object/$priv $value", "grant $priv on $object to $entity at $value", undef, [caller]);
  }

sub revoke_authorisation($$$)
  {
    my ($entity, $object, $priv) = @_;
    $pipe->do("REVOKE_AUTHORISATION $entity $object/$priv", "revoke $priv on $object for $entity", undef, [caller]);
  }

sub get_authorisation($$$$)
  {
    my ($entity, $object, $priv, $value) = @_;
    $pipe->check_cmd_value("GET_AUTHORISATION $entity $object/$priv", $value || "(none)", "get $priv on $object for $entity", undef, [caller]);
  }

sub check_authorisation($$$$$)
  {
    my ($entity, $object, $priv, $value, $bool) = @_;
    $pipe->check_cmd_value("CHECK_AUTHORISATION $entity $object/$priv $value", $bool,
                           "check $priv on $object for $entity against $value", undef, [caller]);
  }

sub drop_entity($)
  {
    my $entity = shift;
    $pipe->do("DROP_ENTITY $entity", "drop entity $entity", undef, [caller]);
  }

sub drop_object($)
  {
    my $object = shift;
    $pipe->do("DROP_OBJECT $object", "drop object $object", undef, [caller]);
  }

sub set_entity_attribute($$$)
  {
    my ($entity, $attribute, $value) = @_;
    $pipe->do("SET_ENTITY_ATTRIBUTE $entity $attribute $value", "set attribute $attribute on $entity to $value", undef, [caller]);
  }

sub set_object_attribute($$$)
  {
    my ($object, $attribute, $value) = @_;
    $pipe->do("SET_OBJECT_ATTRIBUTE $object $attribute $value", "set attribute $attribute on $object to $value", undef, [caller]);
  }

sub get_entity_attribute($$$)
  {
    my ($entity, $attribute, $value) = @_;
    $pipe->check_cmd_value("GET_ENTITY_ATTRIBUTE $entity $attribute", $value || "(null)", "get attribute $attribute on $entity", undef, [caller]);
  }

sub get_object_attribute($$$)
  {
    my ($object, $attribute, $value) = @_;
    $pipe->check_cmd_value("GET_OBJECT_ATTRIBUTE $object $attribute", $value || "(null)", "get attribute $attribute on $object", undef, [caller]);
  }

sub get_list($@)
  {
    my ($query, @values) = @_;
    my $have_values = scalar @values;
    my $value = shift @values;
    if ($pipe->check_cmd_value($query, $value || "(null)", "get list", undef, [caller]))
      {
        while (my $value = shift @values)
          {
            $pipe->check_cmd_value("ITERATOR_NEXT", $value || "(null)", "get list item", undef, [caller]) or last;
          }
        $pipe->check_cmd_value("ITERATOR_NEXT", "(null)", "get list item", undef, [caller]) if $have_values;
      }
    $pipe->do("FREE_ITERATOR", "free iterator", undef, [caller]);
  }

get_list("GET_ENTITY_METHODS alpha");
set_credentials("alpha", "iti");
try_credentials("alpha", "iti");
get_list("GET_ENTITY_METHODS alpha", "alpha\01test");
drop_entity("alpha");
get_list("GET_ENTITY_METHODS alpha");

get_list("GET_ENTITY_ATTRIBUTES alpha");
get_entity_attribute "alpha", "iti", undef;
set_entity_attribute "alpha", "iti", "one";
set_entity_attribute "alpha", "ni", "two";
set_entity_attribute "alpha", "san", "three";
set_entity_attribute "beta", "alit", "one";
get_list("GET_ENTITY_ATTRIBUTES alpha", "alpha\01iti", "alpha\01ni" ,"alpha\01san");
get_list("GET_ENTITY_ATTRIBUTES beta", "beta\01alit");
get_entity_attribute "alpha", "iti", "one";
get_entity_attribute "alpha", "ni", "two";
get_entity_attribute "alpha", "san", "three";
get_entity_attribute "beta", "alit", "one";
drop_entity("alpha");
get_list("GET_ENTITY_ATTRIBUTES alpha");
get_entity_attribute "alpha", "iti", undef;
get_entity_attribute "alpha", "ni", undef;
get_entity_attribute "alpha", "san", undef;

get_list("GET_OBJECT_ATTRIBUTES alpha");
get_object_attribute "thing:alpha", "iti", undef;
set_object_attribute "thing:alpha", "iti", "one";
set_object_attribute "thing:alpha", "ni", "two";
set_object_attribute "thing:alpha", "san", "three";
set_object_attribute "thing:beta", "alit", "one";
get_list("GET_OBJECT_ATTRIBUTES thing:alpha", "thing:alpha\01iti", "thing:alpha\01ni" ,"thing:alpha\01san");
get_list("GET_OBJECT_ATTRIBUTES thing:beta", "thing:beta\01alit");
get_object_attribute "thing:alpha", "iti", "one";
get_object_attribute "thing:alpha", "ni", "two";
get_object_attribute "thing:alpha", "san", "three";
get_object_attribute "thing:beta", "alit", "one";
drop_object("thing:alpha");
get_list("GET_OBJECT_ATTRIBUTES thing:alpha");
get_object_attribute "thing:alpha", "iti", undef;
get_object_attribute "thing:alpha", "ni", undef;
get_object_attribute "thing:alpha", "san", undef;

get_list("GET_ENTITY_AUTHORISATION alpha");
get_list("GET_OBJECT_AUTHORISATION thing:alpha");
get_list("GET_OBJECT_PRIV_AUTHORISATION thing:alpha iti");
get_authorisation "alpha", "thing:alpha", "iti", undef;
set_authorisation "alpha", "thing:alpha", "iti", 10;
set_authorisation "alpha", "thing:alpha", "ni", 20;
set_authorisation "alpha", "thing:alpha", "san", 30;
set_authorisation "beta", "thing:alpha", "iti", 11;
set_authorisation "alpha", "thing:beta", "iti", 12;
get_authorisation "alpha", "thing:alpha", "iti", 10;
get_authorisation "alpha", "thing:alpha", "ni", 20;
get_authorisation "alpha", "thing:alpha", "san", 30;
get_authorisation "beta", "thing:alpha", "iti", 11;
get_authorisation "alpha", "thing:beta", "iti", 12;
check_authorisation "alpha", "thing:alpha", "iti", 9, "YES";
check_authorisation "alpha", "thing:alpha", "iti", 10, "YES";
check_authorisation "alpha", "thing:alpha", "iti", 11, "NO";
get_list("GET_ENTITY_AUTHORISATION alpha", (map {"alpha\01thing:alpha\01$_"} qw/iti ni san/), "alpha\01thing:beta\01iti");
get_list("GET_OBJECT_AUTHORISATION thing:alpha", (map {"alpha\01thing:alpha\01$_"} qw/iti ni san/), "beta\01thing:alpha\01iti");
get_list("GET_OBJECT_PRIV_AUTHORISATION thing:alpha iti", "alpha\01thing:alpha\01iti", "beta\01thing:alpha\01iti");
revoke_authorisation "alpha", "thing:alpha", "ni";
get_authorisation "alpha", "thing:alpha", "ni", undef;
get_list("GET_ENTITY_AUTHORISATION alpha", (map {"alpha\01thing:alpha\01$_"} qw/iti san/), "alpha\01thing:beta\01iti");
get_list("GET_OBJECT_AUTHORISATION thing:alpha", (map {"alpha\01thing:alpha\01$_"} qw/iti san/), "beta\01thing:alpha\01iti");
drop_entity("alpha");
get_list("GET_ENTITY_AUTHORISATION alpha");
get_list("GET_OBJECT_AUTHORISATION thing:alpha", "beta\01thing:alpha\01iti");
get_list("GET_OBJECT_PRIV_AUTHORISATION thing:alpha iti", "beta\01thing:alpha\01iti");
drop_object("thing:alpha");
get_list("GET_ENTITY_AUTHORISATION alpha");
get_list("GET_OBJECT_AUTHORISATION thing:alpha");
get_list("GET_OBJECT_PRIV_AUTHORISATION thing:alpha iti");
get_authorisation "alpha", "thing:alpha", "iti", undef;
get_authorisation "alpha", "thing:alpha", "ni", undef;
get_authorisation "alpha", "thing:alpha", "san", undef;
get_authorisation "beta", "thing:alpha", "iti", undef;
get_authorisation "alpha", "thing:beta", "iti", undef;

verbose "Finished thing run";

$dancer->kill();

1;
