# Copyright (C) 2002 Andrew Suffield <asuffield@users.sourceforge.net>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA

# Note that here, unlike most places, we're checking every message in
# the log, so that we don't miss any errors.

use strict;
use 5.006;
use warnings;
use DejaPerl qw/:framework/;
use Dancer::CmdPipe;
use Dancer::Spawn;

my $dancer = new Dancer::Spawn "main.modules";
$dancer->install();
$dancer->start(modules => [qw/test.main.modules/]);
my $pipe = new Dancer::CmdPipe $dancer, "main.modules", "test.main.modules", sanity_checks => 0;

sub load_module($;$)
  {
    my $module = shift;
    my $version = shift;
    $module = "test.stub.$module";
    $version = '.' unless defined $version;
    $pipe->do("LOAD_MODULE $module $version", "Load module $module" . ($version ne '.' ? ", version $version" : ""),
              undef, [caller]);
  }

sub find_module($;$)
  {
    my $module = shift;
    my $present = shift;
    $present = 1 unless defined $present;
    $module = "test.stub.$module";
    $pipe->check_cmd_value("FIND_MODULE $module", $present ? 'YES' : 'NO', "Check for presence of module $module",
                           undef, [caller]);
  }

sub unload_module($@)
  {
    my $module = shift;
    $module = "test.stub.$module";
    if (scalar @_)
      {
        my $handlers = join "\n", map {"UNLOAD: $_"} @_;
        $pipe->do("UNLOAD_MODULE $module", "Unload module $module",
                  undef, [caller]);
        $pipe->check_value($handlers, "Unload handlers for module $module",
                           undef, [caller]);
      }
    else
      {
        $pipe->do("UNLOAD_MODULE $module", "Unload module $module",
                  undef, [caller]);
      }
  }

sub register_handler($$)
  {
    my $module = shift;
    my $str = shift;
    $module = "test.stub.$module";
    $pipe->do("REGISTER_HANDLER $module $str", "Register unload handler for module $module, string $str",
              undef, [caller]);
  }

my $desc;
sub passmsg() {pass "got $desc";}
sub failmsg(;$) {fail "didn't get $desc",shift || [caller];}
sub failtimeout(;$) {fail "didn't get $desc (timeout)",shift || [caller];}
sub faileof(;$) {fail "didn't get $desc (eof)",shift || [caller];}
sub desc($) {$desc = shift}

sub expect_log_message($$$)
  {
    my $level = shift;
    my $re = shift;
    my $caller = shift;
    my $log;
    do
      {
        $log = $dancer->log(5);
        return failmsg($caller) unless defined $log;
        return faileof($caller) if $log->{eof};
        # And eat the one from the error log
        $dancer->error_log(5) if $log->{facility} eq 'core';
      } while ($log->{facility} ne 'core');
    if ($log->{facility} ne 'core' or $log->{level} ne $level or $log->{msg} !~ $re)
      {
        failmsg($caller);
        note "Got $log->{facility} [$log->{level}]: $log->{msg}";
        note "Expected core [$level]: $re";
        note "Wrong facility" if $log->{facility} ne 'core';
        note "Wrong level" if $log->{level} ne $level;
        note "Wrong message" if $log->{msg} !~ $re;
     }
    passmsg();
  }

sub expect_try_load($;$)
  {
    my $module = shift;
    my $version = shift;
    $module = "test.stub.$module";
    if (defined $version)
      {
        desc("Trying to load module $module-$version");
        expect_log_message(info => qr/\QTrying to load module $module-$version\E/, [caller]);
        desc("Looking for module $module-$version");
        expect_log_message(debug => qr/\QLooking for $module-$version.la or $module-$version.so\E/, [caller]);
        my $file = "./lib/dancerd/$module-$version.so";
        desc("Trying to open file $file");
        expect_log_message(debug => qr/\QTrying to open $file as a module\E/, [caller]);
      }
    else
      {
        error "Not yet implemented";
      }
  }

sub expect_load_success($;$)
  {
    my $module = shift;
    my $version = shift;
    $module = "test.stub.$module";
    $version = defined $version ? " version $version" : "";
    desc("Loaded module $module$version");
    expect_log_message(info => qr/\QSuccessfully loaded module $module$version\E/, [caller]);
  }

sub expect_replace_success($$$)
  {
    my $module = shift;
    my $oldversion = shift;
    my $newversion = shift;
    $module = "test.stub.$module";
    desc("Replacing module $module version $oldversion with $newversion");
    expect_log_message(info => qr/\QModule $module version $newversion successfully replaced version $oldversion\E/, [caller]);
  }

sub expect_fudge_replace_success($$$)
  {
    my $module = shift;
    my $oldversion = shift;
    my $newversion = shift;
    $module = "test.stub.$module";
    desc("Fudging replace of module $module version $oldversion with $newversion");
    expect_log_message(info => qr/\QModule $module (old version, $oldversion) unloaded, completing load of new module\E/, [caller]);
  }

sub expect_unload($)
  {
    my $module = shift;
    $module = "test.stub.$module";
    desc("Unload module $module");
    expect_log_message(info => qr/\QUnloaded module $module\E/, [caller]);
  }

sub expect_unsatisfied_dependencies($)
  {
    my $module = shift;
    $module = "test.stub.$module";
    desc("Unsatisfied dependencies in $module");
    expect_log_message(error => qr/\QDependencies for $module not satisfied, aborting\E/, [caller]);
  }

sub expect_downgrade_dependency($$$$)
  {
    my $frommodule = shift;
    my $tomodule = shift;
    my $depversion = shift;
    my $newversion = shift;
    $frommodule = "test.stub.$frommodule";
    $tomodule = "test.stub.$tomodule";
    desc("Downgrading would break dependency from $frommodule to $tomodule");
    expect_log_message(error => qr/\QModule $frommodule depends on $tomodule version $depversion or later, cannot downgrade to $newversion\E/, [caller]);
  }

sub expect_unload_blocked_by_dependency($$)
  {
    my $frommodule = shift;
    my $tomodule = shift;
    $frommodule = "test.stub.$frommodule";
    $tomodule = "test.stub.$tomodule";
    desc("Unloading would break dependency from $frommodule to $tomodule");
    expect_log_message(error => qr/\QCannot unload module $tomodule, $frommodule depends on it\E/, [caller]);
  }

sub expect_symbol_not_found($$$$)
  {
    my $module = shift;
    $module = "test.stub.$module";
    my $namespace = shift;
    my $version = shift;
    my $symbol = shift;
    desc("Symbol $namespace.$version\:\:$symbol not found");
    expect_log_message(warn => qr/\QCannot import required symbol $namespace.$version\:\:$symbol into $module: symbol not found\E/, [caller]);
  }

sub expect_symbol_processing_failed($)
  {
    my $module = shift;
    $module = "test.stub.$module";
    desc("Symbol processing for $module failed");
    expect_log_message(error => qr/\QFailed to process all symbols for $module, aborting\E/, [caller]);
  }

sub expect_import_symbol($$$$)
  {
    my $module = shift;
    $module = "test.stub.$module";
    my $namespace = shift;
    my $version = shift;
    my $symbol = shift;
    desc("Symbol $namespace.$version\:\:$symbol imported");
    expect_log_message(debug => qr/\QImport symbol $namespace.$version\:\:$symbol into $module\E/, [caller]);
  }

sub expect_export_symbol($$$$)
  {
    my $module = shift;
    $module = "test.stub.$module";
    my $namespace = shift;
    my $version = shift;
    my $symbol = shift;
    desc("Symbol $namespace.$version\:\:$symbol exported");
    expect_log_message(debug => qr/\QExport symbol $namespace.$version\:\:$symbol from $module\E/, [caller]);
  }

sub expect_unimport_symbol($$$$)
  {
    my $module = shift;
    $module = "test.stub.$module";
    my $namespace = shift;
    my $version = shift;
    my $symbol = shift;
    desc("Symbol $namespace.$version\:\:$symbol unimported");
    expect_log_message(debug => qr/\QUnimport symbol $namespace.$version\:\:$symbol from $module\E/, [caller]);
  }

sub expect_unexport_symbol($$$$)
  {
    my $module = shift;
    $module = "test.stub.$module";
    my $namespace = shift;
    my $version = shift;
    my $symbol = shift;
    desc("Symbol $namespace.$version\:\:$symbol unexported");
    expect_log_message(debug => qr/\QUnexport symbol $namespace.$version\:\:$symbol from $module\E/, [caller]);
  }

sub expect_required_symbol($$$$$)
  {
    my $module = shift;
    $module = "test.stub.$module";
    my $namespace = shift;
    my $version = shift;
    my $symbol = shift;
    my $required_by_module = shift;
    $required_by_module = "test.stub.$required_by_module";
    desc("Symbol $namespace.$version\:\:$symbol in $module required by $required_by_module");
    expect_log_message(warn => qr/\QSymbol $namespace.$version\:\:$symbol in module $module required by module $required_by_module\E/, [caller]);
  }

sub expect_upgrade_blocked_by_required_symbols($$$)
  {
    my $module = shift;
    my $oldversion = shift;
    my $newversion = shift;
    $module = "test.stub.$module";
    desc("Cannot replacing module $module version $oldversion with $newversion; blocked by required symbols");
    expect_log_message(error => qr/\QCannot load new module $module version $newversion, it does not export some required symbols that the old version $oldversion is providing\E/, [caller]);
  }

sub expect_unsatisfied_dependency($;$$)
  {
    my $module = shift;
    my $needversion = shift;
    my $haveversion = shift;
    $module = "test.stub.$module";
    if (defined $needversion)
      {
        desc("Unsatisfied dependency on $module version $needversion");
        if (defined $haveversion)
          {
            expect_log_message(error => qr/\QDependency on $module version $needversion not satisfied (version $haveversion is loaded)\E/, [caller]);
          }
        else
          {
            expect_log_message(error => qr/\QDependency on $module version $needversion not satisfied\E/, [caller]);
          }
      }
    else
      {
        desc("Unsatisfied dependency on $module");
        expect_log_message(error => qr/\QDependency on $module not satisfied\E/, [caller]);
      }
  }

# Load a module
load_module("a", "1.0");
expect_try_load("a", "1.0");
expect_load_success("a", "1.0");
find_module("a");

# Upgrade to version 1.1
load_module("a", "1.1");
expect_try_load("a", "1.1");
expect_replace_success("a", "1.0", "1.1");
find_module("a");

# Downgrade to version 1.0
load_module("a", "1.0");
expect_try_load("a", "1.0");
expect_fudge_replace_success("a", "1.1", "1.0");
expect_load_success("a", "1.0");
find_module("a");

# Unload the module
unload_module("a");
expect_unload("a");
find_module("a", 0);

# Load b 1.0
load_module("b", "1.0");
expect_try_load("b", "1.0");
expect_load_success("b", "1.0");
find_module("b");

# Try to upgrade to b 1.1, which depends on a 1.1
load_module("b", "1.1");
expect_try_load("b", "1.1");
expect_unsatisfied_dependency("a", "1.1");
expect_unsatisfied_dependencies("b");
find_module("b");

# Install a 1.0
load_module("a", "1.0");
expect_try_load("a", "1.0");
expect_load_success("a", "1.0");
find_module("a");

# Try upgrading b to 1.1 again
load_module("b", "1.1");
expect_try_load("b", "1.1");
expect_unsatisfied_dependency("a", "1.1", "1.0");
expect_unsatisfied_dependencies("b");
find_module("b");

# Now upgrade a to 1.1
load_module("a", "1.1");
expect_try_load("a", "1.1");
expect_replace_success("a", "1.0", "1.1");
find_module("a");

# And upgrade b to 1.1 for real this time
load_module("b", "1.1");
expect_try_load("b", "1.1");
expect_replace_success("b", "1.0", "1.1");
find_module("b");

# Now try to downgrade a to 1.0
load_module("a", "1.0");
expect_try_load("a", "1.0");
expect_downgrade_dependency("b", "a", "1.1", "1.0");
expect_unsatisfied_dependencies("a");
find_module("a");

# Now try to unload a
unload_module("a");
expect_unload_blocked_by_dependency("b", "a");
find_module("a");

# Downgrade b to 1.0
load_module("b", "1.0");
expect_try_load("b", "1.0");
expect_replace_success("b", "1.1", "1.0");
find_module("b");

# Downgrade a to 1.0
load_module("a", "1.0");
expect_try_load("a", "1.0");
expect_fudge_replace_success("a", "1.1", "1.0");
expect_load_success("a", "1.0");
find_module("a");

register_handler("a", "a");
register_handler("b", "b");

# Unload a
unload_module("a", "a");
expect_unload("a");
find_module("a", 0);

# Unload b
unload_module("b", "b");
expect_unload("b");
find_module("b", 0);

# Try to load d 1.0
load_module("d", "1.0");
expect_try_load("d", "1.0");
expect_symbol_not_found("d", "c", 0, "alpha");
expect_symbol_processing_failed("d");
find_module("d", 0);

# Load a 1.1
load_module("a", "1.1");
expect_try_load("a", "1.1");
expect_load_success("a", "1.1");
find_module("a");

# Try to load c 1.0
load_module("c", "1.0");
expect_try_load("c", "1.0");
expect_unsatisfied_dependency("b");
expect_unsatisfied_dependencies("c");

# Load b 1.0
load_module("b", "1.0");
expect_try_load("b", "1.0");
expect_load_success("b", "1.0");
find_module("b");

# Load c 1.0 for real
load_module("c", "1.0");
expect_try_load("c", "1.0");
expect_export_symbol("c", "c", 0, "alpha");
expect_export_symbol("c", "c", 0, "beta");
expect_load_success("c", "1.0");
find_module("c");

# Try to upgrade to c 1.1
load_module("c", "1.1");
expect_try_load("c", "1.1");
expect_unsatisfied_dependency("b", "1.1", "1.0");
expect_unsatisfied_dependencies("c");
find_module("c");

# Install d 1.0
load_module("d", "1.0");
expect_try_load("d", "1.0");
expect_import_symbol("d", "c", 0, "alpha");
expect_export_symbol("d", "d", 0, "alpha");
expect_export_symbol("d", "d", 0, "beta");
expect_load_success("d", "1.0");
find_module("d");

# Upgrade to d 1.1
load_module("d", "1.1");
expect_try_load("d", "1.1");
expect_import_symbol("d", "c", 0, "alpha");
expect_import_symbol("d", "c", 0, "beta");
expect_unexport_symbol("d", "d", 0, "alpha");
expect_export_symbol("d", "d", 0, "alpha");
expect_export_symbol("d", "d", 0, "gamma");
expect_replace_success("d", "1.0", "1.1");
expect_unimport_symbol("d", "c", 0, "alpha");
expect_unexport_symbol("d", "d", 0, "beta");
find_module("d");

# Upgrade b to 1.1
load_module("b", "1.1");
expect_try_load("b", "1.1");
expect_replace_success("b", "1.0", "1.1");
find_module("b");

# Try to upgrade c to 1.1
load_module("c", "1.1");
expect_try_load("c", "1.1");
expect_required_symbol("c", "c", 0, "beta", "d");
expect_upgrade_blocked_by_required_symbols("c", "1.0", "1.1");
find_module("c");

$dancer->kill();

1;
