# Copyright (C) 2002 Andrew Suffield <asuffield@users.sourceforge.net>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA

use strict;
use 5.006;
use warnings;
use DejaPerl qw/:framework/;
use Dancer::CmdPipe;
use Dancer::Spawn;
use Dancer::Message;
use DGMTP::Spawn;
use IPC::Open3;
use IO::Handle;

my $dancer = new Dancer::Spawn "protocol.dgmtp";
$dancer->install();
$dancer->put_config("protocol.max-unknown-send-queue", "integer", 4);
$dancer->put_config("transport.classes", "string", "tclass1,tclass2");
$dancer->put_config("transport.classes.tclass1.transport", "string", "test");
$dancer->put_config("transport.classes.tclass2.transport", "string", "ipv4.tcp");
$dancer->put_config("transport.classes.tclass2.bind_address", "integer", unpack("I", v127.0.0.1));
$dancer->put_config("transport.classes.tclass2.bind_port", "integer", unpack("S", pack("n", 6546)));
$dancer->start(modules => [qw/database.db config.database transport protocol protocol.dgmtp transport.ipv4.tcp
                              test.transport test.message test.protocol/]);

my $trans = new Dancer::CmdPipe $dancer, "transport", "test.transport", sanity_checks => 0;
my $proto = new Dancer::CmdPipe $dancer, "protocol", "test.protocol", sanity_checks => 0;
my $msg = new Dancer::CmdPipe $dancer, "message", "test.message", sanity_checks => 0;

no warnings 'redefine';
sub check_message(%)
  {
    (new Dancer::Message @_)->check($msg, [caller]);
  }
use warnings 'redefine';

verbose "Starting test run...";

###########################
# Checking identification #
###########################

$trans->check_cmd_value("REGISTER", "INIT_TCLASS tclass1", "register transport");
# Check the identify function works as expected
# These should cause the connection to be dropped with failed identification
$trans->do("CONNECT tclass1 conn1", "create connection 1");
$msg->check_cmd_value("PEEK_MESSAGE", "NO", "no messages waiting");
$trans->check_cmd_value("READ conn1 \1", "ABORT conn1", "send incorrect handshake to conn1 transport");
$dancer->expect_log(3, {facility => 'protocol', level => 'info',
                        msg => "Dropping session MOOOOOOOOOOOOOOOO, unrecognised protocol"});

# This should do the same, but take a different code path
$trans->do("CONNECT tclass1 conn1", "create connection 1");
$msg->check_cmd_value("PEEK_MESSAGE", "NO", "no messages waiting");
$trans->do("READ conn1 \0", "send null byte to conn1 transport");
$trans->check_cmd_value("READ conn1 \1", "ABORT conn1", "send incorrect handshake to conn1 transport");
$dancer->expect_log(3, {facility => 'protocol', level => 'info',
                        msg => "Dropping session MOOOOOOOOOOOOOOOO, unrecognised protocol"});

$trans->check_cmd_value("UNREGISTER", "DEACTIVATE_TCLASS tclass1 tclass1", "unregister transport");

# Now spawn a dgmtpc to talk to it
my $dgmtpc = new DGMTP::Spawn "protocol.dgmtp", "127.0.0.1", 6546;
$dgmtpc->start() or die "Failed to start dgmtpc";

is($dgmtpc->readline('stdout'), "<Handshake complete>");
$proto->check_value("NEW_SESSION", "new session from protocol");
$proto->do("NAME_SESSION conn1", "name session conn1");
$msg->check_value("NEW_SESSION", "new session from message");

$msg->do("REQUEST_ALL_FROM_SESSION 0 conn1", "request all from conn1");

$msg->check_cmd_value("PEEK_MESSAGE", "NO", "no messages waiting");

######################
# Check inbound data #
######################

$msg->check_cmd_value("REGISTER_TYPE type", "1", "register message type");
$dgmtpc->write('stdin', "type: 'str1' 'str2' +5 1 'str3'\n");
$msg->check_value("MESSAGE_HANDLER: conn1", "handled message from conn1");
check_message type => 'type', str => [qw/str1 str2 str3/], uint32 => [1], sint32 => [5];

#######################
# Check outbound data #
#######################

sub send_message($%)
  {
    my $conn = shift;
    (new Dancer::Message @_)->send($proto, $conn, [caller]);
  }

send_message 'conn1', origin => 'origin', dest => 'dest', type => 'type',
  str => [qw/iti ni san/], uint32 => [1], sint32 => [5];
is($dgmtpc->readline('stdout'), "<origin -> dest> type: 'iti' 'ni' 'san' 1 +5");

$dgmtpc->close('stdin');
is($dgmtpc->readline('stderr'), "Closing connection");
$dgmtpc->kill();
$msg->check_value("MESSAGE_HANDLER: conn1", "NULL message handled");
check_message type => '(null)';

##########################
# Check error conditions #
##########################

sub start_dgmtp
  {
    $dgmtpc = new DGMTP::Spawn "protocol.dgmtp", "127.0.0.1", 6546;
    $dgmtpc->start() or die "Failed to start dgmtpc";
    is($dgmtpc->readline('stdout'), "<Handshake complete>");
    $proto->check_value("NEW_SESSION", "new session from protocol");
    $proto->do("NAME_SESSION conn1", "name session conn1");
    $msg->check_value("NEW_SESSION", "new session from message");
    $msg->do("REQUEST_ALL_FROM_SESSION 0 conn1", "request all from conn1");
    $msg->check_cmd_value("PEEK_MESSAGE", "NO", "no messages waiting");
  }

start_dgmtp();
$dgmtpc->write('stdin', "<>: 'ERROR'\n");
$msg->check_value("MESSAGE_HANDLER: conn1", "NULL message handled");
check_message type => '(null)';
$msg->check_cmd_value("PEEK_MESSAGE", "NO", "no messages waiting");
is($dgmtpc->readline('stderr'), "Connection closed by remote host");
$dgmtpc->kill();

start_dgmtp();
$dgmtpc->write('stdin', "<>: 'ERROR' 'alpha'\n");
$msg->check_value("MESSAGE_HANDLER: conn1", "protocol_error message handled");
$msg->check_value("MESSAGE_HANDLER: conn1", "NULL message handled");
check_message type => 'protocol error', str => ['alpha'];
check_message type => '(null)';
$msg->check_cmd_value("PEEK_MESSAGE", "NO", "no messages waiting");
is($dgmtpc->readline('stderr'), "Connection closed by remote host");
$dgmtpc->kill();

start_dgmtp();
$dgmtpc->write('stdin', "<>: 'UNKNOWN'\n");
$msg->check_value("MESSAGE_HANDLER: conn1", "NULL message handled");
check_message type => '(null)';
$msg->check_cmd_value("PEEK_MESSAGE", "NO", "no messages waiting");
is($dgmtpc->readline('stderr'), "Connection closed by remote host");
$dgmtpc->kill();

start_dgmtp();
$dgmtpc->write('stdin', "<origin -> dest> unknown:\n");
$msg->check_value("MESSAGE_HANDLER: conn1", "NULL message handled");
check_message type => '(null)';
$msg->check_cmd_value("PEEK_MESSAGE", "NO", "no messages waiting");
is($dgmtpc->readline('stderr'), "Connection closed by remote host");
$dgmtpc->kill();

########################
# Outbound connections #
########################

$dgmtpc = new DGMTP::Spawn "protocol.dgmtp", "127.0.0.1", 6547, "listen";
$dgmtpc->start() or die "Failed to start dgmtpc";
$proto->do("BEGIN_CONNECTION dgmtp ipv4.tcp tclass2 127.0.0.1:6547", "begin new connection");
is($dgmtpc->readline('stdout'), "<Handshake complete>");
$proto->check_value("NEW_OUTGOING_SESSION", "new outgoing from protocol");
$msg->check_value("NEW_OUTGOING_SESSION", "new session from message");
$msg->do("REQUEST_ALL_FROM_SESSION 0 conn1", "request all from conn1");
$msg->check_cmd_value("PEEK_MESSAGE", "NO", "no messages waiting");
$proto->do("SHUTDOWN_SESSION 127.0.0.1:6547", "shutdown session");
is($dgmtpc->readline('stderr'), "Connection closed by remote host");
$dgmtpc->kill();

verbose "Finished test run";

$trans->kill;
$msg->kill;

$dancer->kill();

1;
