# Copyright (C) 2002 Andrew Suffield <asuffield@users.sourceforge.net>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA

use strict;
use 5.006;
use warnings;
use DejaPerl qw/:framework/;
use Dancer::CmdPipe;
use Dancer::Spawn;
use Dancer::Message;

my $dancer = new Dancer::Spawn "message";
$dancer->install();
$dancer->put_config("protocol.max-unknown-send-queue", "integer", 4);
$dancer->put_config("transport.classes", "string", "tclass1");
$dancer->put_config("transport.classes.tclass1.transport", "string", "test");

$dancer->start(modules => [qw/basic.logfile database.db config.database transport protocol
                              test.transport test.message test.protocol/]);

my $trans = new Dancer::CmdPipe $dancer, "transport", "test.transport", sanity_checks => 0;
my $msg = new Dancer::CmdPipe $dancer, "message", "test.message", sanity_checks => 0;
my $proto = new Dancer::CmdPipe $dancer, "protocol", "test.protocol", sanity_checks => 0;

verbose "Starting test run...";

# Setup a couple of sessions
$trans->check_cmd_value("REGISTER", "INIT_TCLASS tclass1", "register transport");
$proto->do("REGISTER_PROTOCOL", "register protocol");
$proto->do("IDENTIFY_RESULT 1", "set identify result to 1 (ours)");
$trans->do("CONNECT tclass1 conn1", "create connection 1");
$trans->do("CONNECT tclass1 conn2", "create connection 2");
$trans->do("READ conn1 alpha", "send alpha to conn1 transport");
$proto->check_value("IDENTIFY: alpha", "identify from protocol");
$proto->check_value("NEW_SESSION", "new session from protocol");
$proto->check_value("INIT_SESSION", "init session from protocol");
$proto->check_value("READ_DATA: alpha", "read alpha from protocol");
$proto->do("NAME_SESSION 0", "name session 0");
$msg->check_value("NEW_SESSION", "new session from message");
$trans->do("READ conn2 alpha", "send alpha to conn2 transport");
$proto->check_value("IDENTIFY: alpha", "identify from protocol");
$proto->check_value("NEW_SESSION", "new session from protocol");
$proto->check_value("INIT_SESSION", "init session from protocol");
$proto->check_value("READ_DATA: alpha", "read alpha from protocol");
$proto->do("NAME_SESSION 1", "name session 0");
$msg->check_value("NEW_SESSION", "new session from message");

# Setup a few message types
$msg->check_cmd_value("REGISTER_TYPE alpha", "0", "register message type alpha");
$msg->check_cmd_value("REGISTER_TYPE beta", "1", "register message type alpha");
$msg->check_cmd_value("REGISTER_TYPE gamma", "2", "register message type alpha");

# Okay, first we check out the type_from_all form
$msg->do("REQUEST_TYPE_FROM_ALL alpha alpha", "request all 'alpha' messages");
my $alpha1 = new Dancer::Message type => 'alpha', origin => 'gamma', dest => 'delta',
  str => ["iti", "ni", "san"], sint32 => [0, 1, -2], uint32 => [3, 4, 5];
my $beta1 = new Dancer::Message type => 'beta', origin => 'epsilon', dest => 'eta',
  str => ["iti", "ni", "san"], sint32 => [0, 1, -2], uint32 => [3, 4, 5];
my $null = new Dancer::Message type => '(null)';
$proto->do("READ_MESSAGE 0 $alpha1", "read message 1 on conn1");
$msg->check_value("MESSAGE_HANDLER: alpha", "message 1 handled");
$alpha1->check($msg);
my $alpha2 = new Dancer::Message type => 'alpha';
my $beta2 = new Dancer::Message type => 'beta';
$proto->do("READ_MESSAGE 0 $alpha2", "read message 2 on conn1");
$msg->check_value("MESSAGE_HANDLER: alpha", "message 2 handled");
$alpha2->check($msg);
$proto->do("READ_MESSAGE 0 $beta1", "read message 1 on conn1");
$msg->check_cmd_value("PEEK_MESSAGE", "NO", "no messages waiting");

$msg->do("REQUEST_TYPE_FROM_ALL beta beta", "request all 'beta' messages");
$proto->do("READ_MESSAGE 0 $beta2", "read message 2 on conn1");
$msg->check_value("MESSAGE_HANDLER: beta", "message 2 handled");
$beta2->check($msg);

$msg->do("IGNORE_TYPE_FROM_ALL alpha alpha", "ignore all 'alpha' messages");
$proto->do("READ_MESSAGE 0 $alpha1", "read message 1 on conn1");
$msg->check_cmd_value("PEEK_MESSAGE", "NO", "no messages waiting");

$proto->do("READ_MESSAGE 1 $beta1", "read message 1 on conn2");
$msg->check_value("MESSAGE_HANDLER: beta", "message 1 handled");
$beta1->check($msg);

$msg->do("IGNORE_TYPE_FROM_ALL beta beta", "ignore all 'beta' messages");
$proto->do("READ_MESSAGE 0 $beta1", "read message 1 on conn1");
$msg->check_cmd_value("PEEK_MESSAGE", "NO", "no messages waiting");

# Next we check type_from_session
$msg->do("REQUEST_TYPE_FROM_SESSION alpha 0 alpha0", "request 'alpha' messages from conn1");
$proto->do("READ_MESSAGE 0 $alpha1", "read message 1 on conn1");
$msg->check_value("MESSAGE_HANDLER: alpha0", "message 1 handled");
$alpha1->check($msg);
$proto->do("READ_MESSAGE 0 $beta1", "read message 1 on conn1");
$msg->check_cmd_value("PEEK_MESSAGE", "NO", "no messages waiting");
$proto->do("READ_MESSAGE 1 $alpha1", "read message 1 on conn2");
$msg->check_cmd_value("PEEK_MESSAGE", "NO", "no messages waiting");

$msg->do("REQUEST_TYPE_FROM_ALL alpha alpha", "request all 'alpha' messages");
$proto->do("READ_MESSAGE 0 $alpha1", "read message 1 on conn1");
# We'll get two of these (requested down two paths)
$msg->check_value("MESSAGE_HANDLER: alpha", "message 1 handled");
$msg->check_value("MESSAGE_HANDLER: alpha0", "message 1 handled");
$alpha1->check($msg);
$alpha1->check($msg);
# One of these
$proto->do("READ_MESSAGE 1 $alpha1", "read message 1 on conn2");
$msg->check_value("MESSAGE_HANDLER: alpha", "message 1 handled");
$alpha1->check($msg);
# None of these
$proto->do("READ_MESSAGE 0 $beta1", "read message 1 on conn1");
$msg->check_cmd_value("PEEK_MESSAGE", "NO", "no messages waiting");

$msg->do("REQUEST_TYPE_FROM_SESSION beta 1 beta1", "request 'alpha' messages from conn1");
$proto->do("READ_MESSAGE 1 $alpha1", "read message 1 on conn2");
$msg->check_value("MESSAGE_HANDLER: alpha", "message 1 handled");
$alpha1->check($msg);
$proto->do("READ_MESSAGE 1 $beta1", "read message 1 on conn2");
$msg->check_value("MESSAGE_HANDLER: beta1", "message 1 handled");
$beta1->check($msg);

$msg->do("IGNORE_TYPE_FROM_ALL alpha alpha", "ignore all 'alpha' messages");
$msg->do("IGNORE_TYPE_FROM_SESSION alpha 0 alpha0", "ignore 'alpha' messages from conn1");
$proto->do("READ_MESSAGE 0 $alpha1", "read message 1 on conn1");
$msg->check_cmd_value("PEEK_MESSAGE", "NO", "no messages waiting");

$msg->do("IGNORE_TYPE_FROM_SESSION beta 1 beta1", "ignore 'alpha' messages from conn1");

# Now all_from_session
$msg->do("REQUEST_ALL_FROM_SESSION 0 all0", "request all messages from conn1");
$proto->do("READ_MESSAGE 0 $alpha1", "read message 1 on conn1");
$msg->check_value("MESSAGE_HANDLER: all0", "message 1 handled");
$alpha1->check($msg);
$proto->do("READ_MESSAGE 0 $beta1", "read message 1 on conn1");
$msg->check_value("MESSAGE_HANDLER: all0", "message 1 handled");
$beta1->check($msg);
$proto->do("READ_MESSAGE 1 $alpha1", "read message 1 on conn2");
$msg->check_cmd_value("PEEK_MESSAGE", "NO", "no messages waiting");

$msg->do("IGNORE_ALL_FROM_SESSION 0 all0", "ignore all messages from conn1");
$proto->do("READ_MESSAGE 0 $alpha1", "read message 1 on conn1");
$msg->check_cmd_value("PEEK_MESSAGE", "NO", "no messages waiting");

# Now all three together
$msg->do("REQUEST_ALL_FROM_SESSION 0 all0", "request all messages from conn1");
$msg->do("REQUEST_TYPE_FROM_SESSION alpha 0 alpha0", "request 'alpha' messages from conn1");
$msg->do("REQUEST_TYPE_FROM_ALL alpha alpha", "request all 'alpha' messages");
$proto->do("READ_MESSAGE 0 $alpha1", "read message 1 on conn1");
$msg->check_value("MESSAGE_HANDLER: alpha", "message 1 handled");
$msg->check_value("MESSAGE_HANDLER: alpha0", "message 1 handled");
$msg->check_value("MESSAGE_HANDLER: all0", "message 1 handled");
$alpha1->check($msg);
$alpha1->check($msg);
$alpha1->check($msg);

# Now send a NULL message (indicates close of session)
$proto->do("READ_MESSAGE 0 :", "read NULL message");
$msg->check_value("MESSAGE_HANDLER: all0", "NULL message handled");
$null->check($msg);
$proto->do("READ_MESSAGE 0 $alpha1", "read message 1 on conn1");
# This should be the only one that is now handled - the others should stop on NULL
$msg->check_value("MESSAGE_HANDLER: alpha", "message 1 handled");
$alpha1->check($msg);

$msg->check_cmd_value("PEEK_MESSAGE", "NO", "no messages waiting");

$msg->do("REQUEST_ALL_FROM_SESSION 1 all1", "request all messages from conn2");
# Now close the session from the protocol side
$proto->do("SHUTDOWN_SESSION 1", "shutdown session for conn2");
$trans->check_value("CLOSE conn2", "got close from conn2 transport");
# All further messages should be ignored
$proto->do("READ_MESSAGE 1 $alpha1", "read message 1 on conn2");
$msg->check_cmd_value("PEEK_MESSAGE", "NO", "no messages waiting");

# Close the connection for real
$trans->do("CLOSE conn2", "close conn2");
$proto->check_value("CLOSE_SESSION", "got close session from conn2 protocol");

# A NULL message will be emitted
$msg->check_value("MESSAGE_HANDLER: all1", "NULL message handled");
$null->check($msg);
$msg->check_cmd_value("PEEK_MESSAGE", "NO", "no messages waiting");

# Open a new connection
$trans->do("CONNECT tclass1 conn3", "create connection 3");
$trans->do("READ conn3 alpha", "send alpha to conn3 transport");
$proto->check_value("IDENTIFY: alpha", "identify from protocol");
$proto->check_value("NEW_SESSION", "new session from protocol");
$proto->check_value("INIT_SESSION", "init session from protocol");
$proto->check_value("READ_DATA: alpha", "read alpha from protocol");
$proto->do("NAME_SESSION conn3", "name session conn3");
$msg->check_value("NEW_SESSION", "new session from message");

# Close the session while within a message handler
$trans->do("AUTOCLOSE_ON", "turn on autoclose");
$msg->do("AUTOQUIT_ON", "turn on autoclose");
$msg->do("REQUEST_ALL_FROM_SESSION 2 all3", "request all messages from conn3");
$proto->check_cmd_value("READ_MESSAGE conn3 $alpha1", "CLOSE_SESSION", "read message 1 on conn3");
$msg->check_value("MESSAGE_HANDLER: alpha", "message 1 handled");
# Note that all3 has NOT received a copy of this message, because the
# session was shut down during the 'alpha' handler
$msg->check_value("MESSAGE_HANDLER: all3", "NULL message handled");
$alpha1->check($msg);
$null->check($msg);
$trans->check_value("CLOSE conn3", "connection 3 closed");
$msg->check_cmd_value("PEEK_MESSAGE", "NO", "no messages waiting");

verbose "Finished test run";

$trans->kill();
$msg->kill();
$proto->kill();

$dancer->kill();

1;
