# Copyright (C) 2002 Andrew Suffield <asuffield@users.sourceforge.net>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA

use strict;
use 5.006;
use warnings;
use DejaPerl qw/:framework/;
use Dancer::CmdPipe;
use Dancer::Spawn;
use Dancer::Message;

my $dancer = new Dancer::Spawn "rfc1459.client";
$dancer->install();
$dancer->put_config("protocol.max-unknown-send-queue", "integer", 4);
$dancer->put_config("transport.classes", "string", "tclass1");
$dancer->put_config("transport.classes.tclass1.transport", "string", "test");
$dancer->start(modules => [qw/database.db config.database transport protocol protocol.rfc1459.client
                              test.transport test.message test.protocol/]);

my $trans = new Dancer::CmdPipe $dancer, "transport", "test.transport", sanity_checks => 0;
my $proto = new Dancer::CmdPipe $dancer, "protocol", "test.protocol", sanity_checks => 0;
my $msg = new Dancer::CmdPipe $dancer, "message", "test.message", sanity_checks => 0;

no warnings 'redefine';
sub check_message(%)
  {
    (new Dancer::Message @_)->check($msg, [caller]);
  }

sub read_message($$$;$)
  {
    my ($conn, $msg, $desc, $reply) = @_;
    if ($reply)
      {
        $trans->check_cmd_value("READ $conn :$msg\n", $reply, "send $desc to $conn transport", undef, [caller]);
      }
    else
      {
        $trans->do("READ $conn :$msg\n", "send $desc to $conn transport", undef, [caller]);
      }
  }
use warnings 'redefine';

verbose "Starting test run...";

# We need to ask for this one explicitly so that we get the first messages after connecting
$msg->do("REQUEST_TYPE_FROM_ALL all-real set real name", "request all 'set real name' messages");
$msg->do("REQUEST_TYPE_FROM_ALL all-name set name", "request all 'set name' messages");

###########################
# Checking identification #
###########################

# Setup a couple of sessions
$trans->check_cmd_value("REGISTER", "INIT_TCLASS tclass1", "register transport");
# Check the identify function works as expected
# These should cause the connection to be dropped with failed identification
$trans->do("CONNECT tclass1 conn1", "create connection 1");
$msg->check_cmd_value("PEEK_MESSAGE", "NO", "no messages waiting");
read_message("conn1", "SERVER x", "SERVER", "ABORT conn1");
$dancer->expect_log(3, {facility => 'protocol', level => 'info',
                        msg => "Dropping session MOOOOOOOOOOOOOOOO, unrecognised protocol"});
$trans->do("CONNECT tclass1 conn1", "create connection 1");
$msg->check_cmd_value("PEEK_MESSAGE", "NO", "no messages waiting");
read_message("conn1", "\01", "invalid message (character set)", "ABORT conn1");
$dancer->expect_log(3, {facility => 'protocol', level => 'info',
                        msg => "Dropping session MOOOOOOOOOOOOOOOO, unrecognised protocol"});
foreach (511..514)
  {
    $trans->do("CONNECT tclass1 conn1", "create connection 1");
    $msg->check_cmd_value("PEEK_MESSAGE", "NO", "no messages waiting");
    read_message("conn1", "x" x $_, "invalid message (length $_)", "ABORT conn1");
    $dancer->expect_log(3, {facility => 'protocol', level => 'info',
                            msg => "Dropping session MOOOOOOOOOOOOOOOO, unrecognised protocol"});
  }

# USER is accepted
$trans->do("CONNECT tclass1 conn1", "create connection 1");
read_message("conn1", "USER x x x :Real name 1", "USER");
$proto->check_value("NEW_SESSION", "new session from protocol");
$proto->do("NAME_SESSION 0", "name session 0");
$msg->check_value("NEW_SESSION", "new session from message");
$msg->check_value("MESSAGE_HANDLER: all-real", "handled real name from conn1");

# As is NICK
$trans->do("CONNECT tclass1 conn2", "create connection 2");
read_message("conn2", "NICK name1", "NICK");
$proto->check_value("NEW_SESSION", "new session from protocol");
$proto->do("NAME_SESSION 1", "name session 1");
$msg->check_value("NEW_SESSION", "new session from message");
$msg->check_value("MESSAGE_HANDLER: all-name", "handled real name from conn2");

# Maximum-length message
$trans->do("CONNECT tclass1 conn3", "create connection 3");
read_message("conn3", "USER x x x :" . ('x' x 498), "USER");
$proto->check_value("NEW_SESSION", "new session from protocol");
$proto->do("NAME_SESSION 2", "name session 2");
$msg->check_value("NEW_SESSION", "new session from message");
$msg->check_value("MESSAGE_HANDLER: all-real", "handled real name from conn1");

# Split message
$trans->do("CONNECT tclass1 conn4", "create connection 4");
$trans->do("READ conn4 :US", "send first part of USER to conn4 transport");
$trans->do("READ conn4 :ER x x", "send second part of USER to conn4 transport");
$trans->do("READ conn4 : x :Real name 2\n", "send last part of USER to conn4 transport");
$proto->check_value("NEW_SESSION", "new session from protocol");
$proto->do("NAME_SESSION 3", "name session 3");
$msg->check_value("NEW_SESSION", "new session from message");
$msg->check_value("MESSAGE_HANDLER: all-real", "handled real name from conn1");

# Not the first message
$trans->do("CONNECT tclass1 conn5", "create connection 5");
read_message("conn5", "INVALID", "INVALID");
read_message("conn5", "USER x x x :Real name 3", "USER", "WRITE conn5: 421  INVALID :Unknown command\\r");
$proto->check_value("NEW_SESSION", "new session from protocol");
$proto->do("NAME_SESSION 4", "name session 4");
$msg->check_value("NEW_SESSION", "new session from message");
$msg->check_value("MESSAGE_HANDLER: all-real", "handled real name from conn1");

# An extra one with USER/NICK in the other order (we use 2-5 in testing 'reject set real name')
$trans->do("CONNECT tclass1 conn6", "create connection 6");
read_message("conn6", "NICK name2", "NICK");
$proto->check_value("NEW_SESSION", "new session from protocol");
$proto->do("NAME_SESSION 5", "name session 5");
$msg->check_value("NEW_SESSION", "new session from message");
$msg->check_value("MESSAGE_HANDLER: all-name", "handled real name from conn6");

# And one for testing close-duringessage-parse
$trans->do("CONNECT tclass1 conn7", "create connection 7");
read_message("conn7", "NICK name3", "NICK");
$proto->check_value("NEW_SESSION", "new session from protocol");
$proto->do("NAME_SESSION 6", "name session 6");
$msg->check_value("NEW_SESSION", "new session from message");
$msg->check_value("MESSAGE_HANDLER: all-name", "handled real name from conn7");

# We can drop this now and just hook all the messages
$msg->do("IGNORE_TYPE_FROM_ALL all-real set real name", "ignore all 'set real name' messages");
$msg->do("IGNORE_TYPE_FROM_ALL all-name set name", "ignore all 'set real name' messages");
$msg->do("REQUEST_ALL_FROM_SESSION 0 conn1", "request all from conn1");
$msg->do("REQUEST_ALL_FROM_SESSION 1 conn2", "request all from conn2");
$msg->do("REQUEST_ALL_FROM_SESSION 2 conn3", "request all from conn3");
$msg->do("REQUEST_ALL_FROM_SESSION 3 conn4", "request all from conn4");
$msg->do("REQUEST_ALL_FROM_SESSION 4 conn5", "request all from conn5");
$msg->do("REQUEST_ALL_FROM_SESSION 5 conn6", "request all from conn6");

# Pick up all the messages we generated during that little session
check_message type => 'set real name', str => ["Real name 1"];
check_message type => 'set name', str => ["name1"];
check_message type => 'set real name', str => ['x' x 498];
check_message type => 'set real name', str => ["Real name 2"];
check_message type => 'set real name', str => ["Real name 3"];
check_message type => 'set name', str => ["name2"];
check_message type => 'set name', str => ["name3"];
$msg->check_cmd_value("PEEK_MESSAGE", "NO", "no messages waiting");

###################################
# Check incoming message handlers #
###################################

# Okay, now let's run over all the message types
read_message("conn1", "NICK nick1", "NICK");
$msg->check_value("MESSAGE_HANDLER: conn1", "handled message from conn1");
$msg->check_value("MESSAGE_HANDLER: conn1", "handled message from conn1");
check_message type => 'set name', str => ["nick1"];
check_message type => 'done handshake';

read_message("conn1", "WHOIS nick2", "WHOIS");
$msg->check_value("MESSAGE_HANDLER: conn1", "handled message from conn1");
check_message type => 'query node information', str => ["nick2"];

read_message("conn1", "QUIT", "QUIT");
$msg->check_value("MESSAGE_HANDLER: conn1", "handled message from conn1");
check_message type => 'close session', str => [undef];

read_message("conn1", "QUIT reason", "QUIT");
$msg->check_value("MESSAGE_HANDLER: conn1", "handled message from conn1");
check_message type => 'close session', str => [undef, "reason"];

read_message("conn1", "VERSION", "VERSION");
$msg->check_value("MESSAGE_HANDLER: conn1", "handled message from conn1");
check_message type => 'query node software information';

read_message("conn1", "LICENSE", "LICENSE");
$msg->check_value("MESSAGE_HANDLER: conn1", "handled message from conn1");
check_message type => 'query node license information';

read_message("conn1", "INFO", "INFO");
$msg->check_value("MESSAGE_HANDLER: conn1", "handled message from conn1");
check_message type => 'query node author information';

read_message("conn1", "BEGINAUTH type name", "BEGINAUTH");
$msg->check_value("MESSAGE_HANDLER: conn1", "handled message from conn1");
check_message type => 'begin auth', str => ['type', 'name'];

read_message("conn1", "CANCELAUTH", "CANCELAUTH");
$msg->check_value("MESSAGE_HANDLER: conn1", "handled message from conn1");
check_message type => 'cancel auth';

read_message("conn1", "CANCELAUTH reason", "CANCELAUTH");
$msg->check_value("MESSAGE_HANDLER: conn1", "handled message from conn1");
check_message type => 'cancel auth', str => ['reason'];

read_message("conn1", "ACCEPTAUTH reason", "ACCEPTAUTH");
$msg->check_value("MESSAGE_HANDLER: conn1", "handled message from conn1");
check_message type => 'accept auth';

read_message("conn1", "AUTHDATA", "AUTHDATA");
$msg->check_value("MESSAGE_HANDLER: conn1", "handled message from conn1");
check_message type => 'auth data';

read_message("conn1", "AUTHDATA alpha", "AUTHDATA");
$msg->check_value("MESSAGE_HANDLER: conn1", "handled message from conn1");
check_message type => 'auth data', str => ['alpha'];

read_message("conn1", "AUTHDATA beta gamma delta", "AUTHDATA");
$msg->check_value("MESSAGE_HANDLER: conn1", "handled message from conn1");
check_message type => 'auth data', str => ['beta', 'gamma', 'delta'];

read_message("conn1", "PRIVMSG nick1 :message", "PRIVMSG");
$msg->check_value("MESSAGE_HANDLER: conn1", "handled message from conn1");
check_message type => 'private', dest => 'nick1', str => ['message'];

read_message("conn1", "PRIVMSG #channel1 :message", "PRIVMSG");
$msg->check_value("MESSAGE_HANDLER: conn1", "handled message from conn1");
check_message type => 'channel', str => ['channel1', 'message'];

read_message("conn1", "PRIVMSG nick1 :\01message\01", "PRIVMSG");
$msg->check_value("MESSAGE_HANDLER: conn1", "handled message from conn1");
check_message type => 'ctcp query', dest => 'nick1', str => ['message'];

read_message("conn1", "PRIVMSG #channel1 :\01message\01", "PRIVMSG");
$msg->check_value("MESSAGE_HANDLER: conn1", "handled message from conn1");
check_message type => 'channel ctcp query', str => ['channel1', 'message'];

read_message("conn1", "PRIVMSG nick1 :", "PRIVMSG");
$msg->check_value("MESSAGE_HANDLER: conn1", "handled message from conn1");
check_message type => 'private', dest => 'nick1', str => [''];

read_message("conn1", "PRIVMSG nick1 :ascii message: ", "PRIVMSG");
$msg->check_value("MESSAGE_HANDLER: conn1", "handled message from conn1");
check_message type => 'private', dest => 'nick1', str => ['ascii message: ???'];

read_message("conn1", "PRIVMSG #channel1 :ascii message: ", "PRIVMSG");
$msg->check_value("MESSAGE_HANDLER: conn1", "handled message from conn1");
check_message type => 'channel', str => ['channel1', 'ascii message: ???'];

read_message("conn1", "PRIVMSG nick1 :\01ascii message: \01", "PRIVMSG");
$msg->check_value("MESSAGE_HANDLER: conn1", "handled message from conn1");
check_message type => 'ctcp query', dest => 'nick1', str => ['ascii message: ???'];

read_message("conn1", "PRIVMSG #channel1 :\01ascii message: \01", "PRIVMSG");
$msg->check_value("MESSAGE_HANDLER: conn1", "handled message from conn1");
check_message type => 'channel ctcp query', str => ['channel1', 'ascii message: ???'];

# /usr/share/i18n/charmaps/UTF-8.gz for names

# CHARSET does not generate any internal messages; it affects only the
# rfc1459.client module.
read_message("conn1", "CHARSET invalid", "CHARSET",
             "WRITE conn1: : NOTICE CHARSET :The character set 'invalid' is not supported\\r");
read_message("conn1", "CHARSET ISO8859-1", "CHARSET",
             "WRITE conn1: : NOTICE CHARSET :Your character set is now 'ISO8859-1'\\r");

# Now these should be identified as pound signs
read_message("conn1", "PRIVMSG nick1 :iso 8859-1 message: ", "PRIVMSG");
$msg->check_value("MESSAGE_HANDLER: conn1", "handled message from conn1");
# Note: '£' is a UTF-8 pound sign
check_message type => 'private', dest => 'nick1', str => ["iso 8859-1 message: £££"];

# Put it back to ascii, just in case
read_message("conn1", "CHARSET ASCII", "CHARSET",
             "WRITE conn1: : NOTICE CHARSET :Your character set is now 'ASCII'\\r");

# NICK then USER, instead of USER then NICK
read_message("conn6", "USER x x x :Real name 4", "USER");
$msg->check_value("MESSAGE_HANDLER: conn6", "handled message from conn6");
$msg->check_value("MESSAGE_HANDLER: conn6", "handled message from conn6");
check_message type => 'set real name', str => ["Real name 4"];
check_message type => 'done handshake';

# Check the quoting
read_message("conn1", "QUIT :reason", "QUIT with quoted reason");
$msg->check_value("MESSAGE_HANDLER: conn1", "handled message from conn1");
check_message type => 'close session', str => [undef, "reason"];
read_message("conn1", "QUIT :long reason", "QUIT with quoted reason");
$msg->check_value("MESSAGE_HANDLER: conn1", "handled message from conn1");
check_message type => 'close session', str => [undef, "long reason"];
read_message("conn1", "QUIT long reason", "QUIT with unquoted reason");
$msg->check_value("MESSAGE_HANDLER: conn1", "handled message from conn1");
check_message type => 'close session', str => [undef, "long reason"];

# Prefix is ignored
read_message("conn1", ":iti QUIT :reason", "QUIT with prefix");
$msg->check_value("MESSAGE_HANDLER: conn1", "handled message from conn1");
check_message type => 'close session', str => [undef, "reason"];

# Extra spaces after the command are ignored
read_message("conn1", "QUIT    :reason", "QUIT with extra spaces");
$msg->check_value("MESSAGE_HANDLER: conn1", "handled message from conn1");
check_message type => 'close session', str => [undef, "reason"];

# As are extra spaces in the middle
read_message("conn1", "USER x x     x :Real name 1, second time", "USER");
$msg->check_value("MESSAGE_HANDLER: conn1", "handled message from conn1");
check_message type => 'set real name', str => ["Real name 1, second time"];

# Lines ending in \r\n are acceptable
read_message("conn1", "QUIT :reason\r", "QUIT with \\r\\n");
$msg->check_value("MESSAGE_HANDLER: conn1", "handled message from conn1");
check_message type => 'close session', str => [undef, "reason"];

# As are lines ending only in \r
$trans->do("READ conn1 :QUIT :reason\r", "QUIT with only \\r");
$msg->check_value("MESSAGE_HANDLER: conn1", "handled message from conn1");
check_message type => 'close session', str => [undef, "reason"];

# Lines of exactly 510 characters are preserved
read_message("conn1", "QUIT :" . ('x' x 504), "QUIT with 510 characters");
$msg->check_value("MESSAGE_HANDLER: conn1", "handled message from conn1");
check_message type => 'close session', str => [undef, 'x' x 504];

# Lines of more than 510 characters are truncated
foreach (511..514)
  {
    read_message("conn1", "QUIT :" . ('x' x ($_ - 6)), "QUIT with $_ characters");
    $msg->check_value("MESSAGE_HANDLER: conn1", "handled message from conn1");
    check_message type => 'close session', str => [undef, 'x' x 504];
  }

# And the same truncation applies with \r\n
read_message("conn1", "QUIT :" . ('x' x 504) . "\r", "QUIT with 510 characters and \\r\\n");
$msg->check_value("MESSAGE_HANDLER: conn1", "handled message from conn1");
check_message type => 'close session', str => [undef, 'x' x 504];
foreach (511..514)
  {
    read_message("conn1", "QUIT :" . ('x' x ($_ - 6)) . "\r", "QUIT with $_ characters");
    $msg->check_value("MESSAGE_HANDLER: conn1", "handled message from conn1");
    check_message type => 'close session', str => [undef, 'x' x 504];
  }

# And with just \r
$trans->do("READ conn1 :QUIT :" . ('x' x 504) . "\r", "QUIT with 510 characters and only \\r");
$msg->check_value("MESSAGE_HANDLER: conn1", "handled message from conn1");
check_message type => 'close session', str => [undef, 'x' x 504];
foreach (511..514)
  {
    $trans->do("READ conn1 :QUIT :" . ('x' x ($_ - 6)) . "\r", "QUIT with $_ characters and only \\r");
    $msg->check_value("MESSAGE_HANDLER: conn1", "handled message from conn1");
    check_message type => 'close session', str => [undef, 'x' x 504];
  }

# A message with only a prefix is ignored
read_message("conn1", ":iti", "prefix only, no message");
$msg->check_cmd_value("PEEK_MESSAGE", "NO", "no messages waiting");
read_message("conn1", ":iti ", "prefix only, no message");
$msg->check_cmd_value("PEEK_MESSAGE", "NO", "no messages waiting");
read_message("conn1", ":iti  ", "prefix only, no message");
$msg->check_cmd_value("PEEK_MESSAGE", "NO", "no messages waiting");

# Errors
$trans->check_cmd_value("READ conn1 :NICK\n", "WRITE conn1: 461  NICK :Not enough parameters\\r", "NICK without parameters");
$trans->check_cmd_value("READ conn1 :INVALID\n", "WRITE conn1: 421  INVALID :Unknown command\\r", "INVALID message");
$trans->check_cmd_value("READ conn1 :INVALID foo\n", "WRITE conn1: 421  INVALID :Unknown command\\r", "INVALID message with parameter");

#######################################
# Sending messages back the other way #
#######################################

no warnings 'redefine';
sub send_message($%)
  {
    my $conn = shift;
    (new Dancer::Message @_)->send($proto, $conn, [caller]);
  }
use warnings 'redefine';

# This one is ignored
send_message 0, type => 'done handshake';
$dancer->expect_log(3, {facility => 'protocol.rfc1459.client', level => 'error',
                        msg => "Can't send 'done handshake' message: missing origin"});

send_message 0, type => 'reject set name', origin => 'alpha', dest => 'beta', str => ['charlie', undef];
$trans->check_value("WRITE conn1: :alpha 432 beta charlie :Erroneous nickname\\r", "432 numeric");
send_message 0, type => 'reject set name', origin => 'alpha', dest => 'beta', str => ['charlie', 'delta'];
$trans->check_value("WRITE conn1: :alpha 432 beta charlie :delta\\r", "432 numeric");
send_message 0, type => 'reject set name', origin => 'alpha', str => ['charlie', 'delta'];
$trans->check_value("WRITE conn1: :alpha 432 * charlie :delta\\r", "432 numeric");
send_message 0, type => 'reject set name', origin => 'alpha', dest => 'beta', str => ['charlie'];
$trans->check_value("WRITE conn1: :alpha 433 beta charlie :Nickname is already in use\\r", "433 numeric");
send_message 0, type => 'reject set name', origin => 'alpha', str => ['charlie'];
$trans->check_value("WRITE conn1: :alpha 433 * charlie :Nickname is already in use\\r", "433 numeric");

send_message 0, type => 'wrong parameters', origin => 'alpha', dest => 'beta', str => ['charlie'];
$trans->check_value("WRITE conn1: :alpha 461 beta * :charlie\\r", "461 numeric");
send_message 0, type => 'wrong parameters', origin => 'alpha', dest => 'beta', str => [undef];
$trans->check_value("WRITE conn1: :alpha 461 beta * :Not enough parameters\\r", "461 numeric");
send_message 0, type => 'wrong parameters', origin => 'alpha', dest => 'beta';
$trans->check_value("WRITE conn1: :alpha 461 beta * :Not enough parameters\\r", "461 numeric");
send_message 0, type => 'wrong parameters', origin => 'alpha';
$trans->check_value("WRITE conn1: :alpha 461 * * :Not enough parameters\\r", "461 numeric");

send_message 0, type => 'handshake required', origin => 'alpha', dest => 'beta';
$trans->check_value("WRITE conn1: :alpha 451 beta * :Register first\\r", "461 numeric");
send_message 0, type => 'handshake required', origin => 'alpha';
$trans->check_value("WRITE conn1: :alpha 451 * * :Register first\\r", "461 numeric");

# This one kills the connection
send_message 4, type => 'reject set real name', origin => 'alpha', dest => 'beta', str => ['charlie', 'delta'];
$trans->check_value("WRITE conn5: ERROR :Closing link: beta (delta [charlie])\\r", "ERROR message");
$trans->check_value("CLOSE conn5", "got close from conn2 transport");
$trans->do("CLOSE conn5", "close conn5");
$msg->check_value("MESSAGE_HANDLER: conn5", "NULL message handled");
check_message type => '(null)';
send_message 3, type => 'reject set real name', origin => 'alpha', dest => 'beta', str => ['charlie', undef];
$trans->check_value("WRITE conn4: ERROR :Closing link: beta (Invalid username [charlie])\\r", "ERROR message");
$trans->check_value("CLOSE conn4", "got close from conn2 transport");
$trans->do("CLOSE conn4", "close conn4");
$msg->check_value("MESSAGE_HANDLER: conn4", "NULL message handled");
check_message type => '(null)';
send_message 2, type => 'reject set real name', origin => 'alpha', dest => 'beta', str => ['charlie'];
$trans->check_value("WRITE conn3: ERROR :Closing link: beta (Invalid username [charlie])\\r", "ERROR message");
$trans->check_value("CLOSE conn3", "got close from conn2 transport");
$trans->do("CLOSE conn3", "close conn3");
$msg->check_value("MESSAGE_HANDLER: conn3", "NULL message handled");
check_message type => '(null)';
send_message 1, type => 'reject set real name', origin => 'alpha', str => ['charlie'];
$trans->check_value("WRITE conn2: ERROR :Closing link: * (Invalid username [charlie])\\r", "ERROR message");
$trans->check_value("CLOSE conn2", "got close from conn2 transport");
$trans->do("CLOSE conn2", "close conn2");
$msg->check_value("MESSAGE_HANDLER: conn2", "NULL message handled");
check_message type => '(null)';

# This one changes state in the protocol module
send_message 0, type => 'accept handshake', origin => 'alpha', dest => 'beta';
$trans->check_value("WRITE conn1: :alpha 001 beta :Welcome to the Internet Relay Network beta\\r", "001 numeric");
$trans->check_value("WRITE conn1: :alpha 002 beta :Your host is *[*/0], running version 1.1.0-unreleased/protocol.rfc1459.client-1.0\\r", "001 numeric");
# And causes an 'accept handshake' message to go back the other way
$msg->check_value("MESSAGE_HANDLER: conn1", "'accept handshake' message handled");
check_message type => 'accept handshake', origin => 'beta', dest => 'alpha';

# Check that CHARSET works as expected
read_message("conn1", "CHARSET EBCDIC-US", "CHARSET",
             "WRITE conn1: :alpha NOTICE beta :Your character set is now 'EBCDIC-US'\\r");
# Now, we should get EBCDIC '?' characters when we try to send a  symbol
send_message 0, type => 'private', origin => 'beta', dest => 'alpha', str => ["iso 8859-1 message: £££"];
# Fear this obsolete drivel
$trans->check_value("WRITE conn1: :beta PRIVMSG alpha :\@`\@z\@oooooo\\r", "PRIVMSG");

read_message("conn1", "CHARSET ASCII", "CHARSET",
             "WRITE conn1: :alpha NOTICE beta :Your character set is now 'ASCII'\\r");

# Send 'node information'
send_message 0, type => 'node information', origin => 'router.', dest => 'alpha', str => ['beta', 'gamma', 'delta'], uint32 => [1];
$trans->check_value("WRITE conn1: :router. 311 alpha beta beta gamma * :delta\\r", "311 numeric");
$trans->check_value("WRITE conn1: :router. 318 alpha beta :End of /WHOIS list\\r", "318 numeric");
send_message 0, type => 'node information', origin => 'router.', dest => 'alpha', str => ['beta', undef, undef], uint32 => [1];
$trans->check_value("WRITE conn1: :router. 311 alpha beta beta * * :\\r", "311 numeric");
$trans->check_value("WRITE conn1: :router. 318 alpha beta :End of /WHOIS list\\r", "318 numeric");
send_message 0, type => 'node information', origin => 'router.', dest => 'alpha', str => ['beta'], uint32 => [0];
$trans->check_value("WRITE conn1: :router. 401 alpha beta :No such nick/channel\\r", "401 numeric");
send_message 0, type => 'node information', origin => 'router.', dest => 'alpha', str => ['beta', 'gamma'], uint32 => [0];
$trans->check_value("WRITE conn1: :router. 401 alpha beta :gamma\\r", "401 numeric");

# Send 'node software information'
send_message 0, type => 'node software information', origin => 'router.', dest => 'alpha', str => ['beta', 'gamma', 'delta', 'epsilon', 'eta'];
$trans->check_value("WRITE conn1: :router. 351 alpha beta-gamma(delta) router. :eta\\r", "351 numeric");

# Send 'node license information'
send_message 0, type => 'node license information', origin => 'router.', dest => 'alpha', str => ['beta', 'gamma', 'delta'];
$trans->check_value("WRITE conn1: :router. 371 alpha :beta\\r", "371 numeric");
$trans->check_value("WRITE conn1: :router. 371 alpha :gamma\\r", "371 numeric");
$trans->check_value("WRITE conn1: :router. 371 alpha :delta\\r", "371 numeric");
$trans->check_value("WRITE conn1: :router. 374 alpha :End of /LICENSE list\\r", "374 numeric");

# Send 'node author information'
send_message 0, type => 'node author information', origin => 'router.', dest => 'alpha', str => ['beta', 'gamma', 'delta'];
$trans->check_value("WRITE conn1: :router. 371 alpha :beta\\r", "371 numeric");
$trans->check_value("WRITE conn1: :router. 371 alpha :gamma\\r", "371 numeric");
$trans->check_value("WRITE conn1: :router. 371 alpha :delta\\r", "371 numeric");
$trans->check_value("WRITE conn1: :router. 374 alpha :End of /INFO list\\r", "374 numeric");

# 'private'
send_message 0, type => 'private', origin => 'beta', dest => 'alpha', str => ['gamma'];
$trans->check_value("WRITE conn1: :beta PRIVMSG alpha :gamma\\r", "PRIVMSG");

# 'channel'
send_message 0, type => 'channel', origin => 'beta', str => ['alpha', 'gamma'];
$trans->check_value("WRITE conn1: :beta PRIVMSG #alpha :gamma\\r", "PRIVMSG");

# 'ctcp query'
send_message 0, type => 'ctcp query', origin => 'beta', dest => 'alpha', str => ['gamma'];
$trans->check_value("WRITE conn1: :beta PRIVMSG alpha :\01gamma\01\\r", "PRIVMSG");

# 'ctcp response'
send_message 0, type => 'ctcp response', origin => 'beta', dest => 'alpha', str => ['gamma'];
$trans->check_value("WRITE conn1: :beta NOTICE alpha :\01gamma\01\\r", "NOTICE");

# 'channel ctcp query'
send_message 0, type => 'channel ctcp query', origin => 'beta', str => ['alpha', 'gamma'];
$trans->check_value("WRITE conn1: :beta PRIVMSG #alpha :\01gamma\01\\r", "PRIVMSG");

# 'accept auth'
send_message 0, type => 'accept auth', origin => 'router.', dest => 'alpha';
$trans->check_value("WRITE conn1: :router. 051 alpha :Authentication accepted\\r", "051 numeric");

# 'cancel auth'
send_message 0, type => 'cancel auth', origin => 'router.', dest => 'alpha';
$trans->check_value("WRITE conn1: :router. 052 alpha :Authentication cancelled\\r", "052 numeric");
send_message 0, type => 'cancel auth', origin => 'router.', dest => 'alpha', str => ['beta'];
$trans->check_value("WRITE conn1: :router. 052 alpha :beta\\r", "052 numeric");

# 'auth data'
send_message 0, type => 'auth data', origin => 'router.', dest => 'alpha';
$trans->check_value("WRITE conn1: :router. 050 alpha\\r", "050 numeric");
send_message 0, type => 'auth data', origin => 'router.', dest => 'alpha', str => ['beta', 'gamma', 'delta'];
$trans->check_value("WRITE conn1: :router. 050 alpha beta gamma delta\\r", "050 numeric");
# Now the string boundary cases
send_message 0, type => 'auth data', origin => 'router.', dest => 'alpha', str => ['beta', 'gamma', 'delta', 'x' x 473];
$trans->check_value("WRITE conn1: :router. 050 alpha beta gamma delta " . ('x' x 473) . "\\r", "050 numeric");
send_message 0, type => 'auth data', origin => 'router.', dest => 'alpha', str => ['beta', 'gamma', 'delta', 'x' x 474];
$trans->check_value("WRITE conn1: :router. 050 alpha beta gamma delta " . ('x' x 474) . "\\r", "050 numeric");
send_message 0, type => 'auth data', origin => 'router.', dest => 'alpha', str => ['beta', 'gamma', 'delta', 'x' x 475];
$trans->check_value("WRITE conn1: :router. 050 alpha beta gamma delta " . ('x' x 474) . "\\r", "050 numeric");
send_message 0, type => 'auth data', origin => 'router.', dest => 'alpha', str => ['beta', 'gamma', 'delta', 'x' x 476];
$trans->check_value("WRITE conn1: :router. 050 alpha beta gamma delta " . ('x' x 474) . "\\r", "050 numeric");
send_message 0, type => 'auth data', origin => 'router.', dest => 'alpha', str => ['beta', 'gamma', 'delta', 'x' x 473, 'epsilon'];
$trans->check_value("WRITE conn1: :router. 050 alpha beta gamma delta " . ('x' x 473) . "\\r", "050 numeric");
send_message 0, type => 'auth data', origin => 'router.', dest => 'alpha', str => ['beta', 'gamma', 'delta', 'x' x 472, 'epsilon'];
$trans->check_value("WRITE conn1: :router. 050 alpha beta gamma delta " . ('x' x 472) . " e\\r", "050 numeric");

# This is a hash of message names to arrays of test cases

# For each test case, a message is formed by adding an origin and dest
# to the 'send' hash, and the expected response is found by evaling
# the 'receive' string with %args defined as being the full set of
# parameters used to construct the message.

# If the key 'debug' is defined to be 1 in a given test case,
# $dancer->gdb() will be called before anything is sent.

my %messages = ('relaying denied'
                => [{send => {str => ['beta']
                            },
                    receive => ":\$args{origin} 481 \$args{dest} :You can't do that, you're no deity",
                   }],
                'not authenticated'
                => [{send => {uint32 => [0]},
                    receive => ":\$args{origin} 399 \$args{dest} :Authenticate first",
                   }],
                'not authorised'
                => [{send => {uint32 => [0]},
                    receive => ":\$args{origin} 399 \$args{dest} :You are not authorised to perform that operation",
                   }],
                'accept set credentials'
                => [{send => {str => ['beta']},
                    receive => ":\$args{origin} 399 \$args{dest} :Credentials updated",
                   }],
                'reject set credentials'
                => [{send => {str => ['beta']},
                    receive => ":\$args{origin} 399 \$args{dest} :Credentials not updated: invalid authentication method specified",
                   },
                   {send => {str => ['beta', 'gamma']},
                    receive => ":\$args{origin} 399 \$args{dest} :Credentials not updated: gamma",
                   }],
                'accept register user'
                => [{send => {},
                    receive => ":\$args{origin} 399 \$args{dest} :Your username has been registered",
                   }],
                'reject register user'
                => [{send => {},
                    receive => ":\$args{origin} 399 \$args{dest} :User not registered",
                   },
                    {send => {str => ['beta']},
                    receive => ":\$args{origin} 399 \$args{dest} :User not registered: beta",
                   }],
                'registered user'
                => [{send => {},
                    receive => ":\$args{origin} 399 \$args{dest} :This username has been registered. Please authenticate.",
                   }],
                'unregistered user'
                => [{send => {},
                    receive => ":\$args{origin} 399 \$args{dest} :This username has not been registered. Please register.",
                   }],
               );

foreach my $message (keys %messages)
  {
    foreach my $variant (@{$messages{$message}})
      {
        my %args = (type => $message, origin => 'router.', dest => 'alpha', %{$variant->{send}});
        $dancer->gdb() if $variant->{debug};
        send_message 0, %args;
        my $sub = eval "sub {my \%args = \@_; return \"$variant->{receive}\"}";
        die $@ if $@;
        $trans->check_value("WRITE conn1: " . $sub->(%args) . "\\r", "Response to '$message'");
      }
  }

# Errors again, now with a client name
$trans->check_cmd_value("READ conn1 :NICK\n", "WRITE conn1: 461 beta NICK :Not enough parameters\\r", "NICK without parameters");
$trans->check_cmd_value("READ conn1 :INVALID\n", "WRITE conn1: 421 beta INVALID :Unknown command\\r", "INVALID message");
$trans->check_cmd_value("READ conn1 :INVALID foo\n", "WRITE conn1: 421 beta INVALID :Unknown command\\r", "INVALID message with parameter");

# Boundary case for some other random strings
send_message 0, type => 'cancel auth', origin => 'router.', dest => 'alpha', str => ['x' x 489];
$trans->check_value("WRITE conn1: :router. 052 alpha :" . ('x' x 489) . "\\r", "052 numeric");
send_message 0, type => 'cancel auth', origin => 'router.', dest => 'alpha', str => ['x' x 490];
$trans->check_value("WRITE conn1: :router. 052 alpha :" . ('x' x 490) . "\\r", "052 numeric");
send_message 0, type => 'cancel auth', origin => 'router.', dest => 'alpha', str => ['x' x 491];
$trans->check_value("WRITE conn1: :router. 052 alpha :" . ('x' x 490) . "\\r", "052 numeric");
send_message 0, type => 'cancel auth', origin => 'router.', dest => 'alpha', str => ['x' x 492];
$trans->check_value("WRITE conn1: :router. 052 alpha :" . ('x' x 490) . "\\r", "052 numeric");

# Check we didn't miss any
$msg->check_cmd_value("PEEK_MESSAGE", "NO", "no messages waiting");

# Close some sessions in varying states
$proto->do("SHUTDOWN_SESSION 0", "shutdown session for conn1");
$trans->check_value("CLOSE conn1", "got close from conn1 transport");
$trans->do("CLOSE conn1", "close conn1");
$msg->check_value("MESSAGE_HANDLER: conn1", "NULL message handled");
check_message type => '(null)';

$msg->check_cmd_value("PEEK_MESSAGE", "NO", "no messages waiting");

# Close a session from within the message parsing stuff
$msg->do("AUTOQUIT_ON", "turn on autoquit-whenessage-handled");
$trans->do("AUTOCLOSE_ON", "turn on autoclose-when-transport-close");
read_message("conn6", "QUIT reason", "QUIT", "CLOSE conn6");
$msg->check_value("MESSAGE_HANDLER: conn6", "handled message from conn6");
$msg->check_value("MESSAGE_HANDLER: conn6", "NULL message handled");
check_message type => 'close session', str => [undef, "reason"];
check_message type => '(null)';
$msg->check_cmd_value("PEEK_MESSAGE", "NO", "no messages waiting");

verbose "Finished test run";

$trans->kill;
$msg->kill;

$dancer->kill();

1;
