/************************************************************************
 *   This is part of Dancer, a text relay daemon
 *   Copyright (C) 2002 Andrew Suffield <asuffield@users.sourceforge.net>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

/* This is like tail -f, but faster; tail normally only polls the file
 * once per second.
 *
 * It also starts by displaying the entire file, from the start, which
 * happens to be what we want.
 */

#include <dancerd/config.h>

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>
#include <unistd.h>
#include <signal.h>

static void usage(void) noreturn_attribute;
static void
usage(void)
{
  fprintf(stderr, "Usage:\n");
  fprintf(stderr, "fasttail <file>\n");
  exit(1);
}

static int sigint_raised = 0;

static void
sigint_handler(int sig unused_attribute)
{
  sigint_raised = 1;
}

int
main(int argc, char *argv[])
{
  int fd;
  char buf[512];
  struct sigaction sa;

  if (argc != 2)
    usage();

  fd = open(argv[1], O_RDONLY, 0);
  if (fd == -1)
    {
      fprintf(stderr, "Failed to open %s: %s\n", argv[1], strerror(errno));
      exit(1);
    }

  sa.sa_handler = &sigint_handler;
  sa.sa_flags = SA_ONESHOT;
  sigemptyset(&sa.sa_mask);
  sigaction(SIGINT, &sa, NULL);

  while(1)
    {
      ssize_t len = read(fd, buf, 512), buflen = len, wrotelen = 0;
      char *p = buf;
      if ((len == -1) && (errno == EINTR))
        continue;
      if (len == -1)
        {
          if (errno == EINTR)
            continue;
          else
            {
              fprintf(stderr, "Failed to read from %s: %s (%d, %d)", argv[1], strerror(errno), errno, EINTR);
              exit(1);
            }
        }
      while (buflen > 0)
        {
          wrotelen = write(1, p, buflen);
          if (wrotelen == -1)
            {
              if (errno == EINTR)
                continue;
              else
                {
                  fprintf(stderr, "Failed to write to stdout: %s", strerror(errno));
                  exit(1);
                }
            }
          p += wrotelen;
          buflen -= wrotelen;
        }
      /* Exit after SIGINT if there is no data left to read */
      if (sigint_raised && (len == 0))
        {
          exit(0);
        }
      /* This is all it takes to forcibly get the processor usage down to almost nothing */
      if (len == 0)
        {
          usleep(1);
        }
    }
}
