# Copyright (C) 2002 Andrew Suffield <asuffield@users.sourceforge.net>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA

package Dancer::CmdInterpreter;

use strict;
use 5.006;
use warnings;

use Cwd;
use Carp;
use Dancer::CmdScript;
use DejaPerl qw/:framework/;
use Parse::RecDescent;
use Data::Dumper;

use Dancer::CmdScript::Conditional;
use Dancer::CmdScript::Block;
use Dancer::CmdScript::IO;
use Dancer::CmdScript::Function;

use Dancer::CmdScript::Expression::Assign;
use Dancer::CmdScript::Expression::BinOp;
use Dancer::CmdScript::Expression::Atom;

# All the elements in a script need to define these methods:

# interpolate_function - recurse down the tree, and interpolate any
# functions

# interpolate_stereotypes - likewise

# optimise - return a minimal object that has the same effect when
# evaluated

# constant - if true, then eval() may not pay any attention to its
# script parameter, may not have any side effects, and must always
# return the same value. Some constants will be folded out of
# existance entirely.

# eval - given a script parameter to store the current state, evaluate
# the element

BEGIN
  {
    use Exporter;
    our @ISA = qw/Exporter/;
    our $VERSION = "0.0.1";
    our @EXPORT = qw//;
    our @EXPORT_OK = qw//;
  }

my %functions = ();
my %stereotype = ();

sub new($%)
  {
    my $class = shift;
    my %args = @_;

    my $self = {
                category => $args{category},
               };
    bless $self, $class;
    return $self;
  }

sub register_function
  {
    my ($name, $sub, @args) = @_;
    carp "Redefining function $name" if exists $functions{$name};
    $functions{$name} = {name => $name, sub => $sub, args => [@args]};
  }

sub unregister_function
  {
    my $name = shift;
    delete $functions{$name};
  }

sub function
  {
    my $self = shift;
    my $name = shift;
    return $functions{$name};
  }

sub register_stereotype
  {
    my ($name, $open, $listen, @args) = @_;
    carp "Redefining stereotype $name" if exists $stereotype{$name};
    $stereotype{$name} = {
                          name => $name,
                          open => $open,
                          listen => $listen,
                          args => [@args],
                         };
  }

sub unregister_stereotype
  {
    my $name = shift;
    delete $stereotype{$name};
  }

sub stereotype
  {
    my $self = shift;
    my $name = shift;
    return $stereotype{$name};
  }

require Dancer::CmdScript::Call;

$::RD_TRACE = 1 if $ENV{RD_TRACE};

my $grammar = q
  {
 Script: Line(s) Eof {Dancer::CmdScript::Block->new($thisline, @{$item[1]})}

 Eof: /^\Z/

 Line:
     '\}' <commit> <reject>

   | Eof <commit> <reject>

   | '/*' <commit> /([^*]|[*](?=[^\/]))*/ '*/'
       {0}

   | '{' <commit> Line(s?) '}'
       {Dancer::CmdScript::Block->new($thisline, undef, @{$item[3]})}

   | 'if' <commit> Equality 'then' Line <uncommit> 'else' <commit> Line
       {Dancer::CmdScript::If->new($thisline, @item[3,5,8])}

   | 'if' <commit> Equality 'then' Line
       {Dancer::CmdScript::If->new($thisline, @item[3,5])}

   | Statement ';'
       {$item[1]}

   | ';'
       {0}

   | <error>

 Statement:
     '\}' <commit> <reject>

   | ';' <commit> <reject>

   | VariableName /=(?!>)/ <commit> Expression
       {Dancer::CmdScript::Expression::Assign->new($thisline, @item[1,4])}

   | 'readre' <commit> Connection IOData
       {Dancer::CmdScript::Read->new($thisline, $item[3], 1, $item[4])}

   | 'read' <commit> Connection IOData
       {Dancer::CmdScript::Read->new($thisline, $item[3], 0, $item[4])}

   | 'write' <commit> Connection IOData
       {Dancer::CmdScript::Write->new($thisline, $item[3], 0, $item[4])}

   | 'open' <commit> Connection Stereotype ExpressionList(?)
       {Dancer::CmdScript::Open->new($thisline, @item[3,4], @{$item[5]})}

   | 'close' <commit> Connection
       {Dancer::CmdScript::Close->new($thisline, $item[3])}

   | 'listen' <commit> Connection Stereotype ExpressionList(?)
       {Dancer::CmdScript::Listen->new($thisline, @item[3,4], @{$item[5]})}

   | 'accept' <commit> Connection
       {Dancer::CmdScript::Accept->new($thisline, $item[3])}

   | 'eof' <commit> Connection
       {Dancer::CmdScript::EOF->new($thisline, $item[3])}

   | 'use' <commit> ModuleName
       {eval "use $item[3];"; die $@ if $@; return 1;}

   | 'include' <commit> FileName
       {Dancer::CmdScript::Block->new($thisline, $item[3] . ".csi")}

   | 'define' <commit> FunctionName '{' Line(s?) '}'
       {
        Dancer::CmdScript::Function->new($thisline, "define",
                                         [
                                          Dancer::CmdScript::Expression::Value->new($thisline, $item[3]),
                                          Dancer::CmdScript::Block->new($thisline, undef, @{$item[5]})
                                         ]
                                        )
       }

   | Expression

   | <error>

 Expression:
     '\}' <commit> <reject>
   | ';' <commit> <reject>

   | "'" <commit> /(\\\'|[^\'])*/ "'"
       {Dancer::CmdScript::Expression::Value->new($thisline, $item[3])}

   | '"' <commit> /(\\\"|[^\"])*/ '"'
       {Dancer::CmdScript::Expression::Value->new($thisline, $item[3])}

   | Product Add_Op(?)
       {scalar @{$item[2]} ? $item[2]->[0]->($thisline, $item[1]) : $item[1]}

   | '(' Expression ')'
       {$item[2]}

   | <error>

 Add_Op:
     '.' <commit> Product {sub {Dancer::CmdScript::Expression::Concatenate->new(@_, $item[3])}}
   | '+' <commit> Product {sub {Dancer::CmdScript::Expression::Add->new(@_, $item[3])}}
   | '-' <commit> Product {sub {Dancer::CmdScript::Expression::Subtract->new(@_, $item[3])}}

 Product:
     '(' Product ')' {$item[2]}
   | Value Mul_Op(?) {scalar @{$item[2]} ? $item[2]->[0]->($thisline, $item[1]) : $item[1]}

 Mul_Op:
     '*' <commit> Value {sub {Dancer::CmdScript::Expression::Product->new(@_, $item[3])}}
   | '/' <commit> Value {sub {Dancer::CmdScript::Expression::Divide->new(@_, $item[3])}}

 Value:
     Number

   | "'" <commit> /(\\\'|[^\'])*/ "'"
       {Dancer::CmdScript::Expression::Value->new($thisline, $item[3])}

   | '"' <commit> /(\\\"|[^\"])*/ '"'
       {Dancer::CmdScript::Expression::Value->new($thisline, $item[3])}

   | "\$" <commit> VariableName
       {Dancer::CmdScript::Expression::Variable->new($thisline, $item[3])}

   | VariableName '=(?!>)' <commit> Expression
       {Dancer::CmdScript::Expression::Assign->new($thisline, @item[1,4])}

   | '{' <commit> Line(s?) '}'
       {Dancer::CmdScript::Block->new($thisline, undef, @{$item[3]})}

   | '(' Expression ')'
       {$item[2]}

   | /^get\s/ <commit> Connection
       {Dancer::CmdScript::Get->new($thisline, $item[3], 0, undef)}

   | /^call\s/ <commit> FunctionName ExpressionList(?)
       {
         Dancer::CmdScript::Function->new($thisline, "call",
                                          [
                                           Dancer::CmdScript::Expression::Value->new($thisline, $item[3]),
                                           map {@$_} @{$item[4]}
                                          ]
                                         )
       }

   | FunctionName ExpressionList(?)
       {Dancer::CmdScript::Function->new($thisline, $item[1], @{$item[2]})}

 ExpressionList:
     '\}' <commit> <reject>
   | ';' <commit> <reject>
   | /[^;\s]+/ '=>' <commit> ExpressionList(s?)
       {[Dancer::CmdScript::Expression::Value->new($thisline, $item[1]), map {@$_} @{$item[4]}]}
   | Expression ExpressionList(s?)
       {[$item[1], map {@$_} @{$item[2]}]}

 Equality:
     Expression '==' <commit> Expression {Dancer::CmdScript::Expression::Equals->new($thisline, @item[1,4])}
   | Expression '!=' <commit> Expression {Dancer::CmdScript::Expression::NotEquals->new($thisline, @item[1,4])}
   | Expression '<'  <commit> Expression {Dancer::CmdScript::Expression::LessThan->new($thisline, @item[1,4])}
   | Expression '<=' <commit> Expression {Dancer::CmdScript::Expression::LessThanOrEquals->new($thisline, @item[1,4])}
   | Expression '>=' <commit> Expression {Dancer::CmdScript::Expression::GreaterThanOrEquals->new($thisline, @item[1,4])}
   | Expression '>'  <commit> Expression {Dancer::CmdScript::Expression::GreaterThan->new($thisline, @item[1,4])}
   | Expression 'eq' <commit> Expression {Dancer::CmdScript::Expression::StrEquals->new($thisline, @item[1,4])}
   | Expression 'ne' <commit> Expression {Dancer::CmdScript::Expression::StrNotEquals->new($thisline, @item[1,4])}
   | Expression 'lt' <commit> Expression {Dancer::CmdScript::Expression::StrLessThan->new($thisline, @item[1,4])}
   | Expression 'le' <commit> Expression {Dancer::CmdScript::Expression::StrLessThanOrEquals->new($thisline, @item[1,4])}
   | Expression 'ge' <commit> Expression {Dancer::CmdScript::Expression::StrGreaterThanOrEquals->new($thisline, @item[1,4])}
   | Expression 'gt' <commit> Expression {Dancer::CmdScript::Expression::StrGreaterThan->new($thisline, @item[1,4])}
   | Expression
   | <error>

 FunctionName: /[_A-Za-z][_A-Za-z0-9]*/
 ModuleName: /[A-Za-z0-9]+(\:\:[A-Za-z0-9]+)*/ {"Dancer\:\:CmdScript\:\:$item[1]"}
 FileName: /[A-Za-z0-9._\/]+/
 Connection: /[_A-Za-z][_A-Za-z0-9]*/
 Stereotype: /[_A-Za-z][_A-Za-z0-9]*/
 IOData: /(\\\\;|[^;])*/ {Dancer::CmdScript::Expression::Value->new($thisline, $item[1])}
 Number: /\d+/ {Dancer::CmdScript::Expression::Value->new($thisline, $item[1])}
 VariableName: /[_A-Za-z][_A-Za-z0-9]*/

  };

# If a precompiled grammar exists and is older than this script,
# delete it (so that we will regenerate it)
if (-e "$objdir/lib/Dancer/CmdGrammar.pm")
  {
    if ((stat("$objdir/lib/Dancer/CmdGrammar.pm"))[9] < (stat("$srcdir/lib/Dancer/CmdInterpreter.pm"))[9])
      {
        unlink "$objdir/lib/Dancer/CmdGrammar.pm";
      }
  }

# This chunk of magic will precompile the grammar if it has not
# already been precompiled
eval "use Dancer::CmdGrammar;";
if ($@)
  {
    my $olddir = getcwd;
    chdir("$objdir/lib/Dancer");
    Parse::RecDescent->Precompile($grammar, "Dancer::CmdGrammar");
    chdir $olddir;
    eval "use Dancer::CmdGrammar;";
    die $@ if $@;
  }
my $expr_parser = Dancer::CmdGrammar->new() or die "Couldn't create parser";

sub compile_subscript
  {
    my $self = shift;
    my $file = shift;
    open FILE, "<", $file or die "Can't open $file: $!";
    local $/ = undef;
    my $data = <FILE>;
    close FILE;
    my $script;
    my $block = $expr_parser->Script($data) or die "Compilation failed";
    print Dumper($block) if $ENV{CI_DUMP};
    $block->interpolate($self);
    $block = $block->optimise();
    return $block;
  }

sub compile_script
  {
    my $self = shift;
    my $file = shift;
    my $script;
    eval
      {
        my $block = $self->compile_subscript($file);
        $script = new Dancer::CmdScript category => $self->{category}, name => $file, block => $block;
        return $script;
      };
    if ($@)
      {
        # Hash exceptions are errors in the script. Other exceptions are errors in the compiler.
        if (ref $@ eq 'HASH')
          {
            error $@->{msg}, 0, caller => [$self->{category}, $file, $@->{line}];
            return undef;
          }
        else
          {
            die $@;
          }
      }
    else
      {
        return $script;
      }
  }

1;
