# Copyright (C) 2002 Andrew Suffield <asuffield@users.sourceforge.net>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA

package Dancer::CmdPipe;

use strict;
use 5.006;
use warnings;

use Carp;
use Expect;
use DejaPerl ':framework';
use IO::File;
use POSIX;
$Expect::Multiline_Matching = 0;

my $default_timeout = 5;

BEGIN
  {
    use Exporter;
    our @ISA = qw/Exporter/;
    our $VERSION = "0.0.1";
    our @EXPORT = qw//;
  }

sub new($$$$%)
  {
    my $proto = shift;
    my $class = ref($proto) || $proto;
    my $self = {
                read_data => [],
               };

    my $dancer = $self->{dancer} = shift;
    # The name of the dancer installation to work on
    my $dancer_name = $self->{dancer_name} = $dancer->name;
    # The name this object should refer to itself by
    my $pipe_name = $self->{pipe_name} = shift;
    # The name of the module this object will communicate with
    my $module = $self->{module} = shift;

    my %args = @_;
    $self->{no_sanity_checks} = not $args{sanity_checks};

    verbose "Creating Dancer::CmdPipe '$pipe_name' for $dancer_name, module $module";
    verbose "Opening control pipe $pipe_name for $dancer_name", 2;
    my $control_path = "$dancer/var/run/dancerd/$module-control.fifo";
    $self->{control} = new IO::File $control_path, ">" or croak "Failed to open $control_path: $!";
    $self->{control}->autoflush(1);

    verbose "Opening output pipe $pipe_name for $dancer_name", 2;
    my $output_path = "$dancer/var/run/dancerd/$module-output.data";
    $self->{output} = new Expect;
    $self->{output}->raw_pty(1);
    $self->{output}->spawn("$objdir/fasttail", $output_path);
    $self->{output}->log_user($all_flag);
    $self->{output}->log_file($log);

    verbose "Dancer::CmdPipe '$pipe_name' is ready";

    bless ($self, $class);
    return $self;
  }

sub kill()
  {
    my $self = shift;
    return if $self->{dead};
    verbose "Closing pipes for Dancer::CmdPipe '$self->{pipe_name}'";
    close $self->{control} if $self->{control};
    $self->{output}->hard_close() if $self->{output};
    $self->{dead}++;
    return 1;
  }

sub DESTROY
  {
    my $self = shift;
    $self->kill;
  }

sub exp_internal($)
  {
    my $self = shift;
    $self->{internal} = shift;
    $self->{output}->exp_internal($self->{internal});
  }

sub no_sanity_checks($;$)
  {
    my $self = shift;
    if (scalar @_)
      {
        $self->{no_sanity_checks} = shift;
      }
    return $self->{no_sanity_checks};
  }

# Escape anything that needs to be escaped for transport over a cmdpipe
sub _escape($ )
  {
    my $str = shift;
    if ($str =~ /[\0-\31]/)
      {
        $str =~ s/(.)/'\\' . ord($1)/egs;
      }
    else
      {
        $str =~ s/\n/\\n/g;
        $str =~ s/\r/\\r/g;
      }
    return $str;
  }

my $sanity_index = 0;

use Data::Dumper;
sub fetch_data
  {
    my $self = shift;
    my $timeout = shift;
    my $caller = shift;
    my $ret = undef;
    my $data = "";
    while (not defined $ret)
      {
        my (undef, $err) =
          expect($timeout,
                 -i => $self->{dancer}->spawned(),
                 -i => $self->{output},
                 -re => qr/^\*\*DONE\*\*\n/, sub {$ret = 1},
                 -re => qr/^\*\*ERROR\*\*\n/, sub {$ret = 0},
                 -re => qr/^\(segfault\)\n/, sub {error "Simulated segfault from pipe", caller => $caller; $ret = 0},
                 -re => qr/^(.*)\n/, sub {my $exp = shift; $data .= $exp->match()}
                 );
        if ($err)
          {
            if ($err =~ /TIMEOUT/)
              {
                # No data read
                return 0;
              }
            elsif ($err =~ /EOF/)
              {
                error "Unexpected EOF from output pipe", caller => $caller;
                return undef;
              }
            elsif ($err =~ s/^3://)
              {
                unless ($self->{dancer}->spawned()->pid)
                  {
                    error "Dancer has died: $err", caller => $caller;
                    $self->{dancer}->kill();
                  }
                error "fasttail child has died: $err", caller => $caller unless $self->{output}->pid;
              }
            else
              {
                error "Unknown error: $err (\$! eq $!)", caller => $caller;
                return undef;
              }
          }
      }
    if (length $data)
      {
        $data =~ s/\r?\n?$//;
        push @{$self->{read_data}}, $data;
      }
    return 1;
  }

sub require_data
  {
    my $self = shift;
    my $timeout = shift;
    my $caller = shift;

    unless (defined $self->fetch_data($timeout, $caller))
      {
        fail "Error when reading from pipe", $caller;
        return undef;
      }
    unless (scalar @{$self->{read_data}})
      {
        # Still no data waiting? Fail.
        fail "No data to read", $caller;
        return undef;
      }
    return shift @{$self->{read_data}};
  }

sub _check_sanity($$;$)
  {
    my $self = shift;
    my $timeout = shift;
    my $caller = shift || [caller];

    return if $self->{no_sanity_checks};

    $sanity_index++;

    my $control = $self->{control};
    print $control "SANITY\n";

    my $data = $self->require_data($timeout, $caller) or return;

    if ($data eq 'OK')
      {
        pass "sanity check #$sanity_index";
      }
    elsif ($data eq 'ERROR')
      {
        fail "sanity check #$sanity_index", $caller;
      }
    return 1;
  }

# Send a command to the test driver and catch the standard ack
sub do($$$;$$)
  {
    my $self = shift;
    my $command = shift;
    my $description = shift;
    my $timeout = shift || $default_timeout;
    my $caller = shift || [caller];

    my $control = $self->{control};
    print "Sending $command\n" if $self->{internal};
    print $control _escape("$command") . "\n";
    verbose $command if $all_flag;
    # Trim the length in the logs, some of those commands are huge
    if (length $command > 128)
      {
        $command = substr($command, 0, 128) . "...";
      }
    if (my $rc = $self->fetch_data($timeout, $caller))
      {
        pass "Got response to '$command'";
      }
    elsif (defined $rc)
      {
        fail "No response to '$command' from pipe", $caller;
        return undef;
      }
    else
      {
        fail "Error when reading from pipe", $caller;
        return undef;
      }

    $self->_check_sanity($timeout, $caller);
    return 1;
  }

# Send a command to the test driver, check it returns the appropriate value, and catch the ack
sub check_cmd_value($$$$;$)
  {
    my $self = shift;
    my $command = shift;
    my $value = shift;
    my $description = shift;
    my $timeout = shift || $default_timeout;
    my $caller = shift || [caller];

    my $control = $self->{control};
    print "Sending $command\n" if $self->{internal};
    print $control _escape("$command") . "\n";
    verbose $command if $all_flag;

    my $data = $self->require_data($timeout, $caller) or return;

    if ($data eq $value)
      {
        pass "got value for '$description'"
      }
    else
      {
        fail "unexpected value '$data' for '$description' (expected '$value')", $caller;
        return undef;
      }

    $self->_check_sanity($timeout, $caller);
    return 1;
  }

# Check for an arbitrary string from the test driver
sub check_value($$$$;$)
  {
    my $self = shift;
    my $value = shift;
    my $description = shift;
    my $timeout = shift || $default_timeout;
    my $caller = shift || [caller];

    my $data = $self->require_data($timeout, $caller) or return;

    if ($data eq $value)
      {
        pass "got value for '$description'"
      }
    else
      {
        fail "unexpected value '$data' for '$description' (expected '$value')", $caller;
        return undef;
      }

    $self->_check_sanity($timeout, $caller);
    return 1;
  }

# Generally speaking, do not use this function. It is only for the
# occasions where a) this is not used on the code under test, and b)
# there is no realistic manner to predict what the result will be.
sub get_cmd_value($$$;$)
  {
    my $self = shift;
    my $command = shift;
    my $description = shift;
    my $timeout = shift || $default_timeout;
    my $caller = shift || [caller];

    my $value;
    my $bail = 0;
    my $control = $self->{control};
    print "Sending $command\n" if $self->{internal};
    print $control _escape("$command") . "\n";
    verbose $command if $all_flag;

    my $data = $self->require_data($timeout, $caller) or return;

    $self->_check_sanity($timeout, $caller);
    return $data;
  }

# See comments for get_cmd_value
sub get_value($$$;$)
  {
    my $self = shift;
    my $description = shift;
    my $timeout = shift || $default_timeout;
    my $caller = shift || [caller];

    my $data = $self->require_data($timeout, $caller) or return;

    $self->_check_sanity($timeout, $caller);
    return $data;
  }

sub gdb_foo
  {
    $default_timeout = undef;
  }

1;
