# Copyright (C) 2002 Andrew Suffield <asuffield@users.sourceforge.net>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA

package Dancer::CmdScript::Block;

use strict;
use 5.006;
use warnings;

use Dancer::CmdScript::Expression::Atom;

BEGIN
  {
    use Exporter;
    our @ISA = qw/Exporter/;
    our $VERSION = "0.0.1";
    our @EXPORT = qw//;
    our @EXPORT_OK = qw//;
  }

sub new
  {
    my $class = shift;
    my $line = shift;
    my $file = shift;

    my $self = {
                # Number of the line this block was defined on
                line => $line,
                file => $file,
                # Lines this block contains
                lines => [@_],
               };
    bless $self, $class;
    return $self;
  }

sub add_line
  {
    my $self = shift;
    push @{$self->{lines}}, scalar @_ ? [@_] : undef;
  }

sub lines
  {
    my $self = shift;
    return scalar @{$self->{lines}};
  }

sub interpolate
  {
    my $self = shift;
    my $interp = shift;

    if ($self->{file})
      {
        # This block is external; it's a subscript. Recurse back to
        # the interpreter to compile it.

        unless (-r $self->{file})
          {
            foreach my $x (@INC)
              {
                $x .= "/" . $self->{file};
                if (-r $x)
                  {
                    $self->{file} = $x;
                    last;
                  }
              }
          }

        my $block = $interp->compile_subscript($self->{file});
        $self->{lines} = [$block];
      }

    $_->interpolate($interp) foreach grep {ref $_} @{$self->{lines}};
  }

sub optimise
  {
    my $self = shift;
    for (my $i = 0; $i < scalar @{$self->{lines}}; $i++)
      {
        next unless ref $self->{lines}->[$i];
        $self->{lines}->[$i] = $self->{lines}->[$i]->optimise();
      }
    #######################
    # Local optimisations #
    #######################
    my @lines = ();
    # Discard all uninteresting lines (any constant lines before the end are not interesting)
    for (my $i = 0; $i < (scalar @{$self->{lines}} - 1); $i++)
      {
        next unless $self->{lines}->[$i];
        next if $self->{lines}->[$i]->constant();
        push @lines, $self->{lines}->[$i];
      }
    push @lines, $self->{lines}->[-1];

    # If there is only one line, return that
    # We must build the stack frame, so we can't do this
    # return $lines[0] if scalar @lines == 1 unless $self->{file};

    # If there are no lines, return true
    return Dancer::CmdScript::Expression::Value->new($self->{line}, 1) unless scalar @lines;

    # Otherwise return our optimised form
    $self->{lines} = \@lines;
    return $self;
  }

sub constant
  {
    my $self = shift;
    # An empty block is constant and returns true
    return 1 unless scalar @{$self->{lines}};
    # A block is not constant if it contains more than one line
    return 0 if scalar @{$self->{lines}} > 1;
    # Otherwise it is as constant as its last line
    return $self->{lines}->[0]->constant();
  }

sub eval
  {
    my $self = shift;
    my $script = shift;
    my $frame = shift || $self->{file};
    my $line = shift || $self->{line};
    my $last_value = 1;
    if ($frame)
      {
        push @{$script->stack()}, [undef, $frame, $line];
      }
    for (my $i = 0; $i < scalar @{$self->{lines}}; $i++)
      {
        $last_value = $self->{lines}->[$i]->eval($script);
      }
    if ($frame)
      {
        pop @{$script->stack()};
      }
    return $last_value;
  }

1;
