# Copyright (C) 2002 Andrew Suffield <asuffield@users.sourceforge.net>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA

package Dancer::CmdScript::Stereotype::CmdPipe;

use strict;
use 5.006;
use warnings;

use Carp;
use Dancer::CmdInterpreter;
use DejaPerl qw/:framework/;
use Expect;
$Expect::Multiline_Matching = 0;

BEGIN
  {
    use Exporter;
    our @ISA = qw/Exporter/;
    our $VERSION = "0.0.1";
    our @EXPORT = qw//;
    our @EXPORT_OK = qw//;
  }

my %parsers = ();

sub register_parser
  {
    my $name = shift;
    my $read_sub = shift;
    my $write_sub = shift;
    $parsers{$name} = {read => $read_sub, write => $write_sub};
  }

sub simple_parser
  {
    my $expr = shift;
    my $regexp = shift;
    return {type => 'matcher', matcher => sub{0}} unless defined $expr;
    return {type => 'matcher', matcher => $regexp
            ? sub {return "/^$expr$/" unless defined $_[0]; $_[0] =~ /$expr/}
            : sub {return $expr     unless defined $_[0]; $_[0] eq $expr}};
  }

sub open
  {
    my $script = shift;
    die "Not enough parameters" if scalar @_ < 3;
    my $name = shift;
    my $spawn_name = $script->prefix() . shift;
    my $module_name = shift;
    my $spawn = $script->spawns()->{$spawn_name} or die "Spawn $spawn_name does not exist";
    my $sanity_checks = 0;
    my $parser = {read => \&simple_parser, write => sub {$_[0]}};

    while (my $arg = shift)
      {
        if ($arg eq 'sanity_checks')
          {
            $sanity_checks = 1;
          }
        if ($arg =~ /^parse:(.*)$/)
          {
            die "Unknown parser $parser" unless exists $parsers{$1};
            $parser = $parsers{$1};
          }
        else
          {
            die "Unrecognised argument '$arg'";
          }
      }

    my $path = $spawn->path();

    verbose "Opening control pipe for $spawn_name/$module_name", 2;
    my $control_path = $path . "/var/run/dancerd/${module_name}-control.fifo";
    my $control = new IO::File $control_path, ">" or croak "Failed to open $control_path: $!";
    $control->autoflush(1);

    verbose "Opening output pipe for $spawn_name/$module_name", 2;
    my $output_path = $path . "/var/run/dancerd/${module_name}-output.data";
    my $output = new Expect;
    $output->raw_pty(1);
    $output->spawn("$objdir/fasttail", $output_path);
    $output->log_user($all_flag);
    $output->log_file($log);

    verbose "Command pipe '$name' is ready";

    my $self = {
                name => $name,
                script => $script,
                spawn => $spawn,
                control => $control,
                output => $output,
                parser => $parser,
                default_timeout => 5,
                sanity_checks => $sanity_checks,
                read_data => [],
               };
    bless $self;
    return $self;
  }

sub close
  {
    my $self = shift;
    return unless $self->{script};
    verbose "Closing pipes for command pipe '" . $self->{name} . "'";
    close $self->{control} if $self->{control};
    $self->{output}->hard_close() if $self->{output};
    $self->{dead}++;
    # Make sure we don't get any circular references
    %{$self} = ();
  }

sub trace
  {
    my $script = shift;
    my $name = shift or die "No connection name given";
    $name = $name->eval($script);
    my $value = shift;
    $value = defined $value ? $value->eval($script) : 1;
    my $connection;
    die "Installation $name does not exist" unless $connection = $script->connections()->{$name};
    $connection->{trace} = $value;
    $connection->{output}->exp_internal($value);
  }

sub DESTROY
  {
    my $self = shift;
    $self->close();
  }

sub escape($ )
  {
    my $str = shift;
    $str =~ s/\n/\\n/g;
    $str =~ s/\r/\\r/g;
    return $str;
  }

sub pretty($ )
  {
    my $str = shift;
    $str =~ s/\n/\\n/g;
    $str =~ s/\r/\\r/g;
    return $str;
  }

sub gdb
  {
    my $self = shift;
    $self->{default_timeout} = undef;
  }

sub fetch_data
  {
    my $self = shift;
    my $line = shift;
    my $ret = undef;
    my $data = "";
    while (not defined $ret)
      {
        my (undef, $err) =
          expect($self->{default_timeout},
                 -i => $self->{spawn}->spawned(),
                 -i => $self->{output},
                 -re => qr/^\*\*DONE\*\*\n/, sub {$ret = 1},
                 -re => qr/^\*\*ERROR\*\*\n/, sub {$ret = -1},
                 -re => qr/^\(segfault\)\n/, sub {error "Simulated segfault from pipe", caller =>
                                                  [$self->{script}->category(), $self->{script}->name(), $line];
                                                $ret = undef},
                 -re => qr/^(.*)\n/, sub {my $exp = shift; $data .= $exp->match()}
                 );
        if ($err)
          {
            if ($err =~ /TIMEOUT/)
              {
                # No data read
                return 0;
              }
            elsif ($err =~ /EOF/)
              {
                error "Unexpected EOF from output pipe", caller => [$self->{script}->category(), $self->{script}->name(), $line];
                return undef;
              }
            elsif ($err =~ s/^3://)
              {
                unless ($self->{spawn}->spawned()->pid)
                  {
                    error "Dancer has died: $err",
                      caller => [$self->{script}->category(), $self->{script}->name(), $line];
                    $self->{spawn}->kill();
                  }
                unless ($self->{output}->pid)
                  {
                    error "fasttail child has died: $err",
                      caller => [$self->{script}->category(), $self->{script}->name(), $line];
                  }
              }
            else
              {
                error "Unknown error: $err (\$! eq $!)", caller => [$self->{script}->category(), $self->{script}->name(), $line];
                return undef;
              }
          }
      }
    if (length $data)
      {
        $data =~ s/\r?\n?$//;
        push @{$self->{read_data}}, $data;
      }
    return $ret;
  }

sub _read_query
  {
    my $self = shift;
    my $line = shift;
    my $query = shift;

    print "Sending $query\n" if $self->{trace};
    my $control = $self->{control};
    print $control escape($query) . "\n";
    return 1;
  }

sub _read_match
  {
    my $self = shift;
    my $line = shift;
    my $matcher = shift;

    while (scalar @{$self->{read_data}} == 0)
      {
        # No data waiting. Try to read some.
        my $rc = $self->fetch_data($line);
        if (defined $rc)
          {
            if ($rc == -1)
              {
                fail "Error from daemon while trying to read '" . pretty($matcher->()) . "'",
                  [$self->{script}->category(), $self->{script}->name(), $line];
                return;
              }
            elsif ($rc != 1)
              {
                fail "No response from pipe while trying to read '" . pretty($matcher->()) . "'",
                  [$self->{script}->category(), $self->{script}->name(), $line];
                return;
              }
          }
        else
          {
            fail "Error when reading from pipe", [$self->{script}->category(), $self->{script}->name(), $line];
            return
          }
      }
    unless (scalar @{$self->{read_data}})
      {
        # Still no data waiting? Fail.
        fail "No data to read", [$self->{script}->category(), $self->{script}->name(), $line];
        return;
      }
    my $data = shift @{$self->{read_data}};
    if ($matcher->($data))
      {
        pass "Got '" . pretty($data) . "'";
      }
    else
      {
        fail "Expecting '" . pretty($matcher->()) . "' but got '" . $data . "'",
          [$self->{script}->category(), $self->{script}->name(), $line];
      }
  }

sub read
  {
    my $self = shift;
    my $line = shift;
    my $regexp = shift;
    die "Not enough parameters" if scalar @_ < 1;
    my $expr = shift;
    my @sequence = $self->{parser}{read}->($expr, $regexp);
    foreach my $event (@sequence)
      {
        die "Undefined event" unless defined $event;
        if (exists $event->{condition})
          {
            next unless $event->{condition}->();
          }

        if ($event->{type} eq 'query')
          {
            $self->_read_query($line, $event->{query}) or return;
          }
        elsif ($event->{type} eq 'matcher')
          {
            $self->_read_match($line, $event->{matcher}) or return;
          }
        else
          {
            die "Unknown event type in sequence given by parser";
          }
      }
  }

sub get
  {
    my $self = shift;
    my $line = shift;
    unless (scalar @{$self->{read_data}})
      {
        # No data waiting. Try to read some.
        unless (defined $self->fetch_data($line))
          {
            fail "Error when reading from pipe", [$self->{script}->category(), $self->{script}->name(), $line];
            return;
          }
      }
    unless (scalar @{$self->{read_data}})
      {
        # Still no data waiting? Fail.
        fail "No data to read", [$self->{script}->category(), $self->{script}->name(), $line];
        return;
      }
    return shift @{$self->{read_data}};
  }

sub write
  {
    my $self = shift;
    my $line = shift;
    die "Not enough parameters" if scalar @_ < 1;
    my $expr = shift;
    my $data = $self->{parser}{write}->($expr);
    print "Sending $data\n" if $self->{trace};
    my $control = $self->{control};
    print $control escape($data) . "\n";
    my $rc;
    if (defined ($rc = $self->fetch_data($line)) and $rc == 1)
      {
        pass "Got response to '$data'";
      }
    elsif (defined $rc)
      {
        if ($rc == -1)
          {
            fail "Error from daemon in response to '$data'", [$self->{script}->category(), $self->{script}->name(), $line];
          }
        else
          {
            fail "No response to '$data' from pipe", [$self->{script}->category(), $self->{script}->name(), $line];
          }
      }
    else
      {
        fail "Error when reading from pipe", [$self->{script}->category(), $self->{script}->name(), $line];
      }
  }

Dancer::CmdInterpreter::register_stereotype('cmdpipe', \&open, undef);
Dancer::CmdInterpreter::register_function('trace', \&trace);

1;
