# Copyright (C) 2002 Andrew Suffield <asuffield@users.sourceforge.net>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA

package Dancer::CmdScript::Stereotype::CmdPipe::Message;

use strict;
use 5.006;
use warnings;

use Carp;
use Dancer::CmdInterpreter;
use Dancer::CmdScript::Stereotype::CmdPipe;
use DejaPerl qw/:framework/;
use Expect;
$Expect::Multiline_Matching = 0;

BEGIN
  {
    use Exporter;
    our @ISA = qw/Exporter/;
    our $VERSION = "0.0.1";
    our @EXPORT = qw//;
    our @EXPORT_OK = qw//;
  }

my $messages_pending = 0;

sub parse_read
  {
    my $expr = shift;
    my $regexp = shift;
    if ($expr =~ s/^raw:// or $expr !~ s/^message://)
      {
        if ($regexp)
          {
            return {type => 'matcher', matcher => sub {my $data = shift; return "/$expr/" unless defined $data; return $data =~ /$expr/o;}};
          }
        else
          {
            return {type => 'matcher', matcher => sub {my $data = shift; return $expr unless defined $data; return $data eq $expr;}};
          }
      }
    $expr =~ s/^\s*//;
    $expr =~ s/\s*$//;
    my @fields = split /\s*(\S+?):\s*/, $expr;
    # That split will insert a blank at the start, every time
    shift @fields;
    die "Invalid message string: incorrect number of fields" unless (scalar @fields) % 2 == 0;
    my %fields = @fields;
    die "No message type specified" unless exists $fields{type};
    $fields{origin} ||= "(null)";
    $fields{dest} ||= "(null)";
    $_ =~ s/\n$// foreach @fields{qw/type origin dest/};

    foreach (qw/str uint32 sint32/)
      {
        $fields{$_} = [defined $fields{$_} ? split(/\\,\s*/, $fields{$_}) : ()];
        $_ =~ s/\n$// foreach @{$fields{$_}};
        # Escaped spaces disappear (hack)
        $_ =~ s/\\ //g foreach @{$fields{$_}};
        # All other escapes get interpolated as-is
        $_ =~ s/\\(.)/$1/g foreach @{$fields{$_}};
      }

    # Tidy up $expr for output
    $expr =~ s/ +/ /g;

    return ({type => 'matcher',
             condition => sub
             {
               if ($messages_pending > 0)
                 {
                   $messages_pending--;
                   pass "Got 'SEND_MESSAGE'";
                   return 0;
                 }
               else
                 {
                   return 1;
                 }
             },
             matcher => sub
             {
               my $data = shift;
               return "SEND_MESSAGE" unless defined $data;
               $messages_pending++ while $data =~ s/^\n?SEND_MESSAGE//g;
               return $messages_pending-- > 0;
             }},
            {type => 'query', query => "SHIFT_MSG"},

            $regexp ?
            {type => 'matcher', matcher => sub
             {
               my $data = shift;
               return "/$expr/" unless defined $data;
               return 0 unless $data =~ s/^type:$fields{type} origin:$fields{origin} dest:$fields{dest}//;
               foreach my $type (qw/str uint32 sint32/)
                 {
                   return 0 unless $data =~ s/^\n$type:(\d+) //;
                   return 0 unless $1 == scalar @{$fields{$type}};
                   my $s = join ',', @{$fields{$type}};
                   return 0 unless $data =~ s/$s\n.//;
                 }
               return 1;
             }}
            :
            {type => 'matcher', matcher => sub
               {
                 my $data = shift;
                 return $expr unless defined $data;
                 return 0 unless $data =~ s/^type:\Q$fields{type}\E origin:\Q$fields{origin}\E dest:\Q$fields{dest}\E//;
                 foreach my $type (qw/str uint32 sint32/)
                   {
                     return 0 unless $data =~ s/^\n$type:(\d+) //;
                     return 0 unless $1 == scalar @{$fields{$type}};
                     my $s = join ',', @{$fields{$type}};
                     return 0 unless $data =~ s/\Q$s\E\n.//;
                   }
                 return 1;
               }
             }
           );
  }

sub parse_write
  {
    my $input = shift;
    if ($input =~ s/^raw:// or $input !~ s/^message://)
      {
        return $input;
      }
    $input =~ s/^\s*//;
    $input =~ s/\s*$//;
    my @fields = split /\s*(\S*?):\s*/, $input;
    # That split will insert a blank at the start, every time
    shift @fields;
    my %fields = @fields;
    die "No message type specified" unless exists $fields{type};
    die "No connection specified" unless exists $fields{read} or exists $fields{send};
    foreach (@fields{qw/read send type origin dest/})
      {
        next unless defined $_;
        $_ =~ s/\n$//;
      }
    foreach (qw/str uint32 sint32/)
      {
        $fields{$_} = [$fields{$_} ? split(',\s*', $fields{$_}) : ()];
        $_ =~ s/\n$// foreach @{$fields{$_}};
      }
    my @str;
    push @str, '"' . $fields{type} . '"';
    push @str, ':';
    push @str, '<' . $fields{origin} if defined $fields{origin};
    push @str, '>' . $fields{dest} if defined $fields{dest};
    push @str, map {defined $_ ? "\"$_\"" : "(null)"} @{$fields{str}};
    push @str, @{$fields{uint32}};
    push @str, map {sprintf "%+d", $_} @{$fields{sint32}};
    if (defined $fields{send})
      {
        return "SEND_MESSAGE $fields{send} " . join(' ', @str);
      }
    if (defined $fields{read})
      {
        return "READ_MESSAGE $fields{read} " . join(' ', @str);
      }
  }

Dancer::CmdScript::Stereotype::CmdPipe::register_parser("message", \&parse_read, \&parse_write);
