# Copyright (C) 2002 Andrew Suffield <asuffield@users.sourceforge.net>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA

package Dancer::CmdScript::Stereotype::Log;

use strict;
use 5.006;
use warnings;

use Carp;
use Dancer::CmdInterpreter;
use DejaPerl qw/:framework/;

BEGIN
  {
    use Exporter;
    our @ISA = qw/Exporter/;
    our $VERSION = "0.0.1";
    our @EXPORT = qw//;
    our @EXPORT_OK = qw//;
  }

sub open
  {
    my $script = shift;
    die "Not enough parameters" if scalar @_ < 2;
    my $name = shift;
    my $arg = shift;
    my $spawn_name = $script->prefix() . $arg;
    my $spawn = $script->spawns()->{$spawn_name} or die "Spawn $spawn_name does not exist";
    my $self = {
                name => $name,
                script => $script,
                spawn => $spawn,
               };
    bless $self;
    return $self;
  }

sub close
  {
  }

sub DESTROY
  {
    my $self = shift;
    # Make sure we don't get any circular references
    %{$self} = ();
  }

sub parse_read
  {
    my $expr = shift;
    my $regexp = shift;
    if ($expr =~ /^\s*EOF\s*$/)
      {
        return {eof => 1};
      }
    $expr =~ /^\s*(.*?) \[(.*)\]: (.*?)\s*$/ or die "Unknown log message format: $expr";
    my ($facility, $level, $msg) = ($1, $2, $3);
    if ($regexp)
      {
        return {facility => qr/$facility/, level => $level, msg => qr/$msg/};
      }
    else
      {
        return {facility => $facility, level => $level, msg => $msg};
      }
  }

my $x = 0;
my %log_levels = map {$_ => $x++} qw/debug trace info warning error critical/;

sub read
  {
    my $self = shift;
    my $line_number = shift;
    my $regexp = shift;
    my $expr = shift;
    $expr = parse_read($expr, $regexp);
    my $line;
    my $log_level = $self->{script}->variables()->{log_level} || 'info';
    LOOP:
    {
      do
        {
          $line = $self->{spawn}->log();
          last if not defined $line or $line->{eof};
          # If the message is for the core, take it from error_log as well
          if ($line->{facility} eq 'core')
            {
              my $error_line = $self->{spawn}->error_log();
              if ($line->{level} ne $error_line->{level}
                  or $line->{facility} ne $error_line->{facility}
                  or $line->{msg} ne $error_line->{msg})
                {
                  error "Differing messages from basic.log and error.log";
                  note "Got '$line->{facility} [$line->{level}]: $line->{msg}' from basic.log";
                  note "Got '$error_line->{facility} [$error_line->{level}]: $error_line->{msg}' from error.log";
                }
            }
        } while ($line->{level} and $log_levels{$line->{level}} < $log_levels{$log_level});
    }
    if ($self->{eof})
      {
        if ($expr->{eof})
          {
            pass "Received EOF from log";
            return 1;
          }
        else
          {
            unresolved "Log EOF has already been reached";
            return undef;
          }
      }
    elsif (not defined $line)
      {
        fail "Timeout while waiting for '$expr->{facility} [$expr->{level}]: $expr->{msg}' from log";
        return 0;
      }
    elsif ($line->{eof})
      {
        if ($expr->{eof})
          {
            pass "Received EOF from log";
            return 1;
          }
        else
          {
            fail "Unexpected EOF from log",
               [$self->{script}->category(), $self->{script}->name(), $line];
            return 0;
          }
        $self->{eof} = 1;
      }
    elsif ($regexp)
      {
        if ($line->{facility} =~ /$expr->{facility}/
            and $line->{level} eq $expr->{level}
            and $line->{msg} =~ /$expr->{msg}/)
          {
            pass "Got log message '$line->{facility} [$line->{level}]: $line->{msg}'";
            return 1;
          }
        else
          {
            fail "Didn't get log message '$expr->{facility} [$expr->{level}]: $expr->{msg}'",
               [$self->{script}->category(), $self->{script}->name(), $line];
            note "Got message '$line->{facility} [$line->{level}]: $line->{msg}' instead";
            return 0;
          }
      }
    else
      {
        if ($line->{facility} eq $expr->{facility}
            and $line->{level} eq $expr->{level}
            and $line->{msg} eq $expr->{msg})
          {
            pass "Got log message '$line->{facility} [$line->{level}]: $line->{msg}'";
            return 1;
          }
        else
          {
            fail "Didn't get log message '$expr->{facility} [$expr->{level}]: $expr->{msg}'",
              [$self->{script}->category(), $self->{script}->name(), $line];
            note "Got message '$line->{facility} [$line->{level}]: $line->{msg}' instead";
            return 0;
          }
      }
  }

sub write
  {
    my $self = shift;
    die "Cannot write to a read-only log connection";
  }

sub gdb
  {
  }

Dancer::CmdInterpreter::register_stereotype('log', \&open, undef);

1;
