# Copyright (C) 2002 Andrew Suffield <asuffield@users.sourceforge.net>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA

package Dancer::CmdScript::Stereotype::Socket;

use strict;
use 5.006;
use warnings;

use Carp;
use Dancer::CmdInterpreter;
use DejaPerl qw/:framework/;
use IO::Socket::INET;
use Expect;
$Expect::Multiline_Matching = 0;

BEGIN
  {
    use Exporter;
    our @ISA = qw/Exporter/;
    our $VERSION = "0.0.1";
    our @EXPORT = qw//;
    our @EXPORT_OK = qw//;
  }

sub open
  {
    my $script = shift;
    die "Not enough parameters" if scalar @_ < 2;
    my $name = shift;
    my $host = shift;
    my $port = shift;

    my $socket = new IO::Socket::INET PeerAddr => $host, PeerPort => $port, ReuseAddr => 1, Type => SOCK_STREAM
      or die "Failed to open socket to $host:$port: $!";
    my $exp = init Expect $socket;

    my $self = {
                name => $name,
                script => $script,
                socket => $socket,
                exp => $exp,
                default_timeout => 5,
               };
    bless $self;
    return $self;
  }

sub listen
  {
    my $script = shift;
    die "Not enough parameters" if scalar @_ < 2;
    my $name = shift;
    my $host = shift;
    my $port = shift;

    my $socket = new IO::Socket::INET Listen => 1, LocalAddr => $host, LocalPort => $port,
      ReuseAddr => 1, Type => SOCK_STREAM
      or die "Failed to listen on $host:$port: $!";

    my $self = {
                name => $name,
                script => $script,
                socket => $socket,
                host => $host,
                port => $port,
                listen => 1,
                default_timeout => 5,
               };
    bless $self;
    return $self;
  }

sub close
  {
    my $self = shift;
    return unless $self->{script};
    $self->{exp}->hard_close() if $self->{exp};
    $self->{socket}->close() if $self->{socket};
    # Make sure we don't get any circular references
    %{$self} = ();
  }

sub accept
  {
    my $self = shift;
    die "Not a listening socket" unless $self->{listen};
    $self->{socket} = $self->{socket}->accept()
      or die "Couldn't accept connection on $self->{host}:$self->{port}: $!";
    $self->{exp} = init Expect $self->{socket};
    delete $self->{listen};
    return 1;
  }

sub DESTROY
  {
    my $self = shift;
    $self->close();
  }

sub gdb
  {
    my $self = shift;
    $self->{default_timeout} = undef;
  }

sub pretty($ )
  {
    my $str = shift;
    return unless defined $str;
    $str =~ s/\n/\\n/g;
    $str =~ s/\r/\\r/g;
    return $str;
  }

sub eof
  {
    my $self = shift;
    my $line = shift;
    die "Not valid on a listening socket" if $self->{listen};
    if ($self->{eof})
      {
        pass "Got EOF from socket";
      }
    else
      {
        my (undef, $err) = $self->{exp}->expect($self->{default_timeout},
                                                -re => qr/./);
        if ($err)
          {
            if ($err =~ /TIMEOUT/)
              {
                # No data read
                fail "Didn't get EOF from socket", [$self->{script}->category(), $self->{script}->name(), $line];
              }
            elsif ($err =~ /EOF/)
              {
                pass "Got EOF from socket";
                $self->{exp}->hard_close();
                $self->{eof}++;
                return;
              }
            elsif ($err =~ s/^3://)
              {
                die "Child has died (which is odd, since there is no child): $err" unless $self->{spawn}->pid;
              }
            else
              {
                die "Unknown error: $err (\$! eq $!)";
              }
          }
        else
          {
            fail "Got data instead of EOF from socket", [$self->{script}->category(), $self->{script}->name(), $line];
          }
      }
  }

sub read
  {
    my $self = shift;
    my $line = shift;
    die "Not valid on a listening socket" if $self->{listen};
    my $regexp = shift;
    die "Not enough parameters" if scalar @_ < 1;
    my $expr = shift;

    if ($self->{eof})
      {
        fail "Already had EOF from socket", [$self->{script}->category(), $self->{script}->name(), $line];
        note "Was expecting '$expr'";
        return;
      }

    my $err;
    my $passed = 0;
    if ($regexp)
      {
        (undef, $err) = $self->{exp}->expect($self->{default_timeout},
                                             -re => qr/$expr/, sub
                                             {
                                               my $exp = shift;
                                               pass "Got '" . pretty($exp->match()) . "'";
                                               $passed++;
                                             });
      }
    else
      {
        (undef, $err) = $self->{exp}->expect($self->{default_timeout},
                                             -re => qr/^\Q$expr\E/, sub
                                             {
                                               my $exp = shift;
                                               pass "Got '" . pretty($exp->match()) . "'";
                                               $passed++;
                                             });
      }
    if ($err)
      {
        if ($err =~ /TIMEOUT/)
          {
            # No data read
            fail "No data read";
          }
        elsif ($err =~ /EOF/)
          {
            note "Was expecting '$expr'";
            $self->{exp}->hard_close();
            $self->{eof}++;
            die "EOF from socket, aborting";
          }
        elsif ($err =~ s/^3://)
          {
            die "Child has died (which is odd, since there is no child): $err" unless $self->{spawn}->pid;
          }
        else
          {
            die "Unknown error: $err (\$! eq $!)";
          }
      }
    unless ($passed)
      {
        my $match = $self->{exp}->match();
        fail "Didn't get '$expr'", [$self->{script}->category(), $self->{script}->name(), $line];
        note "Got '" . pretty($match) . "' instead";
      }
  }

sub write
  {
    my $self = shift;
    my $line = shift;
    die "Not valid on a listening socket" if $self->{listen};
    die "Not enough parameters" if scalar @_ < 1;
    my $expr = shift;
    my $exp = $self->{exp};
    print $exp "$expr";
  }

Dancer::CmdInterpreter::register_stereotype('socket', \&open, \&listen);

1;
