# Copyright (C) 2002 Andrew Suffield <asuffield@users.sourceforge.net>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA

package Dancer::Spawn;

use strict;
use 5.006;
use warnings;

use Carp qw/cluck croak confess/;
use Cwd;
use DejaPerl qw/:framework/;
use File::Path;
use File::Temp qw/tempfile/;
use Expect;
use Dancer::CmdPipe;
use overload '""' => \&stringify;
$Expect::Multiline_Matching = 0;

BEGIN
  {
    use Exporter;
    our @ISA = qw/Exporter/;
    our $VERSION = "0.0.1";
    our @EXPORT = qw//;
    our @EXPORT_OK = qw//;
  }

my $work_dir = "$objdir/test-work";
my $install_tmp_dir = "$work_dir/install-tmp";
if (-e $install_tmp_dir and not $ENV{DANCER_SKIP_INSTALL})
  {
    verbose "Removing old tree $install_tmp_dir...";
    rmtree($install_tmp_dir, 0, 1);
    die "Couldn't remove tree $install_tmp_dir" if -e $install_tmp_dir;
  }

sub path
  {
    my $self = shift;
    return $self->{path} . $::prefix;
  }

sub name
  {
    my $self = shift;
    return $self->{name};
  }

sub stringify
  {
    my $self = shift;
    my ($a, $b) = @_;
    $a ||= "";
    $b ||= "";
    return $a . $self->path() . $b;
  }

sub new
  {
    my $class = shift;
    my $name = shift;
    my $self = {
                name => $name,
                path => "$work_dir/dancer-$name",
                traced => 0,
                installed => 0,
                spawned => undef,
                log => undef,
                connections => [],
                default_timeout => 5,
                pipes => {},
                on_kill => [],
                log_magic => [],
                error_log_magic => [],
               };
    bless $self, $class;
    return $self;
  }

sub _check_expect($$$;$ )
  {
    my $exp = shift;
    my $re = shift;
    my $description = shift;
    my $timeout = shift || 5;
    my $got = 0;
    my (undef, $err, undef, undef, undef) =
      $exp->expect($timeout, -re => $re, sub {pass "got $description"; $got = 1});
    if ($err)
      {
        if ($err =~ /TIMEOUT/)
          {
            fail "no $description (timeout)";
          }
        elsif ($err =~ /EOF/)
          {
            fail "no $description (eof)";
          }
        elsif ($err =~ s/^3://)
          {
            return 1 if $got;
            croak $err;
          }
        else
          {
            die "Unknown error: $err (\$! eq $!)";
          }
        return 0;
      }
    confess "Not reached" unless $got;
    return 1;
  }

sub installed
  {
    my $self = shift;
    return $self->{installed};
  }

sub spawned
  {
    my $self = shift;
    return $self->{spawned};
  }

sub install
  {
    my $self = shift;
    if (-e $self->{path})
      {
	verbose "Removing old tree " . $self->{path};
        rmtree($self->{path}, 0, 1);
        die "Couldn't remove tree " . $self->{path} if -e $self->{path};
      }
    unless (-e $install_tmp_dir)
      {
        verbose "Installing dancer into $install_tmp_dir...";
        verbose "make -C $objdir/.. DESTDIR=\"$install_tmp_dir\" install >/dev/null 2>&1", 2;
        system("make -C $objdir/.. DESTDIR=\"$install_tmp_dir\" install >/dev/null 2>&1") == 0
          or die "Failed to install dancer into $install_tmp_dir";
      }
    verbose "Copying dancer installation from $install_tmp_dir to " . $self->{path} . "...";
    system("cp", "-a", $install_tmp_dir, $self->{path});
    $self->{installed} = 1;
  }

sub spawn
  {
    my $self = shift;
    my %args = @_;

    croak "Dancer not installed for $self->{name}" unless $self->{installed};

    unlink $self->path() . "/var/log/dancerd/error.log";
    unlink $self->path() . "/var/log/dancerd/basic.log";
    unlink $self->path() . "/core";
    verbose "Spawning dancer for $self->{name}...";
    my $olddir = getcwd;
    chdir $self->path();
    my $exp = new Expect;
    $exp->log_user($all_flag);
    $exp->log_file($log);
    $exp->raw_pty(1);
    local $ENV{LD_LIBRARY_PATH};
    $ENV{LD_LIBRARY_PATH} = $self->path() . "/lib/dancerd/" .
      ($ENV{LD_LIBRARY_PATH} ? ":$ENV{LD_LIBRARY_PATH}" : "");
    $exp->spawn(qw{sbin/dancerd -d . -n -E -x 0},
                (map {"-m$_"} "basic.logfile", @{$args{modules} || []}),
                @{$args{args} || []});
    chdir $olddir;
    $self->{spawned} = $exp;

    open TOUCH, ">>", $self->path() . "/var/log/dancerd/basic.log";
    close TOUCH;
    verbose "Spawning tail for $self->{name} basic.log...";
    $exp = new Expect;
    $exp->log_user($all_flag);
    $exp->log_file($log);
    $exp->raw_pty(1);
    $exp->spawn("$objdir/fasttail", $self->path() . "/var/log/dancerd/basic.log");
    $self->{log} = $exp;

    if (grep {$_ eq '-e'} @{$args{args}})
      {
        $self->{error_log} = undef;
      }
    else
      {
        open TOUCH, ">>", $self->path() . "/var/log/dancerd/error.log";
        close TOUCH;
        verbose "Spawning tail for $self->{name} error.log...";
        $exp = new Expect;
        $exp->log_user($all_flag);
        $exp->log_file($log);
        $exp->raw_pty(1);
        $exp->spawn("$objdir/fasttail", $self->path() . "/var/log/dancerd/error.log");
        $self->{error_log} = $exp;
      }

    $self->{log_magic} = [];
    $self->{error_log_magic} = [];

    return 1;
  }

sub start
  {
    my $self = shift;
    my %args = @_;
    $self->spawn(@_);
    verbose "Waiting for startup to complete...";
    my $errors = 0;
    my $startup = 0;
    if (defined $self->{error_log})
      {
        while (my $log = $self->error_log(5))
          {
            if ($log->{eof})
              {
                $errors++;
                last;
              }
            # if $log->{level} is in qw/warning error critical/
            if ({map {$_=>1} qw/warning error critical/}->{$log->{level}})
              {
                error "Unexpected $log->{level} for $log->{facility} in error.log: $log->{msg}", 0;
                $errors++;
              }
            if ($log->{facility} eq 'core' and $log->{msg} eq 'Startup complete')
              {
                verbose "Startup complete";
                $startup++;
                last;
              }
          }
      }
    else
      {
        # We can't test it. Assume it went OK and rely on the basic.log check for timing
        $startup++;
      }
    verbose "Waiting for basic.logfile";
    my @expected_log = ({facility => 'basic.logfile', level => 'info', msg => 'basic.log opened'},
                        {facility => 'core', level => 'info', msg => 'Successfully loaded module basic.logfile'}
                       );
    foreach my $expected (@expected_log)
      {
        my $log;
        do
          {
            $log = $self->log(5);
            unless ($log)
              {
                error "Timeout while waiting for '$expected->{facility} [$expected->{level}]: $expected->{msg}' from basic.log";
                $errors++;
                last;
              }
            if ($log->{eof})
              {
                error "EOF from basic.log", 0;
                note "Expected '$expected->{facility} [$expected->{level}]: $expected->{msg}'";
                $errors++;
                last;
              }
          } while ($log->{level} eq 'debug');
        if (grep {$log->{$_} ne $expected->{$_}} qw/facility level msg/)
          {
            error "Unexpected $log->{level} message for $log->{facility} in basic.log: $log->{msg}", 0;
            note "Expected '$expected->{facility} [$expected->{level}]: $expected->{msg}'";
            $errors++;
            last;
          }
      }
    if (defined $self->{error_log})
      {
        verbose "Checking messages in error.log";
        my @expected_error_log = ({facility => 'core', level => 'info', msg => 'Trying to load module basic.logfile'},
                                  {facility => 'core', level => 'info', msg => 'Successfully loaded module basic.logfile'}
                                 );
        foreach my $expected (@expected_error_log)
          {
            my $log;
            do
              {
                $log = $self->error_log(5);
                unless ($log)
                  {
                    error "Timeout while waiting for '$expected->{facility} [$expected->{level}]: $expected->{msg}' from error.log";
                    $errors++;
                    last;
                  }
                if ($log->{eof})
                  {
                    error "EOF from error.log", 0;
                    note "Expected '$expected->{facility} [$expected->{level}]: $expected->{msg}'";
                    $errors++;
                    last;
                  }
              } while ($log->{level} eq 'debug');
            if (grep {$log->{$_} ne $expected->{$_}} qw/facility level msg/)
              {
                error "Unexpected $log->{level} message for $log->{facility} in error.log: $log->{msg}", 0;
                note "Expected '$expected->{facility} [$expected->{level}]: $expected->{msg}'";
                $errors++;
                last;
              }
          }
      }
    verbose "basic.logfile ready";

    # Now wait for all the modules to load
    foreach (@{$args{modules}})
      {
        if ($self->await_module_load($_))
          {
            pass "$_ loaded successfully";
          }
        else
          {
            fail "$_ failed to load";
          }
      }

    if ($startup and not $errors)
      {
        pass "Dancer started successfully";
        return 1;
      }
    else
      {
        if ($startup)
          {
            fail "Errors during startup";
          }
        else
          {
            fail "Dancer failed to start";
          }
        $self->kill;
        die "Dancer failed to start";
      }
  }

sub trace
  {
    my $self = shift;
    return if $self->{traced};
    if (-e $self->path . "/core")
      {
        $self->{traced}++;
        fail "Corefile found for $self->{name}";
        my $path = $self->path;
        my $corefile = "$path/core";
        my $binary = "$path/sbin/dancerd";
        my ($fh, $commands) = tempfile(UNLINK => 1);
        print $fh "bt";
        close $fh;
        system("gdb", "--batch", "--command=$commands", "--core=$corefile", "--cd=$path", $binary);
        note "To debug corefile: gdb --core=$corefile --cd=$path $binary";
        exit 1;
      }
    else
      {
        pass "No corefile after dancer killed";
      }
  }

sub _log
  {
    my $self = shift;
    my $timeout = shift || $self->{default_timeout};
    my $log = shift;
    # Pull one line out of the log, if possible
    return {eof => 1} unless $log;
    my (undef, $err) = expect($timeout,
                              (defined $self->{spawned}) ? (-i => $self->{spawned}) : (),
                              -i => $log,
                              -re => qr/^.*\n/);
    if ($err)
      {
        if ($err =~ /TIMEOUT/)
          {
            return undef;
          }
        elsif ($err =~ /EOF/)
          {
            if (not $self->{spawned}->pid)
              {
                $self->kill;
                croak "Dancer has died: $err";
              }
            return {eof => 1};
          }
        elsif ($err =~ s/^3://)
          {
            if ($self->{spawned} and not $self->{spawned}->pid)
              {
                $self->kill;
                croak "Dancer has died: $err";
              }
            return {eof => 1};
          }
        else
          {
            error "Unknown error: $err (\$! eq $!)";
          }
      }
    my $line = $log->match();
    $line =~ s/\n$//;
    croak "Unrecognised line in $self->{name} log: '$line'" unless $line =~ /^.+ (\S+) \[([^][]+)\]: ([^\r]*)$/;
    return {facility => $1, level => $2, msg => $3};
  }

sub error_log
  {
    my $self = shift;
    my $timeout = shift;
    my $buffer_magic = shift;

    # If buffer magic is on, keep the values to return them properly
    # when buffer magic isn't on.

    # Basically, the two streams are demuxed from the log itself

    if ($buffer_magic)
      {
        my $log = $self->_log($timeout, $self->{error_log});
        return undef unless defined $log;
        push @{$self->{error_log_magic}}, $log;
        return $log;
      }
    else
      {
        return (shift @{$self->{error_log_magic}}) if scalar @{$self->{error_log_magic}};
        return $self->_log($timeout, $self->{error_log});
      }
  }

sub log
  {
    my $self = shift;
    my $timeout = shift;
    my $buffer_magic = shift;

    # If buffer magic is on, keep the values to return them properly
    # when buffer magic isn't on.

    # Basically, the two streams are demuxed from the log itself

    if ($buffer_magic)
      {
        my $log = $self->_log($timeout, $self->{log});
        return undef unless defined $log;
        # Copy the %$log hash, so that the caller can modify it without breaking the next call
        push @{$self->{log_magic}}, $log;
        return $log;
      }
    else
      {
        return (shift @{$self->{log_magic}}) if scalar @{$self->{log_magic}};
        return $self->_log($timeout, $self->{log});
      }
  }

sub expect_log
  {
    my $self = shift;
    my $timeout = shift;
    my $expect = shift;

    my $data = $self->log($timeout);
    unless ($data)
      {
        fail "Didn't get '$expect->{facility} [$expect->{level}]: $expect->{msg}' from log (timeout)";
        return 0;
      }
    if (grep {!exists $data->{$_} or !exists $expect->{$_} or $data->{$_} ne $expect->{$_}}
        (keys %{$expect}, keys %{$data}))
      {
        if ($data->{eof})
          {
            fail "Got EOF from log";
          }
        else
          {
            fail "Got '$data->{facility} [$data->{level}]: $data->{msg}' from log";
          }
        note "Was expecting '$expect->{facility} [$expect->{level}]: $expect->{msg}'";
        return 0;
      }
    else
      {
        if ($data->{eof})
          {
            pass "Got EOF from log";
          }
        else
          {
            pass "Got '$data->{facility} [$data->{level}]: $data->{msg}' from log";
          }
        return 1;
      }
  }

sub await_module_load
  {
    my $self = shift;
    my $module = shift;
    while (my $log = $self->log(5))
      {
        last unless $log;
        last if $log->{eof};
        next if $log->{facility} ne 'core';
        # Take out one line for synchronisation
        $self->error_log();
        foreach my $re (
                        {level => 'info', msg => qr/\QSuccessfully loaded module $module\E.*/,
                         result => 1},
                        {level => 'info', msg => qr/\QModule $module failed to initialise, unloading\E.*\n/,
                         result => 0},
                        {level => 'error', msg => qr/\QModule $module failed to initialise successfully\E.*\n/,
                         result => 0},
                        {level => 'error', msg => qr/\QCouldn\'t load module $module\E.*\n/,
                         result => 0},
                        {level => 'error', msg => qr/\QCannot load a new copy of module $module\E.*\n/,
                         result => 0},
                        {level => 'error',
                         msg => qr/\QModule $module version \E.*\Q does not support in-place reloading\E.*\n/,
                         result => 0},
                        {level => 'error',
                         msg => qr/\QModule $module versions \E.*\Q and \E.*\Q have differing interface versions\E.*\n/,
                         result => 0},
                        {level => 'error', msg => qr/\QDependencies for $module not satisfied, aborting\E.*\n/,
                         result => 0},
                        {level => 'error', msg => qr/\QFailed to process all symbols for $module, aborting\E.*\n/,
                         result => 0},
                        {level => 'error', msg => qr/\QModule $module init function failed, aborting\E.*\n/,
                         result => 0},
                       )
          {
            return $re->{result} if $log->{level} eq $re->{level} and $log->{msg} =~ $re->{msg};
          }
      }
    die "$module didn't load into $self->{name}";
  }

sub on_kill
  {
    my $self = shift;
    my $sub = shift;
    push @{$self->{on_kill}}, $sub;
  }

sub kill
  {
    my $self = shift;
    $_->() foreach @{$self->{on_kill}};
    $self->{on_kill} = [];
    $_->kill foreach values %{$self->{pipes}};
    $self->{pipes} = {};

    while (my $conn = shift @{$self->{connections}})
      {
        verbose "Closing connection to " . $conn->{target};
        $conn->{exp}->hard_close();
      }

    if ($self->{spawned})
      {
        verbose "Killing dancer for $self->{name}";
        my $pid = $self->{spawned}->pid();
        kill TERM => $pid if $pid;
        $self->{spawned}->soft_close();
        $self->{spawned} = undef;
        #kill KILL => $pid;
      }
    $self->trace;

    if ($self->{log})
      {
        verbose "Killing log tail for $self->{name}";
        my $pid = $self->{log}->pid();
        kill INT => $pid if $pid;
        # Scan any log lines for errors.
        my $errors = 0;
        while (my $log = $self->log(1, 1))
          {
            last if $log->{eof};
            # if $log->{level} isn't in qw/debug trace info/
            unless ({map {$_=>1} qw/debug trace info/}->{$log->{level}})
              {
                warning "Unexpected $log->{level} for $log->{facility} in $self->{name}/basic.log: $log->{msg}";
                $errors++;
              }
          }
        if ($errors)
          {
            fail "Unexpected errors in $self->{name} basic.log";
          }
        else
          {
            pass "No unexpected errors in $self->{name} basic.log";
          }
        $self->{log}->soft_close();
        $self->{log} = undef;
      }

    if ($self->{error_log})
      {
        verbose "Killing error log tail for $self->{name}";
        my $pid = $self->{error_log}->pid();
        kill INT => $pid if $pid;
        # Scan any log lines for errors.
        my $errors = 0;
        while (my $log = $self->error_log(1, 1))
          {
            last if $log->{eof};
            # if $log->{level} isn't in qw/debug trace info/
            unless ({map {$_=>1} qw/debug trace info/}->{$log->{level}})
              {
                warning "Unexpected $log->{level} for $log->{facility} in error.log: $log->{msg}";
                $errors++;
              }
          }
        if ($errors)
          {
            fail "Unexpected errors in $self->{name} error.log";
          }
        else
          {
            pass "No unexpected errors in $self->{name} error.log";
          }
        $self->{error_log}->soft_close();
        $self->{error_log} = undef;
      }
    return 1;
  }

sub cmdpipe
  {
    my $self = shift;
    my $module = shift;
    return $self->{pipes}{$module} if $self->{pipes}{$module};
    $self->{pipes}{$module} = new Dancer::CmdPipe $self, "$module for $self->{name}", $module, @_;
    return $self->{pipes}{$module};
  }

sub get_config
  {
    my $self = shift;
    my $key = shift;
    my $exp = new Expect;
    $exp->raw_pty(1);
    $exp->spawn($self->path() . "/sbin/dancer-edit-config",
                "--dancerd=" . $self->path() . "/sbin/dancerd",
                "--dancer-args=-d" . $self->path(),
                "--", "get", $key);
    my ($match, $err) =
      $exp->expect(5,
                   -re => qr/Key '$key' is (\S+) (.*)\n/, sub {},
                  );
    if ($err)
      {
        if ($err =~ /TIMEOUT/)
          {
            fail "failed to get config key $key for $self->{name} (timeout)";
            exit 1;
          }
        elsif ($err =~ /EOF/)
          {
            fail "failed to get config key $key for $self->{name} (eof)";
            exit 1;
          }
        elsif ($err !~ s/^3://)
          {
            error "Unknown error: $err (\$! eq $!)";
            exit 1;
          }
      }
    $exp->match() =~ /Key '$key' is (\S+) (.*)\n/;
    $exp->soft_close();
    return ($1, $2);
  }

sub put_config
  {
    my $self = shift;
    my ($key, $type, $value) = @_;
    my $exp = new Expect;
    $exp->raw_pty(1);
    $exp->log_user($all_flag);
    $exp->log_file($log);
    $exp->spawn($self->path() . "/sbin/dancer-edit-config",
                "--dancerd=" . $self->path . "/sbin/dancerd",
                "--dancer-args=-d" . $self->path,
                "--", "put", $key, $type, $value);
    _check_expect $exp, qr/\QPut key '$key' as $type $value\E\n/, "put config key $key for $self->{name}";
    $exp->soft_close();
  }

sub delete_config
  {
    my $self = shift;
    my $key = @_;
    my $exp = new Expect;
    $exp->raw_pty(1);
    $exp->spawn($self->path() . "/sbin/dancer-edit-config",
                "--dancerd=" . $self->path() . "/sbin/dancerd",
                "--dancer-args=-d" . $self->path(),
                "--", "delete", $key);
    _check_expect $exp, qr/\QDeleted key '$key'\E\n/, "deleted config key $key for $self->{name}";
    $exp->soft_close();
  }

sub gdb
  {
    my $self = shift;
    croak "No dancer running for $self->{name}" unless $self->{spawned};
    print "Now run:\n";
    print "gdb -cd " . $self->path . " sbin/dancerd " . $self->{spawned}->pid() . "\n";
    print "Then come back here and hit enter\n";
    $self->{default_timeout} = undef;
    Dancer::CmdPipe::gdb_foo;
    <STDIN>;
    print "Continuing...\n";
  }

sub DESTROY
  {
    my $self = shift;
    $_->kill foreach values %{$self->{pipes}};
    $self->kill if $self->{spawned};
  }

1;
