/************************************************************************
 *   This is part of Dancer, a text relay daemon
 *   Copyright (C) 2002 Andrew Suffield <asuffield@users.sourceforge.net>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

#include <dancerd/config.h>
#include <dancerd/modules.h>
#include <dancerd/log.h>
#include <dancerd/paths.h>
#include <dancerd/xmalloc.h>
#include <dancerd/xstring.h>
#include <dancerd/hashtable.h>
#include <dancerd/test/cmdpipe.h>

#include <dancerd/auth/auth.h>

#include <stddef.h>
#include <string.h>

SETUP_MODULE("test.auth", MODVERSION);

#include "symtab"

static struct cmdpipe *cp = NULL;

static struct hashtable *handle_hash = NULL;
static struct auth_method *my_method = NULL;
static struct auth_iterator *it = NULL;
static int status_result = AUTH_CONTINUE;
static char *status_error = NULL;
static char *result_data = NULL;
static int result_argc = 0;
static char *result_argv[16] = {NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
                                NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL};

static auth_init_handle init_handle;
static void *
init_handle(struct auth_handle *handle unused_attribute, const char *authname,
            void *credentials, uint32_t len unused_attribute)
{
  print_output_buffer(cp, "INIT_HANDLE: ");
  print_output_buffer(cp, authname);
  print_output_buffer(cp, " ");
  print_output_buffer(cp, credentials);
  print_output_buffer(cp, "\n");
  return xstrdup(authname);
}

static auth_cancel_handle cancel_handle;
static void
cancel_handle(struct auth_handle *handle unused_attribute, void *data)
{
  print_output_buffer(cp, "CANCEL_HANDLE: ");
  print_output_buffer(cp, data);
  print_output_buffer(cp, "\n");
  xfree(data);
}

static auth_peer_accept peer_accept;
static int
peer_accept(struct auth_handle *handle unused_attribute, void *data, char **err)
{
  print_output_buffer(cp, "PEER_ACCEPT: ");
  print_output_buffer(cp, data);
  print_output_buffer(cp, "\n");
  if (err)
    *err = status_error;
  return status_result;
}

static auth_read_data read_data;
static int
read_data(struct auth_handle *handle unused_attribute, void *data, int argc, char **argv, char **err)
{
  int i;
  print_output_buffer(cp, "READ_DATA: ");
  print_output_buffer(cp, data);
  for (i = 0; i < argc; i++)
    {
      print_output_buffer(cp, " ");
      print_output_buffer(cp, argv[i]);
    }
  print_output_buffer(cp, "\n");
  if (err)
    *err = status_error;
  return status_result;
}

static auth_read_data read_data;
static int
get_data(struct auth_handle *handle unused_attribute, void *data, int *argc, char ***argv, char **err)
{
  print_output_buffer(cp, "GET_DATA: ");
  print_output_buffer(cp, data);
  print_output_buffer(cp, "\n");
  if (err)
    *err = status_error;
  *argc = result_argc;
  *argv = result_argv;
  return status_result;
}

static auth_set_credentials set_credentials;
static int
set_credentials(const char *entity, int argc, char **argv, char **data, uint32_t *len, char **err)
{
  int i;
  print_output_buffer(cp, "SET_CREDENTIALS: ");
  print_output_buffer(cp, entity);
  for (i = 0; i < argc; i++)
    {
      print_output_buffer(cp, " ");
      print_output_buffer(cp, argv[i]);
    }
  print_output_buffer(cp, "\n");
  if (err)
    *err = status_error;
  *data = result_data;
  *len = strlen(result_data) + 1;
  return status_result;
}

static int
parse_command(char *data, size_t len unused_attribute)
{
  if (strncasecmp(data, "REGISTER", strlen("REGISTER")) == 0)
    {
      if (my_method)
	return 0;
      else
	{
	  my_method = SYM(auth, register_method)("test", MYHANDLE,
                                                 &init_handle, &cancel_handle,
                                                 &peer_accept, &read_data, &get_data,
                                                 &set_credentials);
	  if (!my_method)
	    return 0;
	  else
	    return 1;
	}
    }
  else if (strncasecmp(data, "UNREGISTER", strlen("UNREGISTER")) == 0)
    {
      if (!my_method)
	return -1;
      else
	{
	  SYM(auth, unregister_method)(my_method);
	  my_method = NULL;
	  return 1;
	}
    }
  else if (strncasecmp(data, "STATUS_RESULT", strlen("STATUS_RESULT")) == 0)
    {
      char *c = data + strlen("STATUS_RESULT");
      if (*c)
        c++;
      if (!my_method)
	return -1;
      else
	{
          status_result = strtoul(c, NULL, 0);
	  return 1;
	}
    }
  else if (strncasecmp(data, "STATUS_ERROR", strlen("STATUS_ERROR")) == 0)
    {
      char *c = data + strlen("STATUS_ERROR");
      if (*c)
        c++;
      if (!my_method)
	return -1;
      else
	{
          xfree(status_error);
          status_error = xstrdup(c);
	  return 1;
	}
    }
  else if (strncasecmp(data, "RESULT_DATA", strlen("RESULT_DATA")) == 0)
    {
      char *c = data + strlen("RESULT_DATA");
      if (*c)
        c++;
      if (!my_method)
	return -1;
      else
	{
          xfree(result_data);
          result_data = xstrdup(c);
	  return 1;
	}
    }
  else if (strncasecmp(data, "RESULT_ARGV", strlen("RESULT_ARGV")) == 0)
    {
      char *c = data + strlen("RESULT_ARGV");
      if (*c)
        c++;
      if (!my_method)
	return -1;
      else
	{
          char *d;
          result_argc = 0;
          while ((d = strsep(&c, " ")))
            {
              xfree(result_argv[result_argc]);
              result_argv[result_argc++] = xstrdup(d);
            }
	  return 1;
	}
    }
  else if (strncasecmp(data, "NEW_HANDLE", strlen("NEW_HANDLE")) == 0)
    {
      char *c = data + strlen("NEW_HANDLE"), *d = NULL;
      if (*c)
	{
          c++;
	  d = strchr(c, ' ');
	}
      if (!my_method)
	return -1;
      else
	{
          *d++ = '\0';
          if (hashtable_lookup(handle_hash, d))
            return 0;
          else
            {
              char *err;
              struct auth_handle *handle = SYM(auth, new_handle)(c, d, &err);
              if (handle)
                {
                  hashtable_insert(handle_hash, d, handle);
                  return 1;
                }
              else
                {
                  print_output_buffer(cp, err);
                  print_output_buffer(cp, "\n");
                  return 0;
                }
            }
	}
    }
  else if (strncasecmp(data, "CANCEL_HANDLE", strlen("CANCEL_HANDLE")) == 0)
    {
      char *c = data + strlen("CANCEL_HANDLE");
      if (*c)
        c++;
      if (!my_method)
	return -1;
      else
	{
          struct auth_handle *handle = hashtable_lookup(handle_hash, c);
          
          if (!handle)
            return 0;
          else
            {
              SYM(auth, cancel_handle)(handle);
              hashtable_delete(handle_hash, c);
              return 1;
            }
	}
    }
  else if (strncasecmp(data, "PEER_ACCEPT", strlen("PEER_ACCEPT")) == 0)
    {
      char *c = data + strlen("PEER_ACCEPT");
      if (*c)
        c++;
      if (!my_method)
	return -1;
      else
	{
          struct auth_handle *handle;
          handle = hashtable_lookup(handle_hash, c);
          
          if (!handle)
            return 0;
          else
            {
              char *err;
              switch (SYM(auth, peer_accept)(handle, &err))
                {
                case AUTH_ACCEPT:
                  print_output_buffer(cp, "ACCEPT\n");
                  hashtable_delete(handle_hash, c);
                  break;
                case AUTH_CANCEL:
                  print_output_buffer(cp, "CANCEL ");
                  print_output_buffer(cp, err);
                  print_output_buffer(cp, "\n");
                  hashtable_delete(handle_hash, c);
                  break;
                case AUTH_CONTINUE:
                  print_output_buffer(cp, "CONTINUE\n");
                  break;
                }
              return 1;
            }
	}
    }
  else if (strncasecmp(data, "READ_DATA", strlen("READ_DATA")) == 0)
    {
      char *c = data + strlen("READ_DATA"), *d = NULL;
      if (*c)
	{
          *c++ = '\0';
	  d = strchr(c, ' ');
	}
      if (!my_method)
	return -1;
      else
	{
          if (!d)
            return 0;
          else
            {
              struct auth_handle *handle;
              *d++ = '\0';
              handle = hashtable_lookup(handle_hash, c);
          
              if (!handle)
                return 0;
              else
                {
                  char *p;
                  char *err;
                  int argc, i;
                  char *argv[16];
                  argc = 0;
                  while ((p = strsep(&d, " ")))
                    {
                      argv[argc++] = xstrdup(p);
                    }
                  switch (SYM(auth, read_data)(handle, argc, argv, &err))
                    {
                    case AUTH_ACCEPT:
                      print_output_buffer(cp, "ACCEPT\n");
                      hashtable_delete(handle_hash, c);
                      break;
                    case AUTH_CANCEL:
                      print_output_buffer(cp, "CANCEL ");
                      print_output_buffer(cp, err);
                      print_output_buffer(cp, "\n");
                      hashtable_delete(handle_hash, c);
                      break;
                    case AUTH_CONTINUE:
                      print_output_buffer(cp, "CONTINUE\n");
                      break;
                    }
                  for (i = 0; i < argc; i++)
                    xfree(argv[i]);
                  return 1;
                }
            }
	}
    }
  else if (strncasecmp(data, "GET_DATA", strlen("GET_DATA")) == 0)
    {
      char *c = data + strlen("GET_DATA");
      if (*c)
        c++;
      if (!my_method)
	return -1;
      else
	{
          struct auth_handle *handle = hashtable_lookup(handle_hash, c);
          
          if (!handle)
            return 0;
          else
            {
              char *err;
              int argc, i;
              char **argv;
              switch (SYM(auth, get_data)(handle, &argc, &argv, &err))
                {
                case AUTH_ACCEPT:
                  print_output_buffer(cp, "ACCEPT\n");
                  hashtable_delete(handle_hash, c);
                  break;
                case AUTH_CANCEL:
                  print_output_buffer(cp, "CANCEL ");
                  print_output_buffer(cp, err);
                  print_output_buffer(cp, "\n");
                  hashtable_delete(handle_hash, c);
                  break;
                case AUTH_CONTINUE:
                  print_output_buffer(cp, "CONTINUE\n");
                  break;
                }
              if (argc)
                print_output_buffer(cp, "DATA:");
              for (i = 0; i < argc; i++)
                {
                  print_output_buffer(cp, " ");
                  print_output_buffer(cp, argv[i]);
                }
              print_output_buffer(cp, "\n");
              return 1;
            }
	}
    }
  else if (strncasecmp(data, "SET_CREDENTIALS", strlen("SET_CREDENTIALS")) == 0)
    {
      char *c = data + strlen("SET_CREDENTIALS"), *d = NULL;
      if (*c)
	{
          *c++ = '\0';
	  d = strchr(c, ' ');
	}
      if (!d)
        return 0;
      else
        {
          char *p;
          char *err;
          int argc, i;
          char *argv[16];
          char *e;
          *d++ = '\0';
          e = d;
          argc = 0;
          while ((p = strsep(&e, " ")))
            {
              argv[argc++] = xstrdup(p);
            }
          if (SYM(auth, set_credentials)(c, d, argc, argv, &err))
            {
              print_output_buffer(cp, "OK\n");
            }
          else
            {
              print_output_buffer(cp, err);
              print_output_buffer(cp, "\n");
            }
          for (i = 0; i < argc; i++)
            xfree(argv[i]);
          return 1;
        }
    }
  else if (strncasecmp(data, "DROP_ENTITY", strlen("DROP_ENTITY")) == 0)
    {
      char *c = data + strlen("DROP_ENTITY");
      if (*c)
        c++;
      if (!*c)
        return 0;
      else
        {
          char *err;
          if (SYM(auth, drop_entity)(c, &err))
            return 1;
          else
            {
              print_output_buffer(cp, err);
              print_output_buffer(cp, "\n");
              return 0;
            }
        }
    }
  else if (strncasecmp(data, "DROP_OBJECT", strlen("DROP_OBJECT")) == 0)
    {
      char *c = data + strlen("DROP_OBJECT");
      if (*c)
        c++;
      if (!*c)
        return 0;
      else
        {
          char *err;
          char *d = strchr(c, ':');
          if (d)
            *d++ = '\0';
          else
            {
              d = c;
              c = NULL;
            }
          if (SYM(auth, drop_object)(c, d, &err))
            return 1;
          else
            {
              print_output_buffer(cp, err);
              print_output_buffer(cp, "\n");
              return 0;
            }
        }
    }
  else if (strncasecmp(data, "GRANT_AUTHORISATION", strlen("GRANT_AUTHORISATION")) == 0)
    {
      char *c = data + strlen("GRANT_AUTHORISATION"), *d = NULL, *e = NULL, *f = NULL;
      char *object_type, *object_name;
      uint32_t value;
      char *err;
      if (*c)
	{
          *c++ = '\0';
	  d = strchr(c, ' ');
	}
      if (!d)
        {
          return 0;
        }
      *d++ = '\0';
      if (!(e = strchr(d, '/')))
        {
          return 0;
        }
      *e++ = '\0';
      if (!(f = strchr(e, ' ')))
        {
          return 0;
        }
      *f++ = '\0';
      value = strtoul(f, NULL, 0);
      object_name = strchr(d, ':');
      if (object_name)
        {
          object_type = d;
          *object_name++ = '\0';
        }
      else
        {
          object_type = NULL;
          object_name = d;
        }
      if (SYM(auth, grant_authorisation)(c, object_type, object_name, e, value, &err))
        {
          return 1;
        }
      else
        {
          print_output_buffer(cp, err);
          print_output_buffer(cp, "\n");
          return 0;
        }
    }
  else if (strncasecmp(data, "CHECK_AUTHORISATION", strlen("CHECK_AUTHORISATION")) == 0)
    {
      char *c = data + strlen("CHECK_AUTHORISATION"), *d = NULL, *e = NULL, *f = NULL;
      char *object_type, *object_name;
      uint32_t value;
      char *err;
      if (*c)
	{
          *c++ = '\0';
	  d = strchr(c, ' ');
	}
      if (!d)
        {
          return 0;
        }
      *d++ = '\0';
      if (!(e = strchr(d, '/')))
        {
          return 0;
        }
      *e++ = '\0';
      if (!(f = strchr(e, ' ')))
        {
          return 0;
        }
      *f++ = '\0';
      value = strtoul(f, NULL, 0);
      object_name = strchr(d, ':');
      if (object_name)
        {
          object_type = d;
          *object_name++ = '\0';
        }
      else
        {
          object_type = NULL;
          object_name = d;
        }
      if (SYM(auth, check_authorisation)(c, object_type, object_name, e, value, &err))
        {
          print_output_buffer(cp, "YES\n");
        }
      else
        {
          print_output_buffer(cp, "NO\n");
        }
      if (err)
        {
          print_output_buffer(cp, err);
          print_output_buffer(cp, "\n");
        }
      return 1;
    }
  else if (strncasecmp(data, "GET_AUTHORISATION", strlen("GET_AUTHORISATION")) == 0)
    {
      char *c = data + strlen("GET_AUTHORISATION"), *d = NULL, *e = NULL;
      char *object_type, *object_name;
      int32_t value;
      char *err;
      if (*c)
	{
          *c++ = '\0';
	  d = strchr(c, ' ');
	}
      if (!d)
        {
          return 0;
        }
      *d++ = '\0';
      if (!(e = strchr(d, '/')))
        {
          return 0;
        }
      *e++ = '\0';
      object_name = strchr(d, ':');
      if (object_name)
        {
          object_type = d;
          *object_name++ = '\0';
        }
      else
        {
          object_type = NULL;
          object_name = d;
        }
      if (SYM(auth, get_authorisation)(c, object_type, object_name, e, &value, &err))
        {
          char buf[64];
          snprintf(buf, 64, "%u\n", value);
          print_output_buffer(cp, buf);
        }
      else
        {
          print_output_buffer(cp, "(none)\n");
        }
      if (err)
        {
          print_output_buffer(cp, err);
          print_output_buffer(cp, "\n");
        }
      return 1;
    }
  else if (strncasecmp(data, "REVOKE_AUTHORISATION", strlen("REVOKE_AUTHORISATION")) == 0)
    {
      char *c = data + strlen("REVOKE_AUTHORISATION"), *d = NULL, *e = NULL;
      char *object_type, *object_name;
      char *err;
      if (*c)
	{
          *c++ = '\0';
	  d = strchr(c, ' ');
	}
      if (!d)
        {
          return 0;
        }
      *d++ = '\0';
      if (!(e = strchr(d, '/')))
        {
          return 0;
        }
      *e++ = '\0';
      object_name = strchr(d, ':');
      if (object_name)
        {
          object_type = d;
          *object_name++ = '\0';
        }
      else
        {
          object_type = NULL;
          object_name = d;
        }
      if (SYM(auth, revoke_authorisation)(c, object_type, object_name, e, &err))
        {
          return 1;
        }
      else
        {
          print_output_buffer(cp, err);
          print_output_buffer(cp, "\n");
          return 0;
        }
    }
  else if (strncasecmp(data, "ITERATOR_NEXT", strlen("ITERATOR_NEXT")) == 0)
    {
      if (!it)
        return -1;
      else
        {
          char *err;
          const char *value = SYM(auth, iterator_next)(it, &err);
          if (value)
            print_output_buffer(cp, value);
          else
            print_output_buffer(cp, "(null)");
          print_output_buffer(cp, "\n");
          if (err)
            {
              print_output_buffer(cp, err);
              print_output_buffer(cp, "\n");
            }
          return 1;
        }
    }
  else if (strncasecmp(data, "FREE_ITERATOR", strlen("FREE_ITERATOR")) == 0)
    {
      SYM(auth, free_iterator)(it);
      it = NULL;
      return 1;
    }
  else if (strncasecmp(data, "GET_ENTITY_METHODS", strlen("GET_ENTITY_METHODS")) == 0)
    {
      char *c = data + strlen("GET_ENTITY_METHODS");
      if (*c)
        c++;
      if (!*c)
        return 0;
      else
        {
          char *value, *err;
          it = SYM(auth, get_entity_methods)(c, &value, &err);
          if (value)
            print_output_buffer(cp, value);
          else
            print_output_buffer(cp, "(null)");
          print_output_buffer(cp, "\n");
          if (err)
            {
              print_output_buffer(cp, err);
              print_output_buffer(cp, "\n");
            }
          return 1;
        }
    }
  else if (strncasecmp(data, "GET_ENTITY_ATTRIBUTES", strlen("GET_ENTITY_ATTRIBUTES")) == 0)
    {
      char *c = data + strlen("GET_ENTITY_ATTRIBUTES");
      if (*c)
        c++;
      if (!*c)
        return 0;
      else
        {
          char *value, *err;
          it = SYM(auth, get_entity_attributes)(c, &value, &err);
          if (value)
            print_output_buffer(cp, value);
          else
            print_output_buffer(cp, "(null)");
          print_output_buffer(cp, "\n");
          if (err)
            {
              print_output_buffer(cp, err);
              print_output_buffer(cp, "\n");
            }
          return 1;
        }
    }
  else if (strncasecmp(data, "GET_ENTITY_AUTHORISATION", strlen("GET_ENTITY_AUTHORISATION")) == 0)
    {
      char *c = data + strlen("GET_ENTITY_AUTHORISATION");
      if (*c)
        c++;
      if (!*c)
        return 0;
      else
        {
          char *value, *err;
          it = SYM(auth, get_entity_authorisation)(c, &value, &err);
          if (value)
            print_output_buffer(cp, value);
          else
            print_output_buffer(cp, "(null)");
          print_output_buffer(cp, "\n");
          if (err)
            {
              print_output_buffer(cp, err);
              print_output_buffer(cp, "\n");
            }
          return 1;
        }
    }
  else if (strncasecmp(data, "GET_OBJECT_ATTRIBUTES", strlen("GET_OBJECT_ATTRIBUTES")) == 0)
    {
      char *c = data + strlen("GET_OBJECT_ATTRIBUTES");
      if (*c)
        c++;
      if (!*c)
        return 0;
      else
        {
          char *value, *err;
          char *object_type, *object_name;
          object_name = strchr(c, ':');
          if (object_name)
            {
              object_type = c;
              *object_name++ = '\0';
            }
          else
            {
              object_type = NULL;
              object_name = c;
            }
          it = SYM(auth, get_object_attributes)(object_type, object_name, &value, &err);
          if (value)
            print_output_buffer(cp, value);
          else
            print_output_buffer(cp, "(null)");
          print_output_buffer(cp, "\n");
          if (err)
            {
              print_output_buffer(cp, err);
              print_output_buffer(cp, "\n");
            }
          return 1;
        }
    }
  else if (strncasecmp(data, "GET_OBJECT_AUTHORISATION", strlen("GET_OBJECT_AUTHORISATION")) == 0)
    {
      char *c = data + strlen("GET_OBJECT_AUTHORISATION");
      if (*c)
        c++;
      if (!*c)
        return 0;
      else
        {
          char *value, *err;
          char *object_type, *object_name;
          object_name = strchr(c, ':');
          if (object_name)
            {
              object_type = c;
              *object_name++ = '\0';
            }
          else
            {
              object_type = NULL;
              object_name = c;
            }
          it = SYM(auth, get_object_authorisation)(object_type, object_name, &value, &err);
          if (value)
            print_output_buffer(cp, value);
          else
            print_output_buffer(cp, "(null)");
          print_output_buffer(cp, "\n");
          if (err)
            {
              print_output_buffer(cp, err);
              print_output_buffer(cp, "\n");
            }
          return 1;
        }
    }
  else if (strncasecmp(data, "GET_OBJECT_PRIV_AUTHORISATION", strlen("GET_OBJECT_PRIV_AUTHORISATION")) == 0)
    {
      char *c = data + strlen("GET_OBJECT_PRIV_AUTHORISATION"), *d = NULL;
      if (*c)
        {
          c++;
          d = strchr(c, ' ');
        }
      if (!*c)
        return 0;
      else
        {
          char *value, *err;
          char *object_type, *object_name;
          if (d)
            *d++ = '\0';
          object_name = strchr(c, ':');
          if (object_name)
            {
              object_type = c;
              *object_name++ = '\0';
            }
          else
            {
              object_type = NULL;
              object_name = c;
            }
          it = SYM(auth, get_object_priv_authorisation)(object_type, object_name, d, &value, &err);
          if (value)
            print_output_buffer(cp, value);
          else
            print_output_buffer(cp, "(null)");
          print_output_buffer(cp, "\n");
          if (err)
            {
              print_output_buffer(cp, err);
              print_output_buffer(cp, "\n");
            }
          return 1;
        }
    }
  else if (strncasecmp(data, "SET_ENTITY_ATTRIBUTE", strlen("SET_ENTITY_ATTRIBUTE")) == 0)
    {
      char *c = data + strlen("SET_ENTITY_ATTRIBUTE"), *d = NULL, *e = NULL;
      char *err;
      if (*c)
	{
          *c++ = '\0';
	  d = strchr(c, ' ');
	}
      if (!d)
        {
          return 0;
        }
      *d++ = '\0';
      if (!(e = strchr(d, ' ')))
        {
          return 0;
        }
      *e++ = '\0';
      if (SYM(auth, set_entity_attribute)(c, d, e, strlen(e) + 1, &err))
        {
          return 1;
        }
      else
        {
          print_output_buffer(cp, err);
          print_output_buffer(cp, "\n");
          return 0;
        }
    }
  else if (strncasecmp(data, "SET_OBJECT_ATTRIBUTE", strlen("SET_OBJECT_ATTRIBUTE")) == 0)
    {
      char *c = data + strlen("SET_OBJECT_ATTRIBUTE"), *d = NULL, *e = NULL;
      char *object_type, *object_name;
      char *err;
      if (*c)
	{
          *c++ = '\0';
	  d = strchr(c, ' ');
	}
      if (!d)
        {
          return 0;
        }
      *d++ = '\0';
      if (!(e = strchr(d, ' ')))
        {
          return 0;
        }
      *e++ = '\0';
      object_name = strchr(c, ':');
      if (object_name)
        {
          object_type = c;
          *object_name++ = '\0';
        }
      else
        {
          object_type = NULL;
          object_name = c;
        }
      if (SYM(auth, set_object_attribute)(object_type, object_name, d, e, strlen(e) + 1, &err))
        {
          return 1;
        }
      else
        {
          print_output_buffer(cp, err);
          print_output_buffer(cp, "\n");
          return 0;
        }
    }
  else if (strncasecmp(data, "GET_ENTITY_ATTRIBUTE", strlen("GET_ENTITY_ATTRIBUTE")) == 0)
    {
      char *c = data + strlen("GET_ENTITY_ATTRIBUTE"), *d = NULL;
      void *value;
      char *err;
      if (*c)
	{
          *c++ = '\0';
	  d = strchr(c, ' ');
	}
      if (!d)
        {
          return 0;
        }
      *d++ = '\0';
      if (SYM(auth, get_entity_attribute)(c, d, &value, NULL, &err))
        {
          print_output_buffer(cp, value);
          print_output_buffer(cp, "\n");
          return 1;
        }
      else
        {
          print_output_buffer(cp, err);
          print_output_buffer(cp, "\n");
          return 1;
        }
    }
  else if (strncasecmp(data, "GET_OBJECT_ATTRIBUTE", strlen("GET_OBJECT_ATTRIBUTE")) == 0)
    {
      char *c = data + strlen("GET_OBJECT_ATTRIBUTE"), *d = NULL;
      char *object_type, *object_name;
      void *value;
      char *err;
      if (*c)
	{
          *c++ = '\0';
	  d = strchr(c, ' ');
	}
      if (!d)
        {
          return 0;
        }
      *d++ = '\0';
      object_name = strchr(c, ':');
      if (object_name)
        {
          object_type = c;
          *object_name++ = '\0';
        }
      else
        {
          object_type = NULL;
          object_name = c;
        }
      if (SYM(auth, get_object_attribute)(object_type, object_name, d, &value, NULL, &err))
        {
          print_output_buffer(cp, value);
          print_output_buffer(cp, "\n");
          return 1;
        }
      else
        {
          print_output_buffer(cp, err);
          print_output_buffer(cp, "\n");
          return 1;
        }
    }
  else
    return 0;
}

static void
pipe_fail(void)
{
  unload_module(MYNAME);
}

int
MODULE_INIT(void *data unused_attribute)
{
  if (!(cp = setup_cmdpipe(MYHANDLE, MYNAME, 32768, &parse_command, &pipe_fail)))
    return 0;

  handle_hash = new_hashtable(&hash_ascii, 17);

  return MODULE_LOAD_OK;
}

int
MODULE_DEINIT(const char *new_version unused_attribute, module_init_func *init unused_attribute)
{
  int i;
  for (i = 0; i < 16; i++)
    xfree(result_argv[i]);
  xfree(status_error);
  free_hashtable(handle_hash);
  if (cp)
    close_cmdpipe(cp);
  cp = NULL;
  return MODULE_UNLOAD_OK;
}
