/************************************************************************
 *   This is part of Dancer, a text relay daemon
 *   Copyright (C) 2002 Andrew Suffield <asuffield@users.sourceforge.net>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

/* WARNING
 * This code is not very solid or reliable
 * It should under no circumstances be used in production
 *
 * The sole reason for this code's existance is to provide a simple
 * and reliable driver for the test suite.
 *
 * DO NOT USE THIS FOR ANY MODULE OUTSIDE OF THE test.* SUBTREE
 */

#define _GNU_SOURCE

#include <dancerd/config.h>
#include <dancerd/fd.h>
#include <dancerd/xmalloc.h>
#include <dancerd/paths.h>
#include <dancerd/xstring.h>
#include <dancerd/log.h>
#include <dancerd/modules.h>

#include <assert.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdio.h>

#include "cmdpipe.h"

#define CONTROL_FD_GROUP 0
#define OUTPUT_FD_GROUP 1

struct cmdpipe
{
  size_t buffer_size;
  /* work_buffer is a chunk of suitably sized scratch space */
  char *output_buffer, *control_buffer, *work_buffer;
  size_t output_buffer_fill, control_buffer_fill;
  int control_pipe_fd, output_pipe_fd;
  char *control_pipe_name, *output_pipe_name;
  cmdpipe_parse_func *pf;
  cmdpipe_fail_func *ff;
  char *name;
  int currently_parsing;
};

static int
output_pipe_write_handler(int fd, void *data)
{
  ssize_t len;
  struct cmdpipe *cp = data;

  if (cp->output_buffer_fill == 0)
    {
      block_fd_writes(fd);
      return 1;
    }
  if ((len = write(fd, cp->output_buffer, cp->output_buffer_fill)) == -1)
    {
      log("cmdpipe", LOG_PRIORITY_ERROR, "Error writing to output pipe: %s", strerror(errno));
      return 0;
    }
  assert((size_t)len <= cp->output_buffer_fill);
  memmove(cp->output_buffer, cp->output_buffer + len, cp->output_buffer_fill - len);
  cp->output_buffer_fill -= len;
  return 1;
}

static void
output_pipe_close_handler(int fd unused_attribute, void *data)
{
  struct cmdpipe *cp = data;

  cp->output_pipe_fd = -1;
  if ((cp->output_pipe_fd = open(cp->output_pipe_name, O_WRONLY | O_NONBLOCK, 0)) == -1)
    {
      if (errno != ENXIO)
	log("cmdpipe", LOG_PRIORITY_ERROR, "Failed to open output fifo %s: %s", cp->output_pipe_name, strerror(errno));
      return;
    }
  register_fd_handler(cp->output_pipe_fd, FD_CLASS_LOCAL, cp->name, OUTPUT_FD_GROUP, cp);
}

void
write_output_buffer(struct cmdpipe *cp, const char *data, size_t len)
{
  char *dest;
  const char *end = cp->output_buffer + cp->buffer_size;
  dest = cp->output_buffer + cp->output_buffer_fill;
  /* output_buffer_fill increases with dest. len decreases with data */
  while (len > 0 && dest < end)
    switch(*data)
      {
      case '\r':
        *dest++ = '\\';
        cp->output_buffer_fill++;
        if (dest < end)
          {
            *dest++ = 'r';
            cp->output_buffer_fill++;
          }
        data++;
        len--;
        break;
      default:
        *dest++ = *data++;
        cp->output_buffer_fill++;
        len--;
        break;
      }
  assert(cp->output_buffer_fill <= cp->buffer_size);
  if (cp->output_pipe_fd == -1)
    {
      if ((cp->output_pipe_fd = open(cp->output_pipe_name, O_WRONLY | O_NONBLOCK, 0)) == -1)
        {
          if (errno != ENXIO)
            log("cmdpipe", LOG_PRIORITY_ERROR, "Failed to open output fifo %s: %s", cp->output_pipe_name, strerror(errno));
          return;
        }
      register_fd_handler(cp->output_pipe_fd, FD_CLASS_LOCAL, cp->name, OUTPUT_FD_GROUP, cp);
    }
  if (cp->output_buffer_fill > 0)
    {
      unblock_fd_writes(cp->output_pipe_fd);
    }
}

void
print_output_buffer(struct cmdpipe *cp, const char *fmt, ...)
{
  static const char null[] = "(null)";
  if (!fmt)
    fmt = null;
  va_list args;
  va_start(args, fmt);
  char *str;
  ssize_t len;
  len = vasprintf(&str, fmt, args);
  assert(len >= 0);
  write_output_buffer(cp, str, len);
  xfree(str);
  va_end(args);
}

/* Write a **DONE** iff we are not currently inside a parse for this
 * command.
 * 
 * The effect of this is that callbacks triggered while parsing don't
 * insert spurious **DONE**s, but do if they're triggered externally,
 * so the semantics on the perl side are intuitive
 */
void
write_output_done(struct cmdpipe *cp)
{
  if (!cp->currently_parsing)
    {
      print_output_buffer(cp, "**DONE**\n");
    }
}

static int
control_pipe_read_handler(int fd, void *data)
{
  struct cmdpipe *cp = data;

  {
    ssize_t len;
    if ((len = read(fd, cp->work_buffer, cp->buffer_size)) == -1)
      {
        log("cmdpipe", LOG_PRIORITY_ERROR, "Error reading from control pipe: %s", strerror(errno));
        return 0;
      }
    if ((len + cp->control_buffer_fill) > cp->buffer_size)
      {
        /* Throw away any surplus data */
        len = cp->buffer_size - cp->control_buffer_fill;
      }
    assert(len >= 0);
    memcpy(cp->control_buffer + cp->control_buffer_fill, cp->work_buffer, (size_t)len);
    cp->control_buffer_fill += len;
    assert(cp->control_buffer_fill <= cp->buffer_size);
  }

  char *c;
  while ((c = memchr(cp->control_buffer, '\n', cp->control_buffer_fill)))
    {
      char *src, *dest = cp->work_buffer;
      size_t destlen = 0;
      for (src = cp->control_buffer; src < min(c, cp->control_buffer + cp->control_buffer_fill); src++)
        switch (*src)
          {
          case '\\':
            src++;
            switch (*src)
              {
              case 'n':
                *dest++ = '\n';
                destlen++;
                break;
              case 'r':
                *dest++ = '\r';
                destlen++;
                break;
              case '0': case '1': case '2': case '3': case '4':
              case '5': case '6': case '7': case '8': case '9':
                {
                  char *end;
                  unsigned long int n = strtoul(src, &end, 0);
                  *dest++ = n;
                  destlen++;
                  src = end - 1;
                }
                break;
              default:
                *dest++ = *src;
                destlen++;
                break;
              }
            break;
          default:
            *dest++ = *src;
            destlen++;
            break;
          }
      /* We have space, so pack in a trailing zero. It saves effort in the test.* modules */
      assert(dest < (cp->work_buffer + cp->buffer_size));
      *dest++ = '\0';
      memmove(cp->control_buffer, c, cp->control_buffer_fill - (c - cp->control_buffer + 1));
      cp->control_buffer_fill -= c - cp->control_buffer + 1;

      cp->currently_parsing = 1;
      switch ((*cp->pf)(cp->work_buffer, destlen))
        {
        case 1:
          print_output_buffer(cp, "**DONE**\n");
          break;
        case -1:
          print_output_buffer(cp, "(segfault)\n");
          break;
        case 0:
        default:
          print_output_buffer(cp, "**ERROR**\n");
          break;
        }
      cp->currently_parsing = 0;
    }

  return 1;
}

static void
control_pipe_close_handler(int fd unused_attribute, void *data)
{
  struct cmdpipe *cp = data;

  cp->control_pipe_fd = -1;
  if ((cp->control_pipe_fd = open(cp->control_pipe_name, O_RDONLY | O_NONBLOCK, 0)) == -1)
    {
      log("cmdpipe", LOG_PRIORITY_ERROR, "Failed to reopen control fifo %s, unloading: %s",
	  cp->control_pipe_name, strerror(errno));
      if (cp->ff)
	(*cp->ff)();
    }
  register_fd_handler(cp->control_pipe_fd, FD_CLASS_LOCAL, cp->name, CONTROL_FD_GROUP, cp);
}

void
close_cmdpipe(struct cmdpipe *cp)
{
  if (cp)
    {
      if (cp->control_pipe_fd != -1)
	{
	  deregister_fd_handler(cp->control_pipe_fd);
	  close(cp->control_pipe_fd);
	}
      if (cp->output_pipe_fd != -1)
	{
	  deregister_fd_handler(cp->output_pipe_fd);
          /* One last write to try and dequeue all we can, don't worry if it fails */
          if (cp->output_buffer[0] && (cp->output_pipe_fd >= 0))
            write(cp->output_pipe_fd, cp->output_buffer, cp->output_buffer_fill);
	  close(cp->output_pipe_fd);
	}
      xfree(cp->output_pipe_name);
      xfree(cp->control_pipe_name);
      xfree(cp->output_buffer);
      xfree(cp->control_buffer);
      xfree(cp->name);
      cp->output_pipe_name = (void *)0xdeadbeef;
      cp->control_pipe_name = (void *)0xdeadbeef;
      cp->output_buffer = (void *)0xdeadbeef;
      cp->control_buffer = (void *)0xdeadbeef;
      xfree(cp);
    }
}

struct cmdpipe *
setup_cmdpipe(struct module *m, const char *name, size_t bufsize, cmdpipe_parse_func *pf, cmdpipe_fail_func *ff)
{
  struct cmdpipe *cp;

  if (!name || !bufsize || !pf)
    return NULL;

  cp = xmalloc(sizeof(struct cmdpipe));

  cp->name = xstrdup(name);
  cp->buffer_size = bufsize;
  cp->output_buffer = xmalloc(bufsize);
  cp->output_buffer_fill = 0;
  cp->control_buffer = xmalloc(bufsize);
  cp->control_buffer_fill = 0;
  cp->work_buffer = xmalloc(bufsize);
  cp->control_pipe_fd = cp->output_pipe_fd = -1;
  cp->pf = pf;
  cp->ff = ff;
  cp->currently_parsing = 0;

  if (xasprintf(&cp->control_pipe_name, "%s%s-control.fifo", rundir, name) == -1)
    abort();
  if (xasprintf(&cp->output_pipe_name, "%s%s-output.data", rundir, name) == -1)
    abort();

  unlink(cp->control_pipe_name);
  if (mkfifo(cp->control_pipe_name, S_IRUSR | S_IWUSR) == -1)
    {
      log("cmdpipe", LOG_PRIORITY_ERROR, "Failed to create control fifo %s: %s", cp->control_pipe_name, strerror(errno));
      close_cmdpipe(cp);
      return NULL;
    }
  unlink(cp->output_pipe_name);

  if ((cp->control_pipe_fd = open(cp->control_pipe_name, O_RDONLY | O_NONBLOCK, 0)) == -1)
    {
      log("cmdpipe", LOG_PRIORITY_ERROR, "Failed to open control fifo %s: %s", cp->control_pipe_name, strerror(errno));
      close_cmdpipe(cp);
      return NULL;
    }
  if ((cp->output_pipe_fd = open(cp->output_pipe_name, O_CREAT | O_WRONLY, S_IRUSR | S_IWUSR)) == -1)
/*     if (errno != ENXIO) */
      {
	log("cmdpipe", LOG_PRIORITY_ERROR, "Failed to open output fifo %s: %s", cp->output_pipe_name, strerror(errno));
	close_cmdpipe(cp);
	return NULL;
      }

  register_fd_handler_group(m, cp->name, CONTROL_FD_GROUP, &control_pipe_read_handler, NULL, &control_pipe_close_handler);
  register_fd_handler_group(m, cp->name, OUTPUT_FD_GROUP, NULL, &output_pipe_write_handler, &output_pipe_close_handler);

  register_fd_handler(cp->control_pipe_fd, FD_CLASS_LOCAL, cp->name, CONTROL_FD_GROUP, cp);
  if (cp->output_pipe_fd != -1)
    register_fd_handler(cp->output_pipe_fd, FD_CLASS_LOCAL, cp->name, OUTPUT_FD_GROUP, cp);

  return cp;
}

struct parsed_args args;

struct parsed_args *
parse_cmdpipe_data(char *data, size_t len)
{
  args.argc = 0;
  char *end = data + len;
  while (data && data < end)
    {
      /* Skip any leading spaces */
      while (*data == ' ')
        data++;

      if (*data == ':')
        {
          /* Nod in the general direction of rfc1459 for this syntactical form */
          args.argv[args.argc] = data + 1;
          args.argl[args.argc] = len - 1;
          args.argc++;
          return &args;
        }
      else if (*data == '"')
        {
          /* Grab the quoted word */
          char *readptr, *writeptr;
          args.argv[args.argc] = data;
          args.argl[args.argc] = 0;
          for (readptr = data + 1, writeptr = data; readptr < end && *readptr != '"'; readptr++)
            switch (*readptr)
              {
              case '\\':
                readptr++;
                switch (*readptr)
                  {
                    /* \n, \\, and \" are explicitly defined here */
                  case 'n':
                    *writeptr++ = '\n';
                    args.argl[args.argc]++;
                    break;
                  case '\\':
                    *writeptr++ = '\\';
                    args.argl[args.argc]++;
                    break;
                  case '"':
                    *writeptr++ = '"';
                    args.argl[args.argc]++;
                    break;
                  case '0': case '1': case '2': case '3': case '4':
                  case '5': case '6': case '7': case '8': case '9':
                    {
                      char *endptr;
                      long int val = strtol(readptr, &endptr, 0);
                      *writeptr++ = val;
                      args.argl[args.argc]++;
                      readptr = endptr;
                      break;
                    }
                  default:
                    /* Any other escaped character is passed through
                     * unchanged, but don't rely on that remaining
                     * true
                     */
                    *writeptr++ = *readptr;
                    args.argl[args.argc]++;
                    break;
                  }
                break;
              default:
                *writeptr++ = *readptr;
                args.argl[args.argc]++;
                break;
              }
          *writeptr = '\0';
          args.argc++;
          len -= (readptr - data) + 1;
          data = readptr + 1;
        }
      else
        {
          char *s;

          s = memchr(data, ' ', end - data);
          if (!s)
            {
              /* One word remains */
              args.argv[args.argc] = data;
              args.argl[args.argc] = len;
              args.argc++;
              return &args;
            }
          else
            {
              /* Grab the next word */
              args.argv[args.argc] = data;
              args.argl[args.argc] = s - data;
              args.argc++;
              *s++ = '\0';
              len -= s - data;
              data = s;
              continue;
            }
        }
    }
  return &args;
}
