/************************************************************************
 *   This is part of Dancer, a text relay daemon
 *   Copyright (C) 2002 Andrew Suffield <asuffield@users.sourceforge.net>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

#include <dancerd/config.h>
#include <dancerd/modules.h>
#include <dancerd/log.h>
#include <dancerd/paths.h>
#include <dancerd/xmalloc.h>
#include <dancerd/hashtable.h>
#include <dancerd/test/cmdpipe.h>

#include <dancerd/database/database.h>

#include <stddef.h>
#include <string.h>
#include <stdint.h>

SETUP_MODULE("test.database", MODVERSION);

IMPORTED_SYMBOL(database, driver_name);
IMPORTED_SYMBOL(database, open_database);
IMPORTED_SYMBOL(database, close_database);
IMPORTED_SYMBOL(database, get_value);
IMPORTED_SYMBOL(database, put_value);
IMPORTED_SYMBOL(database, delete_value);
IMPORTED_SYMBOL(database, check_sanity);
IMPORTED_SYMBOL(database, associate_database);
IMPORTED_SYMBOL(database, get_cursor);
IMPORTED_SYMBOL(database, free_cursor);
IMPORTED_SYMBOL(database, cursor_next);
IMPORTED_SYMBOL(database, cursor_prev);
IMPORTED_SYMBOL(database, cursor_get);
IMPORTED_SYMBOL(database, cursor_next_dup);

#include "symtab"

static struct cmdpipe *cp = NULL;
static struct hashtable *db_hash = NULL;
static struct hashtable *cursor_hash = NULL;
static int associated_return = 1;

static int
associated_key(void *key, uint32_t keylen,
               void *data unused_attribute, uint32_t datalen unused_attribute,
               void **newkey, uint32_t *newkeylen)
{
  void *keyend = strchr(key, '/');
  *newkey = key;
  *newkeylen = keyend ? (uint32_t) ((char *)keyend - (char *)key) : keylen;
  return associated_return;
}

static int
parse_command(char *data, size_t len unused_attribute)
{
  if (strncasecmp(data, "DRIVER_NAME", strlen("DRIVER_NAME")) == 0)
    {
      print_output_buffer(cp, SYM(database, driver_name)());
      print_output_buffer(cp, "\n");
      return 1;
    }
  else if (strncasecmp(data, "OPEN_DATABASE", strlen("OPEN_DATABASE")) == 0)
    {
      char *c = data + strlen("OPEN_DATABASE"), *d = NULL, *e = NULL;
      if (*c)
	{
	  c++;
	  d = strchr(c, '/');
	}
      if (!d)
	return -1;
      else
	{
          struct database *db;
	  *d++ = '\0';
          e = strchr(d, ' ');
          if (!e)
            {
              return -1;
            }
          *e++ = '\0';
          db = hashtable_lookup(db_hash, e);
          if (db)
            {
              return 0;
            }
	  db = SYM(database, open_database)(c, d, NULL, 0);
          hashtable_insert(db_hash, e, db);
	  if (db)
	    return 1;
	  else
	    return 0;
	}
    }
  else if (strncasecmp(data, "CLOSE_DATABASE", strlen("CLOSE_DATABASE")) == 0)
    {
      char *c = data + strlen("CLOSE_DATABASE");
      struct database *db;
      if (*c)
        c++;
      db = hashtable_lookup(db_hash, c);
      if (!db)
	return -1;
      else
	{
	  SYM(database, close_database)(db);
          hashtable_delete(db_hash, c);
	  return 1;
	}
    }
  else if (strncasecmp(data, "PUT_VALUE", strlen("PUT_VALUE")) == 0)
    {
      char *c = data + strlen("PUT_VALUE"), *d = NULL, *e = NULL;
      struct database *db;
      if (*c)
	{
	  c++;
	  d = strchr(c, ' ');
	}
      if (!d)
        {
          return -1;
        }
      else
        {
          *d++ = '\0';
          e = strchr(d, ' ');
        }

      db = hashtable_lookup(db_hash, c);
      if (!e || !db)
	return -1;
      else
	{
	  *e++ = '\0';
	  if (SYM(database, put_value)(db, NULL, d, strlen(d), e, strlen(e) + 1))
	    return 1;
	  else
	    return 0;
	}
    }
  else if (strncasecmp(data, "GET_VALUE", strlen("GET_VALUE")) == 0)
    {
      char *c = data + strlen("GET_VALUE"), *d = NULL;
      struct database *db;
      if (*c)
	{
	  c++;
	  d = strchr(c, ' ');
	}
      if (!d)
        {
          return -1;
        }
      else
        {
          *d++ = '\0';
          db = hashtable_lookup(db_hash, c);
        }
      if (!db)
	return -1;
      else
	{
	  void *db_data;
	  uint32_t db_len;
	  switch (SYM(database, get_value)(db, NULL, d, strlen(d), &db_data, &db_len))
	    {
            case 1:
	      print_output_buffer(cp, db_data);
	      print_output_buffer(cp, "\n");
              return 1;
            case 0:
              print_output_buffer(cp, "(none)\n");
	      return 1;
            default:
              return 0;
	    }
	}
    }
  else if (strncasecmp(data, "DELETE_VALUE", strlen("DELETE_VALUE")) == 0)
    {
      char *c = data + strlen("DELETE_VALUE"), *d = NULL;
      struct database *db;
      if (*c)
	{
	  c++;
	  d = strchr(c, ' ');
	}
      if (!d)
        {
          return -1;
        }
      else
        {
          *d++ = '\0';
          db = hashtable_lookup(db_hash, c);
        }
      if (!db)
	return -1;
      else
	{
	  if (SYM(database, delete_value)(db, NULL, d, strlen(d)))
            return 1;
	  else
	    return 0;
	}
    }
  else if (strncasecmp(data, "ASSOCIATE_DATABASE", strlen("ASSOCIATE_DATABASE")) == 0)
    {
      char *c = data + strlen("ASSOCIATE_DATABASE"), *d = NULL;
      struct database *primary, *secondary;
      if (*c)
	{
	  c++;
	  d = strchr(c, ' ');
	}
      if (!d)
        {
          return -1;
        }
      else
        {
          *d++ = '\0';
          primary = hashtable_lookup(db_hash, c);
          secondary = hashtable_lookup(db_hash, d);
        }
      if (!primary && !secondary)
	return -1;
      else
	{
	  if (SYM(database, associate_database)(primary, secondary, NULL, &associated_key))
            return 1;
	  else
	    return 0;
	}
    }
  else if (strncasecmp(data, "ASSOCIATED_RETURN", strlen("ASSOCIATED_RETURN")) == 0)
    {
      char *c = data + strlen("ASSOCIATED_RETURN");
      if (*c)
        c++;
      if (!c)
	return -1;
      else
	{
          associated_return = strtoul(c, NULL, 0);
	  return 1;
	}
    }
  else if (strncasecmp(data, "GET_CURSOR", strlen("GET_CURSOR")) == 0)
    {
      char *c = data + strlen("GET_CURSOR"), *d = NULL, *e = NULL;
      struct database *db;
      if (*c)
        {
          c++;
          d = strchr(c, ' ');
        }
      else
        {
          return -1;
        }

      if (!d || !*d)
        {
          *d++ = '\0';
          e = strchr(c, ' ');
          db = hashtable_lookup(db_hash, c);
        }
      else
        {
          return -1;
        }

      if (!e || !*e)
        {
          /* Get a cursor to the start of the database */
          void *db_data;
          uint32_t db_len;
          struct database_cursor *cursor = SYM(database, get_cursor)(db, NULL, NULL, 0, &db_data, &db_len);
          if (!cursor)
            {
              return 0;
            }
          hashtable_insert(cursor_hash, d, cursor);
          print_output_buffer(cp, data);
	  print_output_buffer(cp, "\n");
          return 1;
        }

      *e++ = '\0';
      {
        void *db_data;
        uint32_t db_len;
        struct database_cursor *cursor = SYM(database, get_cursor)(db, NULL, e, strlen(e) + 1, &db_data, &db_len);
        if (!cursor)
          {
            return 0;
          }
        hashtable_insert(cursor_hash, d, cursor);
        print_output_buffer(cp, data);
        print_output_buffer(cp, "\n");
      }
      return 1;
    }
  else if (strncasecmp(data, "FREE_CURSOR", strlen("FREE_CURSOR")) == 0)
    {
      char *c = data + strlen("FREE_CURSOR");
      struct database_cursor *cursor;
      if (*c)
        c++;
      cursor = hashtable_lookup(cursor_hash, c);
      if (!cursor)
	return -1;
      else
	{
	  SYM(database, free_cursor)(cursor);
          hashtable_delete(cursor_hash, c);
	  return 1;
	}
    }
  else if (strncasecmp(data, "CURSOR_NEXT", strlen("CURSOR_NEXT")) == 0)
    {
      char *c = data + strlen("CURSOR_NEXT");
      struct database_cursor *cursor;
      if (*c)
        c++;
      cursor = hashtable_lookup(cursor_hash, c);
      if (!cursor)
	return -1;
      else
	{
          void *db_data;
          uint32_t db_len;
	  switch (SYM(database, cursor_next)(cursor, &db_data, &db_len))
            {
            case 1:
              print_output_buffer(cp, db_data);
              print_output_buffer(cp, "\n");
              return 1;
            case 0:
              print_output_buffer(cp, "(none)\n");
              return 1;
            case -1:
            default:
              return 0;
            }
	}
    }
  else if (strncasecmp(data, "CURSOR_PREV", strlen("CURSOR_PREV")) == 0)
    {
      char *c = data + strlen("CURSOR_PREV");
      struct database_cursor *cursor;
      if (*c)
        c++;
      cursor = hashtable_lookup(cursor_hash, c);
      if (!cursor)
	return -1;
      else
	{
          void *db_data;
          uint32_t db_len;
	  switch (SYM(database, cursor_prev)(cursor, &db_data, &db_len))
            {
            case 1:
              print_output_buffer(cp, db_data);
              print_output_buffer(cp, "\n");
              return 1;
            case 0:
              print_output_buffer(cp, "(none)\n");
              return 1;
            case -1:
            default:
              return 0;
            }
	}
    }
  else if (strncasecmp(data, "CURSOR_GET", strlen("CURSOR_GET")) == 0)
    {
      char *c = data + strlen("CURSOR_GET");
      struct database_cursor *cursor;
      if (*c)
        c++;
      cursor = hashtable_lookup(cursor_hash, c);
      if (!cursor)
	return -1;
      else
	{
          void *db_data;
          uint32_t db_len;
	  switch (SYM(database, cursor_get)(cursor, &db_data, &db_len))
            {
            case 1:
              print_output_buffer(cp, db_data);
              print_output_buffer(cp, "\n");
              return 1;
            case 0:
              print_output_buffer(cp, "(none)\n");
              return 1;
            default:
            case -1:
              return 0;
            }
	}
    }
  else if (strncasecmp(data, "CURSOR_NEXT_DUP", strlen("CURSOR_NEXT_DUP")) == 0)
    {
      char *c = data + strlen("CURSOR_NEXT_DUP");
      struct database_cursor *cursor;
      if (*c)
        c++;
      cursor = hashtable_lookup(cursor_hash, c);
      if (!cursor)
	return -1;
      else
	{
          void *db_data;
          uint32_t db_len;
	  switch (SYM(database, cursor_next_dup)(cursor, &db_data, &db_len))
            {
            case 1:
              print_output_buffer(cp, db_data);
              print_output_buffer(cp, "\n");
              return 1;
            case 0:
              print_output_buffer(cp, "(none)\n");
              return 1;
            default:
            case -1:
              return 0;
            }
	}
    }
  else if (strncasecmp(data, "SANITY", strlen("SANITY")) == 0)
    {
      char *c = data + strlen("CLOSE_DATABASE");
      struct database *db = hashtable_lookup(db_hash, c);
      if (!db)
	return -1;
      else
        {
          if (SYM(database, check_sanity)(db))
            print_output_buffer(cp, "OK\n");
          else
            print_output_buffer(cp, "ERROR\n");
        }
      return 1;
    }
  else
    return 0;
}

static void
pipe_fail(void)
{
  unload_module(MYNAME);
}

int
MODULE_INIT(void *data unused_attribute)
{
  if (!(cp = setup_cmdpipe(MYHANDLE, MYNAME, 32768, &parse_command, &pipe_fail)))
    return 0;

  db_hash = new_hashtable(&hash_ascii, 17);
  cursor_hash = new_hashtable(&hash_ascii, 17);

  return MODULE_LOAD_OK;
}

int
MODULE_DEINIT(const char *new_version unused_attribute, module_init_func *init unused_attribute)
{
  free_hashtable(db_hash);
  free_hashtable(cursor_hash);
  if (cp)
    close_cmdpipe(cp);
  cp = NULL;
  return MODULE_UNLOAD_OK;
}
