/************************************************************************
 *   This is part of Dancer, a text relay daemon
 *   Copyright (C) 2002 Andrew Suffield <asuffield@users.sourceforge.net>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

#include <dancerd/config.h>
#include <dancerd/modules.h>
#include <dancerd/log.h>
#include <dancerd/xmalloc.h>
#include <dancerd/xstring.h>
#include <dancerd/llist.h>
#include <dancerd/test/cmdpipe.h>

#include <stddef.h>
#include <string.h>

SETUP_MODULE("test.main.llist", MODVERSION);

static struct cmdpipe *cp = NULL;

static void
output_llist_node(void *data, void *x unused_attribute)
{
  char *c = data;
  print_output_buffer(cp, c);
  print_output_buffer(cp, "\n");
}

static int
parse_command(char *data, size_t len unused_attribute)
{
  static struct llist *test_list = NULL;
  static struct llist_node *test_node = NULL;

  if (strncasecmp(data, "NEW_LLIST", strlen("NEW_LLIST")) == 0)
    {
      test_list = new_llist();
      return 1;
    }
  else if (strncasecmp(data, "DESTROY_LLIST", strlen("DESTROY_LLIST")) == 0)
    {
      if (!test_list)
	return -1;
      else
	{
	  destroy_llist(test_list, &xfree);
	  test_list = NULL;
	  return 1;
	}
    }
  else if (strncasecmp(data, "OUTPUT_LIST", strlen("OUTPUT_LIST")) == 0)
    {
      if (!test_list)
	return -1;
      else
	{
	  foreach_llist(test_list, output_llist_node, NULL);
	  return 1;
	}
    }
  else if (strncasecmp(data, "ADD_HEAD", strlen("ADD_HEAD")) == 0)
    {
      char *c = data + strlen("ADD_HEAD") + 1;
      if (!test_list)
	return -1;
      else
	{
	  add_llist_head(test_list, xstrdup(c));
	  return 1;
	}
    }
  else if (strncasecmp(data, "ADD_TAIL", strlen("ADD_TAIL")) == 0)
    {
      char *c = data + strlen("ADD_TAIL") + 1;
      if (!test_list)
	return -1;
      else
	{
	  add_llist_tail(test_list, xstrdup(c));
	  return 1;
	}
    }
  else if (strncasecmp(data, "INSERT_BEFORE", strlen("INSERT_BEFORE")) == 0)
    {
      char *c = data + strlen("INSERT_BEFORE") + 1;
      if (!test_node)
	return -1;
      else
	{
	  insert_llist_before(test_node, xstrdup(c));
	  return 1;
	}
    }
  else if (strncasecmp(data, "INSERT_AFTER", strlen("INSERT_AFTER")) == 0)
    {
      char *c = data + strlen("INSERT_AFTER") + 1;
      if (!test_node)
	return -1;
      else
	{
	  insert_llist_after(test_node, xstrdup(c));
	  return 1;
	}
    }
  else if (strncasecmp(data, "SEEK_HEAD", strlen("SEEK_HEAD")) == 0)
    {
      if (!test_list)
	return -1;
      else
	{
	  test_node = llist_head(test_list);
	  return 1;
	}
    }
  else if (strncasecmp(data, "SEEK_TAIL", strlen("SEEK_TAIL")) == 0)
    {
      if (!test_list)
	return -1;
      else
	{
	  test_node = llist_tail(test_list);
	  return 1;
	}
    }
  else if (strncasecmp(data, "NEXT", strlen("NEXT")) == 0)
    {
      if (!test_node)
	return -1;
      else
	{
	  test_node = llist_node_next(test_node);
	  return 1;
	}
    }
  else if (strncasecmp(data, "PREV", strlen("PREV")) == 0)
    {
      if (!test_node)
	return -1;
      else
	{
	  test_node = llist_node_prev(test_node);
	  return 1;
	}
    }
  else if (strncasecmp(data, "DELETE", strlen("DELETE")) == 0)
    {
      if (!test_node)
	return -1;
      else
	{
	  delete_llist_node(test_node);
	  test_node = NULL;
	  return 1;
	}
    }
  else if (strncasecmp(data, "OUTPUT", strlen("OUTPUT")) == 0)
    {
      if (!test_node)
	return -1;
      else
	{
	  print_output_buffer(cp, (char *)llist_node_data(test_node));
	  print_output_buffer(cp, "\n");
	  return 1;
	}
    }
  else if (strncasecmp(data, "SANITY", strlen("SANITY")) == 0)
    {
      if (llist_check_sanity(test_list))
	print_output_buffer(cp, "OK\n");
      else
	print_output_buffer(cp, "ERROR\n");
      return 1;
    }
  else
    return 0;
}

static void
pipe_fail(void)
{
  unload_module(MYNAME);
}

int
MODULE_INIT(void *data unused_attribute)
{
  if (!(cp = setup_cmdpipe(MYHANDLE, MYNAME, 32768, &parse_command, &pipe_fail)))
    return 0;

  return MODULE_LOAD_OK;
}

int
MODULE_DEINIT(const char *new_version unused_attribute, module_init_func *init unused_attribute)
{
  if (cp)
    close_cmdpipe(cp);
  cp = NULL;
  return MODULE_UNLOAD_OK;
}
