/************************************************************************
 *   This is part of Dancer, a text relay daemon
 *   Copyright (C) 2002 Andrew Suffield <asuffield@users.sourceforge.net>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

#include <dancerd/config.h>
#include <dancerd/modules.h>
#include <dancerd/log.h>
#include <dancerd/paths.h>
#include <dancerd/xmalloc.h>
#include <dancerd/xstring.h>
#include <dancerd/llist.h>
#include <dancerd/hashtable.h>
#include <dancerd/test/cmdpipe.h>

#include <dancerd/protocol/message.h>

#include <stddef.h>
#include <string.h>
#include <stdio.h>

SETUP_MODULE("test.message", MODVERSION);

static struct cmdpipe *cp = NULL;

IMPORTED_SYMBOL(protocol, register_new_session_handler);
IMPORTED_SYMBOL(message, dup_message);
IMPORTED_SYMBOL(message, free_message);
IMPORTED_SYMBOL(message, get_message_type);
IMPORTED_SYMBOL(message, register_message_type);
IMPORTED_SYMBOL(message, get_message_type_name);
IMPORTED_SYMBOL(message, accept_message);
IMPORTED_SYMBOL(message, request_all_from_session);
IMPORTED_SYMBOL(message, ignore_all_from_session);
IMPORTED_SYMBOL(message, request_type_from_session);
IMPORTED_SYMBOL(message, ignore_type_from_session);
IMPORTED_SYMBOL(message, request_type_from_all);
IMPORTED_SYMBOL(message, ignore_type_from_all);
IMPORTED_SYMBOL(protocol, shutdown_session);

#include "symtab"

static struct llist *messages = NULL;
static struct llist *sessions = NULL;
static struct hashtable *names = NULL;
static struct llist *names_list = NULL;
static int autoquit = 0;

static new_session_handler new_session;
static int
new_session(struct session *session, void *d unused_attribute, int outgoing)
{
  if (outgoing)
    print_output_buffer(cp, "NEW_OUTGOING_SESSION\n");
  else
    print_output_buffer(cp, "NEW_SESSION\n");
  write_output_done(cp);
  add_llist_tail(sessions, session);
  return 1;
}

static void
message_handler(struct session *session, const struct message *msg, void *data)
{
  print_output_buffer(cp, "MESSAGE_HANDLER: ");
  print_output_buffer(cp, data);
  print_output_buffer(cp, "\n");
  write_output_done(cp);
  if (autoquit)
    SYM(protocol, shutdown_session)(session);
  add_llist_tail(messages, SYM(message, dup_message)(msg));
}

static int
parse_command(char *data, size_t len unused_attribute)
{
  if (strncasecmp(data, "REGISTER_TYPE", strlen("REGISTER_TYPE")) == 0)
    {
      char *c = data + strlen("REGISTER_TYPE");
      int type;
      char buf[64];

      if (*c)
        c++;

      type = SYM(message, register_message_type)(c);
      snprintf(buf, 64, "%d\n", type);
      print_output_buffer(cp, buf);
      return 1;
    }
  else if (strncasecmp(data, "GET_TYPE_NAME", strlen("GET_TYPE_NAME")) == 0)
    {
      char *c = data + strlen("GET_TYPE_NAME");
      const char *name;
      int type;

      if (*c)
        c++;

      type = strtoul(c, NULL, 0);
      name = SYM(message, get_message_type_name)(type);
      if (name)
        print_output_buffer(cp, name);
      else
        print_output_buffer(cp, "(null)");
      print_output_buffer(cp, "\n");
      return 1;
    }
  else if (strncasecmp(data, "GET_TYPE", strlen("GET_TYPE")) == 0)
    {
      char *c = data + strlen("GET_TYPE");
      int type;
      char buf[64];

      if (*c)
        c++;

      type = SYM(message, get_message_type)(c);
      snprintf(buf, 64, "%d\n", type);
      print_output_buffer(cp, buf);
      return 1;
    }
  else if (strncasecmp(data, "REQUEST_TYPE_FROM_ALL", strlen("REQUEST_TYPE_FROM_ALL")) == 0)
    {
      char *c = data + strlen("REQUEST_TYPE_FROM_ALL"), *d = NULL;
      int type;
      char *name;

      /* First word is the name to display when catching messages on this one
       * Second word is the message type, by name
       */
      if (*c)
        {
          c++;
          d = strchr(c, ' ');
        }

      if (d && *d)
        {
          *d++ = '\0';
          type = SYM(message, get_message_type)(d);
        }
      else
        {
          return -1;
        }

      if (!(name = hashtable_lookup(names, c)))
        {
          name = xstrdup(c);
          hashtable_insert(names, name, name);
          add_llist_tail(names_list, name);
        }

      SYM(message, request_type_from_all)(MYHANDLE, type, &message_handler, name);
      return 1;
    }
  else if (strncasecmp(data, "IGNORE_TYPE_FROM_ALL", strlen("IGNORE_TYPE_FROM_ALL")) == 0)
    {
      char *c = data + strlen("IGNORE_TYPE_FROM_ALL"), *d = NULL;
      int type;
      char *name;

      /* First word is the name to display when catching messages on this one
       * Second word is the message type, by name
       */
      if (*c)
        {
          c++;
          d = strchr(c, ' ');
        }

      if (d && *d)
        {
          *d++ = '\0';
          type = SYM(message, get_message_type)(d);
        }
      else
        {
          return -1;
        }

      if (!(name = hashtable_lookup(names, c)))
        {
          name = xstrdup(c);
          hashtable_insert(names, name, name);
          add_llist_tail(names_list, name);
        }
      SYM(message, ignore_type_from_all)(MYHANDLE, type, &message_handler, name);
      return 1;
    }
  else if (strncasecmp(data, "REQUEST_TYPE_FROM_SESSION", strlen("REQUEST_TYPE_FROM_SESSION")) == 0)
    {
      char *c = data + strlen("REQUEST_TYPE_FROM_SESSION"), *d = NULL, *e = NULL;
      int type;
      char *name;
      struct session *session;

      /* First word is the message type, by name
       * Second word is the session, by index
       * Third word is the name to display when catching messages on this one
       */
      if (*c)
        {
          c++;
          d = strchr(c, ' ');
        }

      if (d && *d)
        {
          *d++ = '\0';
          type = SYM(message, get_message_type)(c);
          e = strchr(d, ' ');
        }
      else
        {
          return -1;
        }

      if (e && *e)
        {
          int i;
          *e++ = '\0';
          struct llist_node *n;
          i = strtol(d, NULL, 0);
          n = llist_head(sessions);
          while (i--)
            {
              if (!n)
                {
                  return -1;
                        }
              n = llist_node_next(n);
            }
          if (!n)
            {
              return -1;
                }
          session = llist_node_data(n);
        }
      else
        {
          return -1;
        }

      if (!(name = hashtable_lookup(names, e)))
        {
          name = xstrdup(e);
          hashtable_insert(names, name, name);
          add_llist_tail(names_list, name);
        }

      SYM(message, request_type_from_session)(MYHANDLE, session, type, &message_handler, name);
      return 1;
    }
  else if (strncasecmp(data, "IGNORE_TYPE_FROM_SESSION", strlen("IGNORE_TYPE_FROM_SESSION")) == 0)
    {
      char *c = data + strlen("IGNORE_TYPE_FROM_SESSION"), *d = NULL, *e = NULL;
      int type;
      char *name;
      struct session *session;

      /* First word is the message type, by name
       * Second word is the session, by index
       * Third word is the name to display when catching messages on this one
       */
      if (*c)
        {
          c++;
          d = strchr(c, ' ');
        }

      if (d && *d)
        {
          *d++ = '\0';
          type = SYM(message, get_message_type)(c);
          e = strchr(d, ' ');
        }
      else
        {
          return -1;
        }

      if (e && *e)
        {
          int i;
          *e++ = '\0';
          struct llist_node *n;
          i = strtol(d, NULL, 0);
          n = llist_head(sessions);
          while (i--)
            {
              if (!n)
                {
                  return -1;
                        }
              n = llist_node_next(n);
            }
          if (!n)
            {
              return -1;
                }
          session = llist_node_data(n);
        }
      else
        {
          return -1;
        }

      if (!(name = hashtable_lookup(names, e)))
        {
          name = xstrdup(e);
          hashtable_insert(names, name, name);
          add_llist_tail(names_list, name);
        }

      SYM(message, ignore_type_from_session)(MYHANDLE, session, type, &message_handler, name);
      return 1;
    }
  else if (strncasecmp(data, "REQUEST_ALL_FROM_SESSION", strlen("REQUEST_ALL_FROM_SESSION")) == 0)
    {
      char *c = data + strlen("REQUEST_ALL_FROM_SESSION"), *d = NULL;
      char *name;
      struct session *session;

      /* First word is the session, by index
       * Second word is the name to display when catching messages on this one
       */
      if (*c)
        {
          c++;
          d = strchr(c, ' ');
        }
      else
        {
          return -1;
        }

      if (d && *d)
        {
          int i;
          struct llist_node *n;
          *d++ = '\0';
          i = strtol(c, NULL, 0);
          n = llist_head(sessions);
          while (i--)
            {
              if (!n)
                {
                  return -1;
                        }
              n = llist_node_next(n);
            }
          if (!n)
            {
              return -1;
                }
          session = llist_node_data(n);
        }
      else
        {
          return -1;
        }

      if (!(name = hashtable_lookup(names, d)))
        {
          name = xstrdup(d);
          hashtable_insert(names, name, name);
          add_llist_tail(names_list, name);
        }

      SYM(message, request_all_from_session)(MYHANDLE, session, &message_handler, name);
      return 1;
    }
  else if (strncasecmp(data, "IGNORE_ALL_FROM_SESSION", strlen("IGNORE_ALL_FROM_SESSION")) == 0)
    {
      char *c = data + strlen("IGNORE_ALL_FROM_SESSION"), *d = NULL;
      char *name;
      struct session *session;

      /* First word is the session, by index
       * Second word is the name to display when catching messages on this one
       */
      if (*c)
        {
          c++;
          d = strchr(c, ' ');
        }
      else
        {
          return -1;
        }

      if (d && *d)
        {
          int i;
          struct llist_node *n;
          *d++ = '\0';
          i = strtol(c, NULL, 0);
          n = llist_head(sessions);
          while (i--)
            {
              if (!n)
                {
                  return -1;
                        }
              n = llist_node_next(n);
            }
          if (!n)
            {
              return -1;
                }
          session = llist_node_data(n);
        }
      else
        {
          return -1;
        }

      if (!(name = hashtable_lookup(names, d)))
        {
          name = xstrdup(d);
          hashtable_insert(names, name, name);
          add_llist_tail(names_list, name);
        }

      SYM(message, ignore_all_from_session)(MYHANDLE, session, &message_handler, name);
      return 1;
    }
  else if (strncasecmp(data, "DROP_SESSION", strlen("DROP_SESSION")) == 0)
    {
      char *c = data + strlen("DROP_SESSION");
      int i;
      struct llist_node *n;

      if (*c)
        c++;

      i = strtoul(c, NULL, 0);
      n = llist_head(sessions);
      while (i--)
        {
          if (!n)
            {
              return -1;
                }
          n = llist_node_next(n);
        }
      if (!n)
        {
          return -1;
        }
      delete_llist_node(n);
      return 1;
    }
  else if (strncasecmp(data, "POP_MESSAGE", strlen("POP_MESSAGE")) == 0)
    {
      if (!llist_head(messages))
	return -1;
      else
	{
          struct message *msg = llist_node_data(llist_head(messages));
          delete_llist_node(llist_head(messages));
          SYM(message, free_message)(msg);
	  return 1;
	}
    }
  else if (strncasecmp(data, "PEEK_MESSAGE", strlen("PEEK_MESSAGE")) == 0)
    {
      if (llist_head(messages))
        {
          print_output_buffer(cp, "YES\n");
        }
      else
        {
          print_output_buffer(cp, "NO\n");
        }
      return 1;
    }
  else if (strncasecmp(data, "SHIFT_MSG", strlen("SHIFT_MSG")) == 0)
    {
      if (!llist_head(messages))
	return -1;
      else
	{
          char buf[128];
          int i;
          struct message *msg = llist_node_data(llist_head(messages));
          if (!msg)
            {
              print_output_buffer(cp, "(null)\n");
              return 1;
            }
          print_output_buffer(cp, "type:");
          print_output_buffer(cp, SYM(message, get_message_type_name)(msg->type));
          print_output_buffer(cp, " origin:");
          print_output_buffer(cp, msg->origin);
          print_output_buffer(cp, " dest:");
          print_output_buffer(cp, msg->dest);

          print_output_buffer(cp, "\nstr:");
          snprintf(buf, 128, "%d ", msg->strc);
          print_output_buffer(cp, buf);
          for (i = 0; i < (msg->strc - 1); i++)
            {
              print_output_buffer(cp, msg->strv[i]);
              print_output_buffer(cp, ",");
            }
          if (msg->strc)
            print_output_buffer(cp, msg->strv[i]);
          print_output_buffer(cp, "\n.\n");

          print_output_buffer(cp, "uint32:");
          snprintf(buf, 128, "%d ", msg->uint32c);
          print_output_buffer(cp, buf);
          for (i = 0; i < (msg->uint32c - 1); i++)
            {
              snprintf(buf, 128, "%u,", msg->uint32v[i]);
              print_output_buffer(cp, buf);
            }
          if (msg->uint32c)
            {
              snprintf(buf, 128, "%u", msg->uint32v[i]);
              print_output_buffer(cp, buf);
            }
          print_output_buffer(cp, "\n.\n");

          print_output_buffer(cp, "sint32:");
          snprintf(buf, 128, "%d ", msg->sint32c);
          print_output_buffer(cp, buf);
          for (i = 0; i < (msg->sint32c - 1); i++)
            {
              snprintf(buf, 128, "%d,", msg->sint32v[i]);
              print_output_buffer(cp, buf);
            }
          if (msg->sint32c)
            {
              snprintf(buf, 128, "%d", msg->sint32v[i]);
              print_output_buffer(cp, buf);
            }
          print_output_buffer(cp, "\n.\n");

          delete_llist_node(llist_head(messages));
          SYM(message, free_message)(msg);

          return 1;
        }
    }
  else if (strncasecmp(data, "GET_MSG_TYPE", strlen("GET_MSG_TYPE")) == 0)
    {
      if (!llist_head(messages))
	return -1;
      else
	{
          struct message *msg = llist_node_data(llist_head(messages));
          if (msg)
            {
              char buf[128];
              snprintf(buf, 128, "%s\n", SYM(message, get_message_type_name)(msg->type));
              print_output_buffer(cp, buf);
            }
          else
            {
              print_output_buffer(cp, "(null)\n");
            }
	  return 1;
	}
    }
  else if (strncasecmp(data, "GET_ORIGIN", strlen("GET_ORIGIN")) == 0)
    {
      if (!llist_head(messages))
	return -1;
      else
	{
          struct message *msg = llist_node_data(llist_head(messages));
          print_output_buffer(cp, msg ? msg->origin : "(null)");
	  print_output_buffer(cp, "\n");
          return 1;
	}
    }
  else if (strncasecmp(data, "GET_DEST", strlen("GET_DEST")) == 0)
    {
      if (!llist_head(messages))
	return -1;
      else
	{
          struct message *msg = llist_node_data(llist_head(messages));
          print_output_buffer(cp, msg ? msg->dest : "(null)");
	  print_output_buffer(cp, "\n");
          return 1;
	}
    }
  else if (strncasecmp(data, "GET_STRC", strlen("GET_STRC")) == 0)
    {
      if (!llist_head(messages))
	return -1;
      else
	{
          struct message *msg = llist_node_data(llist_head(messages));
          char buf[64];
          snprintf(buf, 64, "%d\n", msg ? msg->strc : 0);
          print_output_buffer(cp, buf);
	  return 1;
	}
    }
  else if (strncasecmp(data, "GET_UINT32C", strlen("GET_UINT32C")) == 0)
    {
      if (!llist_head(messages))
	return -1;
      else
	{
          struct message *msg = llist_node_data(llist_head(messages));
          char buf[64];
          snprintf(buf, 64, "%d\n", msg ? msg->uint32c : 0);
          print_output_buffer(cp, buf);
	  return 1;
	}
    }
  else if (strncasecmp(data, "GET_SINT32C", strlen("GET_SINT32C")) == 0)
    {
      if (!llist_head(messages))
	return -1;
      else
	{
          struct message *msg = llist_node_data(llist_head(messages));
          char buf[64];
          snprintf(buf, 64, "%d\n", msg ? msg->sint32c : 0);
          print_output_buffer(cp, buf);
	  return 1;
	}
    }
  else if (strncasecmp(data, "GET_STRV", strlen("GET_STRV")) == 0)
    {
      char *c = data + strlen("GET_STRV");
      if (*c)
        c++;
      if (!llist_head(messages))
	return -1;
      else
	{
          int i = strtoul(c, NULL, 0);
          struct message *msg = llist_node_data(llist_head(messages));
          print_output_buffer(cp, msg->strv[i]);
	  print_output_buffer(cp, "\n");
	  return 1;
	}
    }
  else if (strncasecmp(data, "GET_UINT32V", strlen("GET_UINT32V")) == 0)
    {
      char *c = data + strlen("GET_UINT32V");
      if (*c)
        c++;
      if (!llist_head(messages))
	return -1;
      else
	{
          int i = strtoul(c, NULL, 0);
          struct message *msg = llist_node_data(llist_head(messages));
          char buf[64];
          snprintf(buf, 64, "%lu\n", (long int) msg->uint32v[i]);
          print_output_buffer(cp, buf);
	  return 1;
	}
    }
  else if (strncasecmp(data, "GET_SINT32V", strlen("GET_SINT32V")) == 0)
    {
      char *c = data + strlen("GET_SINT32V");
      if (*c)
        c++;
      if (!llist_head(messages))
	return -1;
      else
	{
          int i = strtoul(c, NULL, 0);
          struct message *msg = llist_node_data(llist_head(messages));
          char buf[64];
          snprintf(buf, 64, "%ld\n", (long int) msg->sint32v[i]);
          print_output_buffer(cp, buf);
	  return 1;
	}
    }
  else if (strncasecmp(data, "AUTOQUIT_ON", strlen("AUTOQUIT_ON")) == 0)
    {
      autoquit = 1;
      return 1;
    }
  else if (strncasecmp(data, "AUTOQUIT_OFF", strlen("AUTOQUIT_OFF")) == 0)
    {
      autoquit = 0;
      return 1;
    }
  else
    return 0;
}

static void
pipe_fail(void)
{
  unload_module(MYNAME);
}

int
MODULE_INIT(void *data unused_attribute)
{
  if (!(cp = setup_cmdpipe(MYHANDLE, MYNAME, 32768, &parse_command, &pipe_fail)))
    return MODULE_LOAD_FAILED;

  SYM(protocol, register_new_session_handler)(MYHANDLE, &new_session, NULL);
  messages = new_llist();
  sessions = new_llist();
  names = new_hashtable(&hash_ascii, 17);
  names_list = new_llist();

  return MODULE_LOAD_OK;
}

int
MODULE_DEINIT(const char *new_version unused_attribute, module_init_func *init unused_attribute)
{
  free_hashtable(names);
  free_llist(sessions);
  destroy_llist(names_list, xfree);
  destroy_llist(messages, SYM(message, free_message));
  if (cp)
    close_cmdpipe(cp);
  cp = NULL;
  return MODULE_UNLOAD_OK;
}
