/************************************************************************
 *   This is part of Dancer, a text relay daemon
 *   Copyright (C) 2002 Andrew Suffield <asuffield@users.sourceforge.net>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

#include <dancerd/config.h>
#include <dancerd/modules.h>
#include <dancerd/log.h>
#include <dancerd/paths.h>
#include <dancerd/xmalloc.h>
#include <dancerd/xstring.h>
#include <dancerd/hashtable.h>
#include <dancerd/llist.h>
#include <dancerd/test/cmdpipe.h>

#include <dancerd/protocol/protocol.h>

#include <stddef.h>
#include <string.h>
#include <stdio.h>

SETUP_MODULE("test.protocol", MODVERSION);

IMPORTED_SYMBOL(message, build_message);
IMPORTED_SYMBOL(protocol, register_protocol);
IMPORTED_SYMBOL(protocol, unregister_protocol);
IMPORTED_SYMBOL(protocol, shutdown_session);
IMPORTED_SYMBOL(protocol, read_message);
IMPORTED_SYMBOL(protocol, write_data);
IMPORTED_SYMBOL(protocol, data_read);
IMPORTED_SYMBOL(protocol, send_message);
IMPORTED_SYMBOL(protocol, register_new_session_handler);
IMPORTED_SYMBOL(protocol, unregister_new_session_handler);
IMPORTED_SYMBOL(protocol, register_attribute);
IMPORTED_SYMBOL(protocol, get_attribute);
IMPORTED_SYMBOL(protocol, set_attribute);
IMPORTED_SYMBOL(message, get_message_type);
IMPORTED_SYMBOL(message, get_message_type_name);
IMPORTED_SYMBOL(message, dup_message);
IMPORTED_SYMBOL(message, free_message);

#include "symtab"

static struct cmdpipe *cp = NULL;

static struct protocol *my_protocol = NULL;
static int identify_result = 0;
static int name_attribute = -1;

static struct hashtable *sessions = NULL;
static struct llist *new_sessions = NULL;
static struct llist *messages = NULL;

static inline struct message *
parse_message_string(const char *c, int type)
{
  int len = c ? strlen(c) : 0;
  char *d, *tmp;
  int i, j;
  char *origin = NULL, *dest = NULL;
  int strc = 0;
  char *strv[8];
  int uint32c = 0;
  uint32_t uint32v[8];
  int sint32c = 0;
  int32_t sint32v[8];
  struct message *msg;
  /* Remove a trailing \n if one is present */
  if (c && (c[len - 1] == '\n'))
    len--;
  for (i = 0; i < len; i++)
    {
      switch (c[i])
        {
        case ' ':
          break;
        case '<':
          i++;
          d = memchr(&c[i], ' ', len - i);
          if (d)
            {
              tmp = xmalloc(1 + (d - &c[i]));
              memcpy(tmp, &c[i], (d - &c[i]));
              tmp[d - &c[i]] = '\0';
              i += d - &c[i];
            }
          else
            {
              tmp = xstrndup(&c[i], len - i);
              i = len;
            }
          xfree(origin);
          origin = tmp;
          break;
        case '>':
          i++;
          d = memchr(&c[i], ' ', len - i);
          if (d)
            {
              tmp = xmalloc(1 + (d - &c[i]));
              memcpy(tmp, &c[i], (d - &c[i]));
              tmp[d - &c[i]] = '\0';
              i += d - &c[i];
            }
          else
            {
              tmp = xstrndup(&c[i], len - i);
              i = len;
            }
          xfree(dest);
          dest = tmp;
          break;
        case '"':
          for (j = i + 1; j < len; j++)
            switch (c[j])
              {
              case '\\':
                j++;
                break;
              case '"':
                strv[strc] = xmalloc(1 + j - (i + 1));
                memcpy(strv[strc], &c[i + 1], j - (i + 1));
                strv[strc][j - (i + 1)] = '\0';
                strc++;
                i = j;
                /* Cheap way to break out of the for loop */
                j = len;
                break;
              }
          break;
        case '(':
          if (((len - i) >= 6) && (memcmp(&c[i], "(null)", 6) == 0))
            {
              i += 6;
              strv[strc++] = NULL;
            }
          break;
        case '0': case '1': case '2': case '3': case '4':
        case '5': case '6': case '7': case '8': case '9':
          uint32v[uint32c++] = strtoul(&c[i], &d, 0);
          i += d - &c[i];
          break;
        case '+': case '-':
          sint32v[sint32c++] = strtol(&c[i], &d, 0);
          i += d - &c[i];
          break;
        default:
          break;
        }
    }
  msg = SYM(message, build_message)(type, origin, dest,
                                    strc, (const char **)strv, uint32c, uint32v, sint32c, sint32v);
  for (i = 0; i < strc; i++)
    xfree(strv[i]);
  return msg;
}

static protocol_identify_func identify;
static int
identify(struct dbuffer *buf)
{
  char *c = xmalloc(dbuffer_len(buf) + 1);
  dbuffer_get(buf, c, dbuffer_len(buf));
  c[dbuffer_len(buf)] = '\0';
  print_output_buffer(cp, "IDENTIFY: ");
  print_output_buffer(cp, c);
  print_output_buffer(cp, "\n");
  write_output_done(cp);
  xfree(c);
  log(MYNAME, LOG_PRIORITY_DEBUG, "Returning %d for identify", identify_result);
  return identify_result;
}

static protocol_init_session_func init_session;
static void *
init_session(struct session *session unused_attribute, void *d unused_attribute)
{
  print_output_buffer(cp, "INIT_SESSION\n");
  write_output_done(cp);
  return NULL;
}

static protocol_connection_ready_func connection_ready;
static void *
connection_ready(struct session *session unused_attribute)
{
  print_output_buffer(cp, "CONNECTION_READY\n");
  write_output_done(cp);
  return NULL;
}

static protocol_read_data_func read_data;
static void
read_data(struct session *session unused_attribute, void *d unused_attribute,
          struct dbuffer *buf, const void *data, size_t len)
{
  char *c;
  if (data)
    {
      c = xmalloc(len + 1);
      memcpy(c, data, len);
      c[len] = '\0';
    }
  else
    {
      len = dbuffer_len(buf);
      c = xmalloc(len + 1);
      dbuffer_extract(buf, c, len);
      c[len] = '\0';
    }
  print_output_buffer(cp, "READ_DATA: ");
  print_output_buffer(cp, c);
  print_output_buffer(cp, "\n");
  write_output_done(cp);
  xfree(c);
}

static protocol_send_message_func send_message;
static void
send_message(struct session *session unused_attribute, void *d unused_attribute,
             const struct message *msg)
{
  print_output_buffer(cp, "SEND_MESSAGE\n");
  write_output_done(cp);
  add_llist_tail(messages, SYM(message, dup_message)(msg));
}

static protocol_close_session_func close_session;
static void
close_session(struct session *session, void *d unused_attribute)
{
  char *name = SYM(protocol, get_attribute)(session, name_attribute);
  print_output_buffer(cp, "CLOSE_SESSION\n");
  write_output_done(cp);
  if (name)
    hashtable_delete(sessions, name);
  delete_llist_data(new_sessions, session);
  xfree(name);
}

static new_session_handler new_session;
static int
new_session(struct session *session, void *d unused_attribute, int outgoing)
{
  if (outgoing)
    print_output_buffer(cp, "NEW_OUTGOING_SESSION\n");
  else
    {
      print_output_buffer(cp, "NEW_SESSION\n");
      /* We add outgoing sessions in the BEGIN_CONNECTION handler in
       * parse_command(), below. Here we just add incoming ones, since
       * we don't know their name yet
       */
      add_llist_tail(new_sessions, session);
    }
  write_output_done(cp);
  return 1;
}

static int
parse_command(char *data, size_t len)
{
  struct parsed_args *args = parse_cmdpipe_data(data, len);
  char **argv = args->argv;
  size_t *argl = args->argl;
  int argc = args->argc;

  if (strcmp(argv[0], "REGISTER_PROTOCOL") == 0)
    {
      if (my_protocol)
	return 0;
      else
	{
	  my_protocol = SYM(protocol, register_protocol)("test", MYHANDLE,
                                                         &identify, &init_session, &connection_ready,
                                                         &read_data, &send_message,
                                                         &close_session);
          return my_protocol != NULL;
	}
    }
  else if (strcmp(argv[0], "UNREGISTER_PROTOCOL") == 0)
    {
      if (!my_protocol)
        return -1;
      else
	{
	  SYM(protocol, unregister_protocol)(my_protocol);
	  my_protocol = NULL;
          return 1;
	}
    }
  else if (strcmp(argv[0], "BEGIN_CONNECTION") == 0)
    {
      if (argc < 5)
        return -1;
      else
	{
          struct session *session;
          if ((session = SYM(protocol, begin_connection)(argv[1], argv[2], argv[3], argv[4], NULL, 0)))
            {
              hashtable_insert(sessions, argv[4], session);
              return 1;
            }
          else
            return 0;
	}
    }
  else if (strcmp(argv[0], "IDENTIFY_RESULT") == 0)
    {
      if (!my_protocol || argc < 1)
        return -1;
      else
	{
          identify_result = strtoul(argv[1], NULL, 0);
	  return 1;
	}
    }
  else if (strcmp(argv[0], "NAME_SESSION") == 0)
    {
      if (!llist_head(new_sessions) || argc < 1)
        return -1;
      else
	{
          struct session *session = llist_node_data(llist_head(new_sessions));
          if (hashtable_lookup(sessions, argv[1]))
            return 0;
          delete_llist_node(llist_head(new_sessions));
          hashtable_insert(sessions, argv[1], session);
          SYM(protocol, set_attribute)(session, name_attribute, xstrdup(argv[1]));
	  return 1;
	}
    }
  else if (strcmp(argv[0], "WRITE_DATA") == 0)
    {
      if (argc < 2)
	return -1;
      else
	{
          struct session *session = hashtable_lookup(sessions, argv[1]);
          if (!session)
            return -1;
          SYM(protocol, write_data)(session, argv[2], argl[2]);
	  return 1;
	}
    }
  else if (strcmp(argv[0], "READ_DATA") == 0)
    {
      if (argc < 2)
	return -1;
      else
	{
          struct session *session = hashtable_lookup(sessions, argv[1]);
          if (!session)
            return -1;
          SYM(protocol, data_read)(session, argv[2], argl[2]);
	  return 1;
	}
    }
  else if (strcmp(argv[0], "READ_MESSAGE") == 0)
    {
      int type;
      struct session *session;

      if (argc < 1)
        return -1;

      session = hashtable_lookup(sessions, argv[1]);
      if (!session)
        return -1;

      if (argc == 1)
        {
          /* Only one word -> send a NULL message */
          SYM(protocol, read_message)(session, NULL);
          return 1;
        }

      if (argc < 3)
        return -1;

      type = SYM(message, get_message_type)(argv[2]);

      SYM(protocol, read_message)(session, parse_message_string(argv[3], type));
      return 1;
    }
  else if (strcmp(argv[0], "SEND_MESSAGE") == 0)
    {
      int type;
      struct session *session;

      session = hashtable_lookup(sessions, argv[1]);
      if (!session)
        return -1;

      if (argc == 1)
        {
          /* Only one word -> send a NULL message */
          SYM(protocol, read_message)(session, NULL);
          return 1;
        }

      if (argc < 3)
        return -1;

      type = SYM(message, get_message_type)(argv[2]);

      SYM(protocol, send_message)(session, parse_message_string(argv[3], type));
      return 1;
    }
  else if (strcmp(argv[0], "POP_MESSAGE") == 0)
    {
      if (!my_protocol || !llist_head(messages))
	return -1;
      else
	{
          struct message *msg = llist_node_data(llist_head(messages));
          delete_llist_node(llist_head(messages));
          SYM(message, free_message)(msg);
	  return 1;
	}
    }
  else if (strcmp(argv[0], "PEEK_MESSAGE") == 0)
    {
      if (llist_head(messages))
        {
          print_output_buffer(cp, "YES\n");
        }
      else
        {
          print_output_buffer(cp, "NO\n");
        }
      return 1;
    }
  else if (strcmp(argv[0], "GET_MSG_TYPE") == 0)
    {
      if (!llist_head(messages))
	return -1;
      else
	{
          struct message *msg = llist_node_data(llist_head(messages));
          char buf[128];
          snprintf(buf, 128, "%s\n", SYM(message, get_message_type_name)(msg->type));
          print_output_buffer(cp, buf);
	  return 1;
	}
    }
  else if (strcmp(argv[0], "SHIFT_MSG") == 0)
    {
      if (!llist_head(messages))
	return 0;
      else
	{
          char buf[128];
          int i;
          struct message *msg = llist_node_data(llist_head(messages));
          if (!msg)
            {
              print_output_buffer(cp, "(null)\n");
              return 1;
            }
          print_output_buffer(cp, "type:");
          print_output_buffer(cp, SYM(message, get_message_type_name)(msg->type));
          print_output_buffer(cp, " origin:");
          print_output_buffer(cp, msg->origin);
          print_output_buffer(cp, " dest:");
          print_output_buffer(cp, msg->dest);

          print_output_buffer(cp, "\nstr:");
          snprintf(buf, 128, "%d ", msg->strc);
          print_output_buffer(cp, buf);
          for (i = 0; i < (msg->strc - 1); i++)
            {
              print_output_buffer(cp, msg->strv[i]);
              print_output_buffer(cp, ",");
            }
          if (msg->strc)
            print_output_buffer(cp, msg->strv[i]);
          print_output_buffer(cp, "\n.\n");

          print_output_buffer(cp, "uint32:");
          snprintf(buf, 128, "%d ", msg->uint32c);
          print_output_buffer(cp, buf);
          for (i = 0; i < (msg->uint32c - 1); i++)
            {
              snprintf(buf, 128, "%u,", msg->uint32v[i]);
              print_output_buffer(cp, buf);
            }
          if (msg->uint32c)
            {
              snprintf(buf, 128, "%u", msg->uint32v[i]);
              print_output_buffer(cp, buf);
            }
          print_output_buffer(cp, "\n.\n");

          print_output_buffer(cp, "sint32:");
          snprintf(buf, 128, "%d ", msg->sint32c);
          print_output_buffer(cp, buf);
          for (i = 0; i < (msg->sint32c - 1); i++)
            {
              snprintf(buf, 128, "%d,", msg->sint32v[i]);
              print_output_buffer(cp, buf);
            }
          if (msg->sint32c)
            {
              snprintf(buf, 128, "%d", msg->sint32v[i]);
              print_output_buffer(cp, buf);
            }
          print_output_buffer(cp, "\n.\n");

          delete_llist_node(llist_head(messages));
          SYM(message, free_message)(msg);

          return 1;
        }
    }
  else if (strcmp(argv[0], "GET_ORIGIN") == 0)
    {
      if (!llist_head(messages))
	return -1;
      else
	{
          struct message *msg = llist_node_data(llist_head(messages));
          print_output_buffer(cp, msg->origin);
	  print_output_buffer(cp, "\n");
          return 1;
	}
    }
  else if (strcmp(argv[0], "GET_DEST") == 0)
    {
      if (!llist_head(messages))
	return -1;
      else
	{
          struct message *msg = llist_node_data(llist_head(messages));
          print_output_buffer(cp, msg->dest);
	  print_output_buffer(cp, "\n");
          return 1;
	}
    }
  else if (strcmp(argv[0], "GET_STRC") == 0)
    {
      if (!my_protocol || !llist_head(messages))
	return -1;
      else
	{
          struct message *msg = llist_node_data(llist_head(messages));
          char buf[64];
          snprintf(buf, 64, "%d\n", msg->strc);
          print_output_buffer(cp, buf);
	  return 1;
	}
    }
  else if (strcmp(argv[0], "GET_UINT32C") == 0)
    {
      if (!my_protocol || !llist_head(messages))
	return -1;
      else
	{
          struct message *msg = llist_node_data(llist_head(messages));
          char buf[64];
          snprintf(buf, 64, "%d\n", msg->uint32c);
          print_output_buffer(cp, buf);
	  return 1;
	}
    }
  else if (strcmp(argv[0], "GET_SINT32C") == 0)
    {
      if (!my_protocol || !llist_head(messages))
	return -1;
      else
	{
          struct message *msg = llist_node_data(llist_head(messages));
          char buf[64];
          snprintf(buf, 64, "%d\n", msg->sint32c);
          print_output_buffer(cp, buf);
	  return 1;
	}
    }
  else if (strcmp(argv[0], "GET_STRV") == 0)
    {
      if (!my_protocol || !llist_head(messages) || argc < 1)
	return -1;
      else
	{
          int i = strtoul(argv[1], NULL, 0);
          struct message *msg = llist_node_data(llist_head(messages));
          print_output_buffer(cp, msg->strv[i]);
	  print_output_buffer(cp, "\n");
	  return 1;
	}
    }
  else if (strcmp(argv[0], "GET_UINT32V") == 0)
    {
      if (!my_protocol || !llist_head(messages) || argc < 1)
	return -1;
      else
	{
          int i = strtoul(argv[1], NULL, 0);
          struct message *msg = llist_node_data(llist_head(messages));
          char buf[64];
          snprintf(buf, 64, "%lu\n", (long int) msg->uint32v[i]);
          print_output_buffer(cp, buf);
	  return 1;
	}
    }
  else if (strcmp(argv[0], "GET_SINT32V") == 0)
    {
      if (!my_protocol || !llist_head(messages) || argc < 1)
	return -1;
      else
	{
          int i = strtoul(argv[1], NULL, 0);
          struct message *msg = llist_node_data(llist_head(messages));
          char buf[64];
          snprintf(buf, 64, "%ld\n", (long int) msg->sint32v[i]);
          print_output_buffer(cp, buf);
	  return 1;
	}
    }
  else if (strcmp(argv[0], "SHUTDOWN_SESSION") == 0)
    {
      struct session *session;

      if (argc < 1)
        return -1;

      session = hashtable_lookup(sessions, argv[1]);
      if (!session)
        return -1;

      SYM(protocol, shutdown_session)(session);
      return 1;
    }
  else if (strcmp(argv[0], "REGISTER_ATTRIBUTE") == 0)
    {
      char buf[64];
      int i;

      if (argc < 1)
        return -1;

      i = SYM(protocol, register_attribute)(argv[1]);
      snprintf(buf, 64, "%d\n", i);
      print_output_buffer(cp, buf);
      return 1;
    }
  else if (strcmp(argv[0], "SET_ATTRIBUTE") == 0)
    {
      struct session *session;
      int i;

      if (argc < 3)
        return -1;

      session = hashtable_lookup(sessions, argv[1]);
      if (!session)
        return -1;

      i = strtoul(argv[2], NULL, 0);

      SYM(protocol, set_attribute)(session, i, xstrdup(argv[3]));
      return 1;
    }
  else if (strcmp(argv[0], "GET_ATTRIBUTE") == 0)
    {
      struct session *session;
      int i;
      char *c;

      if (argc < 2)
        return -1;

      session = hashtable_lookup(sessions, argv[1]);
      if (!session)
        return -1;

      /* Set i to be the integer form of the second word */
      i = strtoul(argv[2], NULL, 0);

      c = SYM(protocol, get_attribute)(session, i);
      print_output_buffer(cp, c);
      print_output_buffer(cp, "\n");
      return 1;
    }
  else
    return 0;
}

static void
pipe_fail(void)
{
  unload_module(MYNAME);
}

int
MODULE_INIT(void *data unused_attribute)
{
  if (!(cp = setup_cmdpipe(MYHANDLE, MYNAME, 32768, &parse_command, &pipe_fail)))
    return 0;

  sessions = new_hashtable(&hash_ascii, 17);
  new_sessions = new_llist();
  messages = new_llist();
  name_attribute = SYM(protocol, register_attribute)("test.protocol name");
  SYM(protocol, register_new_session_handler)(MYHANDLE, &new_session, NULL);

  return MODULE_LOAD_OK;
}

int
MODULE_DEINIT(const char *new_version unused_attribute, module_init_func *init unused_attribute)
{
  free_llist(new_sessions);
  free_hashtable(sessions);
  destroy_llist(messages, SYM(message, free_message));
  if (cp)
    close_cmdpipe(cp);
  cp = NULL;
  return MODULE_UNLOAD_OK;
}
