/************************************************************************
 *   This is part of Dancer, a text relay daemon
 *   Copyright (C) 2002 Andrew Suffield <asuffield@users.sourceforge.net>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

#include <dancerd/config.h>
#include <dancerd/modules.h>
#include <dancerd/log.h>
#include <dancerd/paths.h>
#include <dancerd/xmalloc.h>
#include <dancerd/xstring.h>
#include <dancerd/hashtable.h>
#include <dancerd/test/cmdpipe.h>

#include <dancerd/transport/transport.h>

#include <stddef.h>
#include <string.h>

SETUP_MODULE("test.transport", MODVERSION);

IMPORTED_SYMBOL(transport, register_transport);
IMPORTED_SYMBOL(transport, unregister_transport);
IMPORTED_SYMBOL(transport, transport_connected);
IMPORTED_SYMBOL(transport, attach_fd);
IMPORTED_SYMBOL(transport, attach_mem);
IMPORTED_SYMBOL(transport, data_read);
IMPORTED_SYMBOL(transport, connection_closed);
IMPORTED_SYMBOL(transport, class_name);

#include "symtab"

static struct cmdpipe *cp = NULL;

static struct transport *my_transport = NULL;

static struct hashtable *tclass_hash, *conn_hash;
static int autoclose = 0;

/* We don't send a **DONE** on this one because it's only ever called
 * from within parse_command; it's easier to treat this as a response
 * to a 'REGISTER' command
 *
 * Likewise for deactivate and deinit
 */
static transport_init_tclass_handler init_tclass;
static void *
init_tclass(struct transport_class *tc, void *d)
{
  char *name = xstrdup(SYM(transport, class_name)(tc));
  xfree(d);
  print_output_buffer(cp, "INIT_TCLASS ");
  print_output_buffer(cp, name);
  print_output_buffer(cp, "\n");
  write_output_done(cp);
  hashtable_insert(tclass_hash, name, tc);
  return name;
}

static transport_deactivate_tclass_handler deactivate_tclass;
static void
deactivate_tclass(struct transport_class *tc, void *d)
{
  print_output_buffer(cp, "DEACTIVATE_TCLASS ");
  print_output_buffer(cp, SYM(transport, class_name)(tc));
  print_output_buffer(cp, " ");
  print_output_buffer(cp, d);
  print_output_buffer(cp, "\n");
  write_output_done(cp);
}

static transport_deinit_tclass_handler deinit_tclass;
static void
deinit_tclass(struct transport_class *tc, void *d)
{
  const char *name = SYM(transport, class_name)(tc);
  print_output_buffer(cp, "DEINIT_TCLASS ");
  print_output_buffer(cp, name);
  print_output_buffer(cp, " ");
  print_output_buffer(cp, d);
  print_output_buffer(cp, "\n");
  write_output_done(cp);
  hashtable_delete(tclass_hash, name);
}

static transport_begin_connection_handler begin_connection;
static void *
begin_connection(struct connection *conn, const char *target,
                 void **argv unused_attribute, int argc unused_attribute)
{
  print_output_buffer(cp, "BEGIN ");
  print_output_buffer(cp, target);
  print_output_buffer(cp, "\n");
  write_output_done(cp);
  hashtable_insert(conn_hash, target, conn);
  return xstrdup(target);
}

static transport_write_handler write_data;
static void
write_data(struct connection *conn unused_attribute, void *d, struct dbuffer *buf,
           const void *data unused_attribute, size_t len)
{
  char *c;
  len = dbuffer_len(buf);
  c = xmalloc(len + 1);
  dbuffer_extract(buf, c, len);
  c[len] = '\0';
  /* Strip a trailing \n */
  if (c[len-1] == '\n')
    c[len-1] = '\0';
  print_output_buffer(cp, "WRITE ");
  print_output_buffer(cp, d);
  print_output_buffer(cp, ": ");
  print_output_buffer(cp, c);
  print_output_buffer(cp, "\n");
  write_output_done(cp);
  xfree(c);
}

static transport_close_handler close_connection;
static void
close_connection(struct connection *conn unused_attribute, void *d)
{
  print_output_buffer(cp, "CLOSE ");
  print_output_buffer(cp, d);
  print_output_buffer(cp, "\n");
  write_output_done(cp);
  if (autoclose)
    {
      hashtable_delete(conn_hash, d);
      xfree(d);
      SYM(transport, connection_closed)(conn);
    }
}

static transport_abort_handler abort_connection;
static void
abort_connection(struct connection *conn, void *d)
{
  print_output_buffer(cp, "ABORT ");
  print_output_buffer(cp, d);
  print_output_buffer(cp, "\n");
  write_output_done(cp);
  hashtable_delete(conn_hash, d);
  xfree(d);
  SYM(transport, connection_closed)(conn);
}

static int
parse_command(char *data, size_t len)
{
  struct parsed_args *args = parse_cmdpipe_data(data, len);
  char **argv = args->argv;
  size_t *argl = args->argl;
  int argc = args->argc;
  
  if (strcmp(argv[0], "REGISTER") == 0)
    {
      if (my_transport)
	return 0;
      else
	{
	  my_transport = SYM(transport, register_transport)("test", MYHANDLE,
							    &init_tclass, &deactivate_tclass, &deinit_tclass,
                                                            &begin_connection,
							    &write_data, &close_connection, &abort_connection);
	  if (!my_transport)
	    return 0;
	  else
	    return 1;
	}
    }
  else if (strcmp(argv[0], "UNREGISTER") == 0)
    {
      if (!my_transport)
	return -1;
      else
	{
	  SYM(transport, unregister_transport)(my_transport);
	  my_transport = NULL;
	  return 1;
	}
    }
  else if (strcmp(argv[0], "CONNECT") == 0)
    {
      if (!my_transport || argc < 2)
	return -1;
      else
	{
          struct connection *conn;
          if (hashtable_lookup(conn_hash, argv[2]))
            return 0;
          else
            {
              if ((conn = SYM(transport, transport_connected)(my_transport,
                                                              hashtable_lookup(tclass_hash, argv[1]),
                                                              xstrdup(argv[2]))))
                {
                  hashtable_insert(conn_hash, argv[2], conn);
                  return 1;
                }
              else
                return 0;
            }
        }
    }
  else if (strcmp(argv[0], "READ") == 0)
    {
      if (!my_transport || argc < 2)
	return -1;
      else
	{
          struct connection *conn;
          if (!(conn = hashtable_lookup(conn_hash, argv[1])))
            {
              return -1;
            }
          else
            {
              SYM(transport, data_read)(conn, argv[2], argl[2]);
              return 1;
            }
        }
    }
  else if (strcmp(argv[0], "CLOSE") == 0)
    {
      if (!my_transport || argc < 1)
	return -1;
      else
	{
          struct connection *conn = hashtable_lookup(conn_hash, argv[1]);
          hashtable_delete(conn_hash, argv[1]);
          SYM(transport, connection_closed)(conn);
          return 1;
        }
    }
  else if (strcmp(argv[0], "FINISH") == 0)
    {
      if (!my_transport || argc < 1)
	return -1;
      else
	{
          struct connection *conn = hashtable_lookup(conn_hash, argv[1]);
          SYM(transport, finished_connection)(conn);
          return 1;
        }
    }
  else if (strcmp(argv[0], "AUTOCLOSE_ON") == 0)
    {
      autoclose = 1;
      return 1;
    }
  else if (strcmp(argv[0], "AUTOCLOSE_OFF") == 0)
    {
      autoclose = 0;
      return 1;
    }
  else
    return 0;
}

static void
pipe_fail(void)
{
  unload_module(MYNAME);
}

int
MODULE_INIT(void *data unused_attribute)
{
  if (!(cp = setup_cmdpipe(MYHANDLE, MYNAME, 32768, &parse_command, &pipe_fail)))
    return 0;

  tclass_hash = new_hashtable(&hash_ascii, 17);
  conn_hash = new_hashtable(&hash_ascii, 17);

  return MODULE_LOAD_OK;
}

int
MODULE_DEINIT(const char *new_version unused_attribute, module_init_func *init unused_attribute)
{
  free_hashtable(tclass_hash);
  free_hashtable(conn_hash);
  if (cp)
    close_cmdpipe(cp);
  cp = NULL;
  return MODULE_UNLOAD_OK;
}
